/*
 * Copyright 2016 - 2017 Systems Group, ETH Zurich
 *
 * This hardware operator is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`define FIFO_SIZE 11

`define BENCH_TPUT_READ_ONLY 8'd1
`define BENCH_TPUT_WRITE_ONLY 8'd2
`define BENCH_TPUT_READ_WRITE 8'd3

`define BENCH_TPUT_RAND_READ_ONLY 8'd4
`define BENCH_TPUT_RAND_WRITE_ONLY 8'd5
`define BENCH_TPUT_RAND_READ_WRITE 8'd6

`define BENCH_LATENCY_READ_ONLY 8'd7
`define BENCH_LATENCY_WRITE_ONLY 8'd8
`define BENCH_LATENCY_READ_WRITE 8'd9

`define BENCH_LATENCY_RAND_READ_ONLY 8'd10
`define BENCH_LATENCY_RAND_WRITE_ONLY 8'd11
`define BENCH_LATENCY_RAND_READ_WRITE 8'd12

`define BENCH_WRITE_DATA_BACK 8'd13

`define MAX_READ_REQUESTS_PENDING 32'd2048
`define MAX_WRITE_REQUESTS_PENDING 32'd2048


module microbench #(parameter DATA_WIDTH_IN  = 4,
              parameter DATA_WIDTH_OUT = 4 ) (
    input   wire                                   clk,				// 200 MHz clock
    input   wire                                   rst_n,			// Stays low as long as AFU is idle
    //-------------------------------------------------//
    input   wire                                   start_um,		// active hight 8 cycles to trigger AFU
    input   wire [511:0]                           um_params,		// params (config data structure)
    output  wire                               	   um_done,			// indicate that we are done
    // TX RD
    output  reg  [57:0]                            um_tx_rd_addr,	// read req. address (accesses 512bits)
    output  reg  [7:0]                             um_tx_rd_tag,	// read req. tag
    output  reg                                    um_tx_rd_valid,	// indicated that tag and addr. are valid
    input   wire                                   um_tx_rd_ready,	// framework tells that we can issue request
    // TX WR
    output  reg  [57:0]                            um_tx_wr_addr,	// address to write result
    output  reg  [7:0]                             um_tx_wr_tag,	// tag to write the result
    output  reg                                    um_tx_wr_valid,	// tell that tag and address are valid
    output  reg  [511:0]                           um_tx_data,		// the data to write
    input   wire                                   um_tx_wr_ready,	// that the framework is read to get data
    // RX RD
    input   wire [7:0]                             um_rx_rd_tag,	// response to the read request: the tag
    input   wire [511:0]                           um_rx_data,		// the requested data
    input   wire                                   um_rx_rd_valid,	// that tag and data are valid
    output  reg                                    um_rx_rd_ready,	// that we are ready for new data
    // RX WR 
    input   wire                                   um_rx_wr_valid,	// that the tag to our write is valid
    input   wire [7:0]                             um_rx_wr_tag		// the tag to confirm the write we issued
);

reg                         um_activated;
reg  [511:0]                um_params_reg;

// vars to generate read requests
wire [31:0]					rd_issued_cnt_inc;
reg [31:0]					rd_issued_cnt;
reg 						rd_issued_done;
reg [63:0]					read_start;

// vars to handle incoming read responses
reg                         rd_done;		// flag to indicate if we are finished with reading in data
reg  [31:0]                 rd_cnt;			// read counter
wire [31:0]                 rd_cnt_inc;


// vars to generate write requests
reg  [31:0]                 wr_cnt;			// write counter
wire [31:0]                 wr_cnt_inc;
reg                         wr_done;
reg [63:0]					write_start;

// vars to control the state
reg 			benchmark_start;
reg [63:0]		benchmark_timeout;		// timeout counter to start the next request
reg [31:0]		num_cl	;				// number of cache lines to read for this experiment

// vars to handle ack responses
reg							ack_done;
reg [31:0]					ack_cnt;
wire [31:0]					ack_cnt_inc;

// global clock counter
reg [63:0] 		cycle_counter;


// to generate pseudo random data
// generated using http://outputlogic.com/?page_id=275 and 65536 as input
// found by https://stackoverflow.com/questions/757151/random-number-generation-on-spartan-3e
reg [16:0] lfsr_small;
wire d0_small,lfsr_equal_small;
xnor(d0_small,lfsr_small[16],lfsr_small[13]);
assign lfsr_equal_small = (lfsr_small == 17'h1FF01);

reg [22:0] lfsr_mid;
wire d0_mid,lfsr_equal_mid;
xnor(d0_mid,lfsr_mid[22],lfsr_mid[17]);
assign lfsr_equal_mid = (lfsr_mid == 23'h7FF803);

reg [26:0] lfsr_big;
wire d0_big,lfsr_equal_big;
xnor(d0_big,lfsr_big[26],lfsr_big[4],lfsr_big[1],lfsr_big[0]);
assign lfsr_equal_big = (lfsr_big == 27'h5FFEAAA);


reg [31:0] rand_addr_offset_read;
reg [31:0] rand_addr_offset_write;
reg [15:0] rand_16_bits;

// benchmarking variables
reg [7:0]	benchmark_part;		// to select which benchmark to run
reg [63:0] 	tput_cycles_read;	// amount of cycles to read data
reg [63:0]  tput_cl_read;		// amount of cache lines we have read

reg [63:0]	tput_cycles_write;	// amount of cycles to write data
reg [63:0]  tput_cl_write;		// amount of cache lines we have written

reg [63:0] 	tput_cycles_read_write;	// amount of cycles to read and write data
reg [63:0] 	tput_cl_read_write;		// amount of cache lines we have read and written back

reg [63:0] 	latency_cycles_read;	// sum of all latencies: from issuing a request, till we get the response
reg [63:0]	latency_cl_read;		// amount of cache lines we have read

reg [63:0]	latency_cycles_write;	// sum of all latencies: from issuing a write, till we get the ack
reg [63:0]	latency_cl_write;		// amount of cache lines we have written

reg [63:0]	latency_cycles_read_write;	// sum of all latencies: from issuing a request, till we get the ack
reg [63:0]	latency_cl_read_write;		// amount of cache lines we have read and written


reg [63:0] tput_rand_cycles_read;
reg[63:0] tput_rand_cl_read;

reg[63:0] tput_rand_cycles_write;
reg[63:0] tput_rand_cl_write;

reg[63:0] tput_rand_cycles_read_write;
reg[63:0] tput_rand_cl_read_write;

reg[63:0] latency_rand_cycles_read;
reg[63:0] latency_rand_cl_read;

reg[63:0] latency_rand_cycles_write;
reg[63:0] latency_rand_cl_write;

reg[63:0] latency_rand_cycles_read_write;
reg[63:0] latency_rand_cl_read_write;

// queue for read-write throughput and latency measurements
reg queue_read_enable;
reg queue_write_enable;
wire queue_buf_empty;
wire queue_buf_empty_next;
wire queue_buf_filled;
wire queue_buf_filled_next;
reg [511:0] queue_buf_in;
wire [511:0] queue_buf_out;
wire [31:0] queue_nelems;
wire [31:0] queue_ntotal_read;
wire [31:0] queue_ntotal_write;

queue #( 512, `FIFO_SIZE) queue (
		clk,						// clock
		~rst_n,						// reset the queue
		queue_write_enable,	// data on "buf_in" that should be written
		queue_read_enable,	// data on "buf_out" is read
		queue_buf_in,			// input data
		queue_buf_out,		// output data
		queue_buf_empty,		// if the buffer is empty
		queue_buf_empty_next,
		queue_buf_filled,		// if the buffer is filled
		queue_buf_filled_next, // if the buffer is filled, if we insert an element "in this cycle"
		queue_nelems,			// number of elements in the buffer
        queue_ntotal_read,
        queue_ntotal_write
);

///////////////////////////// handling the params, and activation, and random number generator //////////////////////////////////////////////////////
always @(posedge clk) begin
	
    if (~rst_n) begin
        // reset	
		lfsr_small <= 0;
		lfsr_mid <= 0;
		lfsr_big <= 0;
        um_activated <= 1'b0;
		cycle_counter <= 0;
		rand_addr_offset_read <= 0;
		rand_addr_offset_write <= 0;
		rand_16_bits <= 0;
    end
    else begin
		lfsr_small <= lfsr_equal_small ? 17'h0 : {lfsr_small[15:0],d0_small};
		lfsr_mid <= lfsr_equal_mid ? 23'h0 : {lfsr_mid[21:0],d0_mid};
		lfsr_big <= lfsr_equal_big ? 27'h0 : {lfsr_big[25:0],d0_big};

		rand_addr_offset_read <= (um_params_reg[159:128] <= 2**17) ? 0 
							: (um_params_reg[159:128] >= 2**17 && um_params_reg[159:128] <= 2**23) ? {15'b0, lfsr_small} 
							: ( um_params_reg[159:128] >= 2**23 && um_params_reg[159:128] <= 2**26) ?  {9'b0, lfsr_mid} : {5'b0, lfsr_big};
		rand_addr_offset_write <= (um_params_reg[159:128] <= 2**17) ? 0 
							: (um_params_reg[159:128] >= 2**17 && um_params_reg[159:128] <= 2**23) ? {15'b0, ~lfsr_small}
							: ( um_params_reg[159:128] >= 2**23 && um_params_reg[159:128] <= 2**26) ? {9'b0, ~lfsr_mid} : {5'b0, ~lfsr_big};

		rand_16_bits <= lfsr_small[15:0];

		cycle_counter <= cycle_counter +1;
		if(um_done) begin
        	um_activated <= 1'b0;
    	end
		else if (start_um ) begin
        	um_activated <= 1'b1;
    	end
	end
end

always @(posedge clk) begin

	if(~rst_n) begin
		um_params_reg <= 512'b0;
	end
	else begin
		if (start_um) begin
		    um_params_reg <= um_params;
		end
	end
end

// we are done with processing, when we have writen back all data
assign um_done = (benchmark_part == `BENCH_WRITE_DATA_BACK) & ( wr_done) ; 


reg				read_hold_valid;

/////////////////////// generating read requests /////////////////////////////////////////
always @(posedge clk) begin
    if ( ~rst_n | um_done | ~um_activated | ~benchmark_start) begin
        // reset of output registers
        um_tx_rd_addr  <= 0;
        um_tx_rd_valid <= 0;
        um_tx_rd_tag   <= 0;
		
		// reset of local variables
		rd_issued_cnt <= 0;
		rd_issued_done <= 0;
		read_start <= 0;

		read_hold_valid <= 0;
    end
    else begin
		um_tx_rd_valid <= 0;

		if( um_tx_rd_valid && !um_tx_rd_ready ) begin
			read_hold_valid <= 1;
		end
		else if( read_hold_valid ) begin
			if( um_tx_rd_ready ) begin
				read_hold_valid <= 0;
				um_tx_rd_valid <= 1;
			end
		end
		else if (~rd_issued_done ) begin
			
            if(  benchmark_part == `BENCH_TPUT_READ_ONLY ) begin
                if( um_tx_rd_ready & ( (rd_issued_cnt - rd_cnt ) < `MAX_READ_REQUESTS_PENDING) ) begin
                    um_tx_rd_addr  <= um_params_reg[63:6] + rd_issued_cnt; // issue next read
					um_tx_rd_valid <= 1;
					rd_issued_cnt <= rd_issued_cnt_inc;	
	
					if(rd_issued_cnt_inc == num_cl ) begin 
						rd_issued_done <= 1;
					end
                end
            end
			else if( (benchmark_part == `BENCH_TPUT_READ_WRITE) ) begin
				// if the framework is ready to read, and we do not have as many requests in flight as we have space left
				if( um_tx_rd_ready & ( ( rd_issued_cnt - rd_cnt  ) <  ( ( 1 << `FIFO_SIZE) -queue_nelems - 16) ) &  ( (rd_issued_cnt - rd_cnt ) < `MAX_READ_REQUESTS_PENDING) ) begin
					um_tx_rd_addr  <= um_params_reg[63:6] + rd_issued_cnt; // issue next read
					um_tx_rd_valid <= 1;
					rd_issued_cnt <= rd_issued_cnt_inc;	
	
					if(rd_issued_cnt_inc == num_cl ) begin 
						rd_issued_done <= 1;
					end
				end
			end
			else if( benchmark_part == `BENCH_LATENCY_READ_ONLY) begin
				if( um_tx_rd_ready & (rd_issued_cnt == rd_cnt ) ) begin // issue request only if there is no pending request
					um_tx_rd_addr  <= um_params_reg[63:6] + rd_issued_cnt; // issue next read
					um_tx_rd_valid <= 1;
					rd_issued_cnt <= rd_issued_cnt_inc;
					
					read_start <= cycle_counter;	
					if(rd_issued_cnt_inc == num_cl ) begin 
						rd_issued_done <= 1;
					end
				end
			end	
			else if( benchmark_part == `BENCH_LATENCY_READ_WRITE) begin
				if( um_tx_rd_ready & ( rd_issued_cnt == ack_cnt) ) begin
					um_tx_rd_addr  <= um_params_reg[63:6] + rd_issued_cnt; // issue next read
					um_tx_rd_valid <= 1;
					rd_issued_cnt <= rd_issued_cnt_inc;
				
					read_start <= cycle_counter;
					if(rd_issued_cnt_inc == num_cl ) begin 
						rd_issued_done <= 1;
					end
				end
			end
            else if( (benchmark_part == `BENCH_TPUT_RAND_READ_ONLY) ) begin
                if( um_tx_rd_ready & ( (rd_issued_cnt - rd_cnt ) < 32'd512)  ) begin
                    um_tx_rd_addr  <= um_params_reg[63:6] + rand_addr_offset_read; // issue next read
					um_tx_rd_valid <= 1;
					rd_issued_cnt <= rd_issued_cnt_inc;

					if(rd_issued_cnt_inc == num_cl ) begin 
						rd_issued_done <= 1;
					end
                end
            end
			else if(  (benchmark_part == `BENCH_TPUT_RAND_READ_WRITE) ) begin
				if( um_tx_rd_ready & ( ( rd_issued_cnt - rd_cnt  ) <  ( ( 1 << `FIFO_SIZE) - queue_nelems -16 ) ) & ( (rd_issued_cnt - rd_cnt ) < `MAX_READ_REQUESTS_PENDING) ) begin
					um_tx_rd_addr  <= um_params_reg[63:6] + rand_addr_offset_read; // issue next read
					um_tx_rd_valid <= 1;
					rd_issued_cnt <= rd_issued_cnt_inc;

					if(rd_issued_cnt_inc == num_cl ) begin 
						rd_issued_done <= 1;
					end
				end
			end
            else if( benchmark_part == `BENCH_LATENCY_RAND_READ_ONLY ) begin
                if( um_tx_rd_ready & (rd_issued_cnt == rd_cnt ) ) begin // issue request only if there is no pending request
					um_tx_rd_addr  <= um_params_reg[63:6] + rand_addr_offset_read; // issue next read
					um_tx_rd_valid <= 1;
					rd_issued_cnt <= rd_issued_cnt_inc;
					
					read_start <= cycle_counter;	
					if(rd_issued_cnt_inc == num_cl ) begin 
						rd_issued_done <= 1;
					end
				end
            end
            else if( benchmark_part == `BENCH_LATENCY_RAND_READ_WRITE) begin
                if( um_tx_rd_ready & ( rd_issued_cnt == ack_cnt) ) begin
					um_tx_rd_addr  <= um_params_reg[63:6] + rand_addr_offset_read; // issue next read
					um_tx_rd_valid <= 1;
					rd_issued_cnt <= rd_issued_cnt_inc;
				
					read_start <= cycle_counter;
					if(rd_issued_cnt_inc == num_cl ) begin 
						rd_issued_done <= 1;
					end
				end
            end
			else begin
				rd_issued_done <= 1;
			end
		end
		
	end
end

assign rd_issued_cnt_inc = rd_issued_cnt + 1'b1; 	// count how many cache lines we have requested so far


/////////////////////// handling read responses /////////////////////////////////////////
always @(posedge clk) begin
    if ( ~rst_n | um_done | ~um_activated | ~benchmark_start) begin

		// reset of local variables
		rd_done <= 1'b0;
		rd_cnt <= 32'b0;
		queue_write_enable <= 0;
		queue_buf_in <= 0;

		if( ~rst_n | um_done | ~um_activated ) begin
			tput_cycles_read <= 0;	
			tput_cl_read <= 0;
			
			latency_cycles_read <= 0;	
			latency_cl_read <= 0;

			tput_rand_cycles_read <= 0;
			tput_rand_cl_read <= 0;
			
			latency_rand_cycles_read <= 0;
			latency_rand_cl_read <= 0;
		end
    end
    else begin

		queue_write_enable <= 0;
		queue_buf_in <= 0;
		um_rx_rd_ready <= 0;

		if( ~rd_done ) begin
			if( benchmark_part == `BENCH_TPUT_READ_ONLY ) begin
				tput_cycles_read <= tput_cycles_read +1;
				um_rx_rd_ready <= 1;
				if(um_rx_rd_valid & um_rx_rd_ready) begin
					tput_cl_read <= tput_cl_read + 1;
					rd_cnt <= rd_cnt_inc;
					if(rd_cnt_inc == num_cl) begin
						rd_done <= 1;
						um_rx_rd_ready <= 0;
					end
				end
			end
			else if ( (benchmark_part == `BENCH_TPUT_READ_WRITE) | (benchmark_part == `BENCH_LATENCY_READ_WRITE) |
					  (benchmark_part == `BENCH_TPUT_RAND_READ_WRITE) | (benchmark_part == `BENCH_LATENCY_RAND_READ_WRITE) ) begin
				if(um_rx_rd_valid & um_rx_rd_ready) begin 
				 	rd_cnt         <= rd_cnt_inc;
					queue_write_enable <= 1;
					queue_buf_in <= um_rx_data;
					if(rd_cnt_inc == num_cl) begin
						rd_done <= 1;
						um_rx_rd_ready <= 0;
					end
				end
				else begin
					queue_write_enable <= 0;
					queue_buf_in <= 0;
				end

				// if there is enough space left in the queue, we can issue the next read
				if( ~queue_buf_filled & ~queue_buf_filled_next ) begin 
					um_rx_rd_ready <= 1;
				end
				else begin
					um_rx_rd_ready <= 0;
				end
			end
			else if( benchmark_part == `BENCH_LATENCY_READ_ONLY) begin
				um_rx_rd_ready <= 1;
				if( um_rx_rd_valid & um_rx_rd_ready) begin
					rd_cnt         <= rd_cnt_inc;
					latency_cl_read <= latency_cl_read +1;
					latency_cycles_read <= latency_cycles_read + ( cycle_counter - read_start);		
					if(rd_cnt_inc == num_cl ) begin 
						rd_done <= 1;
					end
				end
			end	
			else if( ( benchmark_part == `BENCH_TPUT_RAND_READ_ONLY )  ) begin
				tput_rand_cycles_read <= tput_rand_cycles_read +1;
				um_rx_rd_ready <= 1;
				if(um_rx_rd_valid & um_rx_rd_ready) begin
					tput_rand_cl_read <= tput_rand_cl_read + 1;
					rd_cnt <= rd_cnt_inc;
					
					if(rd_cnt_inc == num_cl) begin
						rd_done <= 1;
						um_rx_rd_ready <= 0;
					end
				end
			end
			else if( benchmark_part == `BENCH_LATENCY_RAND_READ_ONLY) begin
				um_rx_rd_ready <= 1;
				if( um_rx_rd_valid & um_rx_rd_ready) begin
					rd_cnt         <= rd_cnt_inc;
					latency_rand_cl_read <= latency_rand_cl_read +1;
					latency_rand_cycles_read <= latency_rand_cycles_read + ( cycle_counter - read_start);		
					if(rd_cnt_inc == num_cl ) begin 
						rd_done <= 1;
					end
				end
			end

		end // end of ~rd_done

	end // end of else

end

assign rd_cnt_inc = rd_cnt + 1'b1;					// count how many cache lines we have read so far

/////////////////////// Process the data and controll the benchmark ////////////////////////////

always@(posedge clk) begin
	if (~rst_n | um_done | ~um_activated ) begin

		// state machine
		benchmark_part <= 0;
		benchmark_start <= 0;
		benchmark_timeout <= 0;	
		num_cl <= 0;
    end
	else begin
		
			// if a benchmark is not running, wait some cycles to start the next
			if( benchmark_start == 0) begin
				if( benchmark_timeout > um_params_reg[223:192] ) begin
					benchmark_timeout <= 0;
					benchmark_start <= 1;
					benchmark_part <= benchmark_part +1;	
						
					if( benchmark_part + 1 <= 8'd6) begin
						num_cl	<= um_params_reg[159:128];
					end		
					else begin			
						num_cl <= um_params_reg[191:160];
					end
				end
				else begin
					benchmark_timeout <= benchmark_timeout + 1;
				end
			end
			else if( (benchmark_part == `BENCH_TPUT_READ_ONLY ) | ( benchmark_part == `BENCH_LATENCY_READ_ONLY) |
                     (benchmark_part == `BENCH_TPUT_RAND_READ_ONLY) | (benchmark_part == `BENCH_LATENCY_RAND_READ_ONLY) ) begin
				if( rd_done) begin
					benchmark_start <= 0;
				end
			end
			else if( (benchmark_part == `BENCH_TPUT_WRITE_ONLY) | ( benchmark_part == `BENCH_TPUT_READ_WRITE ) |
					 (benchmark_part == `BENCH_LATENCY_WRITE_ONLY) | ( benchmark_part == `BENCH_LATENCY_READ_WRITE ) |
                     (benchmark_part == `BENCH_TPUT_RAND_WRITE_ONLY ) | (benchmark_part == `BENCH_TPUT_RAND_READ_WRITE) |
                     (benchmark_part == `BENCH_LATENCY_RAND_WRITE_ONLY) | ( benchmark_part == `BENCH_LATENCY_RAND_READ_WRITE )  ) begin
				if( ack_done ) begin
					benchmark_start <= 0;
				end
			end
			else if( benchmark_part == `BENCH_WRITE_DATA_BACK) begin
				// do nothing - just wait. um_done is assiged via "assign" statement
			end
			else begin
				benchmark_start <= 0;
			end
			
		
	end // end of else ( ~rst_n | ~um_activated )
end // end of always





/////////////////////// Generate write requests ///////////////////////////////

reg				write_hold_valid;

always@(posedge clk) begin
    // Signals need reset
    if( ~rst_n | um_done | ~um_activated | ~benchmark_start ) begin
        um_tx_wr_valid    <= 1'b0;
        wr_done <= 1'b0;
        wr_cnt <= 32'b0;
		um_tx_wr_addr <= 58'b0;
		um_tx_data <= 512'b0;
		um_tx_wr_tag <= 8'b0;
		
		queue_read_enable <= 1'b0;
		
		write_start <= 0;

		write_hold_valid <= 0;
    end 
    else begin	
		queue_read_enable <= 0;
		um_tx_wr_valid <= 1'b0;

		if( um_tx_wr_valid && !um_tx_wr_ready ) begin
			write_hold_valid <= 1;
		end
		else if( write_hold_valid ) begin
			if( um_tx_wr_ready ) begin
				um_tx_wr_valid <= 1;
				write_hold_valid <= 0;
			end
		end
		else if(~wr_done ) begin
			if(benchmark_part == `BENCH_TPUT_WRITE_ONLY) begin
				if( um_tx_wr_ready  & ( ( wr_cnt - ack_cnt) < `MAX_WRITE_REQUESTS_PENDING) ) begin
					um_tx_data <= { 128'b0, rand_16_bits, 128'b0, rand_16_bits, 128'b0, ~rand_16_bits, rand_16_bits, rand_16_bits, ~rand_16_bits, rand_16_bits, rand_16_bits };					
					um_tx_wr_addr     <= um_params_reg[127:70] + wr_cnt; // write to the next position
					um_tx_wr_valid    <= 1'b1;
					wr_cnt            <= wr_cnt_inc;
					um_tx_wr_tag <= um_tx_wr_tag + 1;

					if( wr_cnt_inc == num_cl) begin
						wr_done <= 1;
					end
				end
			end
			else if( (benchmark_part == `BENCH_TPUT_READ_WRITE) | ( benchmark_part == `BENCH_LATENCY_READ_WRITE) ) begin
				
				if( um_tx_wr_ready ) begin // we are allowed to write

					//we can write if: we did not write before, and buf is not empty, or we did write, and there is still next data in the queue
					if( ( ~queue_buf_empty & ~um_tx_wr_valid ) | ( um_tx_wr_valid & ~queue_buf_empty_next) )begin 
						um_tx_data        <= queue_buf_out; // the data to write
						um_tx_wr_addr     <= um_params_reg[127:70] + wr_cnt;	// write to the next position
						um_tx_wr_valid    <= 1'b1;
						wr_cnt            <= wr_cnt_inc;
						queue_read_enable <= 1;

						if( wr_cnt_inc == num_cl) begin
							wr_done <= 1;
						end

					end
				end

			end
			else if(benchmark_part == `BENCH_LATENCY_WRITE_ONLY) begin
				if( um_tx_wr_ready & ( wr_cnt == ack_cnt ) ) begin
					um_tx_data <= { 128'b0, rand_16_bits, 128'b0, rand_16_bits, 128'b0, ~rand_16_bits, rand_16_bits, rand_16_bits, ~rand_16_bits, rand_16_bits, rand_16_bits };					
					um_tx_wr_addr     <= um_params_reg[127:70] + wr_cnt; // write to the next position
					um_tx_wr_valid    <= 1'b1;
					wr_cnt            <= wr_cnt_inc;	
					write_start <= cycle_counter;

					if( wr_cnt_inc == num_cl) begin
						wr_done <= 1;
					end
				end

			end
			else if( benchmark_part == `BENCH_WRITE_DATA_BACK) begin
				if( um_tx_wr_ready ) begin
					wr_cnt            <= wr_cnt_inc;
					um_tx_wr_valid <= 1'b1;
					um_tx_wr_addr     <= um_params_reg[127:70] + wr_cnt;
					if( wr_cnt == 0) begin
						um_tx_data <= { {128'b0}, tput_cl_read_write, tput_cycles_read_write, tput_cl_write, tput_cycles_write, tput_cl_read, tput_cycles_read  };
					end
					if( wr_cnt == 1) begin
						um_tx_data <=  { {128'b0}, latency_cl_read_write, latency_cycles_read_write, latency_cl_write, latency_cycles_write, latency_cl_read, latency_cycles_read};
					end
					if( wr_cnt == 2) begin
						um_tx_data <= { {128'b0}, tput_rand_cl_read_write, tput_rand_cycles_read_write, tput_rand_cl_write, tput_rand_cycles_write, tput_rand_cl_read, tput_rand_cycles_read  };
					end
					if( wr_cnt == 3) begin
						um_tx_data <=  { {128'b0}, latency_rand_cl_read_write, latency_rand_cycles_read_write, latency_rand_cl_write, latency_rand_cycles_write, latency_rand_cl_read, latency_rand_cycles_read};
						wr_done <= 1;
					end
					
				end
			end
			else if( benchmark_part == `BENCH_TPUT_RAND_WRITE_ONLY ) begin
				if( um_tx_wr_ready & ( (wr_cnt -ack_cnt) < `MAX_WRITE_REQUESTS_PENDING)  ) begin
					um_tx_data <= { 128'b0, rand_16_bits, 128'b0, rand_16_bits, 128'b0, ~rand_16_bits, rand_16_bits, rand_16_bits, ~rand_16_bits, rand_16_bits, rand_16_bits };					
					um_tx_wr_addr     <= um_params_reg[127:70] + rand_addr_offset_write;
					um_tx_wr_valid    <= 1'b1;
					wr_cnt            <= wr_cnt_inc;

					if( wr_cnt_inc == num_cl) begin
						wr_done <= 1;
					end
				end
			end
            else if( (benchmark_part == `BENCH_TPUT_RAND_READ_WRITE) | ( benchmark_part == `BENCH_LATENCY_RAND_READ_WRITE) ) begin
                if( um_tx_wr_ready ) begin // we are allowed to write

					//we can write if: we did not write before, and buf is not empty, or we did write, and there is still next data in the queue
					if( ( ~queue_buf_empty & ~um_tx_wr_valid ) | ( um_tx_wr_valid & ~queue_buf_empty_next) )begin 
						um_tx_data        <= queue_buf_out; // the data to write
						um_tx_wr_addr     <= um_params_reg[127:70] + rand_addr_offset_write;
						um_tx_wr_valid    <= 1'b1;
						wr_cnt            <= wr_cnt_inc;
						queue_read_enable <= 1;

						if( wr_cnt_inc == num_cl) begin
							wr_done <= 1;
						end

					end
				end
            end
            else if(benchmark_part == `BENCH_LATENCY_RAND_WRITE_ONLY) begin
				if( um_tx_wr_ready & ( wr_cnt == ack_cnt ) ) begin
					um_tx_data <= { 128'b0, rand_16_bits, 128'b0, rand_16_bits, 128'b0, ~rand_16_bits, rand_16_bits, rand_16_bits, ~rand_16_bits, rand_16_bits, rand_16_bits };					
					um_tx_wr_addr     <= um_params_reg[127:70] + rand_addr_offset_write; // write to the next position
					um_tx_wr_valid    <= 1'b1;
					wr_cnt            <= wr_cnt_inc;	
					write_start <= cycle_counter;

					if( wr_cnt_inc == num_cl) begin
						wr_done <= 1;
					end
				end
			end

		end
    end
end 

assign wr_cnt_inc = wr_cnt + 32'd1;


/////////////////////// Handle write responses ///////////////////////////////
// nothing to do so far - we could check if the writes to memory have succeeded or not, and wait for the final write before we terminate

always@(posedge clk) begin
	if( ~rst_n | um_done | ~um_activated | ~benchmark_start) begin
		ack_done <= 0;
		ack_cnt <= 0;
		
		if( ~rst_n | um_done | ~um_activated ) begin
	
			tput_cycles_write <= 0;
			tput_cl_write <= 0;

			tput_cycles_read_write <= 0;		
			tput_cl_read_write <= 0;

			latency_cycles_write <= 0;		
			latency_cl_write <= 0;

			latency_cycles_read_write <= 0;	
			latency_cl_read_write <= 0;

			tput_rand_cycles_write <= 0;
			tput_rand_cl_write <= 0;

            tput_rand_cycles_read_write <= 0;
            tput_rand_cl_read_write <= 0;

            latency_rand_cycles_write <= 0;
            latency_rand_cl_write <= 0;

            latency_rand_cycles_read_write <= 0;
            latency_rand_cl_read_write <= 0;
		end
	end
	else begin
		if( ~ack_done ) begin

			if(benchmark_part == `BENCH_TPUT_WRITE_ONLY) begin
				tput_cycles_write <= tput_cycles_write + 1;
				if( um_rx_wr_valid ) begin
					ack_cnt <= ack_cnt_inc;
					tput_cl_write <= tput_cl_write+1;
					if(ack_cnt_inc == num_cl) begin
						ack_done <= 1;
					end
				end
			end
			else if( benchmark_part == `BENCH_TPUT_READ_WRITE) begin
				tput_cycles_read_write <= tput_cycles_read_write +1;

				if( um_rx_wr_valid ) begin
					ack_cnt <= ack_cnt_inc;			
					tput_cl_read_write <= tput_cl_read_write +1;
					if(ack_cnt_inc == num_cl) begin
						ack_done <= 1;
					end
				end
			end
			else if(benchmark_part == `BENCH_LATENCY_WRITE_ONLY) begin
				if( um_rx_wr_valid ) begin	
					latency_cl_write <= latency_cl_write +1;
					latency_cycles_write <= latency_cycles_write + ( cycle_counter - write_start);
					ack_cnt <= ack_cnt_inc;
					if(ack_cnt_inc == num_cl) begin
						ack_done <= 1;
					end
				end
		
			end		
			else if( benchmark_part == `BENCH_LATENCY_READ_WRITE) begin
				if( um_rx_wr_valid ) begin	
					latency_cl_read_write <= latency_cl_write +1;
					latency_cycles_read_write <= latency_cycles_read_write + ( cycle_counter - read_start);
					ack_cnt <= ack_cnt_inc;
					if(ack_cnt_inc == num_cl) begin
						ack_done <= 1;
					end
				end
			end
			else if( benchmark_part == `BENCH_TPUT_RAND_WRITE_ONLY ) begin
				tput_rand_cycles_write <= tput_rand_cycles_write + 1;
				if( um_rx_wr_valid ) begin
					ack_cnt <= ack_cnt_inc;
					tput_rand_cl_write <= tput_rand_cl_write+1;
					if(ack_cnt_inc == num_cl) begin
						ack_done <= 1;
					end
				end
			end
            else if( benchmark_part == `BENCH_TPUT_RAND_READ_WRITE) begin
                tput_rand_cycles_read_write <= tput_rand_cycles_read_write +1;

				if( um_rx_wr_valid ) begin
					ack_cnt <= ack_cnt_inc;			
					tput_rand_cl_read_write <= tput_rand_cl_read_write +1;
					if(ack_cnt_inc == num_cl) begin
						ack_done <= 1;
					end
				end
            end
			else if( benchmark_part == `BENCH_LATENCY_RAND_WRITE_ONLY) begin
                if( um_rx_wr_valid ) begin	
					latency_rand_cl_write <= latency_rand_cl_write +1;
					latency_rand_cycles_write <= latency_rand_cycles_write + ( cycle_counter - write_start);
					ack_cnt <= ack_cnt_inc;
					if(ack_cnt_inc == num_cl) begin
						ack_done <= 1;
					end
				end
            end
            else if( benchmark_part == `BENCH_LATENCY_RAND_READ_WRITE) begin
                if( um_rx_wr_valid ) begin	
					latency_rand_cl_read_write <= latency_rand_cl_read_write +1;
					latency_rand_cycles_read_write <= latency_rand_cycles_read_write + ( cycle_counter - read_start);
					ack_cnt <= ack_cnt_inc;
					if(ack_cnt_inc == num_cl) begin
						ack_done <= 1;
					end
				end	
            end

      
		end
	end
end

assign ack_cnt_inc = ack_cnt+ 1'b1;

endmodule






