


`default_nettype none


`include "bram_bits.vh"
`include "l_bits.vh"

`define FIFO_SIZE 7

module mwm_crouch_bram_fifo #(parameter DATA_WIDTH_IN  = 4,
              parameter DATA_WIDTH_OUT = 4 ) (
    input   wire                                   clk,				// 200 MHz clock
    input   wire                                   rst_n,			// Stays low as long as AFU is idle
    //-------------------------------------------------//
    input   wire                                   start_um,		// active hight 8 cycles to trigger AFU
    input   wire [511:0]                           um_params,		// params (config data structure)
    output  wire                                   um_done,			// indicate that we are done
    // TX RD
    output  reg  [57:0]                            um_tx_rd_addr,	// read req. address (accesses 512bits)
    output  reg  [7:0]                             um_tx_rd_tag,	// read req. tag
    output  reg                                    um_tx_rd_valid,	// indicated that tag and addr. are valid
    input   wire                                   um_tx_rd_ready,	// framework tells that we can issue request
    // TX WR
    output  reg  [57:0]                            um_tx_wr_addr,	// address to write result
    output  reg  [7:0]                             um_tx_wr_tag,	// tag to write the result
    output  reg                                    um_tx_wr_valid,	// tell that tag and address are valid
    output  reg  [511:0]                           um_tx_data,		// the data to write
    input   wire                                   um_tx_wr_ready,	// that the framework is read to get data
    // RX RD
    input   wire [7:0]                             um_rx_rd_tag,	// response to the read request: the tag
    input   wire [511:0]                           um_rx_data,		// the requested data
    input   wire                                   um_rx_rd_valid,	// that tag and data are valid
    output  reg                                    um_rx_rd_ready,	// that we are ready for new data
    // RX WR 
    input   wire                                   um_rx_wr_valid,	// that the tag to our write is valid
    input   wire [7:0]                             um_rx_wr_tag		// the tag to confirm the write we issued
);


reg                         um_activated;
reg  [511:0]                um_params_reg;

// vars to generate read requests
wire [31:0]					rd_issued_cnt_inc;
reg [31:0]					rd_issued_cnt;
reg 						rd_issued_done;

// vars to handle incoming read responses
reg                         rd_done;		// flag to indicate if we are finished with reading in data
reg  [31:0]                 rd_cnt;			// read counter
wire [31:0]                 rd_cnt_inc;

// vars to process the edges
reg							process_done;
// wire [31:0]					eu;
// wire [31:0]					ev;
wire [31:0]					ew;
reg [`NUM_LEVELS-1:0]					take_edge;
reg [`NUM_LEVELS-1:0]					edge_level;
reg [`NUM_LEVELS-1:0]					oh_te; // one hot representation of take_edge
reg [31:0]					pos_max_significant_bit;
reg [511:0]					read_buffer;
reg							read_buffer_filled;

reg [2:0]					edge_processing_stage;	// to keep track of the processing stage
wire [2:0] 					edge_processing_stage_inc;

reg [2:0]					edge_cnt;	// counting the edges (to keep track of how many cache lines we have processed) and also to keep track of the index of the current read-index
wire [2:0]					edge_cnt_inc;
reg [31:0]					edge_cl_cnt;	// count how many cache lines we have processed
wire [31:0]					edge_cl_cnt_inc;

reg[`NUM_LEVELS-1:0]					mcm_eu;
reg[`NUM_LEVELS-1:0]					mcm_ev;

reg [2:0]					write_data_buffer_count; // to determine the "index" of the output packet
wire [2:0]					write_data_buffer_count_inc;

// vars to generate write requests
reg  [31:0]                 wr_cnt;			// write counter
wire [31:0]                 wr_cnt_inc;
reg                         wr_done;



reg input_queue_read_enable;
reg input_queue_write_enable;
wire input_queue_buf_empty;
wire input_queue_buf_empty_next;
wire input_queue_buf_filled;
wire input_queue_buf_filled_next;
reg [511:0] input_queue_buf_in;
wire [511:0] input_queue_buf_out;
// wire [`FIFO_SIZE:0] input_queue_nelems;
// wire [31:0] input_queue_ntotal_read;
// wire [31:0] input_queue_ntotal_write;

queue #( 512, `FIFO_SIZE) input_queue (
		.clk				(clk),						// clock
		.rst				(~rst_n),						// reset the queue
		.write_enable		(input_queue_write_enable),	// data on "buf_in" that should be written
		.read_enable		(input_queue_read_enable),	// data on "buf_out" is read
		.buf_in				(input_queue_buf_in),			// input data
		.buf_out			(input_queue_buf_out),		// output data
		.buf_empty			(input_queue_buf_empty),		// if the buffer is empty
		.buf_empty_next		(input_queue_buf_empty_next),
		.buf_filled			(input_queue_buf_filled),		// if the buffer is filled
		.buf_filled_next	(input_queue_buf_filled_next)//, // if the buffer is filled, if we insert an element "in this cycle"
		// input_queue_nelems,			// number of elements in the buffer
        // .ntotal_read		(),
        // .ntotal_write		()
);

reg output_queue_write_enable;
reg output_queue_read_enable;
reg [511:0] output_queue_buf_in;
wire [511:0] output_queue_buf_out;
wire output_queue_buf_empty;
wire output_queue_buf_empty_next;
wire output_queue_buf_filled;
wire output_queue_buf_filled_next;
// wire [`FIFO_SIZE:0] output_queue_nelems;
// wire [31:0] output_queue_ntotal_read;
// wire [31:0] output_queue_ntotal_write;

queue #( 512, `FIFO_SIZE) output_queue (
		clk,						// clock
		~rst_n,						// reset the queue	
		output_queue_write_enable,	
		output_queue_read_enable,	
		output_queue_buf_in,
		output_queue_buf_out,
		output_queue_buf_empty,	
		output_queue_buf_empty_next,
		output_queue_buf_filled,
		output_queue_buf_filled_next
		// output_queue_nelems,
        // output_queue_ntotal_read,
        // output_queue_ntotal_write
);


reg							we;
reg							re;
reg [`ADDR_WIDTH -1:0]		addr;
reg [`NUM_LEVELS-1:0]					din;
wire [`NUM_LEVELS-1:0]					dout;

// holds 2^9 = 512 elements
 singleportbram #(
    `NUM_LEVELS,
    `ADDR_WIDTH
) singleportbram (
    clk,
    ~rst_n,
     we,
  	 re,  
    addr,
    din,
    dout
);


///////////////////////////////////////////////////////////////////////////////////

// handling the params, and activation

always @(posedge clk) begin
    if (~rst_n) begin
        // reset
        um_activated <= 1'b0;
    end
    else if(um_done) begin
        um_activated <= 1'b0;
    end
    else if (start_um ) begin
        um_activated <= 1'b1;
    end
end

always @(posedge clk) begin

	if( ~ rst_n) begin
		um_params_reg <= 512'b0;
	end
	else begin
		if (start_um) begin
		    um_params_reg <= um_params;
		end
	end
end

assign um_done  = (um_activated & (um_params_reg[159:128] == 0)) | (wr_done);

/////////////////////// generating read requests /////////////////////////////////////////
always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
        // reset of output registers
        um_tx_rd_addr  <= 0;
        um_tx_rd_valid <= 0;
        um_tx_rd_tag   <= 0;
		
		// reset of local variables
		rd_issued_cnt <= 0;
		rd_issued_done <= 0;
    end
    else begin
		um_tx_rd_valid <= 0;
		if (~rd_issued_done) begin // if we have not generated all read requests...
		    if( um_tx_rd_ready & ( rd_issued_cnt - rd_cnt  ) <  ( 1 << `FIFO_SIZE) ) begin // if the framework is ready to read, and we do not have as many requests in flight as we have space left
				um_tx_rd_addr  <= um_params_reg[63:6] + rd_issued_cnt; // issue next read
				um_tx_rd_tag <= 0;
				um_tx_rd_valid <= 1;
				rd_issued_cnt <= rd_issued_cnt_inc;
					
				if(rd_issued_cnt_inc == um_params[159:128]) begin // if this was our last read, say that generating read requests is done
					rd_issued_done <= 1;
				end
			end
		end
	end
end

assign rd_issued_cnt_inc = rd_issued_cnt + 1'b1; 	// count how many cache lines we have requested so far


/////////////////////// handling read responses /////////////////////////////////////////
always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin

		// reset of local variables
		rd_done <= 1'b0;
		rd_cnt <= 32'b0;
		input_queue_write_enable <= 0;
		input_queue_buf_in <= 0;
    end
    else begin
		if( ~rd_done) begin
			
			// if we have valid data and we told the framework that we are ready to read: we must read the data
			if(um_rx_rd_valid & um_rx_rd_ready) begin 
		     		rd_cnt         <= rd_cnt_inc;
					input_queue_write_enable <= 1;
					input_queue_buf_in <= um_rx_data;

				if(rd_cnt_inc == um_params[159:128]) begin
					rd_done <= 1;
					um_rx_rd_ready <= 0;
				end
				else begin
					// because we read now, we can read in the next cycle, if there is still space available
					if(  ~input_queue_buf_filled_next & ~input_queue_buf_filled) begin 
						um_rx_rd_ready <= 1;
					end
					else begin
						um_rx_rd_ready <= 0;
					end
				end

		    end
			else begin
				input_queue_write_enable <= 0;
				input_queue_buf_in <= 0;
				// because we do not write to the queeu in this cycle, we require that the queue must not be filled now
				// (and also contain some more space, just to be safe)
				if( ~input_queue_buf_filled & ~input_queue_buf_filled_next ) begin 
					um_rx_rd_ready <= 1;
				end
				else begin
					um_rx_rd_ready <= 0;
				end
			end
		end
		else begin // if rd_done
			input_queue_write_enable <= 0;
			input_queue_buf_in <= 0;
			um_rx_rd_ready <= 0;
		end
	end
end

assign rd_cnt_inc = rd_cnt + 1'b1;					// count how many cache lines we have read so far

/////////////////////// Process the edges ////////////////////////////
generate

always@(posedge clk) begin

	
	if ( (~rst_n) | (um_done) | (~um_activated) ) begin

		// state machine
		process_done <= 1'b0;
		write_data_buffer_count <= 3'b0;
		edge_processing_stage <= 3'b0;
		edge_cnt <= 3'b0;
		edge_cl_cnt <= 0;
		
		// computation
		mcm_eu <= 0;
		mcm_ev <= 0;
		edge_level <= 0;
		
		// input queue
		input_queue_read_enable <= 1'b0;
		// read_buffer <= 512'b0;
		read_buffer_filled <= 1'b0;

		// output queue
		output_queue_write_enable <= 1'b0;
		output_queue_buf_in <= 512'b0;

		// bram value initialization
		// reset bram
		if( (um_done || ~um_activated) && rst_n ) begin
			addr <= addr + 1;
			din <= 0;
			we <= 1'b1;
		end
		else begin
			addr <= 0;
			din <= 0;
			re <= 1'b0;
			we <= 1'b0;
		end
		pos_max_significant_bit <= 0;
    end
	else begin
		re <= 1'b0;
		we <= 1'b0;
		
		if( output_queue_write_enable) begin
			output_queue_write_enable <= 1'b0;
		end

		if( input_queue_read_enable ) begin
			input_queue_read_enable <= 1'b0;
		end

		if( ~process_done) begin
	
				// check if we have processed all cache-lines
				if( edge_cl_cnt == um_params[159:128] ) begin
					//  Then, we must also push remaining edges that we have processed to the queue...
					if( write_data_buffer_count > 3'd0) begin
						if( ~output_queue_buf_filled) begin // but only if we can push to the output queue
							write_data_buffer_count <= 0;
							output_queue_write_enable <= 1'b1;
						end
					end
					else begin
						process_done <= 1'b1;
					end
				end
				// write back: we could not write back when we were processing the edge, so try now again. Happens only when the output queue was filled once
				else if( write_data_buffer_count >= 4) begin
						if( ~output_queue_buf_filled ) begin
							write_data_buffer_count <= 0;
							output_queue_write_enable <= 1;
						end
				end
				// grab new edges for the next cycle. Happens only if the input queue was empty once
				else if( ~read_buffer_filled & ~input_queue_buf_empty) begin 
					read_buffer <= input_queue_buf_out;
					input_queue_read_enable <= 1'b1;	
					read_buffer_filled <= 1'b1;
				end
				else if( read_buffer_filled && read_buffer[63:32] == read_buffer[31:0] ) begin // if the input queue has data, and we have "empty" or invalid edges, throw them away
							
							edge_processing_stage <= 0;
							if( edge_cnt_inc  == 3'd5) begin // if it was the last edge in the cl to read
								edge_cl_cnt <= edge_cl_cnt_inc;
								edge_cnt <= 3'b0;
								
								// get the next edge in the next cycle by telling that we have read an element of the input buffer
								if( ~input_queue_buf_empty) begin
									read_buffer <= input_queue_buf_out;
									input_queue_read_enable <= 1'b1;
									read_buffer_filled <= 1'b1;
								end
								else begin
									read_buffer_filled <= 1'b0;
									input_queue_read_enable <= 1'b0;
								end
							end
							else begin // if this was not the last edge of the read-buffer to process: get the next edges
								read_buffer <= read_buffer>>96;
								edge_cnt <= edge_cnt_inc;
							end

				end
				// if the read buffer has data, and there is empty space in the output queue, we start to process it
				else if( read_buffer_filled && ~output_queue_buf_filled ) begin 
						// load the according 32 bit value for eu
						if(edge_processing_stage == 3'd0) begin
							re <= 1'b1;
							we <= 1'b0;
							addr <= read_buffer[`ADDR_WIDTH - 1:0];	
						end
		
						// load the according 32 bit value for ev
						else if(edge_processing_stage == 3'd1) begin
							// mcm_eu <= dout;
							re <= 1'b1;
							we <= 1'b0;	
							addr <= read_buffer[`ADDR_WIDTH - 1+32:32];
						end

						else if(edge_processing_stage == 3'd2) begin
							mcm_eu <= dout;

		if(`NUM_LEVELS == 4) begin
				edge_level <= { 
				ew >= 1, ew >= 2, ew >= 3, ew >= 4
					};
		end
		if(`NUM_LEVELS == 8) begin
				edge_level <= { 
				ew >= 1, ew >= 2, ew >= 3, ew >= 4, ew >= 7, ew >= 10, ew >= 17, ew >= 27
					};
		end
		if(`NUM_LEVELS == 16) begin
				edge_level <= { 
				ew >= 1, ew >= 2, ew >= 3, ew >= 4, ew >= 7, ew >= 10, ew >= 17, ew >= 27, 
				ew >= 43, ew >= 69, ew >= 110, ew >= 176, ew >= 281, ew >= 450, ew >= 721, ew >= 1153
					};
		end
		if(`NUM_LEVELS == 32) begin
				edge_level <= { 
				ew >= 1, ew >= 2, ew >= 3, ew >= 4, ew >= 7, ew >= 10, ew >= 17, ew >= 27, 
				ew >= 43, ew >= 69, ew >= 110, ew >= 176, ew >= 281, ew >= 450, ew >= 721, ew >= 1153, 
				ew >= 1845, ew >= 2951, ew >= 4722, ew >= 7556, ew >= 12089, ew >= 19343, ew >= 30949, ew >= 49518, 
				ew >= 79228, ew >= 126765, ew >= 202824, ew >= 324519, ew >= 519230, ew >= 830767, ew >= 1329228, ew >= 2126765
					};
		end
		if(`NUM_LEVELS == 64) begin
				edge_level <= { 
				ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 2, ew >= 2, ew >= 2, 
				ew >= 2, ew >= 2, ew >= 3, ew >= 3, ew >= 3, ew >= 3, ew >= 4, ew >= 4, 
				ew >= 5, ew >= 5, ew >= 6, ew >= 6, ew >= 7, ew >= 7, ew >= 8, ew >= 9, 
				ew >= 10, ew >= 11, ew >= 12, ew >= 13, ew >= 14, ew >= 16, ew >= 17, ew >= 19, 
				ew >= 21, ew >= 23, ew >= 26, ew >= 28, ew >= 31, ew >= 34, ew >= 37, ew >= 41, 
				ew >= 45, ew >= 50, ew >= 55, ew >= 60, ew >= 66, ew >= 73, ew >= 80, ew >= 88, 
				ew >= 97, ew >= 107, ew >= 117, ew >= 129, ew >= 142, ew >= 156, ew >= 172, ew >= 189, 
				ew >= 208, ew >= 229, ew >= 252, ew >= 277, ew >= 304, ew >= 335, ew >= 368, ew >= 405
					};
		end
		if(`NUM_LEVELS == 128) begin
				edge_level <= { 
				ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 2, ew >= 2, ew >= 2, 
				ew >= 2, ew >= 2, ew >= 3, ew >= 3, ew >= 3, ew >= 3, ew >= 4, ew >= 4, 
				ew >= 5, ew >= 5, ew >= 6, ew >= 6, ew >= 7, ew >= 7, ew >= 8, ew >= 9, 
				ew >= 10, ew >= 11, ew >= 12, ew >= 13, ew >= 14, ew >= 16, ew >= 17, ew >= 19, 
				ew >= 21, ew >= 23, ew >= 26, ew >= 28, ew >= 31, ew >= 34, ew >= 37, ew >= 41, 
				ew >= 45, ew >= 50, ew >= 55, ew >= 60, ew >= 66, ew >= 73, ew >= 80, ew >= 88, 
				ew >= 97, ew >= 107, ew >= 117, ew >= 129, ew >= 142, ew >= 156, ew >= 172, ew >= 189, 
				ew >= 208, ew >= 229, ew >= 252, ew >= 277, ew >= 304, ew >= 335, ew >= 368, ew >= 405, 
				ew >= 446, ew >= 490, ew >= 539, ew >= 593, ew >= 653, ew >= 718, ew >= 790, ew >= 869, 
				ew >= 956, ew >= 1051, ew >= 1156, ew >= 1272, ew >= 1399, ew >= 1539, ew >= 1693, ew >= 1862, 
				ew >= 2048, ew >= 2253, ew >= 2479, ew >= 2726, ew >= 2999, ew >= 3299, ew >= 3629, ew >= 3992, 
				ew >= 4391, ew >= 4830, ew >= 5313, ew >= 5844, ew >= 6429, ew >= 7072, ew >= 7779, ew >= 8557, 
				ew >= 9412, ew >= 10354, ew >= 11389, ew >= 12528, ew >= 13781, ew >= 15159, ew >= 16675, ew >= 18342, 
				ew >= 20176, ew >= 22194, ew >= 24413, ew >= 26855, ew >= 29540, ew >= 32494, ew >= 35743, ew >= 39318, 
				ew >= 43249, ew >= 47574, ew >= 52332, ew >= 57565, ew >= 63322, ew >= 69654, ew >= 76619, ew >= 84281, 
				ew >= 92709, ew >= 101980, ew >= 112178, ew >= 123396, ew >= 135735, ew >= 149309, ew >= 164240, ew >= 180664
					};
		end
		if(`NUM_LEVELS == 256) begin
				edge_level <= { 
				ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, 
				ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 2, ew >= 2, 
				ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, 
				ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 3, 
				ew >= 3, ew >= 3, ew >= 3, ew >= 3, ew >= 3, ew >= 3, ew >= 3, ew >= 3, 
				ew >= 3, ew >= 3, ew >= 3, ew >= 4, ew >= 4, ew >= 4, ew >= 4, ew >= 4, 
				ew >= 4, ew >= 4, ew >= 4, ew >= 5, ew >= 5, ew >= 5, ew >= 5, ew >= 5, 
				ew >= 5, ew >= 5, ew >= 6, ew >= 6, ew >= 6, ew >= 6, ew >= 6, ew >= 6, 
				ew >= 7, ew >= 7, ew >= 7, ew >= 7, ew >= 7, ew >= 8, ew >= 8, ew >= 8, 
				ew >= 8, ew >= 9, ew >= 9, ew >= 9, ew >= 9, ew >= 10, ew >= 10, ew >= 10, 
				ew >= 11, ew >= 11, ew >= 11, ew >= 12, ew >= 12, ew >= 12, ew >= 13, ew >= 13, 
				ew >= 13, ew >= 14, ew >= 14, ew >= 15, ew >= 15, ew >= 16, ew >= 16, ew >= 17, 
				ew >= 17, ew >= 18, ew >= 18, ew >= 19, ew >= 19, ew >= 20, ew >= 20, ew >= 21, 
				ew >= 22, ew >= 22, ew >= 23, ew >= 24, ew >= 24, ew >= 25, ew >= 26, ew >= 27, 
				ew >= 27, ew >= 28, ew >= 29, ew >= 30, ew >= 31, ew >= 32, ew >= 33, ew >= 34, 
				ew >= 35, ew >= 36, ew >= 37, ew >= 38, ew >= 39, ew >= 40, ew >= 41, ew >= 43, 
				ew >= 44, ew >= 45, ew >= 47, ew >= 48, ew >= 49, ew >= 51, ew >= 53, ew >= 54, 
				ew >= 56, ew >= 57, ew >= 59, ew >= 61, ew >= 63, ew >= 65, ew >= 67, ew >= 69, 
				ew >= 71, ew >= 73, ew >= 75, ew >= 77, ew >= 79, ew >= 82, ew >= 84, ew >= 87, 
				ew >= 89, ew >= 92, ew >= 95, ew >= 98, ew >= 101, ew >= 104, ew >= 107, ew >= 110, 
				ew >= 113, ew >= 117, ew >= 120, ew >= 124, ew >= 127, ew >= 131, ew >= 135, ew >= 139, 
				ew >= 143, ew >= 148, ew >= 152, ew >= 157, ew >= 161, ew >= 166, ew >= 171, ew >= 176, 
				ew >= 182, ew >= 187, ew >= 193, ew >= 199, ew >= 205, ew >= 211, ew >= 217, ew >= 223, 
				ew >= 230, ew >= 237, ew >= 244, ew >= 252, ew >= 259, ew >= 267, ew >= 275, ew >= 283, 
				ew >= 292, ew >= 300, ew >= 309, ew >= 319, ew >= 328, ew >= 338, ew >= 348, ew >= 359, 
				ew >= 369, ew >= 380, ew >= 392, ew >= 404, ew >= 416, ew >= 428, ew >= 441, ew >= 454, 
				ew >= 468, ew >= 482, ew >= 496, ew >= 511, ew >= 527, ew >= 542, ew >= 559, ew >= 575, 
				ew >= 593, ew >= 610, ew >= 629, ew >= 648, ew >= 667, ew >= 687, ew >= 708, ew >= 729, 
				ew >= 751, ew >= 773, ew >= 797, ew >= 820, ew >= 845, ew >= 870, ew >= 897, ew >= 923, 
				ew >= 951, ew >= 980, ew >= 1009, ew >= 1039, ew >= 1070, ew >= 1103, ew >= 1136, ew >= 1170, 
				ew >= 1205, ew >= 1241, ew >= 1278, ew >= 1317, ew >= 1356, ew >= 1397, ew >= 1439, ew >= 1482, 
				ew >= 1526, ew >= 1572, ew >= 1619, ew >= 1668, ew >= 1718, ew >= 1769, ew >= 1822, ew >= 1877
					};
		end
		if(`NUM_LEVELS == 512) begin
				edge_level <= { 
				ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, 
				ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 1, ew >= 2, ew >= 2, 
				ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, 
				ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 2, ew >= 3, 
				ew >= 3, ew >= 3, ew >= 3, ew >= 3, ew >= 3, ew >= 3, ew >= 3, ew >= 3, 
				ew >= 3, ew >= 3, ew >= 3, ew >= 4, ew >= 4, ew >= 4, ew >= 4, ew >= 4, 
				ew >= 4, ew >= 4, ew >= 4, ew >= 5, ew >= 5, ew >= 5, ew >= 5, ew >= 5, 
				ew >= 5, ew >= 5, ew >= 6, ew >= 6, ew >= 6, ew >= 6, ew >= 6, ew >= 6, 
				ew >= 7, ew >= 7, ew >= 7, ew >= 7, ew >= 7, ew >= 8, ew >= 8, ew >= 8, 
				ew >= 8, ew >= 9, ew >= 9, ew >= 9, ew >= 9, ew >= 10, ew >= 10, ew >= 10, 
				ew >= 11, ew >= 11, ew >= 11, ew >= 12, ew >= 12, ew >= 12, ew >= 13, ew >= 13, 
				ew >= 13, ew >= 14, ew >= 14, ew >= 15, ew >= 15, ew >= 16, ew >= 16, ew >= 17, 
				ew >= 17, ew >= 18, ew >= 18, ew >= 19, ew >= 19, ew >= 20, ew >= 20, ew >= 21, 
				ew >= 22, ew >= 22, ew >= 23, ew >= 24, ew >= 24, ew >= 25, ew >= 26, ew >= 27, 
				ew >= 27, ew >= 28, ew >= 29, ew >= 30, ew >= 31, ew >= 32, ew >= 33, ew >= 34, 
				ew >= 35, ew >= 36, ew >= 37, ew >= 38, ew >= 39, ew >= 40, ew >= 41, ew >= 43, 
				ew >= 44, ew >= 45, ew >= 47, ew >= 48, ew >= 49, ew >= 51, ew >= 53, ew >= 54, 
				ew >= 56, ew >= 57, ew >= 59, ew >= 61, ew >= 63, ew >= 65, ew >= 67, ew >= 69, 
				ew >= 71, ew >= 73, ew >= 75, ew >= 77, ew >= 79, ew >= 82, ew >= 84, ew >= 87, 
				ew >= 89, ew >= 92, ew >= 95, ew >= 98, ew >= 101, ew >= 104, ew >= 107, ew >= 110, 
				ew >= 113, ew >= 117, ew >= 120, ew >= 124, ew >= 127, ew >= 131, ew >= 135, ew >= 139, 
				ew >= 143, ew >= 148, ew >= 152, ew >= 157, ew >= 161, ew >= 166, ew >= 171, ew >= 176, 
				ew >= 182, ew >= 187, ew >= 193, ew >= 199, ew >= 205, ew >= 211, ew >= 217, ew >= 223, 
				ew >= 230, ew >= 237, ew >= 244, ew >= 252, ew >= 259, ew >= 267, ew >= 275, ew >= 283, 
				ew >= 292, ew >= 300, ew >= 309, ew >= 319, ew >= 328, ew >= 338, ew >= 348, ew >= 359, 
				ew >= 369, ew >= 380, ew >= 392, ew >= 404, ew >= 416, ew >= 428, ew >= 441, ew >= 454, 
				ew >= 468, ew >= 482, ew >= 496, ew >= 511, ew >= 527, ew >= 542, ew >= 559, ew >= 575, 
				ew >= 593, ew >= 610, ew >= 629, ew >= 648, ew >= 667, ew >= 687, ew >= 708, ew >= 729, 
				ew >= 751, ew >= 773, ew >= 797, ew >= 820, ew >= 845, ew >= 870, ew >= 897, ew >= 923, 
				ew >= 951, ew >= 980, ew >= 1009, ew >= 1039, ew >= 1070, ew >= 1103, ew >= 1136, ew >= 1170, 
				ew >= 1205, ew >= 1241, ew >= 1278, ew >= 1317, ew >= 1356, ew >= 1397, ew >= 1439, ew >= 1482, 
				ew >= 1526, ew >= 1572, ew >= 1619, ew >= 1668, ew >= 1718, ew >= 1769, ew >= 1822, ew >= 1877, 
				ew >= 1933, ew >= 1991, ew >= 2051, ew >= 2113, ew >= 2176, ew >= 2241, ew >= 2309, ew >= 2378, 
				ew >= 2449, ew >= 2523, ew >= 2598, ew >= 2676, ew >= 2757, ew >= 2839, ew >= 2924, ew >= 3012, 
				ew >= 3103, ew >= 3196, ew >= 3292, ew >= 3390, ew >= 3492, ew >= 3597, ew >= 3705, ew >= 3816, 
				ew >= 3930, ew >= 4048, ew >= 4170, ew >= 4295, ew >= 4424, ew >= 4556, ew >= 4693, ew >= 4834, 
				ew >= 4979, ew >= 5128, ew >= 5282, ew >= 5440, ew >= 5604, ew >= 5772, ew >= 5945, ew >= 6123, 
				ew >= 6307, ew >= 6496, ew >= 6691, ew >= 6892, ew >= 7099, ew >= 7311, ew >= 7531, ew >= 7757, 
				ew >= 7989, ew >= 8229, ew >= 8476, ew >= 8730, ew >= 8992, ew >= 9262, ew >= 9540, ew >= 9826, 
				ew >= 10121, ew >= 10424, ew >= 10737, ew >= 11059, ew >= 11391, ew >= 11733, ew >= 12085, ew >= 12447, 
				ew >= 12821, ew >= 13205, ew >= 13601, ew >= 14010, ew >= 14430, ew >= 14863, ew >= 15309, ew >= 15768, 
				ew >= 16241, ew >= 16728, ew >= 17230, ew >= 17747, ew >= 18279, ew >= 18828, ew >= 19392, ew >= 19974, 
				ew >= 20573, ew >= 21191, ew >= 21826, ew >= 22481, ew >= 23156, ew >= 23850, ew >= 24566, ew >= 25303, 
				ew >= 26062, ew >= 26844, ew >= 27649, ew >= 28478, ew >= 29333, ew >= 30213, ew >= 31119, ew >= 32053, 
				ew >= 33014, ew >= 34005, ew >= 35025, ew >= 36076, ew >= 37158, ew >= 38273, ew >= 39421, ew >= 40604, 
				ew >= 41822, ew >= 43076, ew >= 44369, ew >= 45700, ew >= 47071, ew >= 48483, ew >= 49937, ew >= 51435, 
				ew >= 52978, ew >= 54568, ew >= 56205, ew >= 57891, ew >= 59628, ew >= 61416, ew >= 63259, ew >= 65157, 
				ew >= 67111, ew >= 69125, ew >= 71199, ew >= 73334, ew >= 75535, ew >= 77801, ew >= 80135, ew >= 82539, 
				ew >= 85015, ew >= 87565, ew >= 90192, ew >= 92898, ew >= 95685, ew >= 98555, ew >= 101512, ew >= 104557, 
				ew >= 107694, ew >= 110925, ew >= 114253, ew >= 117680, ew >= 121211, ew >= 124847, ew >= 128592, ew >= 132450, 
				ew >= 136424, ew >= 140516, ew >= 144732, ew >= 149074, ew >= 153546, ew >= 158152, ew >= 162897, ew >= 167784, 
				ew >= 172817, ew >= 178002, ew >= 183342, ew >= 188842, ew >= 194508, ew >= 200343, ew >= 206353, ew >= 212544, 
				ew >= 218920, ew >= 225488, ew >= 232252, ew >= 239220, ew >= 246396, ew >= 253788, ew >= 261402, ew >= 269244, 
				ew >= 277321, ew >= 285641, ew >= 294210, ew >= 303037, ew >= 312128, ew >= 321491, ew >= 331136, ew >= 341070, 
				ew >= 351302, ew >= 361841, ew >= 372697, ew >= 383878, ew >= 395394, ew >= 407256, ew >= 419473, ew >= 432058, 
				ew >= 445019, ew >= 458370, ew >= 472121, ew >= 486285, ew >= 500873, ew >= 515899, ew >= 531376, ew >= 547318, 
				ew >= 563737, ew >= 580649, ew >= 598069, ew >= 616011, ew >= 634491, ew >= 653526, ew >= 673132, ew >= 693326, 
				ew >= 714125, ew >= 735549, ew >= 757616, ew >= 780344, ew >= 803754, ew >= 827867, ew >= 852703, ew >= 878284, 
				ew >= 904633, ew >= 931772, ew >= 959725, ew >= 988517, ew >= 1018172, ew >= 1048717, ew >= 1080179, ew >= 1112584, 
				ew >= 1145962, ew >= 1180340, ew >= 1215751, ew >= 1252223, ew >= 1289790, ew >= 1328484, ew >= 1368338, ew >= 1409388, 
				ew >= 1451670, ew >= 1495220, ew >= 1540077, ew >= 1586279, ew >= 1633867, ew >= 1682883, ew >= 1733370, ew >= 1785371, 
				ew >= 1838932, ew >= 1894100, ew >= 1950923, ew >= 2009451, ew >= 2069734, ew >= 2131826, ew >= 2195781, ew >= 2261654, 
				ew >= 2329504, ew >= 2399389, ew >= 2471371, ew >= 2545512, ew >= 2621877, ew >= 2700534, ew >= 2781550, ew >= 2864996, 
				ew >= 2950946, ew >= 3039474, ew >= 3130659, ew >= 3224578, ew >= 3321316, ew >= 3420955, ew >= 3523584, ew >= 3629291
					};
		end



						end
						else if( edge_processing_stage == 3'd3) begin
							mcm_ev <= dout;
							we <= 1'b0;
							re <= 1'b0;
						end
						else if( edge_processing_stage == 3'd4) begin
							we <= 1'b1;
							re <= 1'b0;	
							addr <= read_buffer[`ADDR_WIDTH - 1:0];
							din <= mcm_eu | ( (~mcm_eu) & (~mcm_ev) & (edge_level) ) ;
							take_edge <= (~mcm_eu) & (~mcm_ev) & (edge_level);
                            if( (~mcm_eu) & (~mcm_ev) & (edge_level) ) begin
								output_queue_buf_in <= output_queue_buf_in >> 128;
							end

							oh_te <= -( (~mcm_eu) & (~mcm_ev) & (edge_level) ) & ( (~mcm_eu) & (~mcm_ev) & (edge_level) );
						end
						else if( edge_processing_stage == 3'd5) begin
							we <= 1'b1;
							re <= 1'b0;
							addr <= read_buffer[`ADDR_WIDTH - 1+32:32];
							din <= mcm_ev | take_edge;

		if(`NUM_LEVELS == 4) begin
				pos_max_significant_bit <=  
				( {32{oh_te[3]}} & 1) | ( {32{oh_te[2]}} & 2) | ( {32{oh_te[1]}} & 3) | ( {32{oh_te[0]}} & 4) 
					;
		end
		if(`NUM_LEVELS == 8) begin
				pos_max_significant_bit <=  
				( {32{oh_te[7]}} & 1) | ( {32{oh_te[6]}} & 2) | ( {32{oh_te[5]}} & 3) | ( {32{oh_te[4]}} & 4) | 
				( {32{oh_te[3]}} & 5) | ( {32{oh_te[2]}} & 6) | ( {32{oh_te[1]}} & 7) | ( {32{oh_te[0]}} & 8) 
					;
		end
		if(`NUM_LEVELS == 16) begin
				pos_max_significant_bit <=  
				( {32{oh_te[15]}} & 1) | ( {32{oh_te[14]}} & 2) | ( {32{oh_te[13]}} & 3) | ( {32{oh_te[12]}} & 4) | 
				( {32{oh_te[11]}} & 5) | ( {32{oh_te[10]}} & 6) | ( {32{oh_te[9]}} & 7) | ( {32{oh_te[8]}} & 8) | 
				( {32{oh_te[7]}} & 9) | ( {32{oh_te[6]}} & 10) | ( {32{oh_te[5]}} & 11) | ( {32{oh_te[4]}} & 12) | 
				( {32{oh_te[3]}} & 13) | ( {32{oh_te[2]}} & 14) | ( {32{oh_te[1]}} & 15) | ( {32{oh_te[0]}} & 16) 
					;
		end
		if(`NUM_LEVELS == 32) begin
				pos_max_significant_bit <=  
				( {32{oh_te[31]}} & 1) | ( {32{oh_te[30]}} & 2) | ( {32{oh_te[29]}} & 3) | ( {32{oh_te[28]}} & 4) | 
				( {32{oh_te[27]}} & 5) | ( {32{oh_te[26]}} & 6) | ( {32{oh_te[25]}} & 7) | ( {32{oh_te[24]}} & 8) | 
				( {32{oh_te[23]}} & 9) | ( {32{oh_te[22]}} & 10) | ( {32{oh_te[21]}} & 11) | ( {32{oh_te[20]}} & 12) | 
				( {32{oh_te[19]}} & 13) | ( {32{oh_te[18]}} & 14) | ( {32{oh_te[17]}} & 15) | ( {32{oh_te[16]}} & 16) | 
				( {32{oh_te[15]}} & 17) | ( {32{oh_te[14]}} & 18) | ( {32{oh_te[13]}} & 19) | ( {32{oh_te[12]}} & 20) | 
				( {32{oh_te[11]}} & 21) | ( {32{oh_te[10]}} & 22) | ( {32{oh_te[9]}} & 23) | ( {32{oh_te[8]}} & 24) | 
				( {32{oh_te[7]}} & 25) | ( {32{oh_te[6]}} & 26) | ( {32{oh_te[5]}} & 27) | ( {32{oh_te[4]}} & 28) | 
				( {32{oh_te[3]}} & 29) | ( {32{oh_te[2]}} & 30) | ( {32{oh_te[1]}} & 31) | ( {32{oh_te[0]}} & 32) 
					;
		end
		if(`NUM_LEVELS == 64) begin
				pos_max_significant_bit <=  
				( {32{oh_te[63]}} & 1) | ( {32{oh_te[62]}} & 2) | ( {32{oh_te[61]}} & 3) | ( {32{oh_te[60]}} & 4) | 
				( {32{oh_te[59]}} & 5) | ( {32{oh_te[58]}} & 6) | ( {32{oh_te[57]}} & 7) | ( {32{oh_te[56]}} & 8) | 
				( {32{oh_te[55]}} & 9) | ( {32{oh_te[54]}} & 10) | ( {32{oh_te[53]}} & 11) | ( {32{oh_te[52]}} & 12) | 
				( {32{oh_te[51]}} & 13) | ( {32{oh_te[50]}} & 14) | ( {32{oh_te[49]}} & 15) | ( {32{oh_te[48]}} & 16) | 
				( {32{oh_te[47]}} & 17) | ( {32{oh_te[46]}} & 18) | ( {32{oh_te[45]}} & 19) | ( {32{oh_te[44]}} & 20) | 
				( {32{oh_te[43]}} & 21) | ( {32{oh_te[42]}} & 22) | ( {32{oh_te[41]}} & 23) | ( {32{oh_te[40]}} & 24) | 
				( {32{oh_te[39]}} & 25) | ( {32{oh_te[38]}} & 26) | ( {32{oh_te[37]}} & 27) | ( {32{oh_te[36]}} & 28) | 
				( {32{oh_te[35]}} & 29) | ( {32{oh_te[34]}} & 30) | ( {32{oh_te[33]}} & 31) | ( {32{oh_te[32]}} & 32) | 
				( {32{oh_te[31]}} & 33) | ( {32{oh_te[30]}} & 34) | ( {32{oh_te[29]}} & 35) | ( {32{oh_te[28]}} & 36) | 
				( {32{oh_te[27]}} & 37) | ( {32{oh_te[26]}} & 38) | ( {32{oh_te[25]}} & 39) | ( {32{oh_te[24]}} & 40) | 
				( {32{oh_te[23]}} & 41) | ( {32{oh_te[22]}} & 42) | ( {32{oh_te[21]}} & 43) | ( {32{oh_te[20]}} & 44) | 
				( {32{oh_te[19]}} & 45) | ( {32{oh_te[18]}} & 46) | ( {32{oh_te[17]}} & 47) | ( {32{oh_te[16]}} & 48) | 
				( {32{oh_te[15]}} & 49) | ( {32{oh_te[14]}} & 50) | ( {32{oh_te[13]}} & 51) | ( {32{oh_te[12]}} & 52) | 
				( {32{oh_te[11]}} & 53) | ( {32{oh_te[10]}} & 54) | ( {32{oh_te[9]}} & 55) | ( {32{oh_te[8]}} & 56) | 
				( {32{oh_te[7]}} & 57) | ( {32{oh_te[6]}} & 58) | ( {32{oh_te[5]}} & 59) | ( {32{oh_te[4]}} & 60) | 
				( {32{oh_te[3]}} & 61) | ( {32{oh_te[2]}} & 62) | ( {32{oh_te[1]}} & 63) | ( {32{oh_te[0]}} & 64) 
					;
		end
		if(`NUM_LEVELS == 128) begin
				pos_max_significant_bit <=  
				( {32{oh_te[127]}} & 1) | ( {32{oh_te[126]}} & 2) | ( {32{oh_te[125]}} & 3) | ( {32{oh_te[124]}} & 4) | 
				( {32{oh_te[123]}} & 5) | ( {32{oh_te[122]}} & 6) | ( {32{oh_te[121]}} & 7) | ( {32{oh_te[120]}} & 8) | 
				( {32{oh_te[119]}} & 9) | ( {32{oh_te[118]}} & 10) | ( {32{oh_te[117]}} & 11) | ( {32{oh_te[116]}} & 12) | 
				( {32{oh_te[115]}} & 13) | ( {32{oh_te[114]}} & 14) | ( {32{oh_te[113]}} & 15) | ( {32{oh_te[112]}} & 16) | 
				( {32{oh_te[111]}} & 17) | ( {32{oh_te[110]}} & 18) | ( {32{oh_te[109]}} & 19) | ( {32{oh_te[108]}} & 20) | 
				( {32{oh_te[107]}} & 21) | ( {32{oh_te[106]}} & 22) | ( {32{oh_te[105]}} & 23) | ( {32{oh_te[104]}} & 24) | 
				( {32{oh_te[103]}} & 25) | ( {32{oh_te[102]}} & 26) | ( {32{oh_te[101]}} & 27) | ( {32{oh_te[100]}} & 28) | 
				( {32{oh_te[99]}} & 29) | ( {32{oh_te[98]}} & 30) | ( {32{oh_te[97]}} & 31) | ( {32{oh_te[96]}} & 32) | 
				( {32{oh_te[95]}} & 33) | ( {32{oh_te[94]}} & 34) | ( {32{oh_te[93]}} & 35) | ( {32{oh_te[92]}} & 36) | 
				( {32{oh_te[91]}} & 37) | ( {32{oh_te[90]}} & 38) | ( {32{oh_te[89]}} & 39) | ( {32{oh_te[88]}} & 40) | 
				( {32{oh_te[87]}} & 41) | ( {32{oh_te[86]}} & 42) | ( {32{oh_te[85]}} & 43) | ( {32{oh_te[84]}} & 44) | 
				( {32{oh_te[83]}} & 45) | ( {32{oh_te[82]}} & 46) | ( {32{oh_te[81]}} & 47) | ( {32{oh_te[80]}} & 48) | 
				( {32{oh_te[79]}} & 49) | ( {32{oh_te[78]}} & 50) | ( {32{oh_te[77]}} & 51) | ( {32{oh_te[76]}} & 52) | 
				( {32{oh_te[75]}} & 53) | ( {32{oh_te[74]}} & 54) | ( {32{oh_te[73]}} & 55) | ( {32{oh_te[72]}} & 56) | 
				( {32{oh_te[71]}} & 57) | ( {32{oh_te[70]}} & 58) | ( {32{oh_te[69]}} & 59) | ( {32{oh_te[68]}} & 60) | 
				( {32{oh_te[67]}} & 61) | ( {32{oh_te[66]}} & 62) | ( {32{oh_te[65]}} & 63) | ( {32{oh_te[64]}} & 64) | 
				( {32{oh_te[63]}} & 65) | ( {32{oh_te[62]}} & 66) | ( {32{oh_te[61]}} & 67) | ( {32{oh_te[60]}} & 68) | 
				( {32{oh_te[59]}} & 69) | ( {32{oh_te[58]}} & 70) | ( {32{oh_te[57]}} & 71) | ( {32{oh_te[56]}} & 72) | 
				( {32{oh_te[55]}} & 73) | ( {32{oh_te[54]}} & 74) | ( {32{oh_te[53]}} & 75) | ( {32{oh_te[52]}} & 76) | 
				( {32{oh_te[51]}} & 77) | ( {32{oh_te[50]}} & 78) | ( {32{oh_te[49]}} & 79) | ( {32{oh_te[48]}} & 80) | 
				( {32{oh_te[47]}} & 81) | ( {32{oh_te[46]}} & 82) | ( {32{oh_te[45]}} & 83) | ( {32{oh_te[44]}} & 84) | 
				( {32{oh_te[43]}} & 85) | ( {32{oh_te[42]}} & 86) | ( {32{oh_te[41]}} & 87) | ( {32{oh_te[40]}} & 88) | 
				( {32{oh_te[39]}} & 89) | ( {32{oh_te[38]}} & 90) | ( {32{oh_te[37]}} & 91) | ( {32{oh_te[36]}} & 92) | 
				( {32{oh_te[35]}} & 93) | ( {32{oh_te[34]}} & 94) | ( {32{oh_te[33]}} & 95) | ( {32{oh_te[32]}} & 96) | 
				( {32{oh_te[31]}} & 97) | ( {32{oh_te[30]}} & 98) | ( {32{oh_te[29]}} & 99) | ( {32{oh_te[28]}} & 100) | 
				( {32{oh_te[27]}} & 101) | ( {32{oh_te[26]}} & 102) | ( {32{oh_te[25]}} & 103) | ( {32{oh_te[24]}} & 104) | 
				( {32{oh_te[23]}} & 105) | ( {32{oh_te[22]}} & 106) | ( {32{oh_te[21]}} & 107) | ( {32{oh_te[20]}} & 108) | 
				( {32{oh_te[19]}} & 109) | ( {32{oh_te[18]}} & 110) | ( {32{oh_te[17]}} & 111) | ( {32{oh_te[16]}} & 112) | 
				( {32{oh_te[15]}} & 113) | ( {32{oh_te[14]}} & 114) | ( {32{oh_te[13]}} & 115) | ( {32{oh_te[12]}} & 116) | 
				( {32{oh_te[11]}} & 117) | ( {32{oh_te[10]}} & 118) | ( {32{oh_te[9]}} & 119) | ( {32{oh_te[8]}} & 120) | 
				( {32{oh_te[7]}} & 121) | ( {32{oh_te[6]}} & 122) | ( {32{oh_te[5]}} & 123) | ( {32{oh_te[4]}} & 124) | 
				( {32{oh_te[3]}} & 125) | ( {32{oh_te[2]}} & 126) | ( {32{oh_te[1]}} & 127) | ( {32{oh_te[0]}} & 128) 
					;
		end
		if(`NUM_LEVELS == 256) begin
				pos_max_significant_bit <=  
				( {32{oh_te[255]}} & 1) | ( {32{oh_te[254]}} & 2) | ( {32{oh_te[253]}} & 3) | ( {32{oh_te[252]}} & 4) | 
				( {32{oh_te[251]}} & 5) | ( {32{oh_te[250]}} & 6) | ( {32{oh_te[249]}} & 7) | ( {32{oh_te[248]}} & 8) | 
				( {32{oh_te[247]}} & 9) | ( {32{oh_te[246]}} & 10) | ( {32{oh_te[245]}} & 11) | ( {32{oh_te[244]}} & 12) | 
				( {32{oh_te[243]}} & 13) | ( {32{oh_te[242]}} & 14) | ( {32{oh_te[241]}} & 15) | ( {32{oh_te[240]}} & 16) | 
				( {32{oh_te[239]}} & 17) | ( {32{oh_te[238]}} & 18) | ( {32{oh_te[237]}} & 19) | ( {32{oh_te[236]}} & 20) | 
				( {32{oh_te[235]}} & 21) | ( {32{oh_te[234]}} & 22) | ( {32{oh_te[233]}} & 23) | ( {32{oh_te[232]}} & 24) | 
				( {32{oh_te[231]}} & 25) | ( {32{oh_te[230]}} & 26) | ( {32{oh_te[229]}} & 27) | ( {32{oh_te[228]}} & 28) | 
				( {32{oh_te[227]}} & 29) | ( {32{oh_te[226]}} & 30) | ( {32{oh_te[225]}} & 31) | ( {32{oh_te[224]}} & 32) | 
				( {32{oh_te[223]}} & 33) | ( {32{oh_te[222]}} & 34) | ( {32{oh_te[221]}} & 35) | ( {32{oh_te[220]}} & 36) | 
				( {32{oh_te[219]}} & 37) | ( {32{oh_te[218]}} & 38) | ( {32{oh_te[217]}} & 39) | ( {32{oh_te[216]}} & 40) | 
				( {32{oh_te[215]}} & 41) | ( {32{oh_te[214]}} & 42) | ( {32{oh_te[213]}} & 43) | ( {32{oh_te[212]}} & 44) | 
				( {32{oh_te[211]}} & 45) | ( {32{oh_te[210]}} & 46) | ( {32{oh_te[209]}} & 47) | ( {32{oh_te[208]}} & 48) | 
				( {32{oh_te[207]}} & 49) | ( {32{oh_te[206]}} & 50) | ( {32{oh_te[205]}} & 51) | ( {32{oh_te[204]}} & 52) | 
				( {32{oh_te[203]}} & 53) | ( {32{oh_te[202]}} & 54) | ( {32{oh_te[201]}} & 55) | ( {32{oh_te[200]}} & 56) | 
				( {32{oh_te[199]}} & 57) | ( {32{oh_te[198]}} & 58) | ( {32{oh_te[197]}} & 59) | ( {32{oh_te[196]}} & 60) | 
				( {32{oh_te[195]}} & 61) | ( {32{oh_te[194]}} & 62) | ( {32{oh_te[193]}} & 63) | ( {32{oh_te[192]}} & 64) | 
				( {32{oh_te[191]}} & 65) | ( {32{oh_te[190]}} & 66) | ( {32{oh_te[189]}} & 67) | ( {32{oh_te[188]}} & 68) | 
				( {32{oh_te[187]}} & 69) | ( {32{oh_te[186]}} & 70) | ( {32{oh_te[185]}} & 71) | ( {32{oh_te[184]}} & 72) | 
				( {32{oh_te[183]}} & 73) | ( {32{oh_te[182]}} & 74) | ( {32{oh_te[181]}} & 75) | ( {32{oh_te[180]}} & 76) | 
				( {32{oh_te[179]}} & 77) | ( {32{oh_te[178]}} & 78) | ( {32{oh_te[177]}} & 79) | ( {32{oh_te[176]}} & 80) | 
				( {32{oh_te[175]}} & 81) | ( {32{oh_te[174]}} & 82) | ( {32{oh_te[173]}} & 83) | ( {32{oh_te[172]}} & 84) | 
				( {32{oh_te[171]}} & 85) | ( {32{oh_te[170]}} & 86) | ( {32{oh_te[169]}} & 87) | ( {32{oh_te[168]}} & 88) | 
				( {32{oh_te[167]}} & 89) | ( {32{oh_te[166]}} & 90) | ( {32{oh_te[165]}} & 91) | ( {32{oh_te[164]}} & 92) | 
				( {32{oh_te[163]}} & 93) | ( {32{oh_te[162]}} & 94) | ( {32{oh_te[161]}} & 95) | ( {32{oh_te[160]}} & 96) | 
				( {32{oh_te[159]}} & 97) | ( {32{oh_te[158]}} & 98) | ( {32{oh_te[157]}} & 99) | ( {32{oh_te[156]}} & 100) | 
				( {32{oh_te[155]}} & 101) | ( {32{oh_te[154]}} & 102) | ( {32{oh_te[153]}} & 103) | ( {32{oh_te[152]}} & 104) | 
				( {32{oh_te[151]}} & 105) | ( {32{oh_te[150]}} & 106) | ( {32{oh_te[149]}} & 107) | ( {32{oh_te[148]}} & 108) | 
				( {32{oh_te[147]}} & 109) | ( {32{oh_te[146]}} & 110) | ( {32{oh_te[145]}} & 111) | ( {32{oh_te[144]}} & 112) | 
				( {32{oh_te[143]}} & 113) | ( {32{oh_te[142]}} & 114) | ( {32{oh_te[141]}} & 115) | ( {32{oh_te[140]}} & 116) | 
				( {32{oh_te[139]}} & 117) | ( {32{oh_te[138]}} & 118) | ( {32{oh_te[137]}} & 119) | ( {32{oh_te[136]}} & 120) | 
				( {32{oh_te[135]}} & 121) | ( {32{oh_te[134]}} & 122) | ( {32{oh_te[133]}} & 123) | ( {32{oh_te[132]}} & 124) | 
				( {32{oh_te[131]}} & 125) | ( {32{oh_te[130]}} & 126) | ( {32{oh_te[129]}} & 127) | ( {32{oh_te[128]}} & 128) | 
				( {32{oh_te[127]}} & 129) | ( {32{oh_te[126]}} & 130) | ( {32{oh_te[125]}} & 131) | ( {32{oh_te[124]}} & 132) | 
				( {32{oh_te[123]}} & 133) | ( {32{oh_te[122]}} & 134) | ( {32{oh_te[121]}} & 135) | ( {32{oh_te[120]}} & 136) | 
				( {32{oh_te[119]}} & 137) | ( {32{oh_te[118]}} & 138) | ( {32{oh_te[117]}} & 139) | ( {32{oh_te[116]}} & 140) | 
				( {32{oh_te[115]}} & 141) | ( {32{oh_te[114]}} & 142) | ( {32{oh_te[113]}} & 143) | ( {32{oh_te[112]}} & 144) | 
				( {32{oh_te[111]}} & 145) | ( {32{oh_te[110]}} & 146) | ( {32{oh_te[109]}} & 147) | ( {32{oh_te[108]}} & 148) | 
				( {32{oh_te[107]}} & 149) | ( {32{oh_te[106]}} & 150) | ( {32{oh_te[105]}} & 151) | ( {32{oh_te[104]}} & 152) | 
				( {32{oh_te[103]}} & 153) | ( {32{oh_te[102]}} & 154) | ( {32{oh_te[101]}} & 155) | ( {32{oh_te[100]}} & 156) | 
				( {32{oh_te[99]}} & 157) | ( {32{oh_te[98]}} & 158) | ( {32{oh_te[97]}} & 159) | ( {32{oh_te[96]}} & 160) | 
				( {32{oh_te[95]}} & 161) | ( {32{oh_te[94]}} & 162) | ( {32{oh_te[93]}} & 163) | ( {32{oh_te[92]}} & 164) | 
				( {32{oh_te[91]}} & 165) | ( {32{oh_te[90]}} & 166) | ( {32{oh_te[89]}} & 167) | ( {32{oh_te[88]}} & 168) | 
				( {32{oh_te[87]}} & 169) | ( {32{oh_te[86]}} & 170) | ( {32{oh_te[85]}} & 171) | ( {32{oh_te[84]}} & 172) | 
				( {32{oh_te[83]}} & 173) | ( {32{oh_te[82]}} & 174) | ( {32{oh_te[81]}} & 175) | ( {32{oh_te[80]}} & 176) | 
				( {32{oh_te[79]}} & 177) | ( {32{oh_te[78]}} & 178) | ( {32{oh_te[77]}} & 179) | ( {32{oh_te[76]}} & 180) | 
				( {32{oh_te[75]}} & 181) | ( {32{oh_te[74]}} & 182) | ( {32{oh_te[73]}} & 183) | ( {32{oh_te[72]}} & 184) | 
				( {32{oh_te[71]}} & 185) | ( {32{oh_te[70]}} & 186) | ( {32{oh_te[69]}} & 187) | ( {32{oh_te[68]}} & 188) | 
				( {32{oh_te[67]}} & 189) | ( {32{oh_te[66]}} & 190) | ( {32{oh_te[65]}} & 191) | ( {32{oh_te[64]}} & 192) | 
				( {32{oh_te[63]}} & 193) | ( {32{oh_te[62]}} & 194) | ( {32{oh_te[61]}} & 195) | ( {32{oh_te[60]}} & 196) | 
				( {32{oh_te[59]}} & 197) | ( {32{oh_te[58]}} & 198) | ( {32{oh_te[57]}} & 199) | ( {32{oh_te[56]}} & 200) | 
				( {32{oh_te[55]}} & 201) | ( {32{oh_te[54]}} & 202) | ( {32{oh_te[53]}} & 203) | ( {32{oh_te[52]}} & 204) | 
				( {32{oh_te[51]}} & 205) | ( {32{oh_te[50]}} & 206) | ( {32{oh_te[49]}} & 207) | ( {32{oh_te[48]}} & 208) | 
				( {32{oh_te[47]}} & 209) | ( {32{oh_te[46]}} & 210) | ( {32{oh_te[45]}} & 211) | ( {32{oh_te[44]}} & 212) | 
				( {32{oh_te[43]}} & 213) | ( {32{oh_te[42]}} & 214) | ( {32{oh_te[41]}} & 215) | ( {32{oh_te[40]}} & 216) | 
				( {32{oh_te[39]}} & 217) | ( {32{oh_te[38]}} & 218) | ( {32{oh_te[37]}} & 219) | ( {32{oh_te[36]}} & 220) | 
				( {32{oh_te[35]}} & 221) | ( {32{oh_te[34]}} & 222) | ( {32{oh_te[33]}} & 223) | ( {32{oh_te[32]}} & 224) | 
				( {32{oh_te[31]}} & 225) | ( {32{oh_te[30]}} & 226) | ( {32{oh_te[29]}} & 227) | ( {32{oh_te[28]}} & 228) | 
				( {32{oh_te[27]}} & 229) | ( {32{oh_te[26]}} & 230) | ( {32{oh_te[25]}} & 231) | ( {32{oh_te[24]}} & 232) | 
				( {32{oh_te[23]}} & 233) | ( {32{oh_te[22]}} & 234) | ( {32{oh_te[21]}} & 235) | ( {32{oh_te[20]}} & 236) | 
				( {32{oh_te[19]}} & 237) | ( {32{oh_te[18]}} & 238) | ( {32{oh_te[17]}} & 239) | ( {32{oh_te[16]}} & 240) | 
				( {32{oh_te[15]}} & 241) | ( {32{oh_te[14]}} & 242) | ( {32{oh_te[13]}} & 243) | ( {32{oh_te[12]}} & 244) | 
				( {32{oh_te[11]}} & 245) | ( {32{oh_te[10]}} & 246) | ( {32{oh_te[9]}} & 247) | ( {32{oh_te[8]}} & 248) | 
				( {32{oh_te[7]}} & 249) | ( {32{oh_te[6]}} & 250) | ( {32{oh_te[5]}} & 251) | ( {32{oh_te[4]}} & 252) | 
				( {32{oh_te[3]}} & 253) | ( {32{oh_te[2]}} & 254) | ( {32{oh_te[1]}} & 255) | ( {32{oh_te[0]}} & 256) 
					;
		end
		if(`NUM_LEVELS == 512) begin
				pos_max_significant_bit <=  
				( {32{oh_te[511]}} & 1) | ( {32{oh_te[510]}} & 2) | ( {32{oh_te[509]}} & 3) | ( {32{oh_te[508]}} & 4) | 
				( {32{oh_te[507]}} & 5) | ( {32{oh_te[506]}} & 6) | ( {32{oh_te[505]}} & 7) | ( {32{oh_te[504]}} & 8) | 
				( {32{oh_te[503]}} & 9) | ( {32{oh_te[502]}} & 10) | ( {32{oh_te[501]}} & 11) | ( {32{oh_te[500]}} & 12) | 
				( {32{oh_te[499]}} & 13) | ( {32{oh_te[498]}} & 14) | ( {32{oh_te[497]}} & 15) | ( {32{oh_te[496]}} & 16) | 
				( {32{oh_te[495]}} & 17) | ( {32{oh_te[494]}} & 18) | ( {32{oh_te[493]}} & 19) | ( {32{oh_te[492]}} & 20) | 
				( {32{oh_te[491]}} & 21) | ( {32{oh_te[490]}} & 22) | ( {32{oh_te[489]}} & 23) | ( {32{oh_te[488]}} & 24) | 
				( {32{oh_te[487]}} & 25) | ( {32{oh_te[486]}} & 26) | ( {32{oh_te[485]}} & 27) | ( {32{oh_te[484]}} & 28) | 
				( {32{oh_te[483]}} & 29) | ( {32{oh_te[482]}} & 30) | ( {32{oh_te[481]}} & 31) | ( {32{oh_te[480]}} & 32) | 
				( {32{oh_te[479]}} & 33) | ( {32{oh_te[478]}} & 34) | ( {32{oh_te[477]}} & 35) | ( {32{oh_te[476]}} & 36) | 
				( {32{oh_te[475]}} & 37) | ( {32{oh_te[474]}} & 38) | ( {32{oh_te[473]}} & 39) | ( {32{oh_te[472]}} & 40) | 
				( {32{oh_te[471]}} & 41) | ( {32{oh_te[470]}} & 42) | ( {32{oh_te[469]}} & 43) | ( {32{oh_te[468]}} & 44) | 
				( {32{oh_te[467]}} & 45) | ( {32{oh_te[466]}} & 46) | ( {32{oh_te[465]}} & 47) | ( {32{oh_te[464]}} & 48) | 
				( {32{oh_te[463]}} & 49) | ( {32{oh_te[462]}} & 50) | ( {32{oh_te[461]}} & 51) | ( {32{oh_te[460]}} & 52) | 
				( {32{oh_te[459]}} & 53) | ( {32{oh_te[458]}} & 54) | ( {32{oh_te[457]}} & 55) | ( {32{oh_te[456]}} & 56) | 
				( {32{oh_te[455]}} & 57) | ( {32{oh_te[454]}} & 58) | ( {32{oh_te[453]}} & 59) | ( {32{oh_te[452]}} & 60) | 
				( {32{oh_te[451]}} & 61) | ( {32{oh_te[450]}} & 62) | ( {32{oh_te[449]}} & 63) | ( {32{oh_te[448]}} & 64) | 
				( {32{oh_te[447]}} & 65) | ( {32{oh_te[446]}} & 66) | ( {32{oh_te[445]}} & 67) | ( {32{oh_te[444]}} & 68) | 
				( {32{oh_te[443]}} & 69) | ( {32{oh_te[442]}} & 70) | ( {32{oh_te[441]}} & 71) | ( {32{oh_te[440]}} & 72) | 
				( {32{oh_te[439]}} & 73) | ( {32{oh_te[438]}} & 74) | ( {32{oh_te[437]}} & 75) | ( {32{oh_te[436]}} & 76) | 
				( {32{oh_te[435]}} & 77) | ( {32{oh_te[434]}} & 78) | ( {32{oh_te[433]}} & 79) | ( {32{oh_te[432]}} & 80) | 
				( {32{oh_te[431]}} & 81) | ( {32{oh_te[430]}} & 82) | ( {32{oh_te[429]}} & 83) | ( {32{oh_te[428]}} & 84) | 
				( {32{oh_te[427]}} & 85) | ( {32{oh_te[426]}} & 86) | ( {32{oh_te[425]}} & 87) | ( {32{oh_te[424]}} & 88) | 
				( {32{oh_te[423]}} & 89) | ( {32{oh_te[422]}} & 90) | ( {32{oh_te[421]}} & 91) | ( {32{oh_te[420]}} & 92) | 
				( {32{oh_te[419]}} & 93) | ( {32{oh_te[418]}} & 94) | ( {32{oh_te[417]}} & 95) | ( {32{oh_te[416]}} & 96) | 
				( {32{oh_te[415]}} & 97) | ( {32{oh_te[414]}} & 98) | ( {32{oh_te[413]}} & 99) | ( {32{oh_te[412]}} & 100) | 
				( {32{oh_te[411]}} & 101) | ( {32{oh_te[410]}} & 102) | ( {32{oh_te[409]}} & 103) | ( {32{oh_te[408]}} & 104) | 
				( {32{oh_te[407]}} & 105) | ( {32{oh_te[406]}} & 106) | ( {32{oh_te[405]}} & 107) | ( {32{oh_te[404]}} & 108) | 
				( {32{oh_te[403]}} & 109) | ( {32{oh_te[402]}} & 110) | ( {32{oh_te[401]}} & 111) | ( {32{oh_te[400]}} & 112) | 
				( {32{oh_te[399]}} & 113) | ( {32{oh_te[398]}} & 114) | ( {32{oh_te[397]}} & 115) | ( {32{oh_te[396]}} & 116) | 
				( {32{oh_te[395]}} & 117) | ( {32{oh_te[394]}} & 118) | ( {32{oh_te[393]}} & 119) | ( {32{oh_te[392]}} & 120) | 
				( {32{oh_te[391]}} & 121) | ( {32{oh_te[390]}} & 122) | ( {32{oh_te[389]}} & 123) | ( {32{oh_te[388]}} & 124) | 
				( {32{oh_te[387]}} & 125) | ( {32{oh_te[386]}} & 126) | ( {32{oh_te[385]}} & 127) | ( {32{oh_te[384]}} & 128) | 
				( {32{oh_te[383]}} & 129) | ( {32{oh_te[382]}} & 130) | ( {32{oh_te[381]}} & 131) | ( {32{oh_te[380]}} & 132) | 
				( {32{oh_te[379]}} & 133) | ( {32{oh_te[378]}} & 134) | ( {32{oh_te[377]}} & 135) | ( {32{oh_te[376]}} & 136) | 
				( {32{oh_te[375]}} & 137) | ( {32{oh_te[374]}} & 138) | ( {32{oh_te[373]}} & 139) | ( {32{oh_te[372]}} & 140) | 
				( {32{oh_te[371]}} & 141) | ( {32{oh_te[370]}} & 142) | ( {32{oh_te[369]}} & 143) | ( {32{oh_te[368]}} & 144) | 
				( {32{oh_te[367]}} & 145) | ( {32{oh_te[366]}} & 146) | ( {32{oh_te[365]}} & 147) | ( {32{oh_te[364]}} & 148) | 
				( {32{oh_te[363]}} & 149) | ( {32{oh_te[362]}} & 150) | ( {32{oh_te[361]}} & 151) | ( {32{oh_te[360]}} & 152) | 
				( {32{oh_te[359]}} & 153) | ( {32{oh_te[358]}} & 154) | ( {32{oh_te[357]}} & 155) | ( {32{oh_te[356]}} & 156) | 
				( {32{oh_te[355]}} & 157) | ( {32{oh_te[354]}} & 158) | ( {32{oh_te[353]}} & 159) | ( {32{oh_te[352]}} & 160) | 
				( {32{oh_te[351]}} & 161) | ( {32{oh_te[350]}} & 162) | ( {32{oh_te[349]}} & 163) | ( {32{oh_te[348]}} & 164) | 
				( {32{oh_te[347]}} & 165) | ( {32{oh_te[346]}} & 166) | ( {32{oh_te[345]}} & 167) | ( {32{oh_te[344]}} & 168) | 
				( {32{oh_te[343]}} & 169) | ( {32{oh_te[342]}} & 170) | ( {32{oh_te[341]}} & 171) | ( {32{oh_te[340]}} & 172) | 
				( {32{oh_te[339]}} & 173) | ( {32{oh_te[338]}} & 174) | ( {32{oh_te[337]}} & 175) | ( {32{oh_te[336]}} & 176) | 
				( {32{oh_te[335]}} & 177) | ( {32{oh_te[334]}} & 178) | ( {32{oh_te[333]}} & 179) | ( {32{oh_te[332]}} & 180) | 
				( {32{oh_te[331]}} & 181) | ( {32{oh_te[330]}} & 182) | ( {32{oh_te[329]}} & 183) | ( {32{oh_te[328]}} & 184) | 
				( {32{oh_te[327]}} & 185) | ( {32{oh_te[326]}} & 186) | ( {32{oh_te[325]}} & 187) | ( {32{oh_te[324]}} & 188) | 
				( {32{oh_te[323]}} & 189) | ( {32{oh_te[322]}} & 190) | ( {32{oh_te[321]}} & 191) | ( {32{oh_te[320]}} & 192) | 
				( {32{oh_te[319]}} & 193) | ( {32{oh_te[318]}} & 194) | ( {32{oh_te[317]}} & 195) | ( {32{oh_te[316]}} & 196) | 
				( {32{oh_te[315]}} & 197) | ( {32{oh_te[314]}} & 198) | ( {32{oh_te[313]}} & 199) | ( {32{oh_te[312]}} & 200) | 
				( {32{oh_te[311]}} & 201) | ( {32{oh_te[310]}} & 202) | ( {32{oh_te[309]}} & 203) | ( {32{oh_te[308]}} & 204) | 
				( {32{oh_te[307]}} & 205) | ( {32{oh_te[306]}} & 206) | ( {32{oh_te[305]}} & 207) | ( {32{oh_te[304]}} & 208) | 
				( {32{oh_te[303]}} & 209) | ( {32{oh_te[302]}} & 210) | ( {32{oh_te[301]}} & 211) | ( {32{oh_te[300]}} & 212) | 
				( {32{oh_te[299]}} & 213) | ( {32{oh_te[298]}} & 214) | ( {32{oh_te[297]}} & 215) | ( {32{oh_te[296]}} & 216) | 
				( {32{oh_te[295]}} & 217) | ( {32{oh_te[294]}} & 218) | ( {32{oh_te[293]}} & 219) | ( {32{oh_te[292]}} & 220) | 
				( {32{oh_te[291]}} & 221) | ( {32{oh_te[290]}} & 222) | ( {32{oh_te[289]}} & 223) | ( {32{oh_te[288]}} & 224) | 
				( {32{oh_te[287]}} & 225) | ( {32{oh_te[286]}} & 226) | ( {32{oh_te[285]}} & 227) | ( {32{oh_te[284]}} & 228) | 
				( {32{oh_te[283]}} & 229) | ( {32{oh_te[282]}} & 230) | ( {32{oh_te[281]}} & 231) | ( {32{oh_te[280]}} & 232) | 
				( {32{oh_te[279]}} & 233) | ( {32{oh_te[278]}} & 234) | ( {32{oh_te[277]}} & 235) | ( {32{oh_te[276]}} & 236) | 
				( {32{oh_te[275]}} & 237) | ( {32{oh_te[274]}} & 238) | ( {32{oh_te[273]}} & 239) | ( {32{oh_te[272]}} & 240) | 
				( {32{oh_te[271]}} & 241) | ( {32{oh_te[270]}} & 242) | ( {32{oh_te[269]}} & 243) | ( {32{oh_te[268]}} & 244) | 
				( {32{oh_te[267]}} & 245) | ( {32{oh_te[266]}} & 246) | ( {32{oh_te[265]}} & 247) | ( {32{oh_te[264]}} & 248) | 
				( {32{oh_te[263]}} & 249) | ( {32{oh_te[262]}} & 250) | ( {32{oh_te[261]}} & 251) | ( {32{oh_te[260]}} & 252) | 
				( {32{oh_te[259]}} & 253) | ( {32{oh_te[258]}} & 254) | ( {32{oh_te[257]}} & 255) | ( {32{oh_te[256]}} & 256) | 
				( {32{oh_te[255]}} & 257) | ( {32{oh_te[254]}} & 258) | ( {32{oh_te[253]}} & 259) | ( {32{oh_te[252]}} & 260) | 
				( {32{oh_te[251]}} & 261) | ( {32{oh_te[250]}} & 262) | ( {32{oh_te[249]}} & 263) | ( {32{oh_te[248]}} & 264) | 
				( {32{oh_te[247]}} & 265) | ( {32{oh_te[246]}} & 266) | ( {32{oh_te[245]}} & 267) | ( {32{oh_te[244]}} & 268) | 
				( {32{oh_te[243]}} & 269) | ( {32{oh_te[242]}} & 270) | ( {32{oh_te[241]}} & 271) | ( {32{oh_te[240]}} & 272) | 
				( {32{oh_te[239]}} & 273) | ( {32{oh_te[238]}} & 274) | ( {32{oh_te[237]}} & 275) | ( {32{oh_te[236]}} & 276) | 
				( {32{oh_te[235]}} & 277) | ( {32{oh_te[234]}} & 278) | ( {32{oh_te[233]}} & 279) | ( {32{oh_te[232]}} & 280) | 
				( {32{oh_te[231]}} & 281) | ( {32{oh_te[230]}} & 282) | ( {32{oh_te[229]}} & 283) | ( {32{oh_te[228]}} & 284) | 
				( {32{oh_te[227]}} & 285) | ( {32{oh_te[226]}} & 286) | ( {32{oh_te[225]}} & 287) | ( {32{oh_te[224]}} & 288) | 
				( {32{oh_te[223]}} & 289) | ( {32{oh_te[222]}} & 290) | ( {32{oh_te[221]}} & 291) | ( {32{oh_te[220]}} & 292) | 
				( {32{oh_te[219]}} & 293) | ( {32{oh_te[218]}} & 294) | ( {32{oh_te[217]}} & 295) | ( {32{oh_te[216]}} & 296) | 
				( {32{oh_te[215]}} & 297) | ( {32{oh_te[214]}} & 298) | ( {32{oh_te[213]}} & 299) | ( {32{oh_te[212]}} & 300) | 
				( {32{oh_te[211]}} & 301) | ( {32{oh_te[210]}} & 302) | ( {32{oh_te[209]}} & 303) | ( {32{oh_te[208]}} & 304) | 
				( {32{oh_te[207]}} & 305) | ( {32{oh_te[206]}} & 306) | ( {32{oh_te[205]}} & 307) | ( {32{oh_te[204]}} & 308) | 
				( {32{oh_te[203]}} & 309) | ( {32{oh_te[202]}} & 310) | ( {32{oh_te[201]}} & 311) | ( {32{oh_te[200]}} & 312) | 
				( {32{oh_te[199]}} & 313) | ( {32{oh_te[198]}} & 314) | ( {32{oh_te[197]}} & 315) | ( {32{oh_te[196]}} & 316) | 
				( {32{oh_te[195]}} & 317) | ( {32{oh_te[194]}} & 318) | ( {32{oh_te[193]}} & 319) | ( {32{oh_te[192]}} & 320) | 
				( {32{oh_te[191]}} & 321) | ( {32{oh_te[190]}} & 322) | ( {32{oh_te[189]}} & 323) | ( {32{oh_te[188]}} & 324) | 
				( {32{oh_te[187]}} & 325) | ( {32{oh_te[186]}} & 326) | ( {32{oh_te[185]}} & 327) | ( {32{oh_te[184]}} & 328) | 
				( {32{oh_te[183]}} & 329) | ( {32{oh_te[182]}} & 330) | ( {32{oh_te[181]}} & 331) | ( {32{oh_te[180]}} & 332) | 
				( {32{oh_te[179]}} & 333) | ( {32{oh_te[178]}} & 334) | ( {32{oh_te[177]}} & 335) | ( {32{oh_te[176]}} & 336) | 
				( {32{oh_te[175]}} & 337) | ( {32{oh_te[174]}} & 338) | ( {32{oh_te[173]}} & 339) | ( {32{oh_te[172]}} & 340) | 
				( {32{oh_te[171]}} & 341) | ( {32{oh_te[170]}} & 342) | ( {32{oh_te[169]}} & 343) | ( {32{oh_te[168]}} & 344) | 
				( {32{oh_te[167]}} & 345) | ( {32{oh_te[166]}} & 346) | ( {32{oh_te[165]}} & 347) | ( {32{oh_te[164]}} & 348) | 
				( {32{oh_te[163]}} & 349) | ( {32{oh_te[162]}} & 350) | ( {32{oh_te[161]}} & 351) | ( {32{oh_te[160]}} & 352) | 
				( {32{oh_te[159]}} & 353) | ( {32{oh_te[158]}} & 354) | ( {32{oh_te[157]}} & 355) | ( {32{oh_te[156]}} & 356) | 
				( {32{oh_te[155]}} & 357) | ( {32{oh_te[154]}} & 358) | ( {32{oh_te[153]}} & 359) | ( {32{oh_te[152]}} & 360) | 
				( {32{oh_te[151]}} & 361) | ( {32{oh_te[150]}} & 362) | ( {32{oh_te[149]}} & 363) | ( {32{oh_te[148]}} & 364) | 
				( {32{oh_te[147]}} & 365) | ( {32{oh_te[146]}} & 366) | ( {32{oh_te[145]}} & 367) | ( {32{oh_te[144]}} & 368) | 
				( {32{oh_te[143]}} & 369) | ( {32{oh_te[142]}} & 370) | ( {32{oh_te[141]}} & 371) | ( {32{oh_te[140]}} & 372) | 
				( {32{oh_te[139]}} & 373) | ( {32{oh_te[138]}} & 374) | ( {32{oh_te[137]}} & 375) | ( {32{oh_te[136]}} & 376) | 
				( {32{oh_te[135]}} & 377) | ( {32{oh_te[134]}} & 378) | ( {32{oh_te[133]}} & 379) | ( {32{oh_te[132]}} & 380) | 
				( {32{oh_te[131]}} & 381) | ( {32{oh_te[130]}} & 382) | ( {32{oh_te[129]}} & 383) | ( {32{oh_te[128]}} & 384) | 
				( {32{oh_te[127]}} & 385) | ( {32{oh_te[126]}} & 386) | ( {32{oh_te[125]}} & 387) | ( {32{oh_te[124]}} & 388) | 
				( {32{oh_te[123]}} & 389) | ( {32{oh_te[122]}} & 390) | ( {32{oh_te[121]}} & 391) | ( {32{oh_te[120]}} & 392) | 
				( {32{oh_te[119]}} & 393) | ( {32{oh_te[118]}} & 394) | ( {32{oh_te[117]}} & 395) | ( {32{oh_te[116]}} & 396) | 
				( {32{oh_te[115]}} & 397) | ( {32{oh_te[114]}} & 398) | ( {32{oh_te[113]}} & 399) | ( {32{oh_te[112]}} & 400) | 
				( {32{oh_te[111]}} & 401) | ( {32{oh_te[110]}} & 402) | ( {32{oh_te[109]}} & 403) | ( {32{oh_te[108]}} & 404) | 
				( {32{oh_te[107]}} & 405) | ( {32{oh_te[106]}} & 406) | ( {32{oh_te[105]}} & 407) | ( {32{oh_te[104]}} & 408) | 
				( {32{oh_te[103]}} & 409) | ( {32{oh_te[102]}} & 410) | ( {32{oh_te[101]}} & 411) | ( {32{oh_te[100]}} & 412) | 
				( {32{oh_te[99]}} & 413) | ( {32{oh_te[98]}} & 414) | ( {32{oh_te[97]}} & 415) | ( {32{oh_te[96]}} & 416) | 
				( {32{oh_te[95]}} & 417) | ( {32{oh_te[94]}} & 418) | ( {32{oh_te[93]}} & 419) | ( {32{oh_te[92]}} & 420) | 
				( {32{oh_te[91]}} & 421) | ( {32{oh_te[90]}} & 422) | ( {32{oh_te[89]}} & 423) | ( {32{oh_te[88]}} & 424) | 
				( {32{oh_te[87]}} & 425) | ( {32{oh_te[86]}} & 426) | ( {32{oh_te[85]}} & 427) | ( {32{oh_te[84]}} & 428) | 
				( {32{oh_te[83]}} & 429) | ( {32{oh_te[82]}} & 430) | ( {32{oh_te[81]}} & 431) | ( {32{oh_te[80]}} & 432) | 
				( {32{oh_te[79]}} & 433) | ( {32{oh_te[78]}} & 434) | ( {32{oh_te[77]}} & 435) | ( {32{oh_te[76]}} & 436) | 
				( {32{oh_te[75]}} & 437) | ( {32{oh_te[74]}} & 438) | ( {32{oh_te[73]}} & 439) | ( {32{oh_te[72]}} & 440) | 
				( {32{oh_te[71]}} & 441) | ( {32{oh_te[70]}} & 442) | ( {32{oh_te[69]}} & 443) | ( {32{oh_te[68]}} & 444) | 
				( {32{oh_te[67]}} & 445) | ( {32{oh_te[66]}} & 446) | ( {32{oh_te[65]}} & 447) | ( {32{oh_te[64]}} & 448) | 
				( {32{oh_te[63]}} & 449) | ( {32{oh_te[62]}} & 450) | ( {32{oh_te[61]}} & 451) | ( {32{oh_te[60]}} & 452) | 
				( {32{oh_te[59]}} & 453) | ( {32{oh_te[58]}} & 454) | ( {32{oh_te[57]}} & 455) | ( {32{oh_te[56]}} & 456) | 
				( {32{oh_te[55]}} & 457) | ( {32{oh_te[54]}} & 458) | ( {32{oh_te[53]}} & 459) | ( {32{oh_te[52]}} & 460) | 
				( {32{oh_te[51]}} & 461) | ( {32{oh_te[50]}} & 462) | ( {32{oh_te[49]}} & 463) | ( {32{oh_te[48]}} & 464) | 
				( {32{oh_te[47]}} & 465) | ( {32{oh_te[46]}} & 466) | ( {32{oh_te[45]}} & 467) | ( {32{oh_te[44]}} & 468) | 
				( {32{oh_te[43]}} & 469) | ( {32{oh_te[42]}} & 470) | ( {32{oh_te[41]}} & 471) | ( {32{oh_te[40]}} & 472) | 
				( {32{oh_te[39]}} & 473) | ( {32{oh_te[38]}} & 474) | ( {32{oh_te[37]}} & 475) | ( {32{oh_te[36]}} & 476) | 
				( {32{oh_te[35]}} & 477) | ( {32{oh_te[34]}} & 478) | ( {32{oh_te[33]}} & 479) | ( {32{oh_te[32]}} & 480) | 
				( {32{oh_te[31]}} & 481) | ( {32{oh_te[30]}} & 482) | ( {32{oh_te[29]}} & 483) | ( {32{oh_te[28]}} & 484) | 
				( {32{oh_te[27]}} & 485) | ( {32{oh_te[26]}} & 486) | ( {32{oh_te[25]}} & 487) | ( {32{oh_te[24]}} & 488) | 
				( {32{oh_te[23]}} & 489) | ( {32{oh_te[22]}} & 490) | ( {32{oh_te[21]}} & 491) | ( {32{oh_te[20]}} & 492) | 
				( {32{oh_te[19]}} & 493) | ( {32{oh_te[18]}} & 494) | ( {32{oh_te[17]}} & 495) | ( {32{oh_te[16]}} & 496) | 
				( {32{oh_te[15]}} & 497) | ( {32{oh_te[14]}} & 498) | ( {32{oh_te[13]}} & 499) | ( {32{oh_te[12]}} & 500) | 
				( {32{oh_te[11]}} & 501) | ( {32{oh_te[10]}} & 502) | ( {32{oh_te[9]}} & 503) | ( {32{oh_te[8]}} & 504) | 
				( {32{oh_te[7]}} & 505) | ( {32{oh_te[6]}} & 506) | ( {32{oh_te[5]}} & 507) | ( {32{oh_te[4]}} & 508) | 
				( {32{oh_te[3]}} & 509) | ( {32{oh_te[2]}} & 510) | ( {32{oh_te[1]}} & 511) | ( {32{oh_te[0]}} & 512) 
					;
		end

						end
						// write back
						else if( edge_processing_stage == 3'd6) begin
							

							if( take_edge ) begin // we have data to write back
								output_queue_buf_in[511:384] <= { pos_max_significant_bit, read_buffer[95:64], read_buffer[63:32], read_buffer[31:0]};

								if(write_data_buffer_count_inc == 3'd4 && ~output_queue_buf_filled ) begin // if we wrote the last edge into the buffer, put it on output queue if there is space
									write_data_buffer_count <= 0;
									output_queue_write_enable <= 1;
								end
								else begin
									write_data_buffer_count <= write_data_buffer_count_inc;	
									output_queue_write_enable <= 0;
								end
							end

							edge_processing_stage <= 3'd0; // reset the edge processing stage
							if( edge_cnt_inc  == 3'd5) begin // if it was the last edge in the cl to read
								edge_cl_cnt <= edge_cl_cnt_inc;
								edge_cnt <= 0;
								
								// get the next edge in the next cycle by telling that we have read an element of the input buffer
								if( ~input_queue_buf_empty) begin
									read_buffer <= input_queue_buf_out;
									input_queue_read_enable <= 1'b1;
									read_buffer_filled <= 1'b1;
								end
								else begin
									input_queue_read_enable <= 1'b0;
									read_buffer_filled <= 1'b0;
								end
							end
							else begin // if this was not the last edge of the read-buffer to process: get the next edges
								read_buffer <= read_buffer>>96;
								edge_cnt <= edge_cnt_inc;
							end
						end // end of if( edge_processing_stage == 4) begin


						if(edge_processing_stage < 3'd6) begin // if not in the last stage, we can simply increment the stage
							edge_processing_stage <= edge_processing_stage_inc;
						end

					end // end of if( ~input_queue_buf_empty && ~output_queue_buf_filled ) begin
					
			
		end // end of if( ~process_done)
		
	end // end of else ( ~rst_n | um_done | ~um_activated )


end // end of always

endgenerate

assign edge_cl_cnt_inc = edge_cl_cnt + 1'b1;
assign edge_processing_stage_inc = edge_processing_stage + 1'b1;
assign edge_cnt_inc = edge_cnt + 1'b1;
assign write_data_buffer_count_inc = write_data_buffer_count + 1'b1; // increment the number of edges in the write data buffer


assign ew = read_buffer[95:64];
/*
// vars for debugging
reg[31:0]                   n_debug_part;
*/

/////////////////////// Generate write requests ///////////////////////////////

always@(posedge clk) begin
    // Signals need reset
    if(~rst_n | um_done | ~um_activated) begin
        um_tx_wr_valid    <= 1'b0;
        wr_done <= 1'b0;
        wr_cnt <= 32'b0;
		um_tx_wr_addr <= 58'b0;
		um_tx_data <= 512'b0;
		um_tx_wr_tag <= 8'b0;
		output_queue_read_enable <= 1'b0;
        // n_debug_part <= 0;
    end 
    else begin	
	
		if( output_queue_read_enable) begin
			output_queue_read_enable <= 0;
		end

		if(~wr_done) begin
			um_tx_wr_valid <= 1'b0;
			
			// if processing is done, the output queue empty, and nothing is written at the moment, we have nothing to write. So terminate
			if( process_done && output_queue_buf_empty && ~output_queue_write_enable) begin
				 wr_done <= 1;
                /*
				if( n_debug_part >= 1) begin
				    wr_done <= 1;
                end
                else begin
                    if( n_debug_part == 0 && um_tx_wr_ready) begin
                        um_tx_wr_valid <= 1'b1;
                        um_tx_data<= { {32'd123321}, {16'b0}, {4{rd_issued_done}}, {4{rd_done}}, {4{process_done}}, {4{wr_done}},      // 64 bits
                                        rd_issued_cnt, rd_cnt, edge_cl_cnt, wr_cnt,                                  // 128                   
                                        input_queue_ntotal_read, input_queue_ntotal_write, output_queue_ntotal_read, output_queue_ntotal_write,  // 128 bits
                                        {160'b0}, {32'd123321} };
                        n_debug_part <= n_debug_part+1;
						um_tx_wr_addr     <= um_params_reg[127:70] + wr_cnt;
						wr_cnt            <= wr_cnt_inc;
                    end
                end
				*/
			end
			else begin
				if( um_tx_wr_ready ) begin // we are allowed to write

					//we can write if: we did not write before, and buf is not empty, or we did write, and there is still a next package
					if( ( ~output_queue_buf_empty & ~um_tx_wr_valid ) | ( um_tx_wr_valid & ~output_queue_buf_empty_next) )begin 
						um_tx_data        <= output_queue_buf_out; // the data to write
						um_tx_wr_addr     <= um_params_reg[127:70] + wr_cnt;	// write to the next position
						um_tx_wr_tag      <= um_rx_rd_tag; // copy the tag
						um_tx_wr_valid    <= 1'b1;
						wr_cnt            <= wr_cnt_inc;
						output_queue_read_enable <= 1;
					end
				end
		    end
		end
    end
end 

assign wr_cnt_inc = wr_cnt + 32'd1;


/////////////////////// Handle write responses ///////////////////////////////
// nothing to do so far - we could check if the writes to memory have succeeded or not, and wait for the final write before we terminate

always@(posedge clk) begin
	if(~rst_n | um_done | ~um_activated) begin
		
	end
	else begin

	end

end



endmodule



`default_nettype wire



