
/**
 * Fifo queue, based on a ring buffer (see also http://electrosofts.com/verilog/fifo.html )
 *
 */




module queue #( parameter BUF_NBITS = 512, parameter BUF_NELEMS = 7)
( 
		input wire clk,							// clock
		input wire rst,							// reset the queue
		input wire write_enable,				// data on "buf_in" that should be written
		input wire read_enable,					// data on "buf_out" is read
		input wire [BUF_NBITS-1:0] buf_in,		// input data
		output reg [BUF_NBITS-1:0] buf_out,		// output data
		output wire buf_empty,					// if the buffer is empty
		output wire buf_empty_next,
		output wire buf_filled,					// if the buffer is filled
		output wire buf_filled_next //,			// if the buffer is filled, if we insert an element "in this cycle"
		// output reg [BUF_NELEMS:0] nelems,				// number of elements in the buffer
        // output reg [31:0] ntotal_read,          // for debugging
        // output reg [31:0] ntotal_write          // for debugging
);



`ifdef VENDOR_XILINX
    (* ram_extract = "yes", ram_style = "block" *)
    reg  [BUF_NBITS-1:0]         buf_mem[0:2**BUF_NELEMS-1];
`else
(* ramstyle = "no_rw_check" *) reg  [BUF_NBITS-1:0] buf_mem[0:2**BUF_NELEMS-1];
`endif

reg [BUF_NELEMS -1:0] rd_ptr;
reg [BUF_NELEMS -1:0] wr_ptr;

reg [BUF_NELEMS:0] nelems;

// handle full and empty
assign buf_empty = (nelems == 0);
assign buf_empty_next = (nelems -1 == 0);

assign buf_filled = ( nelems >= (1 << BUF_NELEMS) ) ;
assign buf_filled_next = (nelems + 4 >= ( 1 << BUF_NELEMS) );


// update of number of elements
always @(posedge clk)
begin
	if( rst) begin
		nelems <= 0;
	end
	else if( ( !buf_filled && write_enable ) && ( !buf_empty && read_enable) ) begin // read and write
		nelems <= nelems;
	end
	else if( !buf_filled && write_enable ) begin	// only write
		nelems <= nelems + 1'b1;;
	end
	else if( !buf_empty && read_enable) begin // only read
		nelems <= nelems - 1'b1;;
	end
	else begin
		nelems <= nelems; // nothing
	end
end


// assignement of output data
// always assign the data

always @(posedge clk)
begin
	if( rst) begin
	    buf_out <= 0;
	end
	else begin
		if(rd_ptr == wr_ptr) begin
			buf_out <= buf_in;
		end
		else begin
        	buf_out <= buf_mem[ rd_ptr ];
		end
	end
end


// assignement of input data

always @(posedge clk)
begin
	if(rst) begin
		// no reset of the buffer is required, as we have the n_elems counter, which is reset to 0 to indicate that the data is not valid anymore
	end
	else begin
		if( write_enable && !buf_filled) begin
			buf_mem[wr_ptr] <= buf_in;
		end
		else begin
			buf_mem[wr_ptr] <= buf_mem[wr_ptr];
		end
	end
end

// assignement of the pointers
always @(posedge clk)
begin
	if( rst) begin
		wr_ptr <= 0;	
		rd_ptr <= 0;	
		// ntotal_write <= 0;
		// ntotal_read <= 0;
	end
	else begin
		if( write_enable && !buf_filled) begin	
			wr_ptr <= wr_ptr +1; // overflow handled by the width of the register
			// ntotal_write <= ntotal_write +1'b1;
		end	
		else begin
			wr_ptr <= wr_ptr;
		end

		if( read_enable && !buf_empty) begin
			rd_ptr <= rd_ptr+ 1; // overflow handled by the width of the register
			// ntotal_read <= ntotal_read + 1'b1;
		end
		else begin	
			rd_ptr <= rd_ptr;
		end
	end
end


endmodule



