

`default_nettype none


/**
	The BRAM-MCM-BIT-Receiver gets the cache-lines, and puts them into the cache-line queue. From there, they are unwrapped, and put into the queue for the edge-processor

	TODO: TEST WITH (1, 100, 1) (100, 101, 1) TO CHECK, THAT WE GET THE CORRECT BITS FOR THE DOUBLE-BUFFER RAM!
	TODO: ALSO SMALL TEST OF (1, 100, 1), (50, 100, 1) TO CHECK, THAT WE GET THE CORRECT BITS FOR THE V-VALUE!!
*/

module bram_mcm_bit_receiver #(parameter BRAM_MCM_BITS_REQUEST_TAG = 5, parameter OVERFLOW_PADDING= 4, parameter L_BITS = 4, parameter K_BITS = 4, parameter BRAM_QUEUE_LENGTH_BITS = 6) (
	//---------------------- GENERAL VARIABLES---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input wire 										um_done,						// true, if we are done processing (global)
	input wire 										um_activated,					// true, if the afu is activated

	output reg										bram_mcm_bit_receiver_done,
	output reg										bram_mcm_bit_receiver_epoch_done,
	input wire		                                global_epoch_change,
	input wire										bram_mcm_bit_requester_done,

	//---------------------- GENERAL READ INPUT VARIABLES---------------------------//
	input   wire [7:0]                         	    um_rx_rd_tag,	// response to the read request: the tag
    input   wire [511:0]                       	    um_rx_data,		// the requested data
    input   wire                               	    um_rx_rd_valid,	// that tag and data are valid

	//---------------------- VARIABLES TO INDICATE THE CURRENT INPUT SPACE WE HAVE LEFT ---------------------------//
	output reg [31:0]								nelems_space, 						// the number of space left in our input queue
	output reg [31:0]								num_resp_received,

	//---------------------- QUEUE VARIABLES FOR OUTPUT - TO THE BRAM ---------------------------//
	output reg										out_queue_write_enable,
	output reg[2**L_BITS -1:0]						out_queue_buf_in,
	input wire										in_queue_buf_filled_next,

	output reg[511:0]								debug_bram_mcm_bit_receiver
);



reg									bram_mcm_m_axis_ready;
wire								bram_mcm_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	bram_mcm_data_count;
wire [511:0]							bram_mcm_m_axis_data;
reg  [511:0]							bram_mcm_data_buffer;
reg										bram_mcm_data_buffer_valid;

reg			bram_mcm_s_axis_write_enable;
reg	[511:0]	bram_mcm_s_axis_data;


nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(512), .PROG_FULL_THRESH(0 ) ) bram_mcm_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(bram_mcm_s_axis_write_enable),
    .s_axis_ready(),
    .s_axis_data(bram_mcm_s_axis_data),
    .s_axis_last(1'b0),
    .m_axis_valid(bram_mcm_m_axis_valid),
    .m_axis_ready(bram_mcm_m_axis_ready),
    .m_axis_data(bram_mcm_m_axis_data),
    .m_axis_last(),
    .data_count(bram_mcm_data_count)
);

reg [31:0] ntotal_read;
reg [31:0] ntotal_write;


// ----- put the incoming mcm-bits (cache lines) in a BRAM queue ----- //
always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin

        // reset of output registers
        nelems_space <= 0;
		num_resp_received <= 0;
		ntotal_write <= 0;
		
		bram_mcm_s_axis_write_enable <= 0;
    end
    else begin
		bram_mcm_s_axis_write_enable <= 0;
		nelems_space <= (2**BRAM_QUEUE_LENGTH_BITS) - ( ntotal_write - ntotal_read);

		if( ~bram_mcm_bit_receiver_done) begin
			if( um_rx_rd_valid & ( um_rx_rd_tag == BRAM_MCM_BITS_REQUEST_TAG) ) begin
				bram_mcm_s_axis_write_enable <= 1;
				bram_mcm_s_axis_data <= um_rx_data;

				ntotal_write <= ntotal_write + 1 ;
				num_resp_received <= num_resp_received  + 1;
			end
		end
	end
end

reg [31:0] 		offset_cnt;
reg [511:0]		mcm_cache_line;
reg				mcm_cache_line_valid;

reg [511:0]		mcm_cache_line_overflow;
reg				mcm_cache_line_overflow_valid;

reg [31:0]		edges_count;
reg [31:0] 		expected_vertices_count;

reg[511:0]		debug_mcm_cache_line;

reg 			debug_error_flag;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
        
		// local variables
		offset_cnt <= ( (2**K_BITS *2**L_BITS) % 512) >> L_BITS ;
		mcm_cache_line <= 0;
		edges_count <= 0;
		expected_vertices_count<= 2**K_BITS;
		mcm_cache_line_valid <= 0;
		mcm_cache_line_overflow_valid <= 0;
		bram_mcm_bit_receiver_epoch_done <= 0;

		// queue variables
		// drain the queues
		if( (um_done || ~um_activated) && rst_n ) begin
			bram_mcm_m_axis_ready <= 1;
		end
		else begin
			bram_mcm_m_axis_ready <= 0;
		end
		ntotal_read <= 0;

		// outgoing variables
		out_queue_write_enable <= 0;
		bram_mcm_bit_receiver_done <= 0;
		out_queue_buf_in<= 0;	// not required to reset - not a state signal

		debug_mcm_cache_line <= 0;
		debug_error_flag <= 0;
	end
	else begin
		out_queue_write_enable <= 0;

        if( ~bram_mcm_bit_receiver_done ) begin
		
			// only set the epoch to 0 in here.
			// The MCM-BIT-RECEIVER and the requester are not active in the last epoch (since the double-buffer is not used next).
			// Therefore, when we are done, the epoch is done too! Is required for the others to start their last epoch!
			bram_mcm_bit_receiver_epoch_done <= 0;   

			// if there is no epoch change, and we have processed as many edges as expected
			if( ~global_epoch_change && ( edges_count == expected_vertices_count) ) begin

				// done, if the requester is done
				if( bram_mcm_bit_requester_done ) begin
					bram_mcm_bit_receiver_done <= 1;
				end
				
				//certainly the epoch is done
				bram_mcm_bit_receiver_epoch_done <= 1;
				
				// since the epoch is done, we do not care about the queue anymore, and just put ready high, to make sure that it is empty
				// (should actually not be necessary, but we do just to make sure)
				bram_mcm_m_axis_ready <= 1;
			end
			else if( global_epoch_change ) begin
				expected_vertices_count <= expected_vertices_count + 2**K_BITS;
				mcm_cache_line_valid <= 0;	// purge the old cache-lines
				mcm_cache_line_overflow_valid <= 0;
			end
			else if( ~mcm_cache_line_valid ) begin
				bram_mcm_m_axis_ready <= 1;

				// test for mcm_cache_line_overflow_valid not required, since it is empty for sure!
				if( mcm_cache_line_overflow_valid ) begin
					debug_error_flag<= 1;
				end
				if( bram_mcm_m_axis_ready && bram_mcm_m_axis_valid ) begin
					mcm_cache_line <= bram_mcm_m_axis_data  >> ( offset_cnt << L_BITS );
					debug_mcm_cache_line <= bram_mcm_m_axis_data;
					ntotal_read <= ntotal_read +1;
					mcm_cache_line_valid <= 1;
				end
			end
			else begin
				if( ~in_queue_buf_filled_next && ( edges_count < expected_vertices_count ) ) begin
					out_queue_write_enable <= 1;
					out_queue_buf_in <= mcm_cache_line[2**L_BITS-1:0];
					edges_count <= edges_count +1;

					// get a new cache-line
					if( (offset_cnt +1) == (512 >> L_BITS) ) begin
							offset_cnt <= 0;
							if( mcm_cache_line_overflow_valid || (  bram_mcm_m_axis_ready && bram_mcm_m_axis_valid ) ) begin
								ntotal_read <= ntotal_read +1;
								mcm_cache_line_valid <= 1;
							end
			
							if( mcm_cache_line_overflow_valid ) begin
								mcm_cache_line <= mcm_cache_line_overflow;
								if( bram_mcm_m_axis_ready && bram_mcm_m_axis_valid ) begin
									mcm_cache_line_overflow <= bram_mcm_m_axis_data;
									bram_mcm_m_axis_ready <= 0;
								end
								else begin
									mcm_cache_line_overflow_valid <= 0;
									bram_mcm_m_axis_ready <= 1;
								end
							end
							else if( bram_mcm_m_axis_ready && bram_mcm_m_axis_valid ) begin
								mcm_cache_line <= bram_mcm_m_axis_data;
								bram_mcm_m_axis_ready <= 1;
							end
							else begin
								mcm_cache_line_valid <= 0;
								bram_mcm_m_axis_ready <= 1;
							end
					end
					else begin
							mcm_cache_line <= mcm_cache_line >> (2**L_BITS);
							offset_cnt <= offset_cnt + 1;

							// handle reads from queue
							if( bram_mcm_m_axis_ready && bram_mcm_m_axis_valid ) begin
								if( ~mcm_cache_line_valid || mcm_cache_line_overflow_valid ) begin
									debug_error_flag <= 1;
								end
								
								mcm_cache_line_overflow <= bram_mcm_m_axis_data;
								mcm_cache_line_overflow_valid <= 1;
								bram_mcm_m_axis_ready <= 0;
							end
							else if( ~mcm_cache_line_overflow_valid ) begin
								bram_mcm_m_axis_ready <= 1;
							end
					end
				end
				else begin

					// handle reads from queue
					if( bram_mcm_m_axis_ready && bram_mcm_m_axis_valid ) begin
						if( ~mcm_cache_line_valid || mcm_cache_line_overflow_valid ) begin
							debug_error_flag <= 1;
						end
						mcm_cache_line_overflow <= bram_mcm_m_axis_data;
						mcm_cache_line_overflow_valid <= 1;
						bram_mcm_m_axis_ready <= 0;
					end
					else if( ~mcm_cache_line_overflow_valid ) begin
						bram_mcm_m_axis_ready <= 1;
					end
				end




			end
        end

    end


end


always @(posedge clk)
begin
    debug_bram_mcm_bit_receiver <= {416'b0, ntotal_read, ntotal_write, edges_count };
end

endmodule



`default_nettype wire






	
