


/**
	The MCM-Bit-Requester for the double buffered BRAM can issue requests as long as: We request for the next epoch of the edge-processor, and there are still requests to do for this epoch.

	NOTE: The first epoch is not double buffered, so we can diretly start with requesting bits for the second epoch. 
	NOTE: This also means, that in the first cache-line, we must issue one request less than in the others, since the first cl has bits for the first epoch, which is skipped.

	NOTE: THE DESIGN IS CURRENTLY LIMITED TO MAX L_BITS = 9 which is 512 levels

*/
module bram_mcm_bit_requester #(parameter BRAM_MCM_BITS_REQUEST_TAG = 5, parameter L_BITS = 4, parameter K_BITS = 4) (
	//---------------------- GENERAL VARIABLES---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input wire 										um_done,						// true, if we are done processing (global)
	input wire 										um_activated,					// true, if the afu is activated

	output reg										bram_mcm_bit_requester_done,
	output reg										epoch_done,

	input wire [31:0]								num_epochs,
	input wire [31:0]								global_epoch,
	input wire 										global_epoch_change,

	//---------------------- ADDRESS VARIABLES---------------------------//
	input wire [57:0]								mcm_bit_base_address,			// the base read address

	//---------------------- QUEUE VARIABLES FOR ISSUING THE REQUEST - TO REQUEST INTERFACE ---------------------------//
	output reg									   	req_queue_write_enable,				// wheter we want to write to the queue or not
	output reg [65:0] 							   	req_queue_buf_in,					// the queue input buffer (65:8 is the request address, 7:0 the tag we assign to the request)
	input wire 									   	req_queue_buf_filled_next,			// 1, if the queue is filled soon

	output reg[511:0]								debug_bram_mcm_bit_requester
	
);



// TODO: FIX BY USING A COUNTER WHICH TELLS HOW MANY CLS TO REQUEST!

reg has_pending_request;
reg [31:0] num_cl_requested;
reg [31:0] num_usefull_bits_requested; // the "base" pointer
reg [31:0] num_usefull_bits_to_request; // to where we must go
reg [31:0] num_current_bits;			// the pointer with respect to the cache-line
reg [31:0] num_requests_in_epoch;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		// local variables
		num_cl_requested <= ((2**K_BITS * 2** L_BITS) >> 9 );
		num_usefull_bits_requested  <= (2**K_BITS * 2** L_BITS);
		num_usefull_bits_to_request <= 2* (2**K_BITS * 2** L_BITS);
		num_current_bits <= ( (2**K_BITS * 2** L_BITS) >> 9) << 9;
		has_pending_request <= 1;
		
		// output variables
        bram_mcm_bit_requester_done <= 0;

		req_queue_write_enable <= 0;
		req_queue_buf_in <= 0;
		epoch_done <= 0;
		num_requests_in_epoch <= 0;
	end
	else begin

		req_queue_write_enable <= 0;
		
		if( ~bram_mcm_bit_requester_done ) begin

			if( ~has_pending_request ) begin
				epoch_done <= 1;
			end
			else begin
				epoch_done<= 0;
			end

			
			if( global_epoch_change ) begin
				num_cl_requested <= num_usefull_bits_requested >> 9;
				num_usefull_bits_to_request <= num_usefull_bits_requested + (2**K_BITS * 2** L_BITS);
				num_current_bits <= ( num_usefull_bits_requested >> 9 ) << 9; // the number of bits, with respect to the start of the cache-line

				has_pending_request <= 1;
				num_requests_in_epoch <= 0;
			end
			else if( ~has_pending_request && ( global_epoch+1  == num_epochs ) ) begin
				bram_mcm_bit_requester_done <= 1;
			end
			// if there is space in the request queue, and we still have bits to request, then request them
			else if( ~req_queue_buf_filled_next && has_pending_request ) begin
				
				req_queue_write_enable <= 1;
				req_queue_buf_in <= { mcm_bit_base_address + num_cl_requested, 8'b0} + BRAM_MCM_BITS_REQUEST_TAG;
				num_cl_requested <= num_cl_requested + 1;
				num_requests_in_epoch<= num_requests_in_epoch  + 1;

				// update our state holder
				if( num_current_bits + 512 >= num_usefull_bits_to_request ) begin
					has_pending_request <= 0;
					num_usefull_bits_requested <= num_usefull_bits_to_request; // we have requested as many bits as we had to
				end
				else begin
					num_current_bits <= num_current_bits  + 512; // we have requested 512 bits more
				end

			end

		end
    end
end


always @(posedge clk)
begin
	debug_bram_mcm_bit_requester <= {384'b0, num_current_bits, num_usefull_bits_to_request, num_usefull_bits_requested, num_cl_requested };
end



endmodule



