
`default_nettype none

// `include "level_values.vh"

module edge_processor #(parameter BRAM_QUEUE_LENGTH_BITS  = 7, parameter K_BITS = 4, parameter L_BITS = 4, parameter OVERFLOW_PADDING = 4, parameter EPSILON = 0.8, parameter BRAM_QUEUE_LENGTH_BITS_LARGE = 11 ) (
	//---------------------- GENERAL VARIABLES---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input 	wire 									um_done,						// true, if we are done processing (global)
	input 	wire 									um_activated,					// true, if the afu is activated
	output reg 										edge_processor_done,					// true, if we are done processing data
	input wire[31:0]								num_edges,

	input wire										mcm_bit_receiver_epoch_done,
	input wire										bram_mcm_bit_receiver_epoch_done,
	
	input wire[31:0]								global_epoch,
	input wire										global_epoch_change,

	output reg										epoch_done,

	//---------------------- VARIABLES FOR THE MCM-BIT BIT QUEUE (FROM THE BRAM-BIT-RECEIVER) ---------------------------//
	input wire 											queue_mcmbram_write_enable,
	input wire[2**L_BITS -1:0]			   	       		queue_mcmbram_buf_in,		
	output reg											queue_mcmbram_buf_filled_next,

	//---------------------- VARIABLES FOR THE OFFSET QUEUE ---------------------------//
	input wire 											queue_offset_write_enable,
	input wire[(9 - L_BITS)+1:0]				   	    queue_offset_buf_in,		
	output reg											queue_offset_buf_filled_next,

    //---------------------- VARIABLES FOR THE MCM-BIT INPUT QUEUE ---------------------------//
    input wire 											queue_mcmin_write_enable,					
	input wire [512-1:0]					   	        queue_mcmin_buf_in,		
	output reg											queue_mcmin_buf_filled_next,

    //---------------------- VARIABLES FOR THE EDGE READY QUEUE ---------------------------//
    input wire 											queue_edgein_write_enable,
	input wire [128-1:0]					   	        queue_edgein_buf_in,                   // (epoch, u, v, w)
	output reg											queue_edgein_buf_filled_next,

	//---------------------- FOR THE MCM OUTPUT QUEUE TO RAM ---------------------------//
	output reg 										queue_mcmout_write_enable,
	output reg [32+512-1:0]						    queue_mcmout_buf_in,  
	input wire										queue_mcmout_buf_filled_next,

    //---------------------- FOR THE EDGE OUTPUT QUEUE TO RAM ---------------------------//
	output reg 										queue_edgeout_write_enable,
	output reg [128 -1:0]						    queue_edgeout_buf_in,   
	input wire										queue_edgeout_buf_filled_next,

	output reg [511:0]								debug_edge_processor

);

//---------------------- QUEUE WITH BITS FOR THE DOUBLE-BUFFER BRAM ---------------------------//
reg									mcm_bram_s_axis_write_enable;
reg [2**L_BITS-1:0]					mcm_bram_s_axis_data;
reg									mcm_bram_m_axis_ready;
wire								mcm_bram_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	mcm_bram_data_count;
wire [2**L_BITS-1:0]						mcm_bram_m_axis_data;
reg  [2**L_BITS-1:0]						mcm_bram_data_buffer_1;
reg  [2**L_BITS-1:0]						mcm_bram_data_buffer_2;
reg											mcm_bram_data_buffer_1_valid;
reg											mcm_bram_data_buffer_2_valid;
nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA( (2**L_BITS) ), .PROG_FULL_THRESH(0 ) ) mcm_bram_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(mcm_bram_s_axis_write_enable),
    .s_axis_ready(),
    .s_axis_data(mcm_bram_s_axis_data),
    .s_axis_last(1'b0),
    .m_axis_valid(mcm_bram_m_axis_valid),
    .m_axis_ready(mcm_bram_m_axis_ready),
    .m_axis_data(mcm_bram_m_axis_data),
    .m_axis_last(),
    .data_count(mcm_bram_data_count)
);

//---------------------- QUEUE WITH BITS INCOMING MATCHING BITS ---------------------------//
reg									mcm_in_s_axis_write_enable;
reg [511:0]							mcm_in_s_axis_data;
reg									mcm_in_m_axis_ready;
wire								mcm_in_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS_LARGE-1:0]	mcm_in_data_count;
wire [511:0]						mcm_in_m_axis_data;
reg  [511:0]						mcm_in_data_buffer_1;
reg  [511:0]						mcm_in_data_buffer_2;
reg											mcm_in_data_buffer_1_valid;
reg											mcm_in_data_buffer_2_valid;
nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS_LARGE), .BUS_SIZE_DATA(512), .PROG_FULL_THRESH(0 ) ) mcm_in_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(mcm_in_s_axis_write_enable),
    .s_axis_ready(),
    .s_axis_data(mcm_in_s_axis_data),
    .s_axis_last(1'b0),
    .m_axis_valid(mcm_in_m_axis_valid),
    .m_axis_ready(mcm_in_m_axis_ready),
    .m_axis_data(mcm_in_m_axis_data),
    .m_axis_last(),
    .data_count(mcm_in_data_count)
);

//---------------------- QUEUE WITH OFFSET DATA ---------------------------//
reg									offset_s_axis_write_enable;
reg [(9 - L_BITS)+1:0]							offset_s_axis_data;
reg									offset_m_axis_ready;
wire								offset_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS_LARGE-1:0]	offset_data_count;
wire [(9 - L_BITS)+1:0]						offset_m_axis_data;
reg  [(9 - L_BITS)+1:0]						offset_data_buffer_1;
reg  [(9 - L_BITS)+1:0]						offset_data_buffer_2;
reg											offset_data_buffer_1_valid;
reg											offset_data_buffer_2_valid;
nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS_LARGE), .BUS_SIZE_DATA((9 - L_BITS+2)), .PROG_FULL_THRESH(0 ) ) offset_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(offset_s_axis_write_enable),
    .s_axis_ready(),
    .s_axis_data(offset_s_axis_data),
    .s_axis_last(1'b0),
    .m_axis_valid(offset_m_axis_valid),
    .m_axis_ready(offset_m_axis_ready),
    .m_axis_data(offset_m_axis_data),
    .m_axis_last(),
    .data_count(offset_data_count)
);


//---------------------- QUEUE WITH EDGES ---------------------------//
reg									edge_s_axis_write_enable;
reg [127:0]				edge_s_axis_data;
reg									edge_m_axis_ready;
wire								edge_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS_LARGE-1:0]	edge_data_count;
wire [127:0]						edge_m_axis_data;
reg  [127:0]						edge_data_buffer_1;
reg  [127:0]						edge_data_buffer_2;
reg											edge_data_buffer_1_valid;
reg											edge_data_buffer_2_valid;
nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS_LARGE), .BUS_SIZE_DATA(128), .PROG_FULL_THRESH(0 ) ) edge_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(edge_s_axis_write_enable),
    .s_axis_ready(),
    .s_axis_data(edge_s_axis_data),
    .s_axis_last(1'b0),
    .m_axis_valid(edge_m_axis_valid),
    .m_axis_ready(edge_m_axis_ready),
    .m_axis_data(edge_m_axis_data),
    .m_axis_last(),
    .data_count(edge_data_count)
);


reg 	bram_wu_ru_we;
reg 	bram_wu_ru_re;
reg 	[K_BITS-1:0] 		bram_wu_ru_raddr;
reg 	[K_BITS-1:0] 		bram_wu_ru_waddr;
reg 	[2**L_BITS-1:0]		bram_wu_ru_din;
wire 	[2**L_BITS-1:0]		bram_wu_ru_dout;
spl_sdp_mem_t #( .DATA_WIDTH(2**L_BITS), .ADDR_WIDTH((K_BITS)) ) bram_wu_ru (
    .clk(clk),
    .we(bram_wu_ru_we),
	.re(bram_wu_ru_re),
    .raddr(bram_wu_ru_raddr),
    .waddr(bram_wu_ru_waddr),
    .din(bram_wu_ru_din),
    .dout(bram_wu_ru_dout)
);

reg 	bram_wu_rv_we;
reg 	bram_wu_rv_re;
reg 	[K_BITS-1:0] 		bram_wu_rv_raddr;
reg 	[K_BITS-1:0] 		bram_wu_rv_waddr;
reg 	[2**L_BITS-1:0]		bram_wu_rv_din;
wire 	[2**L_BITS-1:0]		bram_wu_rv_dout;
spl_sdp_mem_t #( .DATA_WIDTH(2**L_BITS), .ADDR_WIDTH((K_BITS)) ) bram_wu_rv (
    .clk(clk),
    .we(bram_wu_rv_we),
	.re(bram_wu_rv_re),
    .raddr(bram_wu_rv_raddr),
    .waddr(bram_wu_rv_waddr),
    .din(bram_wu_rv_din),
    .dout(bram_wu_rv_dout)
);

reg 	bram_wv_ru_we;
reg 	bram_wv_ru_re;
reg 	[K_BITS-1:0] 		bram_wv_ru_raddr;
reg 	[K_BITS-1:0] 		bram_wv_ru_waddr;
reg 	[2**L_BITS-1:0]		bram_wv_ru_din;
wire 	[2**L_BITS-1:0]		bram_wv_ru_dout;
spl_sdp_mem_t #( .DATA_WIDTH(2**L_BITS), .ADDR_WIDTH((K_BITS)) ) bram_wv_ru (
    .clk(clk),
    .we(bram_wv_ru_we),
	.re(bram_wv_ru_re),
    .raddr(bram_wv_ru_raddr),
    .waddr(bram_wv_ru_waddr),
    .din(bram_wv_ru_din),
    .dout(bram_wv_ru_dout)
);

reg 	bram_wv_rv_we;
reg 	bram_wv_rv_re;
reg 	[K_BITS-1:0] 		bram_wv_rv_raddr;
reg 	[K_BITS-1:0] 		bram_wv_rv_waddr;
reg 	[2**L_BITS-1:0]		bram_wv_rv_din;
wire 	[2**L_BITS-1:0]		bram_wv_rv_dout;
spl_sdp_mem_t #( .DATA_WIDTH(2**L_BITS), .ADDR_WIDTH((K_BITS)) ) bram_wv_rv (
    .clk(clk),
    .we(bram_wv_rv_we),
	.re(bram_wv_rv_re),
    .raddr(bram_wv_rv_raddr),
    .waddr(bram_wv_rv_waddr),
    .din(bram_wv_rv_din),
    .dout(bram_wv_rv_dout)
);

// the double-buffer part

reg 	bram_2_wu_ru_we;
reg 	bram_2_wu_ru_re;
reg 	[K_BITS-1:0] 		bram_2_wu_ru_raddr;
reg 	[K_BITS-1:0] 		bram_2_wu_ru_waddr;
reg 	[2**L_BITS-1:0]		bram_2_wu_ru_din;
wire 	[2**L_BITS-1:0]		bram_2_wu_ru_dout;
spl_sdp_mem_t #( .DATA_WIDTH(2**L_BITS), .ADDR_WIDTH((K_BITS)) ) bram_2_wu_ru (
    .clk(clk),
    .we(bram_2_wu_ru_we),
	.re(bram_2_wu_ru_re),
    .raddr(bram_2_wu_ru_raddr),
    .waddr(bram_2_wu_ru_waddr),
    .din(bram_2_wu_ru_din),
    .dout(bram_2_wu_ru_dout)
);

reg 	bram_2_wu_rv_we;
reg 	bram_2_wu_rv_re;
reg 	[K_BITS-1:0] 		bram_2_wu_rv_raddr;
reg 	[K_BITS-1:0] 		bram_2_wu_rv_waddr;
reg 	[2**L_BITS-1:0]		bram_2_wu_rv_din;
wire 	[2**L_BITS-1:0]		bram_2_wu_rv_dout;
spl_sdp_mem_t #( .DATA_WIDTH(2**L_BITS), .ADDR_WIDTH((K_BITS)) ) bram_2_wu_rv (
    .clk(clk),
    .we(bram_2_wu_rv_we),
	.re(bram_2_wu_rv_re),
    .raddr(bram_2_wu_rv_raddr),
    .waddr(bram_2_wu_rv_waddr),
    .din(bram_2_wu_rv_din),
    .dout(bram_2_wu_rv_dout)
);

reg 	bram_2_wv_ru_we;
reg 	bram_2_wv_ru_re;
reg 	[K_BITS-1:0] 		bram_2_wv_ru_raddr;
reg 	[K_BITS-1:0] 		bram_2_wv_ru_waddr;
reg 	[2**L_BITS-1:0]		bram_2_wv_ru_din;
wire 	[2**L_BITS-1:0]		bram_2_wv_ru_dout;
spl_sdp_mem_t #( .DATA_WIDTH(2**L_BITS), .ADDR_WIDTH((K_BITS)) ) bram_2_wv_ru (
    .clk(clk),
    .we(bram_2_wv_ru_we),
	.re(bram_2_wv_ru_re),
    .raddr(bram_2_wv_ru_raddr),
    .waddr(bram_2_wv_ru_waddr),
    .din(bram_2_wv_ru_din),
    .dout(bram_2_wv_ru_dout)
);

reg 	bram_2_wv_rv_we;
reg 	bram_2_wv_rv_re;
reg 	[K_BITS-1:0] 		bram_2_wv_rv_raddr;
reg 	[K_BITS-1:0] 		bram_2_wv_rv_waddr;
reg 	[2**L_BITS-1:0]		bram_2_wv_rv_din;
wire 	[2**L_BITS-1:0]		bram_2_wv_rv_dout;
spl_sdp_mem_t #( .DATA_WIDTH(2**L_BITS), .ADDR_WIDTH((K_BITS)) ) bram_2_wv_rv (
    .clk(clk),
    .we(bram_2_wv_rv_we),
	.re(bram_2_wv_rv_re),
    .raddr(bram_2_wv_rv_raddr),
    .waddr(bram_2_wv_rv_waddr),
    .din(bram_2_wv_rv_din),
    .dout(bram_2_wv_rv_dout)
);


reg [2**K_BITS -1:0] buf_mem_valid;
reg [31:0]							mcmbram_ntotal_read;
reg [31:0]							mcmbram_ntotal_write;

reg [31:0]							offset_ntotal_read;
reg [31:0]							offset_ntotal_write;

reg [31:0]							mcmin_ntotal_read;
reg [31:0]							mcmin_ntotal_write;

reg [31:0]							edgein_ntotal_read;
reg [31:0]							edgein_ntotal_write;

reg [K_BITS -1:0] bram1_wr_ptr;
reg [K_BITS -1:0] bram2_wr_ptr;


// ----- WRITE TO THE QUEUES ----- //
always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		mcm_bram_s_axis_write_enable <= 0;
		offset_s_axis_write_enable <= 0;
		mcm_in_s_axis_write_enable <= 0;
		edge_s_axis_write_enable <= 0;

		mcmbram_ntotal_write <= 0;
		offset_ntotal_write <= 0;
		mcmin_ntotal_write <= 0;
		edgein_ntotal_write <= 0;

		queue_mcmbram_buf_filled_next <= 0;
		queue_offset_buf_filled_next <= 0;
		queue_mcmin_buf_filled_next <= 0;
		queue_edgein_buf_filled_next <= 0;

	end
	else begin
		mcm_bram_s_axis_write_enable <= 0;
		mcm_in_s_axis_write_enable <= 0;
		offset_s_axis_write_enable <= 0;
		edge_s_axis_write_enable <= 0;

		if( queue_mcmbram_write_enable) begin
			mcm_bram_s_axis_data <= queue_mcmbram_buf_in;
			mcm_bram_s_axis_write_enable <= 1;
			mcmbram_ntotal_write <= mcmbram_ntotal_write +1;
		end
		queue_mcmbram_buf_filled_next <=  (mcm_bram_data_count + OVERFLOW_PADDING) >= (2**BRAM_QUEUE_LENGTH_BITS);

		if( queue_offset_write_enable ) begin
			offset_s_axis_data <= queue_offset_buf_in;
			offset_s_axis_write_enable <= 1;
			offset_ntotal_write <= offset_ntotal_write +1;
		end
		queue_offset_buf_filled_next <= (offset_data_count +OVERFLOW_PADDING ) >= ( 2**BRAM_QUEUE_LENGTH_BITS_LARGE);
		
		if( queue_mcmin_write_enable) begin
			mcm_in_s_axis_data <= queue_mcmin_buf_in;
			mcm_in_s_axis_write_enable <= 1;
			mcmin_ntotal_write <= mcmin_ntotal_write +1;
		end
		queue_mcmin_buf_filled_next <= ( mcm_in_data_count + OVERFLOW_PADDING ) >= ( 2**BRAM_QUEUE_LENGTH_BITS_LARGE);

		if( queue_edgein_write_enable ) begin
			edge_s_axis_data <= queue_edgein_buf_in;
			edge_s_axis_write_enable <= 1;
			edgein_ntotal_write <= edgein_ntotal_write +1;
		end
		queue_edgein_buf_filled_next <= ( edge_data_count +OVERFLOW_PADDING ) >= ( 2**BRAM_QUEUE_LENGTH_BITS_LARGE);

	end
end

reg bram_selector;
reg [31:0]          		stage_1_edges_processed;
reg	[31:0]					stage_2_edges_processed;
reg [31:0]					stage_3_edges_processed;
reg [31:0]					stage_4_edges_processed;
reg [31:0]					stage_5_edges_processed;
reg [31:0]					stage_6_edges_processed;

reg [31:0] num_double_bram_writes;

// ----- State handling ----- //
always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
        edge_processor_done <= 0;
		epoch_done <= 0;
		bram_selector <= 0;
	end
	else begin

		if(global_epoch_change ) begin
			bram_selector <= bram_selector  + 1;
		end

		epoch_done <= 0;
		if( stage_1_edges_processed == stage_2_edges_processed &&
			stage_1_edges_processed == stage_3_edges_processed &&
			stage_1_edges_processed == stage_4_edges_processed && 
			stage_1_edges_processed == stage_5_edges_processed &&
			stage_1_edges_processed == stage_6_edges_processed ) begin

			if( num_edges == stage_6_edges_processed ) begin
				edge_processor_done <= 1;
			end

				if( ~queue_edgeout_buf_filled_next && ~queue_mcmout_buf_filled_next ) begin
					if( mcmbram_ntotal_read == mcmbram_ntotal_write && 
						offset_ntotal_read == offset_ntotal_write && 
						mcmin_ntotal_read == mcmin_ntotal_write &&
						edgein_ntotal_read == edgein_ntotal_write ) begin

						
						if( num_double_bram_writes == ( (global_epoch +1) << K_BITS ) && mcm_bit_receiver_epoch_done && bram_mcm_bit_receiver_epoch_done) begin 
							epoch_done <= 1;
						end

					end
				end
		end
	end
end


// ----- 1.) Extract the mcm-bits for v ----- //

reg [512-1:0] 				stage_1_last_cl;
// reg [2**L_BITS-1: 0] 		stage_1_last_mcm_bits_v;
// reg [9:0] 					stage_1_last_offset;
reg 						stage_1_done;

// to stage 2:
reg [512-1:0] 				stage_1_cl;
reg [2**L_BITS-1: 0] 		stage_1_mcm_bits_v;
reg [9:0] 					stage_1_offset;					// old: (9 - L_BITS) -1
reg [128-1:0] 				stage_1_edge;
reg 						stage_1_valid;
reg [K_BITS-1:0]			stage_1_bram_address_u;
reg [K_BITS-1:0]			stage_1_bram_address_v;
reg							stage_1_bram_address_v_valid;

reg 						debug_reuse_offset;

reg [512-1:0]				stage_1_debug_full_shifted_cl;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
        
		// state variables
        stage_1_edges_processed <= 0;
		stage_1_done <= 0;
		stage_1_offset <= 0;

        // queue params
		edgein_ntotal_read <= 0;
		mcmin_ntotal_read <= 0;
		offset_ntotal_read <= 0;

        // vars for stage 1 going to stage 2
		stage_1_cl <= 0;
		stage_1_valid <= 0;
        stage_1_mcm_bits_v <= 0;
        // stage_1_last_mcm_bits_v <= 0;
		stage_1_edge <= 0;

		stage_1_debug_full_shifted_cl <= 0;
	end
	else begin
        stage_1_valid <= 0;

		if( edge_processor_done || stage_1_done || queue_edgeout_buf_filled_next || queue_mcmout_buf_filled_next ) begin
				// do nothing
		end
		else if( stage_1_edges_processed == num_edges ) begin
			stage_1_done <= 1;
		end
        else begin
            if(  edge_data_buffer_1_valid  &&  offset_data_buffer_1_valid) begin // we need data from a) edges, b) offset

                if( offset_data_buffer_1[ (9 - L_BITS)+1 ]  == 1 ) begin // if the offset is reused (same "address" bits)
                    debug_reuse_offset <= 1;
					stage_1_edges_processed <= stage_1_edges_processed +1;
					stage_1_valid <= 1;

                    // mcm-bits for v
				    stage_1_mcm_bits_v <= stage_1_last_cl >> ( {8'b0, offset_data_buffer_1[(9 - L_BITS):0] } << L_BITS  ); // stage_1_last_mcm_bits_v;
                    stage_1_cl <= stage_1_last_cl;

                    // edge
                    stage_1_edge <= edge_data_buffer_1;

                    // offset
                    stage_1_offset <= offset_data_buffer_1[(9 - L_BITS):0];

				    // handle the queue update
			        offset_ntotal_read <= offset_ntotal_read +1;

					edgein_ntotal_read <= edgein_ntotal_read +1;
                    stage_1_bram_address_u <= edge_data_buffer_1[95:64] -1; // -  ( edge_data_buffer_1[127:96] << K_BITS  );
					stage_1_bram_address_v <= edge_data_buffer_1[63:32] -1; // -  ( edge_data_buffer_1[127:96] << K_BITS  );
					stage_1_bram_address_v_valid <= ( (edge_data_buffer_1[63:32] -1) >=  (edge_data_buffer_1[127:96] << K_BITS  ) ) &&
													( (edge_data_buffer_1[63:32] -1) <  ( (edge_data_buffer_1[127:96] +1) << K_BITS  ) );
		        end
                else begin
					if( mcm_in_data_buffer_1_valid ) begin
						debug_reuse_offset <= 0;

						stage_1_edges_processed <= stage_1_edges_processed +1;
		                stage_1_valid <= 1;

						stage_1_mcm_bits_v <=       mcm_in_data_buffer_1 >> ( {8'b0, offset_data_buffer_1} << L_BITS  ); // extend the bit range, otherwise, the 1 bit overflows, and we have just 0's
		                // stage_1_last_mcm_bits_v <=  mcm_in_data_buffer_1 >> ( {8'b0, offset_data_buffer_1} << L_BITS  );

						stage_1_debug_full_shifted_cl <= mcm_in_data_buffer_1 >> ( {8'b0, offset_data_buffer_1} << L_BITS  ); 
		            
		                stage_1_cl <= mcm_in_data_buffer_1;
		                stage_1_last_cl <= mcm_in_data_buffer_1 ;

		                // edge
		                stage_1_edge <= edge_data_buffer_1;

		                // offset
		                stage_1_offset <= offset_data_buffer_1[(9 - L_BITS):0];
		                // stage_1_last_offset <= offset_data_buffer_1;
					    offset_ntotal_read <= offset_ntotal_read +1;

					    edgein_ntotal_read <= edgein_ntotal_read +1;
					    mcmin_ntotal_read <= mcmin_ntotal_read+ 1;

						stage_1_bram_address_u <= edge_data_buffer_1[95:64] -1; // -  ( edge_data_buffer_1[127:96] << K_BITS  );
						stage_1_bram_address_v <= edge_data_buffer_1[63:32] -1; // -  ( edge_data_buffer_1[127:96] << K_BITS  );
						stage_1_bram_address_v_valid <= ( (edge_data_buffer_1[63:32] -1) >=  (edge_data_buffer_1[127:96] << K_BITS  ) ) &&
													( (edge_data_buffer_1[63:32] -1) <  ( (edge_data_buffer_1[127:96] +1) << K_BITS  ) );
					end

                end 

            end
        end
   
    end
end



always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		mcm_in_data_buffer_1_valid <= 0;
		mcm_in_data_buffer_2_valid <= 0;

		edge_data_buffer_1_valid <= 0;
		edge_data_buffer_2_valid <= 0;

		offset_data_buffer_1_valid <= 0;
		offset_data_buffer_2_valid <= 0;

		// drain queues
		if( (um_done || ~um_activated) && rst_n ) begin
			mcm_in_m_axis_ready <= 1;
			edge_m_axis_ready <= 1;
			offset_m_axis_ready <= 1;
		end
		else begin
			mcm_in_m_axis_ready <= 0;
			edge_m_axis_ready <= 0;
			offset_m_axis_ready <= 0;
		end
	end
	else begin

			// case for mcm-in
			if(  !(edge_processor_done || stage_1_done || queue_edgeout_buf_filled_next || queue_mcmout_buf_filled_next )
				&&
				!(stage_1_edges_processed == num_edges)
				&&
				 edge_data_buffer_1_valid  &&  offset_data_buffer_1_valid 
				&& !( offset_data_buffer_1[ (9 - L_BITS)+1 ]  == 1  )
				&& mcm_in_data_buffer_1_valid 
				) begin
					// handle consume
					if( mcm_in_data_buffer_2_valid ) begin
						mcm_in_data_buffer_1 <= mcm_in_data_buffer_2;
						if( mcm_in_m_axis_ready && mcm_in_m_axis_valid ) begin
							mcm_in_data_buffer_2 <= mcm_in_m_axis_data;
							mcm_in_m_axis_ready <= 0;
						end
						else begin
							mcm_in_data_buffer_2_valid <= 0;
							mcm_in_m_axis_ready <= 1;
						end
					end
					else begin
						if( mcm_in_m_axis_ready && mcm_in_m_axis_valid ) begin
							mcm_in_data_buffer_1  <= mcm_in_m_axis_data;
						end
						else begin
							mcm_in_data_buffer_1_valid <= 0;
							mcm_in_m_axis_ready <= 1;
						end
					end
			end
			else begin
					// handle add
					if( mcm_in_m_axis_ready && mcm_in_m_axis_valid ) begin
						if( ~mcm_in_data_buffer_1_valid) begin
							mcm_in_data_buffer_1  <= mcm_in_m_axis_data;
							mcm_in_data_buffer_1_valid <= 1;
						end
						else if( ~mcm_in_data_buffer_2_valid ) begin
							mcm_in_data_buffer_2 <= mcm_in_m_axis_data;
							mcm_in_data_buffer_2_valid <= 1;
							mcm_in_m_axis_ready <= 0;
						end
					end
					else if( ~mcm_in_data_buffer_1_valid || ~mcm_in_data_buffer_2_valid ) begin
						mcm_in_m_axis_ready <= 1;
					end
			end

			// case for edge
			if( !(edge_processor_done || stage_1_done || queue_edgeout_buf_filled_next || queue_mcmout_buf_filled_next )
				&&
				!(stage_1_edges_processed == num_edges)
				&&
					 ( 		edge_data_buffer_1_valid  &&  offset_data_buffer_1_valid 
							&& (
									 (  offset_data_buffer_1[ (9 - L_BITS)+1 ]  == 1 ) || 
									( !(offset_data_buffer_1[ (9 - L_BITS)+1 ]  == 1  ) && mcm_in_data_buffer_1_valid )
								)
					)
				) begin

					// handle consume
					if( edge_data_buffer_2_valid ) begin
						edge_data_buffer_1 <= edge_data_buffer_2;
						if( edge_m_axis_ready && edge_m_axis_valid ) begin
							edge_data_buffer_2 <= edge_m_axis_data;
							edge_m_axis_ready <= 0;
						end
						else begin
							edge_data_buffer_2_valid <= 0;
							edge_m_axis_ready <= 1;
						end
					end
					else begin
						if( edge_m_axis_ready && edge_m_axis_valid ) begin
							edge_data_buffer_1  <= edge_m_axis_data;
						end
						else begin
							edge_data_buffer_1_valid <= 0;
							edge_m_axis_ready <= 1;
						end
					end


			end
			else begin
					// handle add
					if( edge_m_axis_ready && edge_m_axis_valid ) begin
						if( ~edge_data_buffer_1_valid) begin
							edge_data_buffer_1  <= edge_m_axis_data;
							edge_data_buffer_1_valid <= 1;
						end
						else if( ~edge_data_buffer_2_valid ) begin
							edge_data_buffer_2 <= edge_m_axis_data;
							edge_data_buffer_2_valid <= 1;
							edge_m_axis_ready <= 0;
						end
					end
					else if( ~edge_data_buffer_1_valid || ~edge_data_buffer_2_valid ) begin
						edge_m_axis_ready <= 1;
					end
			end


			// case for offset
			if( !(edge_processor_done || stage_1_done || queue_edgeout_buf_filled_next || queue_mcmout_buf_filled_next )
				&&
				!(stage_1_edges_processed == num_edges)
				&&
					 ( 		edge_data_buffer_1_valid  &&  offset_data_buffer_1_valid 
							&& (
									 (  offset_data_buffer_1[ (9 - L_BITS)+1 ]  == 1 )  || 
									( !(offset_data_buffer_1[ (9 - L_BITS)+1 ]  == 1 ) && mcm_in_data_buffer_1_valid )
								)
					)
				) begin

					// handle consume
					if( offset_data_buffer_2_valid ) begin
						offset_data_buffer_1 <= offset_data_buffer_2;
						if( offset_m_axis_ready && offset_m_axis_valid ) begin
							offset_data_buffer_2 <= offset_m_axis_data;
							offset_m_axis_ready <= 0;
						end
						else begin
							offset_data_buffer_2_valid <= 0;
							offset_m_axis_ready <= 1;
						end
					end
					else begin
						if( offset_m_axis_ready && offset_m_axis_valid ) begin
							offset_data_buffer_1  <= offset_m_axis_data;
						end
						else begin
							offset_data_buffer_1_valid <= 0;
							offset_m_axis_ready <= 1;
						end
					end
			end
			else begin
					// handle add
					if( offset_m_axis_ready && offset_m_axis_valid ) begin
						if( ~offset_data_buffer_1_valid) begin
							offset_data_buffer_1  <= offset_m_axis_data;
							offset_data_buffer_1_valid <= 1;
						end
						else if( ~offset_data_buffer_2_valid ) begin
							offset_data_buffer_2 <= offset_m_axis_data;
							offset_data_buffer_2_valid <= 1;
							offset_m_axis_ready <= 0;
						end
					end
					else if( ~offset_data_buffer_1_valid || ~offset_data_buffer_2_valid ) begin
						offset_m_axis_ready <= 1;
					end
			end
	end
end


// ----- 2.) Request the the matching bits of v from BRAM----- //
reg 					stage_2_valid;
reg 					stage_2_done;

reg [512-1:0] 				stage_2_cl;
reg [9:0] 					stage_2_offset;
reg [128-1:0] 				stage_2_edge;
reg [2**L_BITS-1: 0] 		stage_2_mcm_bits_v;


reg [K_BITS-1:0]			stage_2_bram_address_u;
reg [K_BITS-1:0]			stage_2_bram_address_v;
reg							stage_2_bram_address_v_valid;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin

		stage_2_edges_processed <= 0;
		stage_2_done <= 0;
		stage_2_valid <= 0;

		bram_wu_ru_re <= 0;
		bram_wu_rv_re <= 0;
		bram_wv_ru_re <= 0;
		bram_wv_rv_re <= 0;

		bram_2_wu_ru_re <= 0;
		bram_2_wu_rv_re <= 0;
		bram_2_wv_ru_re <= 0;
		bram_2_wv_rv_re <= 0;
	end
	else begin
		bram_wu_ru_re <= 0;
		bram_wu_rv_re <= 0;
		bram_wv_ru_re <= 0;
		bram_wv_rv_re <= 0;

		bram_2_wu_ru_re <= 0;
		bram_2_wu_rv_re <= 0;
		bram_2_wv_ru_re <= 0;
		bram_2_wv_rv_re <= 0;


		stage_2_valid <= 0;
		if( stage_2_edges_processed == num_edges ) begin
			stage_2_done <= 1;
		end
		
		if( edge_processor_done || stage_2_done || ~stage_1_valid ) begin
				// do nothing
		end
		else begin
			stage_2_edges_processed <= stage_2_edges_processed + 1;
			stage_2_valid <= 1;

			stage_2_cl <= stage_1_cl;
            stage_2_offset <= stage_1_offset;
            stage_2_edge <= stage_1_edge;
			stage_2_mcm_bits_v <= stage_1_mcm_bits_v;
            
			stage_2_bram_address_u <= stage_1_bram_address_u;
			stage_2_bram_address_v <= stage_1_bram_address_v;
			stage_2_bram_address_v_valid <= stage_1_bram_address_v_valid;

            if( bram_selector ) begin

				bram_wu_ru_raddr <= stage_1_bram_address_u;
				bram_wu_ru_re <= 1;
				bram_wv_ru_raddr <= stage_1_bram_address_u;
				bram_wv_ru_re <= 1;
				
				if( stage_1_bram_address_v_valid ) begin
					bram_wu_rv_raddr <= stage_1_bram_address_v;
					bram_wu_rv_re <= 1;
					bram_wv_rv_raddr <= stage_1_bram_address_v;
					bram_wv_rv_re <= 1;
				end
			end
			else begin
				bram_2_wu_ru_raddr <= stage_1_bram_address_u;
				bram_2_wu_ru_re <= 1;
				bram_2_wv_ru_raddr <= stage_1_bram_address_u;
				bram_2_wv_ru_re <= 1;

				if( stage_1_bram_address_v_valid ) begin
					bram_2_wu_rv_raddr <= stage_1_bram_address_v;
					bram_2_wu_rv_re <= 1;
					bram_2_wv_rv_raddr <= stage_1_bram_address_v;
					bram_2_wv_rv_re <= 1;
				end
			end
        

		end
	end
end


// ----- 2b.) pass data (wait, till data arrives from bram) ----- //
reg 					stage_2b_valid;
reg 					stage_2b_done;
reg [31:0]				stage_2b_edges_processed;

reg [512-1:0] 				stage_2b_cl;
reg [9:0] 					stage_2b_offset;
reg [128-1:0] 				stage_2b_edge;
reg [2**L_BITS-1: 0] 		stage_2b_mcm_bits_v;


reg [K_BITS-1:0]			stage_2b_bram_address_u;
reg [K_BITS-1:0]			stage_2b_bram_address_v;
reg							stage_2b_bram_address_v_valid;


always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin

		stage_2b_edges_processed <= 0;
		stage_2b_done <= 0;
		stage_2b_valid <= 0;
		stage_2b_edge <= 0;
	end
	else begin
		stage_2b_valid <= 0;
		if( stage_2b_edges_processed == num_edges ) begin
			stage_2b_done <= 1;
		end
		
		if( edge_processor_done || stage_2b_done || ~stage_2_valid ) begin
				// do nothing
		end
		else begin
			stage_2b_edges_processed <= stage_2b_edges_processed + 1;
			stage_2b_valid <= 1;

			stage_2b_cl <= stage_2_cl;
            stage_2b_offset <= stage_2_offset;
            stage_2b_edge <= stage_2_edge;
			stage_2b_mcm_bits_v <= stage_2_mcm_bits_v;
            
			stage_2b_bram_address_u <= stage_2_bram_address_u;
			stage_2b_bram_address_v <= stage_2_bram_address_v;
			stage_2b_bram_address_v_valid <= stage_2_bram_address_v_valid;
		end
	end
end

// ----- 3.) Merge data from BRAM, select the correct data, compute the correct level ----- //

reg stage_3_done;

// to stage 4:
reg [512-1:0] 			stage_3_cl;
reg [2**L_BITS-1: 0] 	stage_3_mcm_bits_u;
reg [2**L_BITS-1: 0] 	stage_3_mcm_bits_v;
reg [2**L_BITS-1:0] 	stage_3_edge_level;
reg 					stage_3_valid;
reg [9:0] stage_3_offset;			// old: (9 - L_BITS) -1
reg [128-1:0] 			stage_3_edge;
reg [K_BITS-1:0]		stage_3_bram_address_u;
reg [K_BITS-1:0]		stage_3_bram_address_v;
reg						stage_3_bram_address_v_valid;


always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin

		stage_3_edges_processed <= 0;
		stage_3_done <= 0;
		stage_3_valid <= 0;
	end
	else begin
        stage_3_valid <= 0;
		if( stage_3_edges_processed == num_edges ) begin
			stage_3_done <= 1;
		end
		
		if( edge_processor_done || stage_3_done || ~stage_2b_valid ) begin
				// do nothing
		end
        else begin
			stage_3_edges_processed <= stage_3_edges_processed + 1;
			stage_3_valid <= 1;

			stage_3_cl <= stage_2b_cl;
            stage_3_offset <= stage_2b_offset;
            stage_3_edge <= stage_2b_edge;            
			stage_3_bram_address_u <= stage_2b_bram_address_u;
			stage_3_bram_address_v <= stage_2b_bram_address_v;
			stage_3_bram_address_v_valid <= stage_2b_bram_address_v_valid;

			if( bram_selector ) begin

				stage_3_mcm_bits_u <= bram_wu_ru_dout | bram_wv_ru_dout;
				
				if( stage_2b_bram_address_v_valid ) begin
					stage_3_mcm_bits_v <= bram_wu_rv_dout | bram_wv_rv_dout;
				end
				else begin
					stage_3_mcm_bits_v <= stage_2b_mcm_bits_v;
				end
			end
			else begin
				stage_3_mcm_bits_u <= bram_2_wu_ru_dout | bram_2_wv_ru_dout;

				if( stage_2b_bram_address_v_valid ) begin
					stage_3_mcm_bits_v <= bram_2_wu_rv_dout | bram_2_wv_rv_dout;
				end
				else begin
					stage_3_mcm_bits_v <= stage_2b_mcm_bits_v;
				end
			end

        end
    end
end


// Since the level-computation did not work on harp2 using an initialized array, we hardcode everything in here
// generated using open-office
// ="stage_3_edge_level[" & $A$4-C4 &"] <= stage_2b_edge[31:0] >= "&F4 &";"
// where A4 contains 2**L_BITS -1, C4 is the index, and F4 holds the value we compare to

generate 

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
	end
	else begin
		if( edge_processor_done || stage_3_done || ~stage_2b_valid ) begin
			
		end
		else begin
			// such that we have: {	ew >=  L0, ew >= L1, ... } with L0 = 0, L1 = (1+EPSILON), ...
			if( L_BITS <= 5 ) begin		// L_BITS = 2, 3, 4, 5 - EPSILON = 0.6 : 1.6
				if( L_BITS == 2 ) begin	// 4 Levels
					stage_3_edge_level[3] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[2] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[1] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[0] <= stage_2b_edge[31:0] >= 32'd4;
				end
				else if (L_BITS == 3) begin
					stage_3_edge_level[7] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[6] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[5] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[4] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[3] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[2] <= stage_2b_edge[31:0] >= 32'd10;
					stage_3_edge_level[1] <= stage_2b_edge[31:0] >= 32'd17;
					stage_3_edge_level[0] <= stage_2b_edge[31:0] >= 32'd27;
				end
				else if( L_BITS == 4 ) begin
					stage_3_edge_level[15] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[14] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[13] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[12] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[11] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[10] <= stage_2b_edge[31:0] >= 32'd10;
					stage_3_edge_level[9] <= stage_2b_edge[31:0] >= 32'd17;
					stage_3_edge_level[8] <= stage_2b_edge[31:0] >= 32'd27;
					stage_3_edge_level[7] <= stage_2b_edge[31:0] >= 32'd43;
					stage_3_edge_level[6] <= stage_2b_edge[31:0] >= 32'd69;
					stage_3_edge_level[5] <= stage_2b_edge[31:0] >= 32'd110;
					stage_3_edge_level[4] <= stage_2b_edge[31:0] >= 32'd176;
					stage_3_edge_level[3] <= stage_2b_edge[31:0] >= 32'd281;
					stage_3_edge_level[2] <= stage_2b_edge[31:0] >= 32'd450;
					stage_3_edge_level[1] <= stage_2b_edge[31:0] >= 32'd721;
					stage_3_edge_level[0] <= stage_2b_edge[31:0] >= 32'd1153;
				end
				else if( L_BITS == 5 ) begin
					stage_3_edge_level[31] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[30] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[29] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[28] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[27] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[26] <= stage_2b_edge[31:0] >= 32'd10;
					stage_3_edge_level[25] <= stage_2b_edge[31:0] >= 32'd17;
					stage_3_edge_level[24] <= stage_2b_edge[31:0] >= 32'd27;
					stage_3_edge_level[23] <= stage_2b_edge[31:0] >= 32'd43;
					stage_3_edge_level[22] <= stage_2b_edge[31:0] >= 32'd69;
					stage_3_edge_level[21] <= stage_2b_edge[31:0] >= 32'd110;
					stage_3_edge_level[20] <= stage_2b_edge[31:0] >= 32'd176;
					stage_3_edge_level[19] <= stage_2b_edge[31:0] >= 32'd281;
					stage_3_edge_level[18] <= stage_2b_edge[31:0] >= 32'd450;
					stage_3_edge_level[17] <= stage_2b_edge[31:0] >= 32'd721;
					stage_3_edge_level[16] <= stage_2b_edge[31:0] >= 32'd1153;
					stage_3_edge_level[15] <= stage_2b_edge[31:0] >= 32'd1845;
					stage_3_edge_level[14] <= stage_2b_edge[31:0] >= 32'd2951;
					stage_3_edge_level[13] <= stage_2b_edge[31:0] >= 32'd4722;
					stage_3_edge_level[12] <= stage_2b_edge[31:0] >= 32'd7556;
					stage_3_edge_level[11] <= stage_2b_edge[31:0] >= 32'd12089;
					stage_3_edge_level[10] <= stage_2b_edge[31:0] >= 32'd19343;
					stage_3_edge_level[9] <= stage_2b_edge[31:0] >= 32'd30949;
					stage_3_edge_level[8] <= stage_2b_edge[31:0] >= 32'd49518;
					stage_3_edge_level[7] <= stage_2b_edge[31:0] >= 32'd79228;
					stage_3_edge_level[6] <= stage_2b_edge[31:0] >= 32'd126765;
					stage_3_edge_level[5] <= stage_2b_edge[31:0] >= 32'd202824;
					stage_3_edge_level[4] <= stage_2b_edge[31:0] >= 32'd324519;
					stage_3_edge_level[3] <= stage_2b_edge[31:0] >= 32'd519230;
					stage_3_edge_level[2] <= stage_2b_edge[31:0] >= 32'd830767;
					stage_3_edge_level[1] <= stage_2b_edge[31:0] >= 32'd1329228;
					stage_3_edge_level[0] <= stage_2b_edge[31:0] >= 32'd2126765;
				end
			end
			else if( L_BITS <= 7) begin // L_BITS = 6, 7  EPSILON = 0.1 : 1.6
				if( L_BITS == 6 ) begin
					stage_3_edge_level[63] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[62] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[61] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[60] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[59] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[58] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[57] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[56] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[55] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[54] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[53] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[52] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[51] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[50] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[49] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[48] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[47] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[46] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[45] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[44] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[43] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[42] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[41] <= stage_2b_edge[31:0] >= 32'd8;
					stage_3_edge_level[40] <= stage_2b_edge[31:0] >= 32'd9;
					stage_3_edge_level[39] <= stage_2b_edge[31:0] >= 32'd10;
					stage_3_edge_level[38] <= stage_2b_edge[31:0] >= 32'd11;
					stage_3_edge_level[37] <= stage_2b_edge[31:0] >= 32'd12;
					stage_3_edge_level[36] <= stage_2b_edge[31:0] >= 32'd13;
					stage_3_edge_level[35] <= stage_2b_edge[31:0] >= 32'd14;
					stage_3_edge_level[34] <= stage_2b_edge[31:0] >= 32'd16;
					stage_3_edge_level[33] <= stage_2b_edge[31:0] >= 32'd17;
					stage_3_edge_level[32] <= stage_2b_edge[31:0] >= 32'd19;
					stage_3_edge_level[31] <= stage_2b_edge[31:0] >= 32'd21;
					stage_3_edge_level[30] <= stage_2b_edge[31:0] >= 32'd23;
					stage_3_edge_level[29] <= stage_2b_edge[31:0] >= 32'd26;
					stage_3_edge_level[28] <= stage_2b_edge[31:0] >= 32'd28;
					stage_3_edge_level[27] <= stage_2b_edge[31:0] >= 32'd31;
					stage_3_edge_level[26] <= stage_2b_edge[31:0] >= 32'd34;
					stage_3_edge_level[25] <= stage_2b_edge[31:0] >= 32'd37;
					stage_3_edge_level[24] <= stage_2b_edge[31:0] >= 32'd41;
					stage_3_edge_level[23] <= stage_2b_edge[31:0] >= 32'd45;
					stage_3_edge_level[22] <= stage_2b_edge[31:0] >= 32'd50;
					stage_3_edge_level[21] <= stage_2b_edge[31:0] >= 32'd55;
					stage_3_edge_level[20] <= stage_2b_edge[31:0] >= 32'd60;
					stage_3_edge_level[19] <= stage_2b_edge[31:0] >= 32'd66;
					stage_3_edge_level[18] <= stage_2b_edge[31:0] >= 32'd73;
					stage_3_edge_level[17] <= stage_2b_edge[31:0] >= 32'd80;
					stage_3_edge_level[16] <= stage_2b_edge[31:0] >= 32'd88;
					stage_3_edge_level[15] <= stage_2b_edge[31:0] >= 32'd97;
					stage_3_edge_level[14] <= stage_2b_edge[31:0] >= 32'd107;
					stage_3_edge_level[13] <= stage_2b_edge[31:0] >= 32'd117;
					stage_3_edge_level[12] <= stage_2b_edge[31:0] >= 32'd129;
					stage_3_edge_level[11] <= stage_2b_edge[31:0] >= 32'd142;
					stage_3_edge_level[10] <= stage_2b_edge[31:0] >= 32'd156;
					stage_3_edge_level[9] <= stage_2b_edge[31:0] >= 32'd172;
					stage_3_edge_level[8] <= stage_2b_edge[31:0] >= 32'd189;
					stage_3_edge_level[7] <= stage_2b_edge[31:0] >= 32'd208;
					stage_3_edge_level[6] <= stage_2b_edge[31:0] >= 32'd229;
					stage_3_edge_level[5] <= stage_2b_edge[31:0] >= 32'd252;
					stage_3_edge_level[4] <= stage_2b_edge[31:0] >= 32'd277;
					stage_3_edge_level[3] <= stage_2b_edge[31:0] >= 32'd304;
					stage_3_edge_level[2] <= stage_2b_edge[31:0] >= 32'd335;
					stage_3_edge_level[1] <= stage_2b_edge[31:0] >= 32'd368;
					stage_3_edge_level[0] <= stage_2b_edge[31:0] >= 32'd405;
				end
				else if( L_BITS == 7 ) begin
					stage_3_edge_level[127] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[126] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[125] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[124] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[123] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[122] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[121] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[120] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[119] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[118] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[117] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[116] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[115] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[114] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[113] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[112] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[111] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[110] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[109] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[108] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[107] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[106] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[105] <= stage_2b_edge[31:0] >= 32'd8;
					stage_3_edge_level[104] <= stage_2b_edge[31:0] >= 32'd9;
					stage_3_edge_level[103] <= stage_2b_edge[31:0] >= 32'd10;
					stage_3_edge_level[102] <= stage_2b_edge[31:0] >= 32'd11;
					stage_3_edge_level[101] <= stage_2b_edge[31:0] >= 32'd12;
					stage_3_edge_level[100] <= stage_2b_edge[31:0] >= 32'd13;
					stage_3_edge_level[99] <= stage_2b_edge[31:0] >= 32'd14;
					stage_3_edge_level[98] <= stage_2b_edge[31:0] >= 32'd16;
					stage_3_edge_level[97] <= stage_2b_edge[31:0] >= 32'd17;
					stage_3_edge_level[96] <= stage_2b_edge[31:0] >= 32'd19;
					stage_3_edge_level[95] <= stage_2b_edge[31:0] >= 32'd21;
					stage_3_edge_level[94] <= stage_2b_edge[31:0] >= 32'd23;
					stage_3_edge_level[93] <= stage_2b_edge[31:0] >= 32'd26;
					stage_3_edge_level[92] <= stage_2b_edge[31:0] >= 32'd28;
					stage_3_edge_level[91] <= stage_2b_edge[31:0] >= 32'd31;
					stage_3_edge_level[90] <= stage_2b_edge[31:0] >= 32'd34;
					stage_3_edge_level[89] <= stage_2b_edge[31:0] >= 32'd37;
					stage_3_edge_level[88] <= stage_2b_edge[31:0] >= 32'd41;
					stage_3_edge_level[87] <= stage_2b_edge[31:0] >= 32'd45;
					stage_3_edge_level[86] <= stage_2b_edge[31:0] >= 32'd50;
					stage_3_edge_level[85] <= stage_2b_edge[31:0] >= 32'd55;
					stage_3_edge_level[84] <= stage_2b_edge[31:0] >= 32'd60;
					stage_3_edge_level[83] <= stage_2b_edge[31:0] >= 32'd66;
					stage_3_edge_level[82] <= stage_2b_edge[31:0] >= 32'd73;
					stage_3_edge_level[81] <= stage_2b_edge[31:0] >= 32'd80;
					stage_3_edge_level[80] <= stage_2b_edge[31:0] >= 32'd88;
					stage_3_edge_level[79] <= stage_2b_edge[31:0] >= 32'd97;
					stage_3_edge_level[78] <= stage_2b_edge[31:0] >= 32'd107;
					stage_3_edge_level[77] <= stage_2b_edge[31:0] >= 32'd117;
					stage_3_edge_level[76] <= stage_2b_edge[31:0] >= 32'd129;
					stage_3_edge_level[75] <= stage_2b_edge[31:0] >= 32'd142;
					stage_3_edge_level[74] <= stage_2b_edge[31:0] >= 32'd156;
					stage_3_edge_level[73] <= stage_2b_edge[31:0] >= 32'd172;
					stage_3_edge_level[72] <= stage_2b_edge[31:0] >= 32'd189;
					stage_3_edge_level[71] <= stage_2b_edge[31:0] >= 32'd208;
					stage_3_edge_level[70] <= stage_2b_edge[31:0] >= 32'd229;
					stage_3_edge_level[69] <= stage_2b_edge[31:0] >= 32'd252;
					stage_3_edge_level[68] <= stage_2b_edge[31:0] >= 32'd277;
					stage_3_edge_level[67] <= stage_2b_edge[31:0] >= 32'd304;
					stage_3_edge_level[66] <= stage_2b_edge[31:0] >= 32'd335;
					stage_3_edge_level[65] <= stage_2b_edge[31:0] >= 32'd368;
					stage_3_edge_level[64] <= stage_2b_edge[31:0] >= 32'd405;
					stage_3_edge_level[63] <= stage_2b_edge[31:0] >= 32'd446;
					stage_3_edge_level[62] <= stage_2b_edge[31:0] >= 32'd490;
					stage_3_edge_level[61] <= stage_2b_edge[31:0] >= 32'd539;
					stage_3_edge_level[60] <= stage_2b_edge[31:0] >= 32'd593;
					stage_3_edge_level[59] <= stage_2b_edge[31:0] >= 32'd653;
					stage_3_edge_level[58] <= stage_2b_edge[31:0] >= 32'd718;
					stage_3_edge_level[57] <= stage_2b_edge[31:0] >= 32'd790;
					stage_3_edge_level[56] <= stage_2b_edge[31:0] >= 32'd869;
					stage_3_edge_level[55] <= stage_2b_edge[31:0] >= 32'd956;
					stage_3_edge_level[54] <= stage_2b_edge[31:0] >= 32'd1051;
					stage_3_edge_level[53] <= stage_2b_edge[31:0] >= 32'd1156;
					stage_3_edge_level[52] <= stage_2b_edge[31:0] >= 32'd1272;
					stage_3_edge_level[51] <= stage_2b_edge[31:0] >= 32'd1399;
					stage_3_edge_level[50] <= stage_2b_edge[31:0] >= 32'd1539;
					stage_3_edge_level[49] <= stage_2b_edge[31:0] >= 32'd1693;
					stage_3_edge_level[48] <= stage_2b_edge[31:0] >= 32'd1862;
					stage_3_edge_level[47] <= stage_2b_edge[31:0] >= 32'd2048;
					stage_3_edge_level[46] <= stage_2b_edge[31:0] >= 32'd2253;
					stage_3_edge_level[45] <= stage_2b_edge[31:0] >= 32'd2479;
					stage_3_edge_level[44] <= stage_2b_edge[31:0] >= 32'd2726;
					stage_3_edge_level[43] <= stage_2b_edge[31:0] >= 32'd2999;
					stage_3_edge_level[42] <= stage_2b_edge[31:0] >= 32'd3299;
					stage_3_edge_level[41] <= stage_2b_edge[31:0] >= 32'd3629;
					stage_3_edge_level[40] <= stage_2b_edge[31:0] >= 32'd3992;
					stage_3_edge_level[39] <= stage_2b_edge[31:0] >= 32'd4391;
					stage_3_edge_level[38] <= stage_2b_edge[31:0] >= 32'd4830;
					stage_3_edge_level[37] <= stage_2b_edge[31:0] >= 32'd5313;
					stage_3_edge_level[36] <= stage_2b_edge[31:0] >= 32'd5844;
					stage_3_edge_level[35] <= stage_2b_edge[31:0] >= 32'd6429;
					stage_3_edge_level[34] <= stage_2b_edge[31:0] >= 32'd7072;
					stage_3_edge_level[33] <= stage_2b_edge[31:0] >= 32'd7779;
					stage_3_edge_level[32] <= stage_2b_edge[31:0] >= 32'd8557;
					stage_3_edge_level[31] <= stage_2b_edge[31:0] >= 32'd9412;
					stage_3_edge_level[30] <= stage_2b_edge[31:0] >= 32'd10354;
					stage_3_edge_level[29] <= stage_2b_edge[31:0] >= 32'd11389;
					stage_3_edge_level[28] <= stage_2b_edge[31:0] >= 32'd12528;
					stage_3_edge_level[27] <= stage_2b_edge[31:0] >= 32'd13781;
					stage_3_edge_level[26] <= stage_2b_edge[31:0] >= 32'd15159;
					stage_3_edge_level[25] <= stage_2b_edge[31:0] >= 32'd16675;
					stage_3_edge_level[24] <= stage_2b_edge[31:0] >= 32'd18342;
					stage_3_edge_level[23] <= stage_2b_edge[31:0] >= 32'd20176;
					stage_3_edge_level[22] <= stage_2b_edge[31:0] >= 32'd22194;
					stage_3_edge_level[21] <= stage_2b_edge[31:0] >= 32'd24413;
					stage_3_edge_level[20] <= stage_2b_edge[31:0] >= 32'd26855;
					stage_3_edge_level[19] <= stage_2b_edge[31:0] >= 32'd29540;
					stage_3_edge_level[18] <= stage_2b_edge[31:0] >= 32'd32494;
					stage_3_edge_level[17] <= stage_2b_edge[31:0] >= 32'd35743;
					stage_3_edge_level[16] <= stage_2b_edge[31:0] >= 32'd39318;
					stage_3_edge_level[15] <= stage_2b_edge[31:0] >= 32'd43249;
					stage_3_edge_level[14] <= stage_2b_edge[31:0] >= 32'd47574;
					stage_3_edge_level[13] <= stage_2b_edge[31:0] >= 32'd52332;
					stage_3_edge_level[12] <= stage_2b_edge[31:0] >= 32'd57565;
					stage_3_edge_level[11] <= stage_2b_edge[31:0] >= 32'd63322;
					stage_3_edge_level[10] <= stage_2b_edge[31:0] >= 32'd69654;
					stage_3_edge_level[9] <= stage_2b_edge[31:0] >= 32'd76619;
					stage_3_edge_level[8] <= stage_2b_edge[31:0] >= 32'd84281;
					stage_3_edge_level[7] <= stage_2b_edge[31:0] >= 32'd92709;
					stage_3_edge_level[6] <= stage_2b_edge[31:0] >= 32'd101980;
					stage_3_edge_level[5] <= stage_2b_edge[31:0] >= 32'd112178;
					stage_3_edge_level[4] <= stage_2b_edge[31:0] >= 32'd123396;
					stage_3_edge_level[3] <= stage_2b_edge[31:0] >= 32'd135735;
					stage_3_edge_level[2] <= stage_2b_edge[31:0] >= 32'd149309;
					stage_3_edge_level[1] <= stage_2b_edge[31:0] >= 32'd164240;
					stage_3_edge_level[0] <= stage_2b_edge[31:0] >= 32'd180664;
				end
			end
			else if( L_BITS <= 9) begin // EPSILON = 0.03 : 1.03
				if( L_BITS == 8  ) begin
					stage_3_edge_level[255] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[254] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[253] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[252] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[251] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[250] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[249] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[248] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[247] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[246] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[245] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[244] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[243] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[242] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[241] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[240] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[239] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[238] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[237] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[236] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[235] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[234] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[233] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[232] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[231] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[230] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[229] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[228] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[227] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[226] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[225] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[224] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[223] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[222] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[221] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[220] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[219] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[218] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[217] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[216] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[215] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[214] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[213] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[212] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[211] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[210] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[209] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[208] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[207] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[206] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[205] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[204] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[203] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[202] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[201] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[200] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[199] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[198] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[197] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[196] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[195] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[194] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[193] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[192] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[191] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[190] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[189] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[188] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[187] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[186] <= stage_2b_edge[31:0] >= 32'd8;
					stage_3_edge_level[185] <= stage_2b_edge[31:0] >= 32'd8;
					stage_3_edge_level[184] <= stage_2b_edge[31:0] >= 32'd8;
					stage_3_edge_level[183] <= stage_2b_edge[31:0] >= 32'd8;
					stage_3_edge_level[182] <= stage_2b_edge[31:0] >= 32'd9;
					stage_3_edge_level[181] <= stage_2b_edge[31:0] >= 32'd9;
					stage_3_edge_level[180] <= stage_2b_edge[31:0] >= 32'd9;
					stage_3_edge_level[179] <= stage_2b_edge[31:0] >= 32'd9;
					stage_3_edge_level[178] <= stage_2b_edge[31:0] >= 32'd10;
					stage_3_edge_level[177] <= stage_2b_edge[31:0] >= 32'd10;
					stage_3_edge_level[176] <= stage_2b_edge[31:0] >= 32'd10;
					stage_3_edge_level[175] <= stage_2b_edge[31:0] >= 32'd11;
					stage_3_edge_level[174] <= stage_2b_edge[31:0] >= 32'd11;
					stage_3_edge_level[173] <= stage_2b_edge[31:0] >= 32'd11;
					stage_3_edge_level[172] <= stage_2b_edge[31:0] >= 32'd12;
					stage_3_edge_level[171] <= stage_2b_edge[31:0] >= 32'd12;
					stage_3_edge_level[170] <= stage_2b_edge[31:0] >= 32'd12;
					stage_3_edge_level[169] <= stage_2b_edge[31:0] >= 32'd13;
					stage_3_edge_level[168] <= stage_2b_edge[31:0] >= 32'd13;
					stage_3_edge_level[167] <= stage_2b_edge[31:0] >= 32'd13;
					stage_3_edge_level[166] <= stage_2b_edge[31:0] >= 32'd14;
					stage_3_edge_level[165] <= stage_2b_edge[31:0] >= 32'd14;
					stage_3_edge_level[164] <= stage_2b_edge[31:0] >= 32'd15;
					stage_3_edge_level[163] <= stage_2b_edge[31:0] >= 32'd15;
					stage_3_edge_level[162] <= stage_2b_edge[31:0] >= 32'd16;
					stage_3_edge_level[161] <= stage_2b_edge[31:0] >= 32'd16;
					stage_3_edge_level[160] <= stage_2b_edge[31:0] >= 32'd17;
					stage_3_edge_level[159] <= stage_2b_edge[31:0] >= 32'd17;
					stage_3_edge_level[158] <= stage_2b_edge[31:0] >= 32'd18;
					stage_3_edge_level[157] <= stage_2b_edge[31:0] >= 32'd18;
					stage_3_edge_level[156] <= stage_2b_edge[31:0] >= 32'd19;
					stage_3_edge_level[155] <= stage_2b_edge[31:0] >= 32'd19;
					stage_3_edge_level[154] <= stage_2b_edge[31:0] >= 32'd20;
					stage_3_edge_level[153] <= stage_2b_edge[31:0] >= 32'd20;
					stage_3_edge_level[152] <= stage_2b_edge[31:0] >= 32'd21;
					stage_3_edge_level[151] <= stage_2b_edge[31:0] >= 32'd22;
					stage_3_edge_level[150] <= stage_2b_edge[31:0] >= 32'd22;
					stage_3_edge_level[149] <= stage_2b_edge[31:0] >= 32'd23;
					stage_3_edge_level[148] <= stage_2b_edge[31:0] >= 32'd24;
					stage_3_edge_level[147] <= stage_2b_edge[31:0] >= 32'd24;
					stage_3_edge_level[146] <= stage_2b_edge[31:0] >= 32'd25;
					stage_3_edge_level[145] <= stage_2b_edge[31:0] >= 32'd26;
					stage_3_edge_level[144] <= stage_2b_edge[31:0] >= 32'd27;
					stage_3_edge_level[143] <= stage_2b_edge[31:0] >= 32'd27;
					stage_3_edge_level[142] <= stage_2b_edge[31:0] >= 32'd28;
					stage_3_edge_level[141] <= stage_2b_edge[31:0] >= 32'd29;
					stage_3_edge_level[140] <= stage_2b_edge[31:0] >= 32'd30;
					stage_3_edge_level[139] <= stage_2b_edge[31:0] >= 32'd31;
					stage_3_edge_level[138] <= stage_2b_edge[31:0] >= 32'd32;
					stage_3_edge_level[137] <= stage_2b_edge[31:0] >= 32'd33;
					stage_3_edge_level[136] <= stage_2b_edge[31:0] >= 32'd34;
					stage_3_edge_level[135] <= stage_2b_edge[31:0] >= 32'd35;
					stage_3_edge_level[134] <= stage_2b_edge[31:0] >= 32'd36;
					stage_3_edge_level[133] <= stage_2b_edge[31:0] >= 32'd37;
					stage_3_edge_level[132] <= stage_2b_edge[31:0] >= 32'd38;
					stage_3_edge_level[131] <= stage_2b_edge[31:0] >= 32'd39;
					stage_3_edge_level[130] <= stage_2b_edge[31:0] >= 32'd40;
					stage_3_edge_level[129] <= stage_2b_edge[31:0] >= 32'd41;
					stage_3_edge_level[128] <= stage_2b_edge[31:0] >= 32'd43;
					stage_3_edge_level[127] <= stage_2b_edge[31:0] >= 32'd44;
					stage_3_edge_level[126] <= stage_2b_edge[31:0] >= 32'd45;
					stage_3_edge_level[125] <= stage_2b_edge[31:0] >= 32'd47;
					stage_3_edge_level[124] <= stage_2b_edge[31:0] >= 32'd48;
					stage_3_edge_level[123] <= stage_2b_edge[31:0] >= 32'd49;
					stage_3_edge_level[122] <= stage_2b_edge[31:0] >= 32'd51;
					stage_3_edge_level[121] <= stage_2b_edge[31:0] >= 32'd53;
					stage_3_edge_level[120] <= stage_2b_edge[31:0] >= 32'd54;
					stage_3_edge_level[119] <= stage_2b_edge[31:0] >= 32'd56;
					stage_3_edge_level[118] <= stage_2b_edge[31:0] >= 32'd57;
					stage_3_edge_level[117] <= stage_2b_edge[31:0] >= 32'd59;
					stage_3_edge_level[116] <= stage_2b_edge[31:0] >= 32'd61;
					stage_3_edge_level[115] <= stage_2b_edge[31:0] >= 32'd63;
					stage_3_edge_level[114] <= stage_2b_edge[31:0] >= 32'd65;
					stage_3_edge_level[113] <= stage_2b_edge[31:0] >= 32'd67;
					stage_3_edge_level[112] <= stage_2b_edge[31:0] >= 32'd69;
					stage_3_edge_level[111] <= stage_2b_edge[31:0] >= 32'd71;
					stage_3_edge_level[110] <= stage_2b_edge[31:0] >= 32'd73;
					stage_3_edge_level[109] <= stage_2b_edge[31:0] >= 32'd75;
					stage_3_edge_level[108] <= stage_2b_edge[31:0] >= 32'd77;
					stage_3_edge_level[107] <= stage_2b_edge[31:0] >= 32'd79;
					stage_3_edge_level[106] <= stage_2b_edge[31:0] >= 32'd82;
					stage_3_edge_level[105] <= stage_2b_edge[31:0] >= 32'd84;
					stage_3_edge_level[104] <= stage_2b_edge[31:0] >= 32'd87;
					stage_3_edge_level[103] <= stage_2b_edge[31:0] >= 32'd89;
					stage_3_edge_level[102] <= stage_2b_edge[31:0] >= 32'd92;
					stage_3_edge_level[101] <= stage_2b_edge[31:0] >= 32'd95;
					stage_3_edge_level[100] <= stage_2b_edge[31:0] >= 32'd98;
					stage_3_edge_level[99] <= stage_2b_edge[31:0] >= 32'd101;
					stage_3_edge_level[98] <= stage_2b_edge[31:0] >= 32'd104;
					stage_3_edge_level[97] <= stage_2b_edge[31:0] >= 32'd107;
					stage_3_edge_level[96] <= stage_2b_edge[31:0] >= 32'd110;
					stage_3_edge_level[95] <= stage_2b_edge[31:0] >= 32'd113;
					stage_3_edge_level[94] <= stage_2b_edge[31:0] >= 32'd117;
					stage_3_edge_level[93] <= stage_2b_edge[31:0] >= 32'd120;
					stage_3_edge_level[92] <= stage_2b_edge[31:0] >= 32'd124;
					stage_3_edge_level[91] <= stage_2b_edge[31:0] >= 32'd127;
					stage_3_edge_level[90] <= stage_2b_edge[31:0] >= 32'd131;
					stage_3_edge_level[89] <= stage_2b_edge[31:0] >= 32'd135;
					stage_3_edge_level[88] <= stage_2b_edge[31:0] >= 32'd139;
					stage_3_edge_level[87] <= stage_2b_edge[31:0] >= 32'd143;
					stage_3_edge_level[86] <= stage_2b_edge[31:0] >= 32'd148;
					stage_3_edge_level[85] <= stage_2b_edge[31:0] >= 32'd152;
					stage_3_edge_level[84] <= stage_2b_edge[31:0] >= 32'd157;
					stage_3_edge_level[83] <= stage_2b_edge[31:0] >= 32'd161;
					stage_3_edge_level[82] <= stage_2b_edge[31:0] >= 32'd166;
					stage_3_edge_level[81] <= stage_2b_edge[31:0] >= 32'd171;
					stage_3_edge_level[80] <= stage_2b_edge[31:0] >= 32'd176;
					stage_3_edge_level[79] <= stage_2b_edge[31:0] >= 32'd182;
					stage_3_edge_level[78] <= stage_2b_edge[31:0] >= 32'd187;
					stage_3_edge_level[77] <= stage_2b_edge[31:0] >= 32'd193;
					stage_3_edge_level[76] <= stage_2b_edge[31:0] >= 32'd199;
					stage_3_edge_level[75] <= stage_2b_edge[31:0] >= 32'd205;
					stage_3_edge_level[74] <= stage_2b_edge[31:0] >= 32'd211;
					stage_3_edge_level[73] <= stage_2b_edge[31:0] >= 32'd217;
					stage_3_edge_level[72] <= stage_2b_edge[31:0] >= 32'd223;
					stage_3_edge_level[71] <= stage_2b_edge[31:0] >= 32'd230;
					stage_3_edge_level[70] <= stage_2b_edge[31:0] >= 32'd237;
					stage_3_edge_level[69] <= stage_2b_edge[31:0] >= 32'd244;
					stage_3_edge_level[68] <= stage_2b_edge[31:0] >= 32'd252;
					stage_3_edge_level[67] <= stage_2b_edge[31:0] >= 32'd259;
					stage_3_edge_level[66] <= stage_2b_edge[31:0] >= 32'd267;
					stage_3_edge_level[65] <= stage_2b_edge[31:0] >= 32'd275;
					stage_3_edge_level[64] <= stage_2b_edge[31:0] >= 32'd283;
					stage_3_edge_level[63] <= stage_2b_edge[31:0] >= 32'd292;
					stage_3_edge_level[62] <= stage_2b_edge[31:0] >= 32'd300;
					stage_3_edge_level[61] <= stage_2b_edge[31:0] >= 32'd309;
					stage_3_edge_level[60] <= stage_2b_edge[31:0] >= 32'd319;
					stage_3_edge_level[59] <= stage_2b_edge[31:0] >= 32'd328;
					stage_3_edge_level[58] <= stage_2b_edge[31:0] >= 32'd338;
					stage_3_edge_level[57] <= stage_2b_edge[31:0] >= 32'd348;
					stage_3_edge_level[56] <= stage_2b_edge[31:0] >= 32'd359;
					stage_3_edge_level[55] <= stage_2b_edge[31:0] >= 32'd369;
					stage_3_edge_level[54] <= stage_2b_edge[31:0] >= 32'd380;
					stage_3_edge_level[53] <= stage_2b_edge[31:0] >= 32'd392;
					stage_3_edge_level[52] <= stage_2b_edge[31:0] >= 32'd404;
					stage_3_edge_level[51] <= stage_2b_edge[31:0] >= 32'd416;
					stage_3_edge_level[50] <= stage_2b_edge[31:0] >= 32'd428;
					stage_3_edge_level[49] <= stage_2b_edge[31:0] >= 32'd441;
					stage_3_edge_level[48] <= stage_2b_edge[31:0] >= 32'd454;
					stage_3_edge_level[47] <= stage_2b_edge[31:0] >= 32'd468;
					stage_3_edge_level[46] <= stage_2b_edge[31:0] >= 32'd482;
					stage_3_edge_level[45] <= stage_2b_edge[31:0] >= 32'd496;
					stage_3_edge_level[44] <= stage_2b_edge[31:0] >= 32'd511;
					stage_3_edge_level[43] <= stage_2b_edge[31:0] >= 32'd527;
					stage_3_edge_level[42] <= stage_2b_edge[31:0] >= 32'd542;
					stage_3_edge_level[41] <= stage_2b_edge[31:0] >= 32'd559;
					stage_3_edge_level[40] <= stage_2b_edge[31:0] >= 32'd575;
					stage_3_edge_level[39] <= stage_2b_edge[31:0] >= 32'd593;
					stage_3_edge_level[38] <= stage_2b_edge[31:0] >= 32'd610;
					stage_3_edge_level[37] <= stage_2b_edge[31:0] >= 32'd629;
					stage_3_edge_level[36] <= stage_2b_edge[31:0] >= 32'd648;
					stage_3_edge_level[35] <= stage_2b_edge[31:0] >= 32'd667;
					stage_3_edge_level[34] <= stage_2b_edge[31:0] >= 32'd687;
					stage_3_edge_level[33] <= stage_2b_edge[31:0] >= 32'd708;
					stage_3_edge_level[32] <= stage_2b_edge[31:0] >= 32'd729;
					stage_3_edge_level[31] <= stage_2b_edge[31:0] >= 32'd751;
					stage_3_edge_level[30] <= stage_2b_edge[31:0] >= 32'd773;
					stage_3_edge_level[29] <= stage_2b_edge[31:0] >= 32'd797;
					stage_3_edge_level[28] <= stage_2b_edge[31:0] >= 32'd820;
					stage_3_edge_level[27] <= stage_2b_edge[31:0] >= 32'd845;
					stage_3_edge_level[26] <= stage_2b_edge[31:0] >= 32'd870;
					stage_3_edge_level[25] <= stage_2b_edge[31:0] >= 32'd897;
					stage_3_edge_level[24] <= stage_2b_edge[31:0] >= 32'd923;
					stage_3_edge_level[23] <= stage_2b_edge[31:0] >= 32'd951;
					stage_3_edge_level[22] <= stage_2b_edge[31:0] >= 32'd980;
					stage_3_edge_level[21] <= stage_2b_edge[31:0] >= 32'd1009;
					stage_3_edge_level[20] <= stage_2b_edge[31:0] >= 32'd1039;
					stage_3_edge_level[19] <= stage_2b_edge[31:0] >= 32'd1070;
					stage_3_edge_level[18] <= stage_2b_edge[31:0] >= 32'd1103;
					stage_3_edge_level[17] <= stage_2b_edge[31:0] >= 32'd1136;
					stage_3_edge_level[16] <= stage_2b_edge[31:0] >= 32'd1170;
					stage_3_edge_level[15] <= stage_2b_edge[31:0] >= 32'd1205;
					stage_3_edge_level[14] <= stage_2b_edge[31:0] >= 32'd1241;
					stage_3_edge_level[13] <= stage_2b_edge[31:0] >= 32'd1278;
					stage_3_edge_level[12] <= stage_2b_edge[31:0] >= 32'd1317;
					stage_3_edge_level[11] <= stage_2b_edge[31:0] >= 32'd1356;
					stage_3_edge_level[10] <= stage_2b_edge[31:0] >= 32'd1397;
					stage_3_edge_level[9] <= stage_2b_edge[31:0] >= 32'd1439;
					stage_3_edge_level[8] <= stage_2b_edge[31:0] >= 32'd1482;
					stage_3_edge_level[7] <= stage_2b_edge[31:0] >= 32'd1526;
					stage_3_edge_level[6] <= stage_2b_edge[31:0] >= 32'd1572;
					stage_3_edge_level[5] <= stage_2b_edge[31:0] >= 32'd1619;
					stage_3_edge_level[4] <= stage_2b_edge[31:0] >= 32'd1668;
					stage_3_edge_level[3] <= stage_2b_edge[31:0] >= 32'd1718;
					stage_3_edge_level[2] <= stage_2b_edge[31:0] >= 32'd1769;
					stage_3_edge_level[1] <= stage_2b_edge[31:0] >= 32'd1822;
					stage_3_edge_level[0] <= stage_2b_edge[31:0] >= 32'd1877;
				end
				else if( L_BITS == 9 ) begin
					stage_3_edge_level[511] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[510] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[509] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[508] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[507] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[506] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[505] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[504] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[503] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[502] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[501] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[500] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[499] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[498] <= stage_2b_edge[31:0] >= 32'd1;
					stage_3_edge_level[497] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[496] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[495] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[494] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[493] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[492] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[491] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[490] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[489] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[488] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[487] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[486] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[485] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[484] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[483] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[482] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[481] <= stage_2b_edge[31:0] >= 32'd2;
					stage_3_edge_level[480] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[479] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[478] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[477] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[476] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[475] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[474] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[473] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[472] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[471] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[470] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[469] <= stage_2b_edge[31:0] >= 32'd3;
					stage_3_edge_level[468] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[467] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[466] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[465] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[464] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[463] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[462] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[461] <= stage_2b_edge[31:0] >= 32'd4;
					stage_3_edge_level[460] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[459] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[458] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[457] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[456] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[455] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[454] <= stage_2b_edge[31:0] >= 32'd5;
					stage_3_edge_level[453] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[452] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[451] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[450] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[449] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[448] <= stage_2b_edge[31:0] >= 32'd6;
					stage_3_edge_level[447] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[446] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[445] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[444] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[443] <= stage_2b_edge[31:0] >= 32'd7;
					stage_3_edge_level[442] <= stage_2b_edge[31:0] >= 32'd8;
					stage_3_edge_level[441] <= stage_2b_edge[31:0] >= 32'd8;
					stage_3_edge_level[440] <= stage_2b_edge[31:0] >= 32'd8;
					stage_3_edge_level[439] <= stage_2b_edge[31:0] >= 32'd8;
					stage_3_edge_level[438] <= stage_2b_edge[31:0] >= 32'd9;
					stage_3_edge_level[437] <= stage_2b_edge[31:0] >= 32'd9;
					stage_3_edge_level[436] <= stage_2b_edge[31:0] >= 32'd9;
					stage_3_edge_level[435] <= stage_2b_edge[31:0] >= 32'd9;
					stage_3_edge_level[434] <= stage_2b_edge[31:0] >= 32'd10;
					stage_3_edge_level[433] <= stage_2b_edge[31:0] >= 32'd10;
					stage_3_edge_level[432] <= stage_2b_edge[31:0] >= 32'd10;
					stage_3_edge_level[431] <= stage_2b_edge[31:0] >= 32'd11;
					stage_3_edge_level[430] <= stage_2b_edge[31:0] >= 32'd11;
					stage_3_edge_level[429] <= stage_2b_edge[31:0] >= 32'd11;
					stage_3_edge_level[428] <= stage_2b_edge[31:0] >= 32'd12;
					stage_3_edge_level[427] <= stage_2b_edge[31:0] >= 32'd12;
					stage_3_edge_level[426] <= stage_2b_edge[31:0] >= 32'd12;
					stage_3_edge_level[425] <= stage_2b_edge[31:0] >= 32'd13;
					stage_3_edge_level[424] <= stage_2b_edge[31:0] >= 32'd13;
					stage_3_edge_level[423] <= stage_2b_edge[31:0] >= 32'd13;
					stage_3_edge_level[422] <= stage_2b_edge[31:0] >= 32'd14;
					stage_3_edge_level[421] <= stage_2b_edge[31:0] >= 32'd14;
					stage_3_edge_level[420] <= stage_2b_edge[31:0] >= 32'd15;
					stage_3_edge_level[419] <= stage_2b_edge[31:0] >= 32'd15;
					stage_3_edge_level[418] <= stage_2b_edge[31:0] >= 32'd16;
					stage_3_edge_level[417] <= stage_2b_edge[31:0] >= 32'd16;
					stage_3_edge_level[416] <= stage_2b_edge[31:0] >= 32'd17;
					stage_3_edge_level[415] <= stage_2b_edge[31:0] >= 32'd17;
					stage_3_edge_level[414] <= stage_2b_edge[31:0] >= 32'd18;
					stage_3_edge_level[413] <= stage_2b_edge[31:0] >= 32'd18;
					stage_3_edge_level[412] <= stage_2b_edge[31:0] >= 32'd19;
					stage_3_edge_level[411] <= stage_2b_edge[31:0] >= 32'd19;
					stage_3_edge_level[410] <= stage_2b_edge[31:0] >= 32'd20;
					stage_3_edge_level[409] <= stage_2b_edge[31:0] >= 32'd20;
					stage_3_edge_level[408] <= stage_2b_edge[31:0] >= 32'd21;
					stage_3_edge_level[407] <= stage_2b_edge[31:0] >= 32'd22;
					stage_3_edge_level[406] <= stage_2b_edge[31:0] >= 32'd22;
					stage_3_edge_level[405] <= stage_2b_edge[31:0] >= 32'd23;
					stage_3_edge_level[404] <= stage_2b_edge[31:0] >= 32'd24;
					stage_3_edge_level[403] <= stage_2b_edge[31:0] >= 32'd24;
					stage_3_edge_level[402] <= stage_2b_edge[31:0] >= 32'd25;
					stage_3_edge_level[401] <= stage_2b_edge[31:0] >= 32'd26;
					stage_3_edge_level[400] <= stage_2b_edge[31:0] >= 32'd27;
					stage_3_edge_level[399] <= stage_2b_edge[31:0] >= 32'd27;
					stage_3_edge_level[398] <= stage_2b_edge[31:0] >= 32'd28;
					stage_3_edge_level[397] <= stage_2b_edge[31:0] >= 32'd29;
					stage_3_edge_level[396] <= stage_2b_edge[31:0] >= 32'd30;
					stage_3_edge_level[395] <= stage_2b_edge[31:0] >= 32'd31;
					stage_3_edge_level[394] <= stage_2b_edge[31:0] >= 32'd32;
					stage_3_edge_level[393] <= stage_2b_edge[31:0] >= 32'd33;
					stage_3_edge_level[392] <= stage_2b_edge[31:0] >= 32'd34;
					stage_3_edge_level[391] <= stage_2b_edge[31:0] >= 32'd35;
					stage_3_edge_level[390] <= stage_2b_edge[31:0] >= 32'd36;
					stage_3_edge_level[389] <= stage_2b_edge[31:0] >= 32'd37;
					stage_3_edge_level[388] <= stage_2b_edge[31:0] >= 32'd38;
					stage_3_edge_level[387] <= stage_2b_edge[31:0] >= 32'd39;
					stage_3_edge_level[386] <= stage_2b_edge[31:0] >= 32'd40;
					stage_3_edge_level[385] <= stage_2b_edge[31:0] >= 32'd41;
					stage_3_edge_level[384] <= stage_2b_edge[31:0] >= 32'd43;
					stage_3_edge_level[383] <= stage_2b_edge[31:0] >= 32'd44;
					stage_3_edge_level[382] <= stage_2b_edge[31:0] >= 32'd45;
					stage_3_edge_level[381] <= stage_2b_edge[31:0] >= 32'd47;
					stage_3_edge_level[380] <= stage_2b_edge[31:0] >= 32'd48;
					stage_3_edge_level[379] <= stage_2b_edge[31:0] >= 32'd49;
					stage_3_edge_level[378] <= stage_2b_edge[31:0] >= 32'd51;
					stage_3_edge_level[377] <= stage_2b_edge[31:0] >= 32'd53;
					stage_3_edge_level[376] <= stage_2b_edge[31:0] >= 32'd54;
					stage_3_edge_level[375] <= stage_2b_edge[31:0] >= 32'd56;
					stage_3_edge_level[374] <= stage_2b_edge[31:0] >= 32'd57;
					stage_3_edge_level[373] <= stage_2b_edge[31:0] >= 32'd59;
					stage_3_edge_level[372] <= stage_2b_edge[31:0] >= 32'd61;
					stage_3_edge_level[371] <= stage_2b_edge[31:0] >= 32'd63;
					stage_3_edge_level[370] <= stage_2b_edge[31:0] >= 32'd65;
					stage_3_edge_level[369] <= stage_2b_edge[31:0] >= 32'd67;
					stage_3_edge_level[368] <= stage_2b_edge[31:0] >= 32'd69;
					stage_3_edge_level[367] <= stage_2b_edge[31:0] >= 32'd71;
					stage_3_edge_level[366] <= stage_2b_edge[31:0] >= 32'd73;
					stage_3_edge_level[365] <= stage_2b_edge[31:0] >= 32'd75;
					stage_3_edge_level[364] <= stage_2b_edge[31:0] >= 32'd77;
					stage_3_edge_level[363] <= stage_2b_edge[31:0] >= 32'd79;
					stage_3_edge_level[362] <= stage_2b_edge[31:0] >= 32'd82;
					stage_3_edge_level[361] <= stage_2b_edge[31:0] >= 32'd84;
					stage_3_edge_level[360] <= stage_2b_edge[31:0] >= 32'd87;
					stage_3_edge_level[359] <= stage_2b_edge[31:0] >= 32'd89;
					stage_3_edge_level[358] <= stage_2b_edge[31:0] >= 32'd92;
					stage_3_edge_level[357] <= stage_2b_edge[31:0] >= 32'd95;
					stage_3_edge_level[356] <= stage_2b_edge[31:0] >= 32'd98;
					stage_3_edge_level[355] <= stage_2b_edge[31:0] >= 32'd101;
					stage_3_edge_level[354] <= stage_2b_edge[31:0] >= 32'd104;
					stage_3_edge_level[353] <= stage_2b_edge[31:0] >= 32'd107;
					stage_3_edge_level[352] <= stage_2b_edge[31:0] >= 32'd110;
					stage_3_edge_level[351] <= stage_2b_edge[31:0] >= 32'd113;
					stage_3_edge_level[350] <= stage_2b_edge[31:0] >= 32'd117;
					stage_3_edge_level[349] <= stage_2b_edge[31:0] >= 32'd120;
					stage_3_edge_level[348] <= stage_2b_edge[31:0] >= 32'd124;
					stage_3_edge_level[347] <= stage_2b_edge[31:0] >= 32'd127;
					stage_3_edge_level[346] <= stage_2b_edge[31:0] >= 32'd131;
					stage_3_edge_level[345] <= stage_2b_edge[31:0] >= 32'd135;
					stage_3_edge_level[344] <= stage_2b_edge[31:0] >= 32'd139;
					stage_3_edge_level[343] <= stage_2b_edge[31:0] >= 32'd143;
					stage_3_edge_level[342] <= stage_2b_edge[31:0] >= 32'd148;
					stage_3_edge_level[341] <= stage_2b_edge[31:0] >= 32'd152;
					stage_3_edge_level[340] <= stage_2b_edge[31:0] >= 32'd157;
					stage_3_edge_level[339] <= stage_2b_edge[31:0] >= 32'd161;
					stage_3_edge_level[338] <= stage_2b_edge[31:0] >= 32'd166;
					stage_3_edge_level[337] <= stage_2b_edge[31:0] >= 32'd171;
					stage_3_edge_level[336] <= stage_2b_edge[31:0] >= 32'd176;
					stage_3_edge_level[335] <= stage_2b_edge[31:0] >= 32'd182;
					stage_3_edge_level[334] <= stage_2b_edge[31:0] >= 32'd187;
					stage_3_edge_level[333] <= stage_2b_edge[31:0] >= 32'd193;
					stage_3_edge_level[332] <= stage_2b_edge[31:0] >= 32'd199;
					stage_3_edge_level[331] <= stage_2b_edge[31:0] >= 32'd205;
					stage_3_edge_level[330] <= stage_2b_edge[31:0] >= 32'd211;
					stage_3_edge_level[329] <= stage_2b_edge[31:0] >= 32'd217;
					stage_3_edge_level[328] <= stage_2b_edge[31:0] >= 32'd223;
					stage_3_edge_level[327] <= stage_2b_edge[31:0] >= 32'd230;
					stage_3_edge_level[326] <= stage_2b_edge[31:0] >= 32'd237;
					stage_3_edge_level[325] <= stage_2b_edge[31:0] >= 32'd244;
					stage_3_edge_level[324] <= stage_2b_edge[31:0] >= 32'd252;
					stage_3_edge_level[323] <= stage_2b_edge[31:0] >= 32'd259;
					stage_3_edge_level[322] <= stage_2b_edge[31:0] >= 32'd267;
					stage_3_edge_level[321] <= stage_2b_edge[31:0] >= 32'd275;
					stage_3_edge_level[320] <= stage_2b_edge[31:0] >= 32'd283;
					stage_3_edge_level[319] <= stage_2b_edge[31:0] >= 32'd292;
					stage_3_edge_level[318] <= stage_2b_edge[31:0] >= 32'd300;
					stage_3_edge_level[317] <= stage_2b_edge[31:0] >= 32'd309;
					stage_3_edge_level[316] <= stage_2b_edge[31:0] >= 32'd319;
					stage_3_edge_level[315] <= stage_2b_edge[31:0] >= 32'd328;
					stage_3_edge_level[314] <= stage_2b_edge[31:0] >= 32'd338;
					stage_3_edge_level[313] <= stage_2b_edge[31:0] >= 32'd348;
					stage_3_edge_level[312] <= stage_2b_edge[31:0] >= 32'd359;
					stage_3_edge_level[311] <= stage_2b_edge[31:0] >= 32'd369;
					stage_3_edge_level[310] <= stage_2b_edge[31:0] >= 32'd380;
					stage_3_edge_level[309] <= stage_2b_edge[31:0] >= 32'd392;
					stage_3_edge_level[308] <= stage_2b_edge[31:0] >= 32'd404;
					stage_3_edge_level[307] <= stage_2b_edge[31:0] >= 32'd416;
					stage_3_edge_level[306] <= stage_2b_edge[31:0] >= 32'd428;
					stage_3_edge_level[305] <= stage_2b_edge[31:0] >= 32'd441;
					stage_3_edge_level[304] <= stage_2b_edge[31:0] >= 32'd454;
					stage_3_edge_level[303] <= stage_2b_edge[31:0] >= 32'd468;
					stage_3_edge_level[302] <= stage_2b_edge[31:0] >= 32'd482;
					stage_3_edge_level[301] <= stage_2b_edge[31:0] >= 32'd496;
					stage_3_edge_level[300] <= stage_2b_edge[31:0] >= 32'd511;
					stage_3_edge_level[299] <= stage_2b_edge[31:0] >= 32'd527;
					stage_3_edge_level[298] <= stage_2b_edge[31:0] >= 32'd542;
					stage_3_edge_level[297] <= stage_2b_edge[31:0] >= 32'd559;
					stage_3_edge_level[296] <= stage_2b_edge[31:0] >= 32'd575;
					stage_3_edge_level[295] <= stage_2b_edge[31:0] >= 32'd593;
					stage_3_edge_level[294] <= stage_2b_edge[31:0] >= 32'd610;
					stage_3_edge_level[293] <= stage_2b_edge[31:0] >= 32'd629;
					stage_3_edge_level[292] <= stage_2b_edge[31:0] >= 32'd648;
					stage_3_edge_level[291] <= stage_2b_edge[31:0] >= 32'd667;
					stage_3_edge_level[290] <= stage_2b_edge[31:0] >= 32'd687;
					stage_3_edge_level[289] <= stage_2b_edge[31:0] >= 32'd708;
					stage_3_edge_level[288] <= stage_2b_edge[31:0] >= 32'd729;
					stage_3_edge_level[287] <= stage_2b_edge[31:0] >= 32'd751;
					stage_3_edge_level[286] <= stage_2b_edge[31:0] >= 32'd773;
					stage_3_edge_level[285] <= stage_2b_edge[31:0] >= 32'd797;
					stage_3_edge_level[284] <= stage_2b_edge[31:0] >= 32'd820;
					stage_3_edge_level[283] <= stage_2b_edge[31:0] >= 32'd845;
					stage_3_edge_level[282] <= stage_2b_edge[31:0] >= 32'd870;
					stage_3_edge_level[281] <= stage_2b_edge[31:0] >= 32'd897;
					stage_3_edge_level[280] <= stage_2b_edge[31:0] >= 32'd923;
					stage_3_edge_level[279] <= stage_2b_edge[31:0] >= 32'd951;
					stage_3_edge_level[278] <= stage_2b_edge[31:0] >= 32'd980;
					stage_3_edge_level[277] <= stage_2b_edge[31:0] >= 32'd1009;
					stage_3_edge_level[276] <= stage_2b_edge[31:0] >= 32'd1039;
					stage_3_edge_level[275] <= stage_2b_edge[31:0] >= 32'd1070;
					stage_3_edge_level[274] <= stage_2b_edge[31:0] >= 32'd1103;
					stage_3_edge_level[273] <= stage_2b_edge[31:0] >= 32'd1136;
					stage_3_edge_level[272] <= stage_2b_edge[31:0] >= 32'd1170;
					stage_3_edge_level[271] <= stage_2b_edge[31:0] >= 32'd1205;
					stage_3_edge_level[270] <= stage_2b_edge[31:0] >= 32'd1241;
					stage_3_edge_level[269] <= stage_2b_edge[31:0] >= 32'd1278;
					stage_3_edge_level[268] <= stage_2b_edge[31:0] >= 32'd1317;
					stage_3_edge_level[267] <= stage_2b_edge[31:0] >= 32'd1356;
					stage_3_edge_level[266] <= stage_2b_edge[31:0] >= 32'd1397;
					stage_3_edge_level[265] <= stage_2b_edge[31:0] >= 32'd1439;
					stage_3_edge_level[264] <= stage_2b_edge[31:0] >= 32'd1482;
					stage_3_edge_level[263] <= stage_2b_edge[31:0] >= 32'd1526;
					stage_3_edge_level[262] <= stage_2b_edge[31:0] >= 32'd1572;
					stage_3_edge_level[261] <= stage_2b_edge[31:0] >= 32'd1619;
					stage_3_edge_level[260] <= stage_2b_edge[31:0] >= 32'd1668;
					stage_3_edge_level[259] <= stage_2b_edge[31:0] >= 32'd1718;
					stage_3_edge_level[258] <= stage_2b_edge[31:0] >= 32'd1769;
					stage_3_edge_level[257] <= stage_2b_edge[31:0] >= 32'd1822;
					stage_3_edge_level[256] <= stage_2b_edge[31:0] >= 32'd1877;
					stage_3_edge_level[255] <= stage_2b_edge[31:0] >= 32'd1933;
					stage_3_edge_level[254] <= stage_2b_edge[31:0] >= 32'd1991;
					stage_3_edge_level[253] <= stage_2b_edge[31:0] >= 32'd2051;
					stage_3_edge_level[252] <= stage_2b_edge[31:0] >= 32'd2113;
					stage_3_edge_level[251] <= stage_2b_edge[31:0] >= 32'd2176;
					stage_3_edge_level[250] <= stage_2b_edge[31:0] >= 32'd2241;
					stage_3_edge_level[249] <= stage_2b_edge[31:0] >= 32'd2309;
					stage_3_edge_level[248] <= stage_2b_edge[31:0] >= 32'd2378;
					stage_3_edge_level[247] <= stage_2b_edge[31:0] >= 32'd2449;
					stage_3_edge_level[246] <= stage_2b_edge[31:0] >= 32'd2523;
					stage_3_edge_level[245] <= stage_2b_edge[31:0] >= 32'd2598;
					stage_3_edge_level[244] <= stage_2b_edge[31:0] >= 32'd2676;
					stage_3_edge_level[243] <= stage_2b_edge[31:0] >= 32'd2757;
					stage_3_edge_level[242] <= stage_2b_edge[31:0] >= 32'd2839;
					stage_3_edge_level[241] <= stage_2b_edge[31:0] >= 32'd2924;
					stage_3_edge_level[240] <= stage_2b_edge[31:0] >= 32'd3012;
					stage_3_edge_level[239] <= stage_2b_edge[31:0] >= 32'd3103;
					stage_3_edge_level[238] <= stage_2b_edge[31:0] >= 32'd3196;
					stage_3_edge_level[237] <= stage_2b_edge[31:0] >= 32'd3292;
					stage_3_edge_level[236] <= stage_2b_edge[31:0] >= 32'd3390;
					stage_3_edge_level[235] <= stage_2b_edge[31:0] >= 32'd3492;
					stage_3_edge_level[234] <= stage_2b_edge[31:0] >= 32'd3597;
					stage_3_edge_level[233] <= stage_2b_edge[31:0] >= 32'd3705;
					stage_3_edge_level[232] <= stage_2b_edge[31:0] >= 32'd3816;
					stage_3_edge_level[231] <= stage_2b_edge[31:0] >= 32'd3930;
					stage_3_edge_level[230] <= stage_2b_edge[31:0] >= 32'd4048;
					stage_3_edge_level[229] <= stage_2b_edge[31:0] >= 32'd4170;
					stage_3_edge_level[228] <= stage_2b_edge[31:0] >= 32'd4295;
					stage_3_edge_level[227] <= stage_2b_edge[31:0] >= 32'd4424;
					stage_3_edge_level[226] <= stage_2b_edge[31:0] >= 32'd4556;
					stage_3_edge_level[225] <= stage_2b_edge[31:0] >= 32'd4693;
					stage_3_edge_level[224] <= stage_2b_edge[31:0] >= 32'd4834;
					stage_3_edge_level[223] <= stage_2b_edge[31:0] >= 32'd4979;
					stage_3_edge_level[222] <= stage_2b_edge[31:0] >= 32'd5128;
					stage_3_edge_level[221] <= stage_2b_edge[31:0] >= 32'd5282;
					stage_3_edge_level[220] <= stage_2b_edge[31:0] >= 32'd5440;
					stage_3_edge_level[219] <= stage_2b_edge[31:0] >= 32'd5604;
					stage_3_edge_level[218] <= stage_2b_edge[31:0] >= 32'd5772;
					stage_3_edge_level[217] <= stage_2b_edge[31:0] >= 32'd5945;
					stage_3_edge_level[216] <= stage_2b_edge[31:0] >= 32'd6123;
					stage_3_edge_level[215] <= stage_2b_edge[31:0] >= 32'd6307;
					stage_3_edge_level[214] <= stage_2b_edge[31:0] >= 32'd6496;
					stage_3_edge_level[213] <= stage_2b_edge[31:0] >= 32'd6691;
					stage_3_edge_level[212] <= stage_2b_edge[31:0] >= 32'd6892;
					stage_3_edge_level[211] <= stage_2b_edge[31:0] >= 32'd7099;
					stage_3_edge_level[210] <= stage_2b_edge[31:0] >= 32'd7311;
					stage_3_edge_level[209] <= stage_2b_edge[31:0] >= 32'd7531;
					stage_3_edge_level[208] <= stage_2b_edge[31:0] >= 32'd7757;
					stage_3_edge_level[207] <= stage_2b_edge[31:0] >= 32'd7989;
					stage_3_edge_level[206] <= stage_2b_edge[31:0] >= 32'd8229;
					stage_3_edge_level[205] <= stage_2b_edge[31:0] >= 32'd8476;
					stage_3_edge_level[204] <= stage_2b_edge[31:0] >= 32'd8730;
					stage_3_edge_level[203] <= stage_2b_edge[31:0] >= 32'd8992;
					stage_3_edge_level[202] <= stage_2b_edge[31:0] >= 32'd9262;
					stage_3_edge_level[201] <= stage_2b_edge[31:0] >= 32'd9540;
					stage_3_edge_level[200] <= stage_2b_edge[31:0] >= 32'd9826;
					stage_3_edge_level[199] <= stage_2b_edge[31:0] >= 32'd10121;
					stage_3_edge_level[198] <= stage_2b_edge[31:0] >= 32'd10424;
					stage_3_edge_level[197] <= stage_2b_edge[31:0] >= 32'd10737;
					stage_3_edge_level[196] <= stage_2b_edge[31:0] >= 32'd11059;
					stage_3_edge_level[195] <= stage_2b_edge[31:0] >= 32'd11391;
					stage_3_edge_level[194] <= stage_2b_edge[31:0] >= 32'd11733;
					stage_3_edge_level[193] <= stage_2b_edge[31:0] >= 32'd12085;
					stage_3_edge_level[192] <= stage_2b_edge[31:0] >= 32'd12447;
					stage_3_edge_level[191] <= stage_2b_edge[31:0] >= 32'd12821;
					stage_3_edge_level[190] <= stage_2b_edge[31:0] >= 32'd13205;
					stage_3_edge_level[189] <= stage_2b_edge[31:0] >= 32'd13601;
					stage_3_edge_level[188] <= stage_2b_edge[31:0] >= 32'd14010;
					stage_3_edge_level[187] <= stage_2b_edge[31:0] >= 32'd14430;
					stage_3_edge_level[186] <= stage_2b_edge[31:0] >= 32'd14863;
					stage_3_edge_level[185] <= stage_2b_edge[31:0] >= 32'd15309;
					stage_3_edge_level[184] <= stage_2b_edge[31:0] >= 32'd15768;
					stage_3_edge_level[183] <= stage_2b_edge[31:0] >= 32'd16241;
					stage_3_edge_level[182] <= stage_2b_edge[31:0] >= 32'd16728;
					stage_3_edge_level[181] <= stage_2b_edge[31:0] >= 32'd17230;
					stage_3_edge_level[180] <= stage_2b_edge[31:0] >= 32'd17747;
					stage_3_edge_level[179] <= stage_2b_edge[31:0] >= 32'd18279;
					stage_3_edge_level[178] <= stage_2b_edge[31:0] >= 32'd18828;
					stage_3_edge_level[177] <= stage_2b_edge[31:0] >= 32'd19392;
					stage_3_edge_level[176] <= stage_2b_edge[31:0] >= 32'd19974;
					stage_3_edge_level[175] <= stage_2b_edge[31:0] >= 32'd20573;
					stage_3_edge_level[174] <= stage_2b_edge[31:0] >= 32'd21191;
					stage_3_edge_level[173] <= stage_2b_edge[31:0] >= 32'd21826;
					stage_3_edge_level[172] <= stage_2b_edge[31:0] >= 32'd22481;
					stage_3_edge_level[171] <= stage_2b_edge[31:0] >= 32'd23156;
					stage_3_edge_level[170] <= stage_2b_edge[31:0] >= 32'd23850;
					stage_3_edge_level[169] <= stage_2b_edge[31:0] >= 32'd24566;
					stage_3_edge_level[168] <= stage_2b_edge[31:0] >= 32'd25303;
					stage_3_edge_level[167] <= stage_2b_edge[31:0] >= 32'd26062;
					stage_3_edge_level[166] <= stage_2b_edge[31:0] >= 32'd26844;
					stage_3_edge_level[165] <= stage_2b_edge[31:0] >= 32'd27649;
					stage_3_edge_level[164] <= stage_2b_edge[31:0] >= 32'd28478;
					stage_3_edge_level[163] <= stage_2b_edge[31:0] >= 32'd29333;
					stage_3_edge_level[162] <= stage_2b_edge[31:0] >= 32'd30213;
					stage_3_edge_level[161] <= stage_2b_edge[31:0] >= 32'd31119;
					stage_3_edge_level[160] <= stage_2b_edge[31:0] >= 32'd32053;
					stage_3_edge_level[159] <= stage_2b_edge[31:0] >= 32'd33014;
					stage_3_edge_level[158] <= stage_2b_edge[31:0] >= 32'd34005;
					stage_3_edge_level[157] <= stage_2b_edge[31:0] >= 32'd35025;
					stage_3_edge_level[156] <= stage_2b_edge[31:0] >= 32'd36076;
					stage_3_edge_level[155] <= stage_2b_edge[31:0] >= 32'd37158;
					stage_3_edge_level[154] <= stage_2b_edge[31:0] >= 32'd38273;
					stage_3_edge_level[153] <= stage_2b_edge[31:0] >= 32'd39421;
					stage_3_edge_level[152] <= stage_2b_edge[31:0] >= 32'd40604;
					stage_3_edge_level[151] <= stage_2b_edge[31:0] >= 32'd41822;
					stage_3_edge_level[150] <= stage_2b_edge[31:0] >= 32'd43076;
					stage_3_edge_level[149] <= stage_2b_edge[31:0] >= 32'd44369;
					stage_3_edge_level[148] <= stage_2b_edge[31:0] >= 32'd45700;
					stage_3_edge_level[147] <= stage_2b_edge[31:0] >= 32'd47071;
					stage_3_edge_level[146] <= stage_2b_edge[31:0] >= 32'd48483;
					stage_3_edge_level[145] <= stage_2b_edge[31:0] >= 32'd49937;
					stage_3_edge_level[144] <= stage_2b_edge[31:0] >= 32'd51435;
					stage_3_edge_level[143] <= stage_2b_edge[31:0] >= 32'd52978;
					stage_3_edge_level[142] <= stage_2b_edge[31:0] >= 32'd54568;
					stage_3_edge_level[141] <= stage_2b_edge[31:0] >= 32'd56205;
					stage_3_edge_level[140] <= stage_2b_edge[31:0] >= 32'd57891;
					stage_3_edge_level[139] <= stage_2b_edge[31:0] >= 32'd59628;
					stage_3_edge_level[138] <= stage_2b_edge[31:0] >= 32'd61416;
					stage_3_edge_level[137] <= stage_2b_edge[31:0] >= 32'd63259;
					stage_3_edge_level[136] <= stage_2b_edge[31:0] >= 32'd65157;
					stage_3_edge_level[135] <= stage_2b_edge[31:0] >= 32'd67111;
					stage_3_edge_level[134] <= stage_2b_edge[31:0] >= 32'd69125;
					stage_3_edge_level[133] <= stage_2b_edge[31:0] >= 32'd71199;
					stage_3_edge_level[132] <= stage_2b_edge[31:0] >= 32'd73334;
					stage_3_edge_level[131] <= stage_2b_edge[31:0] >= 32'd75535;
					stage_3_edge_level[130] <= stage_2b_edge[31:0] >= 32'd77801;
					stage_3_edge_level[129] <= stage_2b_edge[31:0] >= 32'd80135;
					stage_3_edge_level[128] <= stage_2b_edge[31:0] >= 32'd82539;
					stage_3_edge_level[127] <= stage_2b_edge[31:0] >= 32'd85015;
					stage_3_edge_level[126] <= stage_2b_edge[31:0] >= 32'd87565;
					stage_3_edge_level[125] <= stage_2b_edge[31:0] >= 32'd90192;
					stage_3_edge_level[124] <= stage_2b_edge[31:0] >= 32'd92898;
					stage_3_edge_level[123] <= stage_2b_edge[31:0] >= 32'd95685;
					stage_3_edge_level[122] <= stage_2b_edge[31:0] >= 32'd98555;
					stage_3_edge_level[121] <= stage_2b_edge[31:0] >= 32'd101512;
					stage_3_edge_level[120] <= stage_2b_edge[31:0] >= 32'd104557;
					stage_3_edge_level[119] <= stage_2b_edge[31:0] >= 32'd107694;
					stage_3_edge_level[118] <= stage_2b_edge[31:0] >= 32'd110925;
					stage_3_edge_level[117] <= stage_2b_edge[31:0] >= 32'd114253;
					stage_3_edge_level[116] <= stage_2b_edge[31:0] >= 32'd117680;
					stage_3_edge_level[115] <= stage_2b_edge[31:0] >= 32'd121211;
					stage_3_edge_level[114] <= stage_2b_edge[31:0] >= 32'd124847;
					stage_3_edge_level[113] <= stage_2b_edge[31:0] >= 32'd128592;
					stage_3_edge_level[112] <= stage_2b_edge[31:0] >= 32'd132450;
					stage_3_edge_level[111] <= stage_2b_edge[31:0] >= 32'd136424;
					stage_3_edge_level[110] <= stage_2b_edge[31:0] >= 32'd140516;
					stage_3_edge_level[109] <= stage_2b_edge[31:0] >= 32'd144732;
					stage_3_edge_level[108] <= stage_2b_edge[31:0] >= 32'd149074;
					stage_3_edge_level[107] <= stage_2b_edge[31:0] >= 32'd153546;
					stage_3_edge_level[106] <= stage_2b_edge[31:0] >= 32'd158152;
					stage_3_edge_level[105] <= stage_2b_edge[31:0] >= 32'd162897;
					stage_3_edge_level[104] <= stage_2b_edge[31:0] >= 32'd167784;
					stage_3_edge_level[103] <= stage_2b_edge[31:0] >= 32'd172817;
					stage_3_edge_level[102] <= stage_2b_edge[31:0] >= 32'd178002;
					stage_3_edge_level[101] <= stage_2b_edge[31:0] >= 32'd183342;
					stage_3_edge_level[100] <= stage_2b_edge[31:0] >= 32'd188842;
					stage_3_edge_level[99] <= stage_2b_edge[31:0] >= 32'd194508;
					stage_3_edge_level[98] <= stage_2b_edge[31:0] >= 32'd200343;
					stage_3_edge_level[97] <= stage_2b_edge[31:0] >= 32'd206353;
					stage_3_edge_level[96] <= stage_2b_edge[31:0] >= 32'd212544;
					stage_3_edge_level[95] <= stage_2b_edge[31:0] >= 32'd218920;
					stage_3_edge_level[94] <= stage_2b_edge[31:0] >= 32'd225488;
					stage_3_edge_level[93] <= stage_2b_edge[31:0] >= 32'd232252;
					stage_3_edge_level[92] <= stage_2b_edge[31:0] >= 32'd239220;
					stage_3_edge_level[91] <= stage_2b_edge[31:0] >= 32'd246396;
					stage_3_edge_level[90] <= stage_2b_edge[31:0] >= 32'd253788;
					stage_3_edge_level[89] <= stage_2b_edge[31:0] >= 32'd261402;
					stage_3_edge_level[88] <= stage_2b_edge[31:0] >= 32'd269244;
					stage_3_edge_level[87] <= stage_2b_edge[31:0] >= 32'd277321;
					stage_3_edge_level[86] <= stage_2b_edge[31:0] >= 32'd285641;
					stage_3_edge_level[85] <= stage_2b_edge[31:0] >= 32'd294210;
					stage_3_edge_level[84] <= stage_2b_edge[31:0] >= 32'd303037;
					stage_3_edge_level[83] <= stage_2b_edge[31:0] >= 32'd312128;
					stage_3_edge_level[82] <= stage_2b_edge[31:0] >= 32'd321491;
					stage_3_edge_level[81] <= stage_2b_edge[31:0] >= 32'd331136;
					stage_3_edge_level[80] <= stage_2b_edge[31:0] >= 32'd341070;
					stage_3_edge_level[79] <= stage_2b_edge[31:0] >= 32'd351302;
					stage_3_edge_level[78] <= stage_2b_edge[31:0] >= 32'd361841;
					stage_3_edge_level[77] <= stage_2b_edge[31:0] >= 32'd372697;
					stage_3_edge_level[76] <= stage_2b_edge[31:0] >= 32'd383878;
					stage_3_edge_level[75] <= stage_2b_edge[31:0] >= 32'd395394;
					stage_3_edge_level[74] <= stage_2b_edge[31:0] >= 32'd407256;
					stage_3_edge_level[73] <= stage_2b_edge[31:0] >= 32'd419473;
					stage_3_edge_level[72] <= stage_2b_edge[31:0] >= 32'd432058;
					stage_3_edge_level[71] <= stage_2b_edge[31:0] >= 32'd445019;
					stage_3_edge_level[70] <= stage_2b_edge[31:0] >= 32'd458370;
					stage_3_edge_level[69] <= stage_2b_edge[31:0] >= 32'd472121;
					stage_3_edge_level[68] <= stage_2b_edge[31:0] >= 32'd486285;
					stage_3_edge_level[67] <= stage_2b_edge[31:0] >= 32'd500873;
					stage_3_edge_level[66] <= stage_2b_edge[31:0] >= 32'd515899;
					stage_3_edge_level[65] <= stage_2b_edge[31:0] >= 32'd531376;
					stage_3_edge_level[64] <= stage_2b_edge[31:0] >= 32'd547318;
					stage_3_edge_level[63] <= stage_2b_edge[31:0] >= 32'd563737;
					stage_3_edge_level[62] <= stage_2b_edge[31:0] >= 32'd580649;
					stage_3_edge_level[61] <= stage_2b_edge[31:0] >= 32'd598069;
					stage_3_edge_level[60] <= stage_2b_edge[31:0] >= 32'd616011;
					stage_3_edge_level[59] <= stage_2b_edge[31:0] >= 32'd634491;
					stage_3_edge_level[58] <= stage_2b_edge[31:0] >= 32'd653526;
					stage_3_edge_level[57] <= stage_2b_edge[31:0] >= 32'd673132;
					stage_3_edge_level[56] <= stage_2b_edge[31:0] >= 32'd693326;
					stage_3_edge_level[55] <= stage_2b_edge[31:0] >= 32'd714125;
					stage_3_edge_level[54] <= stage_2b_edge[31:0] >= 32'd735549;
					stage_3_edge_level[53] <= stage_2b_edge[31:0] >= 32'd757616;
					stage_3_edge_level[52] <= stage_2b_edge[31:0] >= 32'd780344;
					stage_3_edge_level[51] <= stage_2b_edge[31:0] >= 32'd803754;
					stage_3_edge_level[50] <= stage_2b_edge[31:0] >= 32'd827867;
					stage_3_edge_level[49] <= stage_2b_edge[31:0] >= 32'd852703;
					stage_3_edge_level[48] <= stage_2b_edge[31:0] >= 32'd878284;
					stage_3_edge_level[47] <= stage_2b_edge[31:0] >= 32'd904633;
					stage_3_edge_level[46] <= stage_2b_edge[31:0] >= 32'd931772;
					stage_3_edge_level[45] <= stage_2b_edge[31:0] >= 32'd959725;
					stage_3_edge_level[44] <= stage_2b_edge[31:0] >= 32'd988517;
					stage_3_edge_level[43] <= stage_2b_edge[31:0] >= 32'd1018172;
					stage_3_edge_level[42] <= stage_2b_edge[31:0] >= 32'd1048717;
					stage_3_edge_level[41] <= stage_2b_edge[31:0] >= 32'd1080179;
					stage_3_edge_level[40] <= stage_2b_edge[31:0] >= 32'd1112584;
					stage_3_edge_level[39] <= stage_2b_edge[31:0] >= 32'd1145962;
					stage_3_edge_level[38] <= stage_2b_edge[31:0] >= 32'd1180340;
					stage_3_edge_level[37] <= stage_2b_edge[31:0] >= 32'd1215751;
					stage_3_edge_level[36] <= stage_2b_edge[31:0] >= 32'd1252223;
					stage_3_edge_level[35] <= stage_2b_edge[31:0] >= 32'd1289790;
					stage_3_edge_level[34] <= stage_2b_edge[31:0] >= 32'd1328484;
					stage_3_edge_level[33] <= stage_2b_edge[31:0] >= 32'd1368338;
					stage_3_edge_level[32] <= stage_2b_edge[31:0] >= 32'd1409388;
					stage_3_edge_level[31] <= stage_2b_edge[31:0] >= 32'd1451670;
					stage_3_edge_level[30] <= stage_2b_edge[31:0] >= 32'd1495220;
					stage_3_edge_level[29] <= stage_2b_edge[31:0] >= 32'd1540077;
					stage_3_edge_level[28] <= stage_2b_edge[31:0] >= 32'd1586279;
					stage_3_edge_level[27] <= stage_2b_edge[31:0] >= 32'd1633867;
					stage_3_edge_level[26] <= stage_2b_edge[31:0] >= 32'd1682883;
					stage_3_edge_level[25] <= stage_2b_edge[31:0] >= 32'd1733370;
					stage_3_edge_level[24] <= stage_2b_edge[31:0] >= 32'd1785371;
					stage_3_edge_level[23] <= stage_2b_edge[31:0] >= 32'd1838932;
					stage_3_edge_level[22] <= stage_2b_edge[31:0] >= 32'd1894100;
					stage_3_edge_level[21] <= stage_2b_edge[31:0] >= 32'd1950923;
					stage_3_edge_level[20] <= stage_2b_edge[31:0] >= 32'd2009451;
					stage_3_edge_level[19] <= stage_2b_edge[31:0] >= 32'd2069734;
					stage_3_edge_level[18] <= stage_2b_edge[31:0] >= 32'd2131826;
					stage_3_edge_level[17] <= stage_2b_edge[31:0] >= 32'd2195781;
					stage_3_edge_level[16] <= stage_2b_edge[31:0] >= 32'd2261654;
					stage_3_edge_level[15] <= stage_2b_edge[31:0] >= 32'd2329504;
					stage_3_edge_level[14] <= stage_2b_edge[31:0] >= 32'd2399389;
					stage_3_edge_level[13] <= stage_2b_edge[31:0] >= 32'd2471371;
					stage_3_edge_level[12] <= stage_2b_edge[31:0] >= 32'd2545512;
					stage_3_edge_level[11] <= stage_2b_edge[31:0] >= 32'd2621877;
					stage_3_edge_level[10] <= stage_2b_edge[31:0] >= 32'd2700534;
					stage_3_edge_level[9] <= stage_2b_edge[31:0] >= 32'd2781550;
					stage_3_edge_level[8] <= stage_2b_edge[31:0] >= 32'd2864996;
					stage_3_edge_level[7] <= stage_2b_edge[31:0] >= 32'd2950946;
					stage_3_edge_level[6] <= stage_2b_edge[31:0] >= 32'd3039474;
					stage_3_edge_level[5] <= stage_2b_edge[31:0] >= 32'd3130659;
					stage_3_edge_level[4] <= stage_2b_edge[31:0] >= 32'd3224578;
					stage_3_edge_level[3] <= stage_2b_edge[31:0] >= 32'd3321316;
					stage_3_edge_level[2] <= stage_2b_edge[31:0] >= 32'd3420955;
					stage_3_edge_level[1] <= stage_2b_edge[31:0] >= 32'd3523584;
					stage_3_edge_level[0] <= stage_2b_edge[31:0] >= 32'd3629291;
				end
			end
		end
	end
end

endgenerate






// ----- 4.) compute matching ----- //
// to stage 5
reg [512-1:0] stage_4_cl;
reg [2**L_BITS-1: 0] stage_4_mcm_bits_u;
reg [2**L_BITS-1: 0] stage_4_mcm_bits_v;
reg [128-1:0] stage_4_edge;
reg [2**L_BITS-1:0] stage_4_take_edge;
reg stage_4_valid;
reg [9:0] stage_4_offset;
reg [K_BITS-1:0]	stage_4_bram_address_u;
reg [K_BITS-1:0]	stage_4_bram_address_v;
reg 				stage_4_bram_address_v_valid;

reg [K_BITS-1:0] 	stage_4_bram_address_next_v;
reg					stage_4_bram_address_next_v_valid;

// due to the pipelined read "with delay", we have not the most up to date value read from the BRAM. To overcome this problem, we store the last values of u 
reg [2**L_BITS-1:0] stage_4_old_mcm_bits_u_0;
reg [2**L_BITS-1:0] stage_4_old_mcm_bits_u_1;
reg [2**L_BITS-1:0] stage_4_old_mcm_bits_u_2;
reg [2**L_BITS-1:0] stage_4_old_mcm_bits_u_3;
reg [2**L_BITS-1:0] stage_4_old_mcm_bits_u_4;

reg [31:0] stage_4_old_u_0;
reg [31:0] stage_4_old_u_1;
reg [31:0] stage_4_old_u_2;
reg [31:0] stage_4_old_u_3;
reg [31:0] stage_4_old_u_4;

// also keep the last values of v - since the value might come from the BRAM, we also have to store the lastest 2 values
reg [2**L_BITS-1:0] stage_4_old_mcm_bits_v_0;
reg [2**L_BITS-1:0] stage_4_old_mcm_bits_v_1;
reg [2**L_BITS-1:0] stage_4_old_mcm_bits_v_2;
reg [2**L_BITS-1:0] stage_4_old_mcm_bits_v_3;
reg [2**L_BITS-1:0] stage_4_old_mcm_bits_v_4;

reg [31:0] stage_4_old_v_0;
reg [31:0] stage_4_old_v_1;
reg [31:0] stage_4_old_v_2;
reg [31:0] stage_4_old_v_3;
reg [31:0] stage_4_old_v_4;

// stage state variables
reg stage_4_done;

/*
reg [2**L_BITS-1:0] debug_reg_u[0:4];
reg [2**L_BITS-1:0] debug_reg_v[0:4];

reg 				debug_reg_binary_u;
reg					debug_reg_binary_v;
*/


always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
        stage_4_old_mcm_bits_u_0 <= 0;
		stage_4_old_mcm_bits_u_1 <= 0;
		stage_4_old_mcm_bits_u_2 <= 0;
		stage_4_old_mcm_bits_u_3 <= 0;
		stage_4_old_mcm_bits_u_4 <= 0;
        stage_4_old_mcm_bits_v_0 <= 0;
		stage_4_old_mcm_bits_v_1 <= 0;
		stage_4_old_mcm_bits_v_2 <= 0;
		stage_4_old_mcm_bits_v_3 <= 0;
		stage_4_old_mcm_bits_v_4 <= 0;

        stage_4_old_u_0 <= 0;
		stage_4_old_u_1 <= 0;
		stage_4_old_u_2 <= 0;
		stage_4_old_u_3 <= 0;
		stage_4_old_u_4 <= 0;
        stage_4_old_v_0 <= 0;
		stage_4_old_v_1 <= 0;
		stage_4_old_v_2 <= 0;
		stage_4_old_v_3 <= 0;
		stage_4_old_v_4 <= 0;

		stage_4_done <= 0;
		stage_4_edges_processed <= 0;
		stage_4_valid <= 0;
	end
	else begin
		stage_4_valid <= 0;
		if( stage_4_edges_processed == num_edges) begin
			stage_4_done <= 1;
		end

        if( edge_processor_done || stage_4_done || ~stage_3_valid ) begin
		    // if this is the case, then do nothing: we stall the pipeline or are already done
        end
        else begin
				stage_4_edges_processed <= stage_4_edges_processed + 1 ;
	
            	stage_4_valid <= 1;
                stage_4_cl <= stage_3_cl;

				// matching bits for "u"
                stage_4_mcm_bits_u <=  ( stage_3_mcm_bits_u | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
															 | ( ~( stage_3_mcm_bits_u | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
                                        				& ~( stage_3_mcm_bits_v | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
														& stage_3_edge_level );

				stage_4_old_mcm_bits_u_4 <= stage_4_old_mcm_bits_u_3;
				stage_4_old_mcm_bits_u_3 <= stage_4_old_mcm_bits_u_2;
				stage_4_old_mcm_bits_u_2 <= stage_4_old_mcm_bits_u_1;
				stage_4_old_mcm_bits_u_1 <= stage_4_old_mcm_bits_u_0;
                stage_4_old_mcm_bits_u_0 <= ( stage_3_mcm_bits_u | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
															 | ( ~( stage_3_mcm_bits_u | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
                                        				& ~( stage_3_mcm_bits_v | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
														& stage_3_edge_level );


				stage_4_old_u_4 <= stage_4_old_u_3;
				stage_4_old_u_3 <= stage_4_old_u_2;
				stage_4_old_u_2 <= stage_4_old_u_1;
                stage_4_old_u_1 <= stage_4_old_u_0;
				stage_4_old_u_0 <= stage_3_edge[95:64];

                // matching bits for "v"        
                stage_4_mcm_bits_v <=  ( stage_3_mcm_bits_v | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
														 | ( ~( stage_3_mcm_bits_u | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
                                        				& ~( stage_3_mcm_bits_v | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
														& stage_3_edge_level );

				stage_4_old_mcm_bits_v_4 <= stage_4_old_mcm_bits_v_3;
				stage_4_old_mcm_bits_v_3 <= stage_4_old_mcm_bits_v_2;
				stage_4_old_mcm_bits_v_2 <= stage_4_old_mcm_bits_v_1;
                stage_4_old_mcm_bits_v_1 <= stage_4_old_mcm_bits_v_0;
				stage_4_old_mcm_bits_v_0 <= ( stage_3_mcm_bits_v | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
														 | ( ~( stage_3_mcm_bits_u | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
                                        				& ~( stage_3_mcm_bits_v | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
														& stage_3_edge_level );
				
				stage_4_old_v_4 <= stage_4_old_v_3;
				stage_4_old_v_3 <= stage_4_old_v_2;
				stage_4_old_v_2 <= stage_4_old_v_1;
                stage_4_old_v_1 <= stage_4_old_v_0;
				stage_4_old_v_0 <= stage_3_edge[63:32];

				// the take_edge (the level to which we can add this edge) is given by: ~( the union of the mcm-bits for "u" ) & ~(the union of the mcm-bits for "v") & (the level on which the edge is valid)
                stage_4_take_edge <= ( ~( stage_3_mcm_bits_u | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
                                        				& ~( stage_3_mcm_bits_v | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_u_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_2 ) |
																( {2**L_BITS{(stage_4_old_u_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_3 ) |
																( {2**L_BITS{(stage_4_old_u_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_4 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_1 ) |
																( {2**L_BITS{(stage_4_old_v_2 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_2 ) |
																( {2**L_BITS{(stage_4_old_v_3 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_3 ) |
																( {2**L_BITS{(stage_4_old_v_4 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_4 )
                                        						)
														& stage_3_edge_level );

/*
				debug_reg_u <= stage_3_mcm_bits_u | ( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[95:64])}} & stage_4_old_mcm_bits_v_1 ) ;
				debug_reg_v <= stage_3_mcm_bits_v | 
																( {2**L_BITS{(stage_4_old_u_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_0 ) |
                                            					( {2**L_BITS{(stage_4_old_u_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_u_1 ) |
																( {2**L_BITS{(stage_4_old_v_0 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_0 ) |
																( {2**L_BITS{(stage_4_old_v_1 == stage_3_edge[63:32])}} & stage_4_old_mcm_bits_v_1 );
*/

/*
				debug_reg_binary_u[0] <= (stage_4_old_u_1 == stage_3_edge[95:64]);
				debug_reg_binary_u[0] <= (stage_4_old_u_1 == stage_3_edge[95:64]);
				debug_reg_binary_v[0] <= (stage_4_old_v_1 == stage_3_edge[63:32]);
*/

				stage_4_offset <= stage_3_offset;
		        stage_4_edge <= stage_3_edge;

				stage_4_bram_address_u <= stage_3_bram_address_u;
				stage_4_bram_address_v <= stage_3_bram_address_v;
				stage_4_bram_address_v_valid <= stage_3_bram_address_v_valid;

				stage_4_bram_address_next_v_valid <= ( (stage_3_edge [ 63:32] -1 ) >= ( ( stage_3_edge[127:96] +1) << K_BITS  ) )  && ( (stage_3_edge [ 63:32] -1)  <  ( ( stage_3_edge[127:96] +2) << K_BITS ) ) ;
				stage_4_bram_address_next_v <= stage_3_edge [ 63:32] -1; // -  ( ( stage_3_edge[127:96] +1) << K_BITS );
        end
    end

end

// ----- 5.) write u back to BRAM ----- //

// to stage 6
reg [511:0] stage_5_cl;
reg [511:0]	stage_5_cl_original;
reg [9:0] stage_5_offset;
reg [127:0] stage_5_edge;
reg [511:0] stage_5_one_hot_take_edge;

reg stage_5_valid;
reg stage_5_done;

reg [2**L_BITS-1:0] stage_5_debug_bits_written_out;
reg [31:0]		 	stage_5_debug_address_offset;
reg [31:0]		 	stage_5_debug_cache_line_offset;
reg					stage_5_debug_data_valid;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		// local vars
        stage_5_valid <= 0;
		stage_5_done <= 0;
		stage_5_edges_processed <= 0;

		// local buffer vars
		buf_mem_valid <= 0;
		num_double_bram_writes <= 0;
		bram1_wr_ptr <= 0;
		bram2_wr_ptr <= 0;

		if( (um_done || ~um_activated) && rst_n ) begin
			// reset bram
			bram_wu_ru_waddr <= bram_wu_ru_waddr + 1;
			bram_wu_rv_waddr <= bram_wu_rv_waddr + 1;
			bram_wv_ru_waddr <= bram_wv_ru_waddr + 1;
			bram_wv_rv_waddr <= bram_wv_rv_waddr + 1;

			bram_2_wu_ru_waddr <= bram_2_wu_ru_waddr + 1;
			bram_2_wu_rv_waddr <= bram_2_wu_rv_waddr + 1;
			bram_2_wv_ru_waddr <= bram_2_wv_ru_waddr + 1;
			bram_2_wv_rv_waddr <= bram_2_wv_rv_waddr + 1;

			bram_wu_ru_din <= 0;
			bram_wu_rv_din <= 0;
			bram_wv_ru_din <= 0;
			bram_wv_rv_din <= 0;

			bram_2_wu_ru_din <= 0;
			bram_2_wu_rv_din <= 0;
			bram_2_wv_ru_din <= 0;
			bram_2_wv_rv_din <= 0;

			bram_wu_ru_we <= 1;
			bram_wu_rv_we <= 1;
			bram_wv_ru_we <= 1;
			bram_wv_rv_we <= 1;

			bram_2_wu_ru_we <= 1;
			bram_2_wu_rv_we <= 1;
			bram_2_wv_ru_we <= 1;
			bram_2_wv_rv_we <= 1;
		end
		else begin
			// bram value initialization
			bram_wu_ru_waddr <= 0;
			bram_wu_rv_waddr <= 0;
			bram_wv_ru_waddr <= 0;
			bram_wv_rv_waddr <= 0;

			bram_2_wu_ru_waddr <= 0;
			bram_2_wu_rv_waddr <= 0;
			bram_2_wv_ru_waddr <= 0;
			bram_2_wv_rv_waddr <= 0;

			bram_wu_ru_we <= 0;
			bram_wu_rv_we <= 0;
			bram_wv_ru_we <= 0;
			bram_wv_rv_we <= 0;

			bram_2_wu_ru_we <= 0;
			bram_2_wu_rv_we <= 0;
			bram_2_wv_ru_we <= 0;
			bram_2_wv_rv_we <= 0;
		end

		mcmbram_ntotal_read <= 0;

		stage_5_debug_bits_written_out <= 0;
		stage_5_debug_address_offset <= 0;
		stage_5_debug_cache_line_offset <= 0;
		stage_5_debug_data_valid <= 0;
		stage_5_one_hot_take_edge <= 0;

		
	end
    else begin

		bram_wu_ru_we <= 0;
		bram_wu_rv_we <= 0;
		bram_wv_ru_we <= 0;
		bram_wv_rv_we <= 0;

		bram_2_wu_ru_we <= 0;
		bram_2_wu_rv_we <= 0;
		bram_2_wv_ru_we <= 0;
		bram_2_wv_rv_we <= 0;


		stage_5_valid <= 0;

		stage_5_debug_bits_written_out <= 0;
		stage_5_debug_address_offset <= 0;
		stage_5_debug_cache_line_offset <= 0;
		stage_5_debug_data_valid <= 0;

		if( stage_5_edges_processed == num_edges) begin
			stage_5_done <= 1;
		end

		// ignore stage_4_valid, because we also handle the updates of the bram-mcm-bits for the double buffer in this stage!!!
		if( global_epoch_change ) begin
			buf_mem_valid <= 0;
			bram1_wr_ptr <= 0;
			bram2_wr_ptr <= 0;
		end
        else if( edge_processor_done || stage_5_done ) begin 
		    // if this is the case, then do nothing: we stall the pipeline or are already done
		end
        else begin
            
				if( bram_selector ) begin
					if( stage_4_valid ) begin
						bram_wu_ru_we <= 1;
						bram_wu_rv_we <= 1;

						bram_wu_ru_waddr <= stage_4_bram_address_u;
						bram_wu_rv_waddr <= stage_4_bram_address_u;
						
						bram_wu_ru_din <= stage_4_mcm_bits_u;
						bram_wu_rv_din <= stage_4_mcm_bits_u;
						
						if( stage_4_bram_address_v_valid ) begin
							bram_wv_ru_we <= 1;
							bram_wv_rv_we <= 1;
							
							bram_wv_ru_waddr <= stage_4_bram_address_v;
							bram_wv_rv_waddr <= stage_4_bram_address_v;

							bram_wv_ru_din <= stage_4_mcm_bits_v;
							bram_wv_rv_din <= stage_4_mcm_bits_v;
						end
					end

					// write to the bram2, if this is "v" a valid candiate for the next bram
					if( stage_4_valid && stage_4_bram_address_next_v_valid ) begin
							bram_2_wu_ru_we <= 1;
							bram_2_wu_rv_we <= 1;
							bram_2_wv_ru_we <= 1;
							bram_2_wv_rv_we <= 1;
				
							bram_2_wu_ru_din <= stage_4_mcm_bits_v;
							bram_2_wu_rv_din <= stage_4_mcm_bits_v;
							bram_2_wv_ru_din <= stage_4_mcm_bits_v;
							bram_2_wv_rv_din <= stage_4_mcm_bits_v;
			
							bram_2_wu_ru_waddr <= stage_4_bram_address_next_v;
							bram_2_wu_rv_waddr <= stage_4_bram_address_next_v;
							bram_2_wv_ru_waddr <= stage_4_bram_address_next_v;
							bram_2_wv_rv_waddr <= stage_4_bram_address_next_v;
							
							buf_mem_valid[ stage_4_bram_address_next_v  ] <= 1;
					end
					else begin
						if( mcm_bram_data_buffer_1_valid ) begin
							mcmbram_ntotal_read <= mcmbram_ntotal_read +1;
							bram2_wr_ptr <= bram2_wr_ptr +1;
							num_double_bram_writes <= num_double_bram_writes+ 1;

							if( buf_mem_valid [ bram2_wr_ptr ] == 0) begin
								bram_2_wu_ru_we <= 1;
								bram_2_wu_rv_we <= 1;
								bram_2_wv_ru_we <= 1;
								bram_2_wv_rv_we <= 1;

								bram_2_wu_ru_din <= mcm_bram_data_buffer_1;
								bram_2_wu_rv_din <= mcm_bram_data_buffer_1;
								bram_2_wv_ru_din <= mcm_bram_data_buffer_1;
								bram_2_wv_rv_din <= mcm_bram_data_buffer_1;

								bram_2_wu_ru_waddr <= bram2_wr_ptr;
								bram_2_wu_rv_waddr <= bram2_wr_ptr;
								bram_2_wv_ru_waddr <= bram2_wr_ptr;
								bram_2_wv_rv_waddr <= bram2_wr_ptr;
							end
						end
					end
				end // END OF bram_selector
				else begin
					if( stage_4_valid ) begin
						bram_2_wu_ru_we <= 1;
						bram_2_wu_rv_we <= 1;
		
						bram_2_wu_ru_waddr <= stage_4_bram_address_u;
						bram_2_wu_rv_waddr <= stage_4_bram_address_u;

						bram_2_wu_ru_din <= stage_4_mcm_bits_u;
						bram_2_wu_rv_din <= stage_4_mcm_bits_u;
						if( stage_4_bram_address_v_valid ) begin
							bram_2_wv_ru_we <= 1;
							bram_2_wv_rv_we <= 1;
		
							bram_2_wv_ru_waddr <= stage_4_bram_address_v;
							bram_2_wv_rv_waddr <= stage_4_bram_address_v;

							bram_2_wv_ru_din <= stage_4_mcm_bits_v;
							bram_2_wv_rv_din <= stage_4_mcm_bits_v;
						end
					end
					
					// write to the bram1, if this is a valid candiate for the next bram
					if( stage_4_valid && stage_4_bram_address_next_v_valid ) begin

								bram_wu_ru_we <= 1;
								bram_wu_rv_we <= 1;
								bram_wv_ru_we <= 1;
								bram_wv_rv_we <= 1;

								bram_wu_ru_din <= stage_4_mcm_bits_v;
								bram_wu_rv_din <= stage_4_mcm_bits_v;
								bram_wv_ru_din <= stage_4_mcm_bits_v;
								bram_wv_rv_din <= stage_4_mcm_bits_v;

								bram_wu_ru_waddr <= stage_4_bram_address_next_v;
								bram_wu_rv_waddr <= stage_4_bram_address_next_v;
								bram_wv_ru_waddr <= stage_4_bram_address_next_v;
								bram_wv_rv_waddr <= stage_4_bram_address_next_v;

							buf_mem_valid[ stage_4_bram_address_next_v  ] <= 1;
					end
					else begin
						if( mcm_bram_data_buffer_1_valid ) begin
							mcmbram_ntotal_read <= mcmbram_ntotal_read +1;
							bram1_wr_ptr <= bram1_wr_ptr +1;
							num_double_bram_writes <= num_double_bram_writes  + 1;

							if( buf_mem_valid [ bram1_wr_ptr] == 0) begin

								bram_wu_ru_we <= 1;
								bram_wu_rv_we <= 1;
								bram_wv_ru_we <= 1;
								bram_wv_rv_we <= 1;

								bram_wu_ru_din <= mcm_bram_data_buffer_1;
								bram_wu_rv_din <= mcm_bram_data_buffer_1;
								bram_wv_ru_din <= mcm_bram_data_buffer_1;
								bram_wv_rv_din <= mcm_bram_data_buffer_1;

								bram_wu_ru_waddr <= bram1_wr_ptr;
								bram_wu_rv_waddr <= bram1_wr_ptr;
								bram_wv_ru_waddr <= bram1_wr_ptr;
								bram_wv_rv_waddr <= bram1_wr_ptr;
							end
						end
					end
				end

				// update the mcm-cache-line for later write-back
				if( stage_4_valid) begin

					stage_5_edges_processed <= stage_5_edges_processed  + 1;
		            stage_5_cl <= stage_4_cl | ( stage_4_mcm_bits_v << (stage_4_offset << L_BITS) ) ;
					stage_5_cl_original <= stage_4_cl;
		            stage_5_offset <= stage_4_offset;
		            stage_5_edge <= stage_4_edge;
		            stage_5_valid <= 1;
					stage_5_one_hot_take_edge <= stage_4_take_edge; // -stage_4_take_edge & stage_4_take_edge;
					
					stage_5_debug_bits_written_out <= stage_4_mcm_bits_v;
					stage_5_debug_address_offset <=  (( (stage_4_edge[63:32] -1 ) << K_BITS)  >> 9 );
					stage_5_debug_cache_line_offset <= stage_4_offset;
					stage_5_debug_data_valid <= 1;
				end
        end
    end

end

// ----- 5.) Handle the data we get for the double-buffering ----- //
always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin

		// drain queues
		if( (um_done || ~um_activated) && rst_n ) begin
			mcm_bram_m_axis_ready <= 1;
		end
		else begin
			mcm_bram_m_axis_ready <= 0;
		end
		
		mcm_bram_data_buffer_1_valid <= 0;
		mcm_bram_data_buffer_2_valid <= 0;
	end
    else begin
		if( !( global_epoch_change )
			&& 
			!( edge_processor_done || stage_5_done )
			&& 
			( 	(bram_selector && !(stage_4_valid && stage_4_bram_address_next_v_valid) && mcm_bram_data_buffer_1_valid)  
				||
				(~bram_selector && !(stage_4_valid && stage_4_bram_address_next_v_valid ) && mcm_bram_data_buffer_1_valid)
			)
			) begin
			// handle consume

			if( mcm_bram_data_buffer_2_valid ) begin
				mcm_bram_data_buffer_1 <= mcm_bram_data_buffer_2;
				if( mcm_bram_m_axis_ready &&  mcm_bram_m_axis_valid ) begin
					mcm_bram_data_buffer_2 <= mcm_bram_m_axis_data;
					mcm_bram_m_axis_ready <= 0;
				end
				else begin
					mcm_bram_data_buffer_2_valid <= 0;
					mcm_bram_m_axis_ready <= 1;
				end
			end
			else begin
				mcm_bram_m_axis_ready <= 1;
				if( mcm_bram_m_axis_ready &&  mcm_bram_m_axis_valid ) begin
					mcm_bram_data_buffer_1 <= mcm_bram_m_axis_data;
				end
				else begin
					mcm_bram_data_buffer_1_valid <= 0;
				end
			end
		end
		else begin
			// handle add
			if( mcm_bram_m_axis_ready &&  mcm_bram_m_axis_valid ) begin
				if( ~mcm_bram_data_buffer_1_valid ) begin
					mcm_bram_data_buffer_1 <= mcm_bram_m_axis_data;
					mcm_bram_data_buffer_1_valid <= 1;
				end
				else if( ~mcm_bram_data_buffer_2_valid ) begin
					mcm_bram_data_buffer_2 <= mcm_bram_m_axis_data;
					mcm_bram_data_buffer_2_valid <= 1;
					mcm_bram_m_axis_ready <= 0;
				end
			end
			else if( ~mcm_bram_data_buffer_1_valid || ~mcm_bram_data_buffer_2_valid ) begin
				mcm_bram_m_axis_ready <= 1;
			end
		end
    end
end


// ----- 5b) Convert one-hot to the binary value ----- //
reg				stage_5b_valid;
reg [511:0]		stage_5b_cl;
reg [511:0]		stage_5b_cl_original;
reg [127:0]		stage_5b_edge;
wire [8:0]		b1;
wire [8:0]		b2;
wire [8:0]		b3;
wire [8:0]		b4;
wire [8:0]		b5;
wire [8:0]		b6;
wire [8:0]		b7;
wire [8:0]		b8;
wire [8:0]		b9;
wire [8:0]		b10;
wire [8:0]		b11;
wire [8:0]		b12;
wire [8:0]		b13;
wire [8:0]		b14;
wire [8:0]		b15;
wire [8:0]		b16;

wire				v1;
wire				v2;
wire				v3;
wire				v4;
wire				v5;
wire				v6;
wire				v7;
wire				v8;
wire				v9;
wire				v10;
wire				v11;
wire				v12;
wire				v13;
wire				v14;
wire				v15;
wire				v16;
always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		stage_5b_valid <= 0;
	end
	else begin
		stage_5b_valid <= stage_5_valid;
		stage_5b_cl <= stage_5_cl;
		stage_5b_edge <= stage_5_edge;
		stage_5b_cl_original <= stage_5_cl_original;
	end
end

one_hot_to_binary #(.OFFSET(480)) ohtb1( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated),  	.one_hot(stage_5_one_hot_take_edge[31+  0:0+  0]), .binary(b1), .valid(v1));
one_hot_to_binary #(.OFFSET(448)) ohtb2( .clk(clk),	.rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+ 32:0+ 32]), .binary(b2), .valid(v2));
one_hot_to_binary #(.OFFSET(416)) ohtb3( .clk(clk),	.rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+ 64:0+ 64]), .binary(b3), .valid(v3));
one_hot_to_binary #(.OFFSET(384)) ohtb4( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+ 96:0+ 96]), .binary(b4), .valid(v4));
one_hot_to_binary #(.OFFSET(352)) ohtb5( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+128:0+128]), .binary(b5), .valid(v5));
one_hot_to_binary #(.OFFSET(320)) ohtb6( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+160:0+160]), .binary(b6), .valid(v6));
one_hot_to_binary #(.OFFSET(288)) ohtb7( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+192:0+192]), .binary(b7), .valid(v7));
one_hot_to_binary #(.OFFSET(256)) ohtb8( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+224:0+224]), .binary(b8), .valid(v8));
one_hot_to_binary #(.OFFSET(224)) ohtb9( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+256:0+256]), .binary(b9), .valid(v9));
one_hot_to_binary #(.OFFSET(192)) ohtb10( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+288:0+288]), .binary(b10), .valid(v10));
one_hot_to_binary #(.OFFSET(160)) ohtb11( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+320:0+320]), .binary(b11), .valid(v11));
one_hot_to_binary #(.OFFSET(128)) ohtb12( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+352:0+352]), .binary(b12), .valid(v12));
one_hot_to_binary #(.OFFSET(96 )) ohtb13( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+384:0+384]), .binary(b13), .valid(v13));
one_hot_to_binary #(.OFFSET(64 )) ohtb14( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+416:0+416]), .binary(b14), .valid(v14));
one_hot_to_binary #(.OFFSET(32 )) ohtb15( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated),	.one_hot(stage_5_one_hot_take_edge[31+448:0+448]), .binary(b15), .valid(v15));
one_hot_to_binary #(.OFFSET(0  )) ohtb16( .clk(clk), .rst_n(rst_n), .um_done(um_done), .um_activated(um_activated), 	.one_hot(stage_5_one_hot_take_edge[31+480:0+480]), .binary(b16), .valid(v16));


// ----- 6.) write back to RAM: write edges immediately to the cl buffer (write, if we have filled), if they are taken. Write back matching bits, if it belongs to the last edge of the epoch, or it is a new cache line ----- //


reg stage_6_done;
reg[31:0] stage_6_num_edges_written;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		// local vars
		stage_6_done <= 0;
		stage_6_edges_processed <= 0;

		// outgoing vars
		queue_edgeout_write_enable <= 0;
		queue_mcmout_write_enable <= 0;
	
		stage_6_num_edges_written<= 0;

		queue_edgeout_buf_in <= 0;

	end
    else begin
        queue_edgeout_write_enable <= 0;
		queue_mcmout_write_enable <= 0;
		if( stage_6_edges_processed == num_edges) begin
			stage_6_done <= 1;
		end

        if( edge_processor_done || stage_6_done || ~stage_5b_valid ) begin
		    // if this is the case, then do nothing: we stall the pipeline or are already done
        end
        else begin

				stage_6_edges_processed <= stage_6_edges_processed + 1;

                // write back the edge
                if( v1 || v2 || v3 || v4 || v5 || v6 || v7 || v8 || v9 || v10 || v11 || v12 || v13 || v14 || v15 || v16) begin
					stage_6_num_edges_written <= stage_6_num_edges_written + 1;
                    
					
                    queue_edgeout_buf_in[127:96]  <= stage_5b_edge[95:64];	// u
					queue_edgeout_buf_in[95:64] <= stage_5b_edge[63:32];	// v
					queue_edgeout_buf_in[63:32] <= stage_5b_edge[31:0];		// w
					queue_edgeout_buf_in[31:9] <= 0;
					if( v1) begin
						queue_edgeout_buf_in[8:0] <= b1;
					end
					else if( v2 ) begin
						queue_edgeout_buf_in[8:0] <= b2;
					end
					else if (v3) begin
						queue_edgeout_buf_in[8:0] <= b3;
					end
					else if( v4) begin
						queue_edgeout_buf_in[8 :0] <= b4;
					end
					else if(v5) begin
						queue_edgeout_buf_in[8:0] <= b5;
					end
					else if( v6) begin
						queue_edgeout_buf_in[8 :0] <= b6;
					end
					else if(v7) begin
						queue_edgeout_buf_in[8 :0] <= b7;
					end
					else if(v8) begin
						queue_edgeout_buf_in[8 :0] <= b8;
					end
					else if(v9 ) begin
						queue_edgeout_buf_in[8 :0] <= b9;
					end
					else if(v10) begin
						queue_edgeout_buf_in[8 :0] <= b10;
					end
					else if(v11) begin
						queue_edgeout_buf_in[8 :0] <= b11;
					end
					else if(v12) begin
						queue_edgeout_buf_in[8 :0] <= b12;
					end
					else if(v13) begin
						queue_edgeout_buf_in[8 :0] <= b13;
					end
					else if(v14) begin
						queue_edgeout_buf_in[8 :0] <= b14;
					end
					else if(v15) begin
						queue_edgeout_buf_in[8 :0] <= b15;
					end
					else if(v16) begin
						queue_edgeout_buf_in[8 :0] <= b16;
					end
     
                     queue_edgeout_write_enable <= 1;
                end

				if( stage_5b_cl_original != stage_5b_cl) begin
					queue_mcmout_write_enable <= 1;
					queue_mcmout_buf_in <= { stage_5b_cl, stage_5b_edge[63:32] };
				end

		end
    end
end


always @(posedge clk)
begin
	debug_edge_processor <= {	stage_5_one_hot_take_edge[31:0], stage_3_edge_level[31:0],
								mcmbram_ntotal_read, mcmbram_ntotal_write, 
								offset_ntotal_read, offset_ntotal_write, 
								mcmin_ntotal_read, mcmin_ntotal_write, 
								edgein_ntotal_read, edgein_ntotal_write,
								stage_1_edges_processed, 32'b0,
								stage_3_edges_processed, stage_4_edges_processed,
								stage_5_edges_processed, stage_6_edges_processed
								 };
end


endmodule


`default_nettype wire










