
`default_nettype none


/**
	The "edge receiver" receives the edges which have been requested by the "edge requester".
	The incoming data (a cache line) is immediately put into a queue, to make sure that we can receive data all the time.
	Then, we grab the cache line from this queue, and inspect the elements using information we have from the "information queue".
	The "information queue" is used to store information about the request, such we can determine to which of the K "edge queues" the edge must go.
	Further, the "information queue" holds information about the offset and number of valid edges in the cache line we have received.
	To allow the "edge receiver" determine the "K" that the "edge requester" needs to ask for further edges, we count the number of edges received for every K,
	and the "edge requester" passes us the information, how many edges he has requested per K.
	Therefore, the "edge receiver" can determine the amount of edges in flight for every K. This value is added to the lenght of every of the "K" queues. Given
	this data, we determine the K to request next, and pass this information to the "edge requester".
*/


module edge_receiver #(parameter BRAM_QUEUE_LENGTH_BITS  = 7, parameter K_BITS = 3, parameter EDGE_REQUEST_TAG = 2, parameter QUEUE_THRESHOLD = 12) (
	//---------------------- GENERAL VARIABLES---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input 	wire 									um_done,						// true, if we are done processing (global)
	input 	wire 									um_activated,					// true, if the afu is activated
	output  reg 										edge_receiver_done,
	input   wire										edge_requester_done,

	//---------------------- ADDRESS VARIABLES---------------------------//
	input wire [31:0]									num_edges,						// the true amount of edges (NOT IN THE NUMBER OF CACHE LINES!)	

	//---------------------- GENERAL READ INPUT VARIABLES---------------------------//
	input   wire [7:0]                         	    um_rx_rd_tag,	// response to the read request: the tag
    input   wire [511:0]                       	    um_rx_data,		// the requested data
    input   wire                               	    um_rx_rd_valid,	// that tag and data are valid

	//---------------------- VARIABLES FOR THE K MANY EDGE QUEUES ---------------------------//
	output reg[K_BITS-1:0]					queue_write_enable_k,
	output reg 								queue_write_enable,					// wheter we want to write to the queue or not
	output reg [95:0] 						queue_buf_in,						// (epoch, u, v, w) note: could omit the epoch. Given u, epoch = u % K

	//---------------------- VARIABLES TO INDICATE THE CURRENT INPUT SPACE WE HAVE LEFT ---------------------------//
	output reg [31:0]									space_left, 						// the number of space left in our input queue, storing cache lines
	output reg [31:0]									edge_receiver_num_resp_received,
	
	//---------------------- VARIABLES FROM EDGE REQUESTER - containing offset information etc. ---------------------------//
	input wire									   	queue_request_write_enable,
	input wire[127:0]								queue_request_buf_in,
	output reg										queue_request_filled_next,

	// output reg [31:0] 								num_edges_received_per_k[0:2**K_BITS-1],

	output reg[511:0]								debug_edge_receiver
);


reg			debug_queue_overflow_cl;
reg			debug_queue_overflow_info;

reg [31:0] cl_ntotal_read;
reg [31:0] cl_ntotal_write;

wire [BRAM_QUEUE_LENGTH_BITS-1:0] cl_buf_data_count;
reg cl_buf_write_enable;
reg [511:0] 	cl_queue_buf_in;

wire cl_m_axis_valid;
reg cl_m_axis_ready;
wire [511:0] cl_m_axis_data;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(512), .PROG_FULL_THRESH(0 ) ) cl_buf_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(cl_buf_write_enable),
    .s_axis_ready(),
    .s_axis_data(cl_queue_buf_in),
    .s_axis_last(1'b0),
    .m_axis_valid(cl_m_axis_valid),
    .m_axis_ready(cl_m_axis_ready),
    .m_axis_data(cl_m_axis_data),
    .m_axis_last(),
    .data_count(cl_buf_data_count)
);

always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
		
		space_left <= 0;
		edge_receiver_num_resp_received <= 0;
		cl_buf_write_enable <= 0;
		cl_ntotal_write <= 0;
		debug_queue_overflow_cl <= 0;
	end
	else begin
		cl_buf_write_enable<= 0;
		if(~edge_receiver_done ) begin
			space_left <= 2**BRAM_QUEUE_LENGTH_BITS - cl_buf_data_count - 10;
	
			if( um_rx_rd_valid & ( um_rx_rd_tag== EDGE_REQUEST_TAG) ) begin		
				cl_buf_write_enable<= 1;
				cl_queue_buf_in <= um_rx_data;
				cl_ntotal_write <= cl_ntotal_write+1;
				edge_receiver_num_resp_received <= edge_receiver_num_resp_received + 1;
			end

			debug_queue_overflow_cl <= 2**BRAM_QUEUE_LENGTH_BITS - cl_buf_data_count < 2;
		end
	end
end


wire [BRAM_QUEUE_LENGTH_BITS-1:0] info_buf_data_count;
reg info_buf_write_enable;
reg [127:0] 	info_queue_buf_in;
wire info_m_axis_valid;
reg info_m_axis_ready;
wire [127:0] info_m_axis_data;


nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(128), .PROG_FULL_THRESH(0 ) ) info_buf_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(info_buf_write_enable),
    .s_axis_ready(),
    .s_axis_data(info_queue_buf_in),
    .s_axis_last(1'b0),
    .m_axis_valid(info_m_axis_valid),
    .m_axis_ready(info_m_axis_ready),
    .m_axis_data(info_m_axis_data),
    .m_axis_last(),
    .data_count(info_buf_data_count)
);


reg [31:0] info_ntotal_read;
reg [31:0] info_ntotal_write;

always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
		info_ntotal_write <= 0;
		queue_request_filled_next <= 0;
		info_buf_write_enable <= 0;
		debug_queue_overflow_info<= 0;
	end
	else begin
		info_buf_write_enable <= 0;
		if(~edge_receiver_done ) begin
			queue_request_filled_next <= (info_buf_data_count + QUEUE_THRESHOLD + 8) >  (2**BRAM_QUEUE_LENGTH_BITS-1);
			if( queue_request_write_enable ) begin		
				info_buf_write_enable <= 1;
				info_queue_buf_in <= queue_request_buf_in;
				info_ntotal_write <= info_ntotal_write+1;
			end

			debug_queue_overflow_info <= 2**BRAM_QUEUE_LENGTH_BITS - info_buf_data_count < 2;
		end
	end
end




//---------------------- GIVEN INPUT DATA, PUT IT TO THE CORRECT QUEUE FOR THE SORTER ---------------------------//
// note: might require an additional overflow buffer?

// the working buffers
reg [511:0] input_buf;
reg input_buf_empty;
reg [127:0] request_info_buf;
reg			request_info_buf_empty;

// the overflow buffer
reg [511:0] input_buf_overflow_1;
reg input_buf_overflow_1_empty;
reg [127:0] request_info_buf_overflow_1;
reg request_info_buf_overflow_1_empty;

reg [511:0] input_buf_overflow_2;
reg input_buf_overflow_2_empty;
reg [127:0] request_info_buf_overflow_2;
reg request_info_buf_overflow_2_empty;

// temp vars
reg [31:0] u;
reg [31:0] epoch;

// counter vars
reg			debug_assignement_error_flag;
reg			debug_case_error_flag;
reg [31:0] n_processed_edges_flush;
reg [31:0] valid_edge_cnt;
reg [31:0] n_processed_edges;

wire[31:0] all_ones_32_bit;
assign all_ones_32_bit = ~0;

reg [31:0] debug_case;

always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
		// local variables
		input_buf_empty <= 1;
		request_info_buf_empty <= 1;
		input_buf_overflow_1_empty <= 1;
		input_buf_overflow_2_empty <= 1;
		request_info_buf_overflow_1_empty <= 1;
		request_info_buf_overflow_2_empty <= 1;
		
		u <= 0;
		epoch <= 0;
		valid_edge_cnt <= 0;

		queue_write_enable <= 0;
		
		// drain queues
		if( (um_done || ~um_activated) && rst_n ) begin
			cl_m_axis_ready <= 1;
			info_m_axis_ready <= 1;
		end
		else begin
			cl_m_axis_ready <= 0;
			info_m_axis_ready <= 0;
		end
		
		n_processed_edges <= 0;
		n_processed_edges_flush <= 0;
		debug_assignement_error_flag <= 0;
		debug_case_error_flag <= 0;
		debug_case <= 0;
	end
	else begin
		queue_write_enable <= 0;
		
		if( ~edge_receiver_done ) begin


			// case 1: The edge-working buffer and the info-working buffer is filled
			if( ~input_buf_empty && ~request_info_buf_empty) begin
					debug_case <= 1;

					// if there are valid edges in this cache-line: process
					if( request_info_buf[31:0] - valid_edge_cnt > 0 ) begin
						// queue_buf_in[127:96] <=  request_info_buf[127:96]; // epoch
						queue_buf_in[95:64] <= ( request_info_buf[127:96] << K_BITS) + request_info_buf[95:64] + 1; // u
						queue_buf_in[63:32] <=  ( input_buf >> ( (request_info_buf[63:32]  + valid_edge_cnt ) << 6) ); // v
						queue_buf_in[31:0] <=  ( input_buf >> ( (request_info_buf[63:32]  + valid_edge_cnt ) << 6) ) >> 32; // w
						queue_write_enable_k <= request_info_buf[95:64];
						queue_write_enable <= 1;
						n_processed_edges <= n_processed_edges + 1;
						valid_edge_cnt <= valid_edge_cnt + 1;

						// drop out the request-info-buf, and the input-buf, reset the valid-edge-cnt
						if( request_info_buf[31:0] - valid_edge_cnt -1 == 0 ) begin
								valid_edge_cnt <= 0;

								// assignement of request-info
								if( ~request_info_buf_overflow_1_empty ) begin
									request_info_buf <= request_info_buf_overflow_1;
							
									if( ~request_info_buf_overflow_2_empty ) begin
										// now, the buffer 2 is empty, check if we can fill it
										request_info_buf_overflow_1 <= request_info_buf_overflow_2;
										if( info_m_axis_valid && info_m_axis_ready ) begin
											request_info_buf_overflow_2 <= info_m_axis_data;
											info_m_axis_ready <= 0;
										end
										else begin
											request_info_buf_overflow_2_empty<= 1;
											info_m_axis_ready <= 1;
										end
									end
									else begin
										info_m_axis_ready <= 1; // we are ready to fill the buffer 2
										if( info_m_axis_valid && info_m_axis_ready ) begin
											request_info_buf_overflow_1 <= info_m_axis_data;
										end
										else begin
											request_info_buf_overflow_1_empty<= 1;
										end
									end
								end
								else begin
									info_m_axis_ready <= 1;
									if( info_m_axis_valid && info_m_axis_ready ) begin
										request_info_buf <= info_m_axis_data;
									end
									else begin
										request_info_buf_empty<= 1;
									end
								end


								// assignement of edges-data
								if( ~input_buf_overflow_1_empty ) begin
									input_buf <= input_buf_overflow_1;

									// case 1: queue to buffer 2
									if( ~input_buf_overflow_2_empty ) begin
										input_buf_overflow_1 <= input_buf_overflow_2;
										if( cl_m_axis_valid &&  cl_m_axis_ready ) begin
											input_buf_overflow_2 <= cl_m_axis_data;
											cl_m_axis_ready <= 0;
										end
										else begin
											input_buf_overflow_2_empty <= 1;
											cl_m_axis_ready <= 1;
										end
									end
									else begin // case 2: queue to buffer 1
										cl_m_axis_ready <= 1;
										if( cl_m_axis_valid &&  cl_m_axis_ready ) begin
											input_buf_overflow_1 <= cl_m_axis_data;
										end
										else begin
											input_buf_overflow_1_empty<= 1;
										end
									end
								end
								else begin // case 3: queue to working buffer
									cl_m_axis_ready <= 1;
									if( cl_m_axis_valid &&  cl_m_axis_ready ) begin
										input_buf <= cl_m_axis_data;
									end
									else begin
										input_buf_empty <= 1;
									end
								end

						
						end
						else begin
								// handle data assignement: we might have requested data for the overflow buffer, so we need to assign this
								if( info_m_axis_valid && info_m_axis_ready ) begin
									if( request_info_buf_overflow_1_empty) begin
										request_info_buf_overflow_1_empty <= 0;
										request_info_buf_overflow_1 <= info_m_axis_data;
									end
									else if( request_info_buf_overflow_2_empty) begin
										request_info_buf_overflow_2_empty <= 0;
										request_info_buf_overflow_2 <= info_m_axis_data;
										info_m_axis_ready <= 0;
									end
									else begin
										debug_assignement_error_flag <= 1;
									end
								end
								else if(request_info_buf_empty || request_info_buf_overflow_1_empty || request_info_buf_overflow_2_empty ) begin
									info_m_axis_ready <= 1;
								end

								if( cl_m_axis_valid &&  cl_m_axis_ready ) begin
									if( input_buf_overflow_1_empty ) begin
										input_buf_overflow_1_empty <= 0;
										input_buf_overflow_1 <= cl_m_axis_data;
									end
									else if( input_buf_overflow_2_empty ) begin
										input_buf_overflow_2_empty <= 0;
										input_buf_overflow_2 <= cl_m_axis_data;
										cl_m_axis_ready <= 0;
									end
									else begin
										debug_assignement_error_flag <= 1;
									end
								end
								else if( input_buf_empty || input_buf_overflow_1_empty || input_buf_overflow_2_empty ) begin
									cl_m_axis_ready <= 1;
								end

						end

					end
					else begin
								// generate an empty edge, drop the request-info-buf, fetch new data! handle data assignement
								// queue_buf_in[127:96] <= request_info_buf[127:96]; // epoch
								queue_buf_in[95:64] <= ( request_info_buf[127:96] << K_BITS) + request_info_buf[95:64] + 1; // u
								queue_buf_in[63:32] <=  all_ones_32_bit;			//v
								queue_buf_in[31:0] <= 32'b0;						// w
								queue_write_enable_k <= request_info_buf[95:64];
								queue_write_enable <= 1;

								n_processed_edges_flush <= n_processed_edges_flush  + 1;

								// assignement of request-info
								if( ~request_info_buf_overflow_1_empty ) begin
									request_info_buf <= request_info_buf_overflow_1;
							
									if( ~request_info_buf_overflow_2_empty ) begin
										// now, the buffer 2 is empty, check if we can fill it
										request_info_buf_overflow_1 <= request_info_buf_overflow_2;
										if( info_m_axis_valid && info_m_axis_ready ) begin
											request_info_buf_overflow_2 <= info_m_axis_data;
											info_m_axis_ready <= 0;
										end
										else begin
											request_info_buf_overflow_2_empty<= 1;
											info_m_axis_ready <= 1;
										end
									end
									else begin
										info_m_axis_ready <= 1; // we are ready to fill the buffer 2
										if( info_m_axis_valid && info_m_axis_ready ) begin
											request_info_buf_overflow_1 <= info_m_axis_data;
										end
										else begin
											request_info_buf_overflow_1_empty<= 1;
										end
									end
								end
								else begin
									info_m_axis_ready <= 1;
									if( info_m_axis_valid && info_m_axis_ready ) begin
										request_info_buf <= info_m_axis_data;
									end
									else begin
										request_info_buf_empty<= 1;
									end
								end



							// assignement of the cl: we know that the working edge-data is filled, so only handle the assignement to the overflow
							// we know this, since in the case we currently are, we have both filled, but request_info_buf[31:0] == 0, so it generates a flush edge
							if( cl_m_axis_valid &&  cl_m_axis_ready ) begin
								if( input_buf_overflow_1_empty ) begin
									input_buf_overflow_1_empty <= 0;
									input_buf_overflow_1 <= cl_m_axis_data;
								end
								else if( input_buf_overflow_2_empty ) begin
									input_buf_overflow_2_empty <= 0;
									input_buf_overflow_2 <= cl_m_axis_data;
									cl_m_axis_ready <= 0;
								end
								else begin
									debug_assignement_error_flag <= 1;
								end
							end
							else if( input_buf_empty || input_buf_overflow_1_empty || input_buf_overflow_2_empty ) begin
								cl_m_axis_ready <= 1;
							end

					end
				
			end
			// case 2: the info-working buffer is filled, and buffer is not filled for the edges
			else if( ~request_info_buf_empty ) begin
					debug_case <= 2;

					if( request_info_buf[31:0] == 0 ) begin
							// generate an empty edge, drop the request-info-buf, fetch new data! handle data assignement
							// queue_buf_in[127:96] <= request_info_buf[127:96];
							queue_buf_in[95:64] <= ( request_info_buf[127:96] << K_BITS) + request_info_buf[95:64] + 1;
							queue_buf_in[63:32] <=  all_ones_32_bit;
							queue_buf_in[31:0] <= 32'b0;
							queue_write_enable_k <= request_info_buf[95:64];
							queue_write_enable <= 1;

							n_processed_edges_flush <= n_processed_edges_flush  + 1;

							// assignement of request-info
							if( ~request_info_buf_overflow_1_empty ) begin
								request_info_buf <= request_info_buf_overflow_1;
							
								if( ~request_info_buf_overflow_2_empty ) begin
									// now, the buffer 2 is empty, check if we can fill it
									request_info_buf_overflow_1 <= request_info_buf_overflow_2;
									if( info_m_axis_valid && info_m_axis_ready ) begin
										request_info_buf_overflow_2 <= info_m_axis_data;
										info_m_axis_ready <= 0;
									end
									else begin
										request_info_buf_overflow_2_empty<= 1;
										info_m_axis_ready <= 1;
									end
								end
								else begin
									info_m_axis_ready <= 1; // we are ready to fill the buffer 2
									if( info_m_axis_valid && info_m_axis_ready ) begin
										request_info_buf_overflow_1 <= info_m_axis_data;
										info_m_axis_ready <= 0;
									end
									else begin
										request_info_buf_overflow_1_empty<= 1;
									end
								end
							end
							else begin
								info_m_axis_ready <= 1;
								if( info_m_axis_valid && info_m_axis_ready ) begin
									request_info_buf <= info_m_axis_data;
								end
								else begin
									request_info_buf_empty<= 1;
								end
							end

							cl_m_axis_ready <= 1;
							// assignement of the cl: we know that the working data is NOT filled, and so is the overflow buffer, so only handle the assignement to the working part
							if( cl_m_axis_valid &&  cl_m_axis_ready ) begin
								input_buf_empty <= 0;
								input_buf <= cl_m_axis_data;
							end
							if( ~input_buf_empty ) begin
								debug_assignement_error_flag <= 1;
							end
					end
					else begin
							// handle data assignement: we might have requested data for the overflow buffer, so we need to assign this
							if( info_m_axis_valid && info_m_axis_ready ) begin
								if( request_info_buf_overflow_1_empty) begin
									request_info_buf_overflow_1_empty <= 0;
									request_info_buf_overflow_1 <= info_m_axis_data;
								end
								else if( request_info_buf_overflow_2_empty) begin
									request_info_buf_overflow_2_empty <= 0;
									request_info_buf_overflow_2 <= info_m_axis_data;
									info_m_axis_ready <= 0;
								end
								else begin
									debug_assignement_error_flag <= 1;
								end
							end
							else if( request_info_buf_empty|| request_info_buf_overflow_1_empty || request_info_buf_overflow_2_empty ) begin
								info_m_axis_ready <= 1;
							end

							cl_m_axis_ready <= 1;
							// assignement of the cl: we know that the working data is NOT filled, and so is the overflow buffer, so only handle the assignement to the working part
							if( cl_m_axis_valid &&  cl_m_axis_ready ) begin
								input_buf_empty <= 0;
								input_buf <= cl_m_axis_data;
							end
					end
				
			end
			// case 3: all buffers are empty: edge must be empty, since for every edge, there is a request: so we are in here only if for both there is no data!
			else if( request_info_buf_empty && input_buf_empty) begin
					info_m_axis_ready <= 1;
					cl_m_axis_ready <= 1;
					debug_case <= 3;

					if( info_m_axis_valid && info_m_axis_ready ) begin
						request_info_buf_empty <= 0;
						request_info_buf <= info_m_axis_data;
					end
					if( cl_m_axis_valid &&  cl_m_axis_ready ) begin
						input_buf_empty <= 0;
						input_buf <= cl_m_axis_data;
					end
			end
			// case 4: actually, this case should never occur: It means that the edge-buf is valid, but the info-buffer is not
			// this might be a contradiction, but what if there is a case where due to the queue behaviour, we get the edge-data just one cycle before the others? we need to handle this accordingly
			else begin
						debug_case<= 4;
						debug_case_error_flag <= 1;
						if( info_m_axis_valid && info_m_axis_ready ) begin
							if( request_info_buf_empty ) begin
								request_info_buf_empty <= 0;
								request_info_buf <= info_m_axis_data;
							end
							else if( request_info_buf_overflow_1_empty) begin
								request_info_buf_overflow_1_empty <= 0;
								request_info_buf_overflow_1 <= info_m_axis_data;
							end
							else if( request_info_buf_overflow_2_empty) begin
								request_info_buf_overflow_2_empty <= 0;
								request_info_buf_overflow_2 <= info_m_axis_data;
								info_m_axis_ready <= 0;
							end
							else begin
								debug_assignement_error_flag <= 1;
							end
						end
						else if(request_info_buf_empty || request_info_buf_overflow_1_empty || request_info_buf_overflow_2_empty ) begin
							info_m_axis_ready <= 1;
						end


						if( cl_m_axis_valid &&  cl_m_axis_ready ) begin
							if( input_buf_empty ) begin
								input_buf_empty <= 0;
								input_buf <= cl_m_axis_data;
							end
							else if( input_buf_overflow_1_empty ) begin
								input_buf_overflow_1_empty <= 0;
								input_buf_overflow_1 <= cl_m_axis_data;
							end
							else if( input_buf_overflow_2_empty ) begin
								input_buf_overflow_2_empty <= 0;
								input_buf_overflow_2 <= cl_m_axis_data;
								cl_m_axis_ready <= 0;
							end
							else begin
								debug_assignement_error_flag <= 1;
							end
						end
						else if( input_buf_empty || input_buf_overflow_1_empty || input_buf_overflow_2_empty ) begin
							cl_m_axis_ready <= 1;
						end	
						
			end


		end
	end

end



always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
		cl_ntotal_read<= 0;
		info_ntotal_read<= 0;
	end
	else begin
		if( cl_m_axis_valid &&  cl_m_axis_ready ) begin
			cl_ntotal_read<= cl_ntotal_read  + 1;
		end
		
		if( info_m_axis_valid && info_m_axis_ready ) begin
			info_ntotal_read <= info_ntotal_read + 1;
		end
	end
end

reg [31:0]		debug_info_buf_data_count;
reg [31:0]		debug_cl_buf_data_count;
reg [31:0]		debug_case_error_flag_32b;
reg [31:0]		debug_assign_error_flag_32b;
reg [31:0]		debug_info_buf_valid;
reg [31:0]		debug_cl_buf_valid;
reg [31:0]		debug_queue_overflow;

always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
			edge_receiver_done <= 0;

			debug_info_buf_data_count <= 0;
			debug_cl_buf_data_count <= 0;
			debug_case_error_flag_32b<= 0;
			debug_assign_error_flag_32b <= 0;
			debug_info_buf_valid <= 0;
			debug_cl_buf_valid <= 0;
			debug_queue_overflow <= 0;
	end
	else begin
		if( ~edge_receiver_done ) begin
			if( n_processed_edges == num_edges) begin	
				edge_receiver_done <= 1;
			end
			debug_info_buf_data_count <= info_buf_data_count + (!request_info_buf_empty) + (!request_info_buf_overflow_2_empty) + (!request_info_buf_overflow_1_empty);
			debug_cl_buf_data_count <= cl_buf_data_count + (!input_buf_empty) + (!input_buf_overflow_2_empty) + (!input_buf_overflow_1_empty);
			debug_case_error_flag_32b<= debug_case_error_flag;
			debug_assign_error_flag_32b <= debug_assignement_error_flag;

			debug_info_buf_valid[0] <= !request_info_buf_empty;
			debug_info_buf_valid[1] <= !request_info_buf_overflow_1_empty;
			debug_info_buf_valid[2] <= !request_info_buf_overflow_2_empty;

			debug_cl_buf_valid[0] <= !input_buf_empty;
			debug_cl_buf_valid[1] <= !input_buf_overflow_1_empty;
			debug_cl_buf_valid[2] <= !input_buf_overflow_2_empty;

			debug_queue_overflow[0] <= debug_queue_overflow_cl;
			debug_queue_overflow[1] <= debug_queue_overflow_info;
		end
	end
end



/*
// high, if we do not output data, but we could
reg				do_not_work_valid;
reg [31:0]		do_not_work_cycle_counter;
// high, if we hava data, but cannot output data
reg				can_not_output_valid;
reg [31:0]		can_not_output_cycle_counter;
// high, if we do not have data, but could output
reg				can_not_input_valid;
reg [31:0]		can_not_input_cycle_counter;


always @(posedge clk) begin
	if (~rst_n | um_done | ~um_activated) begin
		do_not_work_valid <= 0;
		do_not_work_cycle_counter <= 0;
		can_not_output_valid <= 0;
		can_not_output_cycle_counter <= 0;
		can_not_input_valid <= 0;
		can_not_input_cycle_counter <= 0;
	end
	else begin
		do_not_work_valid <= 0;
		can_not_output_valid <= 0;
		can_not_input_valid <= 0;

		if(  cl_buf_data_count > 0 && !( queue_write_enable) ) begin
			do_not_work_valid<= 1;
			do_not_work_cycle_counter <= do_not_work_cycle_counter + 1;
		end

		if(cl_buf_data_count == 0 && input_buf_empty ) begin
			can_not_input_valid <= 1;
			can_not_input_cycle_counter <= can_not_input_cycle_counter + 1;
		end
	end
end
*/


always @(posedge clk) begin
	debug_edge_receiver<= { 64'b0, 
								debug_cl_buf_valid,	debug_info_buf_valid,
								debug_queue_overflow, debug_assign_error_flag_32b,
								debug_case, debug_case_error_flag_32b,
								debug_cl_buf_data_count, debug_info_buf_data_count,
								n_processed_edges_flush, n_processed_edges, 
								cl_ntotal_read, cl_ntotal_write, 
								info_ntotal_read, info_ntotal_write };
end

endmodule



`default_nettype wire











