`default_nettype none



module edge_requester #(parameter K_BITS = 4, parameter EDGE_REQUEST_TAG = 2, parameter QUEUE_THRESHOLD = 8, parameter BRAM_QUEUE_LENGTH_BITS = 5, parameter SORTER_FIRST_QUEUE_LENGTH_BITS = 7) (
	//---------------------- GENERAL VARIABLES---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input wire 										um_done,						// true, if we are done processing (global)
	input wire 										um_activated,					// true, if the afu is activated
	output reg 										edge_requester_done,
	//---------------------- ADDRESS VARIABLES---------------------------//
	input wire [57:0]								edges_base_addr,					// the base read address
	input wire [31:0]								num_ptrs,							// number of pointers - to determine the termination

	//---------------------- QUEUE TO SEND THE REQUEST TO ---------------------------//
	output reg									   	queue_write_enable,				// wheter we want to write to the queue or not
	output reg [65:0] 							   	queue_buf_in,					// the queue input buffer (65:8 is the request address, 7:0 the tag we assign to the request)
	input wire 									   	queue_buf_filled_next,			// true, if the queue is filled soon, false otherwise

	//---------------------- VARIABLES OF THE K MANY POINTER QUEUES ---------------------------//
	input wire 											queue_ptr_write_enable,
	input wire [127:0] 									queue_ptr_buf_in,
	output reg 											queue_ptr_buf_filled_next,

	//---------------------- QUEUE VARIABLES TO STORE REQUEST INFORMATION (USED TO ASSIGN THE RESPONSE TO THE CORRECT K, and pass offset infos, epoch infos) ---------------------------//
	output reg									   	queue_request_write_enable,
	output reg[127:0]								queue_request_buf_in,
	input wire										queue_request_buf_filled_next,

	input wire										sorter_edge_consumed[0:2**K_BITS-1],

	output reg[511:0]					debug_edge_requester
);


reg 						bram_1_re;
reg 						bram_1_we;
reg 	[K_BITS-1:0] 		bram_1_raddr;
reg 	[K_BITS-1:0] 		bram_1_waddr;
reg 	[199:0]				bram_1_din;
wire 	[199:0]				bram_1_dout;
reg 	[2**K_BITS-1:0]		bram_1_valid;

// 31:0 the address (32 bit)
// 35:32 the offset ( 4 bit)
// 67:36  the number of edges
// 99:68	the epoch
spl_sdp_mem_t #( .DATA_WIDTH(200), .ADDR_WIDTH(K_BITS) ) bram_1 (
    .clk(clk),
    .we(bram_1_we),
	.re(bram_1_re),
    .raddr(bram_1_raddr),
    .waddr(bram_1_waddr),
    .din(bram_1_din),
    .dout(bram_1_dout)
);



// the idea would be as follows: we have multiple brams to store "overflowing" pointers, which were taken from the queue
// we then read from those as well, and pick (if required) those which are valid in decreasing order. and then need to move upwards - which might introduce new wiring problems :/
reg 	[2**K_BITS-1:0]		bram_2_valid;


// 31:0 the address (32 bit)
// 35:32 the offset ( 4 bit)
// 67:36  the number of edges
// 99:68	the epoch

reg	[99:0]							queue1_s_axis_data;
reg									queue1_s_axis_valid;
reg									queue1_m_axis_ready;
wire 								queue1_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	queue1_data_count;
wire [99:0]							queue1_m_axis_data;
reg [K_BITS-3:0]					rd1_k;

reg [31:0]							queue1_write;
reg [31:0]							queue1_read;
reg [31:0]							queue1_real_read;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(100), .PROG_FULL_THRESH(0 ) ) edge_requester_q1_axififo	(
	.clk (clk),
    .rst(~rst_n),
    .s_axis_valid(queue1_s_axis_valid ),
    .s_axis_ready(),
    .s_axis_data(queue1_s_axis_data),
    .s_axis_last(1'b0),
    .m_axis_valid(queue1_m_axis_valid),
    .m_axis_ready(queue1_m_axis_ready),
    .m_axis_data(queue1_m_axis_data),
    .m_axis_last(),
    .data_count(queue1_data_count)
);


reg	[99:0]							queue2_s_axis_data;
reg									queue2_s_axis_valid;
reg									queue2_m_axis_ready;
wire 								queue2_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	queue2_data_count;
wire [99:0]							queue2_m_axis_data;
reg [K_BITS-3:0]					rd2_k;

reg [31:0]							queue2_write;
reg [31:0]							queue2_read;
reg [31:0]							queue2_real_read;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(100), .PROG_FULL_THRESH(0) ) edge_requester_q2_axififo	(
	.clk (clk),
    .rst(~rst_n),
    .s_axis_valid(queue2_s_axis_valid ),
    .s_axis_ready(),
    .s_axis_data(queue2_s_axis_data),
    .s_axis_last(1'b0),
    .m_axis_valid(queue2_m_axis_valid),
    .m_axis_ready(queue2_m_axis_ready),
    .m_axis_data(queue2_m_axis_data),
    .m_axis_last(),
    .data_count(queue2_data_count)
);



reg	[99:0]							queue3_s_axis_data;
reg									queue3_s_axis_valid;
reg									queue3_m_axis_ready;
wire 								queue3_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	queue3_data_count;
wire [99:0]							queue3_m_axis_data;
reg [K_BITS-3:0]					rd3_k;

reg [31:0]							queue3_write;
reg [31:0]							queue3_read;
reg [31:0]							queue3_real_read;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(100), .PROG_FULL_THRESH(0) ) edge_requester_q3_axififo	(
	.clk (clk),
    .rst(~rst_n),
    .s_axis_valid(queue3_s_axis_valid ),
    .s_axis_ready(),
    .s_axis_data(queue3_s_axis_data),
    .s_axis_last(1'b0),
    .m_axis_valid(queue3_m_axis_valid),
    .m_axis_ready(queue3_m_axis_ready),
    .m_axis_data(queue3_m_axis_data),
    .m_axis_last(),
    .data_count(queue3_data_count)
);



reg	[99:0]							queue4_s_axis_data;
reg									queue4_s_axis_valid;
reg									queue4_m_axis_ready;
wire 								queue4_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	queue4_data_count;
wire [99:0]							queue4_m_axis_data;
reg [K_BITS-3:0]					rd4_k;

reg [31:0]							queue4_write;
reg [31:0]							queue4_read;
reg [31:0]							queue4_real_read;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(100), .PROG_FULL_THRESH(0) ) edge_requester_q4_axififo	(
	.clk (clk),
    .rst(~rst_n),
    .s_axis_valid(queue4_s_axis_valid ),
    .s_axis_ready(),
    .s_axis_data(queue4_s_axis_data),
    .s_axis_last(1'b0),
    .m_axis_valid(queue4_m_axis_valid),
    .m_axis_ready(queue4_m_axis_ready),
    .m_axis_data(queue4_m_axis_data),
    .m_axis_last(),
    .data_count(queue4_data_count)
);



//---------------------- Read incoming pointers, place them in the correct position ---------------------------//

// data to the sorter
reg [31:0]						ptr_queue_ntotal_write;
reg	[K_BITS-1:0]				queue_switch;
reg								pointer_queue_assign_error;

always @(posedge clk) begin
	if (~rst_n | um_done | ~um_activated) begin
		ptr_queue_ntotal_write <= 0;
		queue1_write <= 0;
		queue2_write <= 0;
		queue3_write <= 0;
		queue4_write <= 0;

		queue1_s_axis_valid <= 1'b0;
		queue2_s_axis_valid <= 1'b0;
		queue3_s_axis_valid <= 1'b0;
		queue4_s_axis_valid <= 1'b0;

		queue_switch <= 0;
		pointer_queue_assign_error <= 0;
	end
	else begin
		queue1_s_axis_valid <= 1'b0;
		queue2_s_axis_valid <= 1'b0;
		queue3_s_axis_valid <= 1'b0;
		queue4_s_axis_valid <= 1'b0;
		if( queue_ptr_write_enable ) begin
			ptr_queue_ntotal_write <= ptr_queue_ntotal_write + 1'b1;

			// gets for example [0, 3] (include)
			if( queue_switch <= 2**(K_BITS-2)-1) begin
				queue1_write <= queue1_write + 1'b1;
				queue1_s_axis_data[31:0] <= queue_ptr_buf_in[31:0];
				queue1_s_axis_data[35:32] <= queue_ptr_buf_in[35:32];
				queue1_s_axis_data[67:36] <= queue_ptr_buf_in[95:64];
				queue1_s_axis_data[99:68] <= queue_ptr_buf_in[127:96];
				queue1_s_axis_valid <= 1'b1;
			end
			// gets [4, 8] (include)
			else if( queue_switch <= 2**(K_BITS-1)-1 ) begin
				queue2_write <= queue2_write + 1'b1;
				queue2_s_axis_data[31:0] <= queue_ptr_buf_in[31:0];
				queue2_s_axis_data[35:32] <= queue_ptr_buf_in[35:32];
				queue2_s_axis_data[67:36] <= queue_ptr_buf_in[95:64];
				queue2_s_axis_data[99:68] <= queue_ptr_buf_in[127:96];
				queue2_s_axis_valid <= 1'b1;
			end	
			else if( queue_switch <= 2**(K_BITS-1) + 2**(K_BITS-2) -1 ) begin
				queue3_write <= queue3_write + 1'b1;
				queue3_s_axis_data[31:0] <= queue_ptr_buf_in[31:0];
				queue3_s_axis_data[35:32] <= queue_ptr_buf_in[35:32];
				queue3_s_axis_data[67:36] <= queue_ptr_buf_in[95:64];
				queue3_s_axis_data[99:68] <= queue_ptr_buf_in[127:96];
				queue3_s_axis_valid <= 1'b1;
			end
			else if( queue_switch <= 2**(K_BITS)-1 ) begin
				queue4_write <= queue4_write + 1'b1;
				queue4_s_axis_data[31:0] <= queue_ptr_buf_in[31:0];
				queue4_s_axis_data[35:32] <= queue_ptr_buf_in[35:32];
				queue4_s_axis_data[67:36] <= queue_ptr_buf_in[95:64];
				queue4_s_axis_data[99:68] <= queue_ptr_buf_in[127:96];
				queue4_s_axis_valid <= 1'b1;
			end
			else begin
				pointer_queue_assign_error <= 1;
			end

			queue_switch <= queue_switch  + 1;
			
		end
	end
end


//---------------------- DETERMINE THE FILLED STATE ---------------------------//
always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		queue_ptr_buf_filled_next <= 1'b0;
	end
	else begin
		queue_ptr_buf_filled_next <= ( queue1_data_count + QUEUE_THRESHOLD ) > 2**BRAM_QUEUE_LENGTH_BITS ||
									 ( queue2_data_count + QUEUE_THRESHOLD ) > 2**BRAM_QUEUE_LENGTH_BITS ||
									( queue3_data_count + QUEUE_THRESHOLD ) > 2**BRAM_QUEUE_LENGTH_BITS ||
									( queue4_data_count + QUEUE_THRESHOLD ) > 2**BRAM_QUEUE_LENGTH_BITS;
	end
end



//---------------------- Stage 1: get k from sorter and compute the position for the (possible) queue access ---------------------------//
/**
	Approach: 1.) 	try to read from a queue, if we can, and it is valid to request
			  2.) 	check if the k we get from the sorter is valid to request
				3.) try to read from a queue, if we can - despite the fact that it is valid to request
				4.) else, send the k we get from the sorter

*/

// input from the sorter to stage 1
wire [K_BITS-1:0]			k_to_request;
wire [2**K_BITS-1:0]		k_to_request_valid;

// These prevent that we read a pointer from a queue, and "overflow" the according bram
// It requires 3 entries: We see that it is empty, _1 is true in next cycle (stage 2 sees the read), then _2 is true, then _3 is true  (stage 3 can now set the bram_1 or bram_2 accordingly), and then we see it
// therefore, 3 is sufficient and required
reg [K_BITS-1:0]		stage_1_k_queue_requested_1;
reg						stage_1_k_queue_requested_1_valid;
reg [K_BITS-1:0]		stage_1_k_queue_requested_2;
reg						stage_1_k_queue_requested_2_valid;
reg [K_BITS-1:0]		stage_1_k_queue_requested_3;
reg						stage_1_k_queue_requested_3_valid;

// to stage 2
reg [K_BITS-1:0]		stage_1_k_to_request;
reg						stage_1_valid;

// debug
reg [31:0]			stage_1_k_to_request_32b;
reg [31:0]			stage_1_k_to_request_valid_32b;
reg [15:0]			stage_1_valid_debug;
reg					stage_1_queue_requested;

reg 				ptr_add_enable;
reg [K_BITS-1:0]	ptr_added_k;

always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
		stage_1_valid <= 1'b0;
		stage_1_k_to_request_32b <= 0;
		stage_1_k_to_request_valid_32b <= 0;
		
		bram_1_re <= 1'b0;
		// bram_2_re <= 1'b0;

		stage_1_k_queue_requested_1_valid <= 1'b0;
		stage_1_k_queue_requested_2_valid <= 1'b0;
		stage_1_k_queue_requested_3_valid <= 1'b0;

		// drain queues
		if( (um_done || ~um_activated) && rst_n ) begin
			queue1_m_axis_ready <= 1;
			queue2_m_axis_ready <= 1;
			queue3_m_axis_ready <= 1;
			queue4_m_axis_ready <= 1;
		end
		else begin
			queue1_m_axis_ready <= 1'b0;
			queue2_m_axis_ready <= 1'b0;
			queue3_m_axis_ready <= 1'b0;
			queue4_m_axis_ready <= 1'b0;
		end
		rd1_k <= 0;
		rd2_k <= 0;
		rd3_k <= 0;
		rd4_k <= 0;
		queue1_read <= 0;
		queue2_read <= 0;
		queue3_read <= 0;
		queue4_read <= 0;
		queue1_real_read <= 0;
		queue2_real_read <= 0;
		queue3_real_read <= 0;
		queue4_real_read <= 0;

		ptr_add_enable <= 1'b0;
		stage_1_valid_debug<= 0;
	end
	else begin
		// debug
		stage_1_k_to_request_32b <= k_to_request;
		stage_1_k_to_request_valid_32b <= k_to_request_valid;
		stage_1_valid_debug <= stage_1_valid;

		// default state
		stage_1_valid <= 1'b0;
		bram_1_re <= 1'b1; // always read
		// bram_2_re <= 1'b1; // always read
		queue1_m_axis_ready <= 1'b0;
		queue2_m_axis_ready <= 1'b0;
		queue3_m_axis_ready <= 1'b0;
		queue4_m_axis_ready <= 1'b0;
		stage_1_k_queue_requested_1_valid <= 1'b0;
		stage_1_queue_requested <= 0;
		ptr_add_enable <= 1'b0;


		if( ~edge_requester_done && ~queue_buf_filled_next & ~queue_request_buf_filled_next ) begin
			stage_1_valid <= 1'b1;

			// 1.) try to read from the queue 1 (if we can, and it is valid to request)
			if( queue1_write > queue1_read && queue1_data_count > 0 && k_to_request_valid[rd1_k] && !queue2_m_axis_ready && !queue3_m_axis_ready && !queue4_m_axis_ready &&
				( bram_1_valid[rd1_k] == 0 || bram_2_valid[rd1_k] == 0 )
												&& !( stage_1_k_queue_requested_1_valid && stage_1_k_queue_requested_1 == rd1_k ) && 
												 !( stage_1_k_queue_requested_2_valid && stage_1_k_queue_requested_2 == rd1_k ) &&
												 !( stage_1_k_queue_requested_3_valid && stage_1_k_queue_requested_3 == rd1_k )
												  ) begin

				queue1_m_axis_ready <= 1'b1;
				stage_1_k_queue_requested_1_valid <= 1'b1;
				rd1_k <= rd1_k  + 1'b1;
				queue1_read <= queue1_read + 1'b1;
				ptr_add_enable <= 1'b1;

				bram_1_raddr <= rd1_k;
				// bram_2_raddr <= rd1_k;
				ptr_added_k <= rd1_k;
				stage_1_k_to_request <= rd1_k;
				stage_1_k_queue_requested_1 <= rd1_k;
				stage_1_queue_requested <= 1;
			end
			// try queue 2
			else if( queue2_write > queue2_read && queue2_data_count > 0 && k_to_request_valid[rd2_k + 2**(K_BITS-2)] && !queue1_m_axis_ready && !queue3_m_axis_ready && !queue4_m_axis_ready &&
				( bram_1_valid[rd2_k + 2**(K_BITS-2) ] == 0 || bram_2_valid[rd2_k + 2**(K_BITS-2) ] == 0 )
												&& !( stage_1_k_queue_requested_1_valid && stage_1_k_queue_requested_1 == rd2_k + 2**(K_BITS-2)) && 
												 !( stage_1_k_queue_requested_2_valid && stage_1_k_queue_requested_2 == rd2_k + 2**(K_BITS-2) ) &&
												 !( stage_1_k_queue_requested_3_valid && stage_1_k_queue_requested_3 == rd2_k +2**(K_BITS-2) )
												  ) begin


				queue2_m_axis_ready <= 1'b1;
				stage_1_k_queue_requested_1_valid <= 1'b1;
				rd2_k <= rd2_k  + 1'b1;
				queue2_read <= queue2_read + 1'b1;
				ptr_add_enable <= 1'b1;

				bram_1_raddr <= rd2_k + 2**(K_BITS-2);
				// bram_2_raddr <= rd2_k + 2**(K_BITS-2);
				ptr_added_k <= rd2_k + 2**(K_BITS-2);
				stage_1_k_to_request <= rd2_k + 2**(K_BITS-2);
				stage_1_k_queue_requested_1 <= rd2_k + 2**(K_BITS-2);
				stage_1_queue_requested <= 1;

			end
			// try queue 3
			else if( queue3_write > queue3_read && queue3_data_count > 0 && k_to_request_valid[rd3_k + 2**(K_BITS-1) ] && !queue1_m_axis_ready && !queue2_m_axis_ready && !queue4_m_axis_ready &&
				( bram_1_valid[rd3_k + 2**(K_BITS-1) ] == 0 || bram_2_valid[rd3_k + 2**(K_BITS-1)   ] == 0 )
												&& !( stage_1_k_queue_requested_1_valid && stage_1_k_queue_requested_1 == rd3_k + 2**(K_BITS-1)  ) && 
												 !( stage_1_k_queue_requested_2_valid && stage_1_k_queue_requested_2 == rd3_k + 2**(K_BITS-1)  ) &&
												 !( stage_1_k_queue_requested_3_valid && stage_1_k_queue_requested_3 == rd3_k + 2**(K_BITS-1)  )
												  ) begin


				queue3_m_axis_ready <= 1'b1;
				stage_1_k_queue_requested_1_valid <= 1'b1;
				rd3_k <= rd3_k  + 1'b1;
				queue3_read <= queue3_read + 1'b1;
				ptr_add_enable <= 1'b1;

				bram_1_raddr <= rd3_k + 2**(K_BITS-1);
				// bram_2_raddr <= rd3_k + 2**(K_BITS-1);
				ptr_added_k <= rd3_k + 2**(K_BITS-1) ;
				stage_1_k_to_request <= rd3_k + 2**(K_BITS-1) ;
				stage_1_k_queue_requested_1 <= rd3_k + 2**(K_BITS-1) ;
				stage_1_queue_requested <= 1;

			end
			// try queue 4
			else if( queue4_write > queue4_read && queue4_data_count > 0 && k_to_request_valid[rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2)] && !queue1_m_axis_ready && !queue2_m_axis_ready && !queue3_m_axis_ready &&
				( bram_1_valid[rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ] == 0 || bram_2_valid[rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2)] == 0 )
												&& !( stage_1_k_queue_requested_1_valid && stage_1_k_queue_requested_1 == rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ) && 
												 !( stage_1_k_queue_requested_2_valid && stage_1_k_queue_requested_2 == rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ) &&
												 !( stage_1_k_queue_requested_3_valid && stage_1_k_queue_requested_3 == rd4_k +2**(K_BITS-1) + 2**(K_BITS-2) )
												  ) begin


				queue4_m_axis_ready <= 1'b1;
				stage_1_k_queue_requested_1_valid <= 1'b1;
				rd4_k <= rd4_k  + 1'b1;
				queue4_read <= queue4_read + 1'b1;
				ptr_add_enable <= 1'b1;

				bram_1_raddr <= rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ;
				// bram_2_raddr <= rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ;
				ptr_added_k <= rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ;
				stage_1_k_to_request <= rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ;
				stage_1_k_queue_requested_1 <= rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ;
				stage_1_queue_requested <= 1;

			end
			// check out the provided k from the length sorter
			else if( k_to_request_valid[k_to_request] &&  bram_1_valid[k_to_request] ) begin
				bram_1_raddr <= k_to_request;
				// bram_2_raddr <= k_to_request;
				stage_1_k_to_request <= k_to_request;
			end
			// try to read from queue 1
			else if( queue1_write > queue1_read && queue1_data_count > 0 && !queue2_m_axis_ready && !queue3_m_axis_ready && !queue4_m_axis_ready &&
				( bram_1_valid[rd1_k] == 0 || bram_2_valid[rd1_k] == 0 )
												&& !( stage_1_k_queue_requested_1_valid && stage_1_k_queue_requested_1 == rd1_k ) && 
												 !( stage_1_k_queue_requested_2_valid && stage_1_k_queue_requested_2 == rd1_k ) &&
												 !( stage_1_k_queue_requested_3_valid && stage_1_k_queue_requested_3 == rd1_k )
												  ) begin


				queue1_m_axis_ready <= 1'b1;
				stage_1_k_queue_requested_1_valid <= 1'b1;
				rd1_k <= rd1_k  + 1'b1;
				queue1_read <= queue1_read + 1'b1;
				ptr_add_enable <= 1'b1;

				bram_1_raddr <= rd1_k;
				// bram_2_raddr <= rd1_k;
				ptr_added_k <= rd1_k;
				stage_1_k_to_request <= rd1_k;
				stage_1_k_queue_requested_1 <= rd1_k;
				stage_1_queue_requested <= 1;

			end	
			// try to read from queue 2
			else if( queue2_write > queue2_read && queue2_data_count > 0 && !queue1_m_axis_ready && !queue3_m_axis_ready && !queue4_m_axis_ready &&
				( bram_1_valid[rd2_k + 2**(K_BITS-2)] == 0 || bram_2_valid[rd2_k + 2**(K_BITS-2)] == 0 )
												&& !( stage_1_k_queue_requested_1_valid && stage_1_k_queue_requested_1 == rd2_k +2**(K_BITS-2)) && 
												 !( stage_1_k_queue_requested_2_valid && stage_1_k_queue_requested_2 == rd2_k +2**(K_BITS-2)) &&
												 !( stage_1_k_queue_requested_3_valid && stage_1_k_queue_requested_3 == rd2_k +2**(K_BITS-2))
												  ) begin
				queue2_m_axis_ready <= 1'b1;
				stage_1_k_queue_requested_1_valid <= 1'b1;
				rd2_k <= rd2_k  + 1'b1;
				queue2_read <= queue2_read + 1'b1;
				ptr_add_enable <= 1'b1;

				bram_1_raddr <= rd2_k + 2**(K_BITS-2);
				// bram_2_raddr <= rd2_k + 2**(K_BITS-2);
				ptr_added_k <= rd2_k + 2**(K_BITS-2);
				stage_1_k_to_request <= rd2_k + 2**(K_BITS-2);
				stage_1_k_queue_requested_1 <= rd2_k + 2**(K_BITS-2);
				stage_1_queue_requested<= 1;
			end
			// try to read from queue 3
			else if( queue3_write > queue3_read && queue3_data_count > 0 && !queue1_m_axis_ready && !queue2_m_axis_ready && !queue4_m_axis_ready &&
				( bram_1_valid[rd3_k + 2**(K_BITS-1)  ] == 0 || bram_2_valid[rd3_k +  2**(K_BITS-1)  ] == 0 )
												&& !( stage_1_k_queue_requested_1_valid && stage_1_k_queue_requested_1 == rd3_k +  2**(K_BITS-1)  ) && 
												 !( stage_1_k_queue_requested_2_valid && stage_1_k_queue_requested_2 == rd3_k + 2**(K_BITS-1) ) &&
												 !( stage_1_k_queue_requested_3_valid && stage_1_k_queue_requested_3 == rd3_k + 2**(K_BITS-1) )
												  ) begin
				queue3_m_axis_ready <= 1'b1;
				stage_1_k_queue_requested_1_valid <= 1'b1;
				rd3_k <= rd3_k  + 1'b1;
				queue3_read <= queue3_read + 1'b1;
				ptr_add_enable <= 1'b1;

				bram_1_raddr <= rd3_k + 2**(K_BITS-1);
				// bram_2_raddr <= rd3_k + 2**(K_BITS-1) ;
				ptr_added_k <= rd3_k + 2**(K_BITS-1) ;
				stage_1_k_to_request <= rd3_k + 2**(K_BITS-1) ;
				stage_1_k_queue_requested_1 <= rd3_k + 2**(K_BITS-1) ;
				stage_1_queue_requested<= 1;
			end
			// try to read from queue 4
			else if( queue4_write > queue4_read && queue4_data_count > 0 && !queue1_m_axis_ready && !queue2_m_axis_ready && !queue3_m_axis_ready &&
				( bram_1_valid[rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ] == 0 || bram_2_valid[rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ] == 0 )
												&& !( stage_1_k_queue_requested_1_valid && stage_1_k_queue_requested_1 == rd4_k +2**(K_BITS-1) + 2**(K_BITS-2) ) && 
												 !( stage_1_k_queue_requested_2_valid && stage_1_k_queue_requested_2 == rd4_k +2**(K_BITS-1) + 2**(K_BITS-2) ) &&
												 !( stage_1_k_queue_requested_3_valid && stage_1_k_queue_requested_3 == rd4_k +2**(K_BITS-1) + 2**(K_BITS-2) )
												  ) begin
				queue4_m_axis_ready <= 1'b1;
				stage_1_k_queue_requested_1_valid <= 1'b1;
				rd4_k <= rd4_k  + 1'b1;
				queue4_read <= queue4_read + 1'b1;
				ptr_add_enable <= 1'b1;

				bram_1_raddr <= rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ;
				// bram_2_raddr <= rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ;
				ptr_added_k <= rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ;
				stage_1_k_to_request <= rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ;
				stage_1_k_queue_requested_1 <= rd4_k + 2**(K_BITS-1) + 2**(K_BITS-2) ;
				stage_1_queue_requested<= 1;
			end
			// otherwise, assign the stuff form the length sorter
			else begin
				bram_1_raddr <= k_to_request;
				// bram_2_raddr <= k_to_request;
				stage_1_k_to_request <= k_to_request;
			end

		end

		// in the cycle before, they wanted to get data from the queue 1 - make sure it has arrived...
		if( stage_1_k_queue_requested_1_valid && 
					( queue1_m_axis_valid && queue1_m_axis_ready && ( queue1_real_read +1 < queue1_read ) ||
					!( queue1_m_axis_valid && queue1_m_axis_ready) &&  ( queue1_real_read < queue1_read ) ) ) begin
			queue1_m_axis_ready <= 1;
		end
		if( queue1_m_axis_valid && queue1_m_axis_ready ) begin
			queue1_real_read <= queue1_real_read  + 1;
		end

		// in the cycle before, they wanted to get data from the queue 2 - make sure it has arrived...
		if( stage_1_k_queue_requested_1_valid && 
					( queue2_m_axis_valid && queue2_m_axis_ready && ( queue2_real_read +1 < queue2_read ) ||
					!( queue2_m_axis_valid && queue2_m_axis_ready) &&  ( queue2_real_read < queue2_read ) ) ) begin
			queue2_m_axis_ready <= 1;
		end
		if( queue2_m_axis_valid && queue2_m_axis_ready ) begin
			queue2_real_read <= queue2_real_read  + 1;
		end

		// in the cycle before, they wanted to get data from the queue 3 - make sure it has arrived...
		if( stage_1_k_queue_requested_1_valid && 
					( queue3_m_axis_valid && queue3_m_axis_ready && ( queue3_real_read +1 < queue3_read ) ||
					!( queue3_m_axis_valid && queue3_m_axis_ready) &&  ( queue3_real_read < queue3_read ) ) ) begin
			queue3_m_axis_ready <= 1;
		end
		if( queue3_m_axis_valid && queue3_m_axis_ready ) begin
			queue3_real_read <= queue3_real_read  + 1;
		end

		// in the cycle before, they wanted to get data from the queue 4 - make sure it has arrived...
		if( stage_1_k_queue_requested_1_valid && 
					( queue4_m_axis_valid && queue4_m_axis_ready && ( queue4_real_read +1 < queue4_read ) ||
					!( queue4_m_axis_valid && queue4_m_axis_ready) &&  ( queue4_real_read < queue4_read ) ) ) begin
			queue4_m_axis_ready <= 1;
		end
		if( queue4_m_axis_valid && queue4_m_axis_ready ) begin
			queue4_real_read <= queue4_real_read  + 1;
		end



		stage_1_k_queue_requested_2 <= stage_1_k_queue_requested_1;
		stage_1_k_queue_requested_3 <= stage_1_k_queue_requested_2;
		stage_1_k_queue_requested_2_valid <= stage_1_k_queue_requested_1_valid;
		stage_1_k_queue_requested_3_valid <= stage_1_k_queue_requested_2_valid;
	end
end

//---------------------- Stage 2: RECEIVE DATA, IF THE QUEUE HAS VALID BIT ALREADY HIGH ---------------------------//
reg									stage_1b_valid;
reg [K_BITS-1:0] 					stage_1b_k_to_request;

reg									stage_1b_queue_data_valid;
reg [99:0]							stage_1b_queue_data;

reg [15:0]							stage_1b_valid_debug;
reg									stage_1b_queue_requested;

always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
		stage_1b_valid <= 1'b0;
		stage_1b_queue_data_valid <= 1'b0;

		stage_1b_valid_debug <= 0;
	end
	else begin
		stage_1b_queue_data_valid<= 0;
		stage_1b_valid_debug <= stage_1_valid;
		stage_1b_valid <= stage_1_valid;
		stage_1b_k_to_request <= stage_1_k_to_request;
		stage_1b_queue_requested <= stage_1_queue_requested;
	
		if( stage_1_valid && stage_1_queue_requested ) begin
			if( queue1_m_axis_valid && queue1_m_axis_ready  ) begin
				stage_1b_queue_data_valid <= 1'b1;
				stage_1b_queue_data <= queue1_m_axis_data;
			end
			else if( queue2_m_axis_valid && queue2_m_axis_ready ) begin
				stage_1b_queue_data_valid <= 1'b1;
				stage_1b_queue_data <= queue2_m_axis_data;
			end
			else if( queue3_m_axis_valid && queue3_m_axis_ready ) begin
				stage_1b_queue_data_valid <= 1'b1;
				stage_1b_queue_data <= queue3_m_axis_data;
			end
			else if( queue4_m_axis_valid && queue4_m_axis_ready ) begin
				stage_1b_queue_data_valid <= 1'b1;
				stage_1b_queue_data <= queue4_m_axis_data;
			end
		end
	end
end


//---------------------- Stage 2: RECEIVE DATA, IF THE QUEUE HAS VALID BIT ALREADY HIGH ---------------------------//
reg									stage_2_valid;
reg [K_BITS-1:0] 					stage_2_k_to_request;

reg									stage_2_queue_data_valid;
reg [99:0]							stage_2_queue_data;

reg [15:0]							stage_2_valid_debug;

reg [199:0]							stage_2_bram_1_dout;
// reg [99:0]							stage_2_bram_2_dout;

reg									stage_2_received_from_queue;

reg [31:0]							stage_2_debug_num_passed_ptrs;

always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
		stage_2_valid <= 1'b0;
		stage_2_queue_data_valid <= 1'b0;

		stage_2_valid_debug <= 0;
		stage_2_received_from_queue <= 0;

		stage_2_debug_num_passed_ptrs <= 0;
	end
	else begin
		stage_2_valid_debug <= stage_2_valid;

		stage_2_queue_data_valid<= 0;
		stage_2_valid <= stage_1b_valid;
		stage_2_k_to_request <= stage_1b_k_to_request;
		stage_2_received_from_queue <= 0;

		// a problem that might arise: we have read queue data and given to stage 3, but stage 1b sees the data too, and passes it to us
		// therefore, we accept the queue data from stage 1b only, if we have not received data from the queue
		// so: either we get the data and therefore can receive upcoming data too, or stage 1 saw it with the correct k, and we have to ignore the value
		if( stage_1b_queue_data_valid &&  !stage_2_received_from_queue ) begin
			stage_2_queue_data_valid <= 1'b1;
			stage_2_queue_data <= stage_1b_queue_data;
			stage_2_debug_num_passed_ptrs<= stage_2_debug_num_passed_ptrs  + 1;
		end
		else if( stage_1b_queue_requested ) begin
			if( queue1_m_axis_valid && queue1_m_axis_ready ) begin
				stage_2_queue_data_valid <= 1'b1;
				stage_2_queue_data <= queue1_m_axis_data;
				stage_2_received_from_queue <= 1;
				stage_2_debug_num_passed_ptrs<= stage_2_debug_num_passed_ptrs + 1;
			end
			else if( queue2_m_axis_valid && queue2_m_axis_ready ) begin
				stage_2_queue_data_valid <= 1'b1;
				stage_2_queue_data <= queue2_m_axis_data;
				stage_2_received_from_queue <= 1;
				stage_2_debug_num_passed_ptrs <= stage_2_debug_num_passed_ptrs  +1;
			end
			else if( queue3_m_axis_valid && queue3_m_axis_ready ) begin
				stage_2_queue_data_valid <= 1'b1;
				stage_2_queue_data <= queue3_m_axis_data;
				stage_2_received_from_queue <= 1;
				stage_2_debug_num_passed_ptrs <= stage_2_debug_num_passed_ptrs  +1;
			end
			else if( queue4_m_axis_valid && queue4_m_axis_ready ) begin
				stage_2_queue_data_valid <= 1'b1;
				stage_2_queue_data <= queue4_m_axis_data;
				stage_2_received_from_queue <= 1;
				stage_2_debug_num_passed_ptrs <= stage_2_debug_num_passed_ptrs  +1;
			end
		end

		stage_2_bram_1_dout <= bram_1_dout;
		// stage_2_bram_2_dout <= bram_2_dout;
	end
end


//---------------------- Stage 3: PROCESS AND UPDATE  ---------------------------//
// DATA GIVEN DATA FROM A) BRAM AND BIT IS VALID, AND WE DO NOT HAVE MORE UP TO DATE -> THEN PROCESS, B) FROM QUEUE PUT IN BRAM (NO PROCESS TO PREVENT FAN-IN/OUT) C) IF FROM BRAM, AND WE HAVE MORE UP TO-DATE: THEN USE THIS


// How many buffers we need:
// Assume:
// At 0: we write
// At 1: in buffer 1, bram sees the value
// At 2: in buffer 2, it is in the bram. further, we issue a read request
// At 3: in buffer 3, and bram sees read request
// At 4: we see the value at dout in stage 2
// A 5: we get the value from stage 2
reg [199:0] 			stage_3_pointer_buf_1_data;
reg [K_BITS-1:0]	stage_3_pointer_buf_1_k;
reg 				stage_3_pointer_buf_1_valid;

reg [199:0] 			stage_3_pointer_buf_2_data;
reg [K_BITS-1:0]	stage_3_pointer_buf_2_k;
reg 				stage_3_pointer_buf_2_valid;

reg [199:0] 			stage_3_pointer_buf_3_data;
reg [K_BITS-1:0]	stage_3_pointer_buf_3_k;
reg 				stage_3_pointer_buf_3_valid;

reg [199:0] 			stage_3_pointer_buf_4_data;
reg [K_BITS-1:0]	stage_3_pointer_buf_4_k;
reg 				stage_3_pointer_buf_4_valid;


reg 				stage_3_ptr_consume_enable;
reg [K_BITS-1:0] 	stage_3_ptr_consumed_k;

// to stage 4
reg [K_BITS-1:0] 	stage_3_k_to_request;
reg [99:0]			stage_3_pointer_buf;
reg 				stage_3_pointer_buf_valid;
reg					stage_3_request_valid;


reg [15:0]			stage_3_debug_case;
reg [31:0]			stage_3_debug_epoch;
reg [31:0]			stage_3_debug_u;
reg					stage_3_debug_flag;

always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin

		stage_3_k_to_request <= 0;
		stage_3_ptr_consume_enable <= 0;
		stage_3_request_valid <= 0;
		stage_3_pointer_buf_valid <= 0;

		bram_1_valid <= 0;
		bram_1_we <= 0;

		bram_2_valid <= 0;

		stage_3_pointer_buf_1_valid <= 1'b0;
		stage_3_pointer_buf_2_valid <= 1'b0;
		stage_3_pointer_buf_3_valid <= 1'b0;
		stage_3_pointer_buf_4_valid <= 1'b0;


		stage_3_debug_case <= 0;
		stage_3_debug_flag <= 0;
	end
	else begin

		stage_3_debug_flag <= 0;
		stage_3_debug_epoch <= stage_2_queue_data[99:68];
		stage_3_debug_u <= (stage_2_queue_data[99:68] << K_BITS) + stage_2_k_to_request + 1;
		if( ( (stage_2_queue_data[99:68] << K_BITS) + stage_2_k_to_request + 1) == 16 && stage_2_queue_data_valid ) begin
			stage_3_debug_flag <= 1;
		end


		stage_3_debug_case <= 0;

		// default values
		stage_3_k_to_request <= stage_2_k_to_request;
		stage_3_request_valid <= 0;
		stage_3_pointer_buf_valid <= 0;
		bram_1_we <= 0;
		

		stage_3_ptr_consume_enable <= 0;

		// values, which are driven by valid flags or similar
		bram_1_waddr <= stage_2_k_to_request;
		stage_3_ptr_consumed_k <= stage_2_k_to_request;


			// shift values
			stage_3_pointer_buf_1_valid <= 0; 
			stage_3_pointer_buf_1_k <= stage_2_k_to_request;

			stage_3_pointer_buf_2_data <= stage_3_pointer_buf_1_data;
			stage_3_pointer_buf_2_k <= stage_3_pointer_buf_1_k;
			stage_3_pointer_buf_2_valid <= stage_3_pointer_buf_1_valid;

			stage_3_pointer_buf_3_data <= stage_3_pointer_buf_2_data;
			stage_3_pointer_buf_3_k <= stage_3_pointer_buf_2_k;
			stage_3_pointer_buf_3_valid <= stage_3_pointer_buf_2_valid;

			stage_3_pointer_buf_4_data <= stage_3_pointer_buf_3_data;
			stage_3_pointer_buf_4_k <= stage_3_pointer_buf_3_k;
			stage_3_pointer_buf_4_valid <= stage_3_pointer_buf_3_valid;


		if(stage_2_valid) begin


			// buffer 1
			if( bram_1_valid[stage_2_k_to_request] && k_to_request_valid[stage_2_k_to_request] ) begin

				// select the correct value to process (the first ist the most up to date value)
				if( stage_3_pointer_buf_1_valid && stage_3_pointer_buf_1_k == stage_2_k_to_request ) begin

					stage_3_debug_case <= 3;

					if( stage_3_pointer_buf_1_data[67:36] > 0  ) begin
	
						// put data for the request
						stage_3_pointer_buf_valid <= 1;
						stage_3_request_valid <= 1;
						stage_3_pointer_buf <= stage_3_pointer_buf_1_data;

						if( stage_3_pointer_buf_1_data[67:36]  <= 8 - stage_3_pointer_buf_1_data[35:32] ) begin
							stage_3_ptr_consume_enable <= 1'b1;
							stage_3_pointer_buf_2_valid <= 1'b0;

							// update state
							if( bram_2_valid[stage_2_k_to_request] || stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b1;
								bram_1_we <= 1'b1;
							end
							else begin
								bram_1_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b0; // buf_1 is correct
							end

							// move data
							if( bram_2_valid[stage_2_k_to_request] ) begin
								stage_3_pointer_buf_1_data <= stage_3_pointer_buf_1_data >> 100;
								bram_1_din <= stage_3_pointer_buf_1_data >> 100;
							end
							else if( stage_2_queue_data_valid ) begin
								stage_3_pointer_buf_1_data <= stage_2_queue_data;
								bram_1_din <= stage_2_queue_data;
							end
							
							
						end
						else begin
							stage_3_pointer_buf_1_valid <= 1; // buf_1 is correct
							
							// update the buffer buf_1
							stage_3_pointer_buf_1_data <= stage_3_pointer_buf_1_data;
							stage_3_pointer_buf_1_data[31:0] <= stage_3_pointer_buf_1_data[31:0] + 1; // increase the counter to the next cache line
							stage_3_pointer_buf_1_data[35:32] <= 0; // delete the offset
							stage_3_pointer_buf_1_data[67:36] <= stage_3_pointer_buf_1_data[67:36] - 8 + stage_3_pointer_buf_1_data[35:32]; // num_edges - (8-offset)
							stage_3_pointer_buf_1_data[99:68] <= stage_3_pointer_buf_1_data[99:68];

							// update the BRAM
							bram_1_din <= stage_3_pointer_buf_1_data;
							bram_1_din[31:0] <= stage_3_pointer_buf_1_data[31:0] + 1;
							bram_1_din[35:32] <= 0;
							bram_1_din[67:36] <= stage_3_pointer_buf_1_data[67:36] - 8 + stage_3_pointer_buf_1_data[35:32];
							bram_1_din[99:68] <= stage_3_pointer_buf_1_data[99:68];
							bram_1_we <= 1;

							if( stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b1;
								stage_3_pointer_buf_1_data[199:100] <= stage_2_queue_data;
								bram_1_din[199:100] <= stage_2_queue_data;
							end

						end		
					end
					else begin

							if( bram_2_valid[stage_2_k_to_request] || stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b1;
								bram_1_we <= 1'b1;
							end
							else begin
								bram_1_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b0; // buf_1 is correct
							end

							if( bram_2_valid[stage_2_k_to_request] ) begin
								stage_3_pointer_buf_1_data <= stage_3_pointer_buf_1_data >> 100;
								bram_1_din <= stage_3_pointer_buf_1_data >> 100;
							end
							else if( stage_2_queue_data_valid ) begin
								stage_3_pointer_buf_1_data <= stage_2_queue_data;
								bram_1_din <= stage_2_queue_data;
							end


							stage_3_pointer_buf_2_valid <= 0;
							stage_3_ptr_consume_enable <= 1;

							// for stage 3, to generate the info data
							stage_3_pointer_buf_valid <= 1;
							stage_3_pointer_buf[99:68] <= stage_3_pointer_buf_1_data[99:68];
							stage_3_pointer_buf[67:36] <= 0;
							stage_3_pointer_buf[35:32] <= 0;
							stage_3_pointer_buf[31:0] <= 0;
					end

				end
				// buffer 2
				else if( stage_3_pointer_buf_2_valid && stage_3_pointer_buf_2_k == stage_2_k_to_request ) begin

					stage_3_debug_case <= 4;

					if( stage_3_pointer_buf_2_data[67:36] > 0  ) begin
	
						// put data for the request
						stage_3_pointer_buf_valid <= 1;
						stage_3_request_valid <= 1;
						stage_3_pointer_buf <= stage_3_pointer_buf_2_data;

						if( stage_3_pointer_buf_2_data[67:36] <= 8 - stage_3_pointer_buf_2_data[35:32] ) begin

							if( bram_2_valid[stage_2_k_to_request] || stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b1;
								bram_1_we <= 1'b1;
								// data_transfer <= 1;
							end
							else begin
								bram_1_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b0; // buf_1 is correct
							end

							
							if( bram_2_valid[stage_2_k_to_request] ) begin
								stage_3_pointer_buf_1_data <= stage_3_pointer_buf_2_data >> 100;
								bram_1_din <= stage_3_pointer_buf_2_data >> 100;
							end
							else if( stage_2_queue_data_valid ) begin
								stage_3_pointer_buf_1_data <= stage_2_queue_data;
								bram_1_din <= stage_2_queue_data;
							end

							stage_3_pointer_buf_3_valid <= 0;
							stage_3_ptr_consume_enable <= 1;
						end
						else begin
							stage_3_pointer_buf_1_valid <= 1;
							
							// update the buffer buf_1
							stage_3_pointer_buf_1_data <= stage_3_pointer_buf_2_data;
							stage_3_pointer_buf_1_data[31:0] <= stage_3_pointer_buf_2_data[31:0] + 1 ; // increase the counter to the next cache line
							stage_3_pointer_buf_1_data[35:32] <= 0; // delete the offset
							stage_3_pointer_buf_1_data[67:36] <= stage_3_pointer_buf_2_data[67:36] - 8 + stage_3_pointer_buf_2_data[35:32];
							stage_3_pointer_buf_1_data[99:68] <= stage_3_pointer_buf_2_data[99:68];

							// update the BRAM
							bram_1_din <= stage_3_pointer_buf_2_data;
							bram_1_din[31:0] <= stage_3_pointer_buf_2_data[31:0] + 1;
							bram_1_din[35:32] <= 0;
							bram_1_din[67:36] <= stage_3_pointer_buf_2_data[67:36] - 8 + stage_3_pointer_buf_2_data[35:32];
							bram_1_din[99:68] <=  stage_3_pointer_buf_2_data[99:68];
							bram_1_we <= 1;

							if( stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b1;
								stage_3_pointer_buf_1_data[199:100] <= stage_2_queue_data;
								bram_1_din[199:100] <= stage_2_queue_data;
							end

						end		
					end
					else begin

							if( bram_2_valid[stage_2_k_to_request] || stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b1;
								bram_1_we <= 1'b1;
							end
							else begin
								bram_1_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b0; // buf_1 is correct
							end

							if( bram_2_valid[stage_2_k_to_request] ) begin
								stage_3_pointer_buf_1_data <= stage_3_pointer_buf_2_data >> 100;
								bram_1_din <= stage_3_pointer_buf_2_data >> 100;
							end
							else if( stage_2_queue_data_valid ) begin
								stage_3_pointer_buf_1_data <= stage_2_queue_data;
								bram_1_din <= stage_2_queue_data;
							end


						stage_3_pointer_buf_3_valid <= 0;
						stage_3_ptr_consume_enable <= 1;

						// for stage 3, to generate the info data
						stage_3_pointer_buf_valid <= 1;
						stage_3_pointer_buf[99:68] <= stage_3_pointer_buf_2_data[99:68];
						stage_3_pointer_buf[67:36] <= 0;
						stage_3_pointer_buf[35:32] <= 0;
						stage_3_pointer_buf[31:0] <= 0;
					end


				end
				// buffer 3
				else if( stage_3_pointer_buf_3_valid && stage_3_pointer_buf_3_k == stage_2_k_to_request ) begin

					stage_3_debug_case <= 5;
					
					if( stage_3_pointer_buf_3_data[67:36] > 0  ) begin
	
						// put data for the request
						stage_3_pointer_buf_valid <= 1;
						stage_3_request_valid <= 1;
						stage_3_pointer_buf <= stage_3_pointer_buf_3_data;

						if(  stage_3_pointer_buf_3_data[67:36] <= 8 - stage_3_pointer_buf_3_data[35:32] ) begin

							if( bram_2_valid[stage_2_k_to_request] || stage_2_queue_data_valid  ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b1;
								bram_1_we <= 1'b1;
							end
							else begin
								bram_1_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b0; // buf_1 is correct
							end

							
							if( bram_2_valid[stage_2_k_to_request] ) begin
								stage_3_pointer_buf_1_data <= stage_3_pointer_buf_3_data >> 100 ;
								bram_1_din <= stage_3_pointer_buf_3_data >> 100;
							end
							else if( stage_2_queue_data_valid ) begin
								stage_3_pointer_buf_1_data <= stage_2_queue_data;
								bram_1_din <= stage_2_queue_data;
							end


							stage_3_pointer_buf_4_valid <= 0;
							stage_3_ptr_consume_enable <= 1;
						end
						else begin
							stage_3_pointer_buf_1_valid <= 1; // buf_1 is correct
							
							// update the buffer buf_1
							stage_3_pointer_buf_1_data <= stage_3_pointer_buf_3_data;
							stage_3_pointer_buf_1_data[31:0] <= stage_3_pointer_buf_3_data[31:0] + 1 ; // increase the counter to the next cache line
							stage_3_pointer_buf_1_data[35:32] <= 0; // delete the offset
							stage_3_pointer_buf_1_data[67:36] <= stage_3_pointer_buf_3_data[67:36] - 8 + stage_3_pointer_buf_3_data[35:32];
							stage_3_pointer_buf_1_data[99:68] <= stage_3_pointer_buf_3_data[99:68];
							// update the BRAM
							bram_1_din <= stage_3_pointer_buf_3_data;
							bram_1_din[31:0] <= stage_3_pointer_buf_3_data[31:0] + 1;
							bram_1_din[35:32] <= 0;
							bram_1_din[67:36] <= stage_3_pointer_buf_3_data[67:36] - 8 + stage_3_pointer_buf_3_data[35:32];
							bram_1_din[99:68] <=  stage_3_pointer_buf_3_data[99:68];
							bram_1_we <= 1;

							if( stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b1;
								stage_3_pointer_buf_1_data[199:100] <= stage_2_queue_data;
								bram_1_din[199:100] <= stage_2_queue_data;
							end

						end		
					end
					else begin

							if( bram_2_valid[stage_2_k_to_request] || stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b1;
								bram_1_we <= 1'b1;
								// data_transfer <= 1;
							end
							else begin
								bram_1_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b0; // buf_1 is correct
							end

							if( bram_2_valid[stage_2_k_to_request] ) begin
								stage_3_pointer_buf_1_data <= stage_3_pointer_buf_3_data >> 100;
								bram_1_din <= stage_3_pointer_buf_3_data >> 100;
							end
							else if( stage_2_queue_data_valid ) begin
								stage_3_pointer_buf_1_data <= stage_2_queue_data;
								bram_1_din <= stage_2_queue_data;
							end


						stage_3_pointer_buf_4_valid <= 0;
						stage_3_ptr_consume_enable <= 1;

						// for stage 3, to generate the info data
						stage_3_pointer_buf_valid <= 1;
						stage_3_pointer_buf[99:68] <= stage_3_pointer_buf_3_data[99:68];
						stage_3_pointer_buf[67:36] <= 0;
						stage_3_pointer_buf[35:32] <= 0;
						stage_3_pointer_buf[31:0] <= 0;
					end

				end
				// buffer 4
				else if(  stage_3_pointer_buf_4_valid && stage_3_pointer_buf_4_k == stage_2_k_to_request ) begin

					stage_3_debug_case <= 6;

					if( stage_3_pointer_buf_4_data[67:36] > 0  ) begin
	
						// put data for the request
						stage_3_pointer_buf_valid <= 1;
						stage_3_request_valid <= 1;
						stage_3_pointer_buf <= stage_3_pointer_buf_4_data;

						if(  stage_3_pointer_buf_4_data[67:36] <= 8 - stage_3_pointer_buf_4_data[35:32] ) begin

							if( bram_2_valid[stage_2_k_to_request] || stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b1;
								bram_1_we <= 1'b1;
							end
							else begin
								bram_1_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b0; // buf_1 is correct
							end

							
							if( bram_2_valid[stage_2_k_to_request] ) begin
								stage_3_pointer_buf_1_data <= stage_3_pointer_buf_4_data >> 100;
								bram_1_din <= stage_3_pointer_buf_4_data >> 100;
							end
							else if( stage_2_queue_data_valid ) begin
								stage_3_pointer_buf_1_data <= stage_2_queue_data;
								bram_1_din <= stage_2_queue_data;
							end


							stage_3_ptr_consume_enable <= 1;
						end
						else begin
							stage_3_pointer_buf_1_valid <= 1;
							
							// update the buffer buf_1
							stage_3_pointer_buf_1_data <= stage_3_pointer_buf_4_data;
							stage_3_pointer_buf_1_data[31:0] <= stage_3_pointer_buf_4_data[31:0] + 1 ; // increase the counter to the next cache line
							stage_3_pointer_buf_1_data[35:32] <= 0; // delete the offset
							stage_3_pointer_buf_1_data[67:36] <= stage_3_pointer_buf_4_data[67:36] - 8 + stage_3_pointer_buf_4_data[35:32];
							stage_3_pointer_buf_1_data[99:68] <= stage_3_pointer_buf_4_data[99:68];

							// update the BRAM
							bram_1_din <= stage_3_pointer_buf_4_data;
							bram_1_din[31:0] <= stage_3_pointer_buf_4_data[31:0] + 1;
							bram_1_din[35:32] <= 0;
							bram_1_din[67:36] <= stage_3_pointer_buf_4_data[67:36] - 8 + stage_3_pointer_buf_4_data[35:32];
							bram_1_din[99:68] <= stage_3_pointer_buf_4_data[99:68];
							bram_1_we <= 1;

							if( stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b1;
								stage_3_pointer_buf_1_data[199:100] <= stage_2_queue_data;
								bram_1_din[199:100] <= stage_2_queue_data;
							end

						end		
					end
					else begin

							if( bram_2_valid[stage_2_k_to_request]  || stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b1;
								bram_1_we <= 1'b1;
								// data_transfer <= 1;
							end
							else begin
								bram_1_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b0; // buf_1 is correct
							end

							if( bram_2_valid[stage_2_k_to_request] ) begin
								stage_3_pointer_buf_1_data <= stage_3_pointer_buf_4_data >> 100;
								bram_1_din <= stage_3_pointer_buf_4_data >> 100;
							end
							else if( stage_2_queue_data_valid ) begin
								stage_3_pointer_buf_1_data <= stage_2_queue_data;
								bram_1_din <= stage_2_queue_data;
							end

						stage_3_ptr_consume_enable <= 1;

						// for stage 3, to generate the info data
						stage_3_pointer_buf_valid <= 1;
						stage_3_pointer_buf[99:68] <= stage_3_pointer_buf_4_data[99:68];
						stage_3_pointer_buf[67:36] <= 0;
						stage_3_pointer_buf[35:32] <= 0;
						stage_3_pointer_buf[31:0] <= 0;
					end


				end
				// bram-data
				else begin

					stage_3_debug_case <= 7;

					if( stage_2_bram_1_dout[67:36] > 0  ) begin
	
						// put data for the request
						stage_3_pointer_buf_valid <= 1;
						stage_3_request_valid <= 1;
						stage_3_pointer_buf <= stage_2_bram_1_dout;

						if(  stage_2_bram_1_dout[67:36] <= 8 - stage_2_bram_1_dout[35:32] ) begin
							 
							if( bram_2_valid[stage_2_k_to_request] || stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b1;
								bram_1_we <= 1'b1;
								// data_transfer <= 1;
							end
							else begin
								bram_1_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b0; // buf_1 is correct
							end

							
							if( bram_2_valid[stage_2_k_to_request] ) begin
								stage_3_pointer_buf_1_data <= stage_2_bram_1_dout >> 100;
								bram_1_din <= stage_2_bram_1_dout >> 100;
							end
							else if( stage_2_queue_data_valid ) begin
								stage_3_pointer_buf_1_data <= stage_2_queue_data;
								bram_1_din <= stage_2_queue_data;
							end

							
							stage_3_ptr_consume_enable <= 1;
						end
						else begin
							stage_3_pointer_buf_1_valid <= 1; // buf_1 is correct
							
							// update the buffer
							stage_3_pointer_buf_1_data <= stage_2_bram_1_dout;
							stage_3_pointer_buf_1_data[31:0] <= stage_2_bram_1_dout[31:0] + 1 ; // increase the counter to the next cache line
							stage_3_pointer_buf_1_data[35:32] <= 0; // delete the offset
							stage_3_pointer_buf_1_data[67:36] <= stage_2_bram_1_dout[67:36] - 8 + stage_2_bram_1_dout[35:32];
							stage_3_pointer_buf_1_data[99:68] <= stage_2_bram_1_dout[99:68];

							// update the BRAM
							bram_1_din <= stage_2_bram_1_dout;
							bram_1_din[31:0] <= stage_2_bram_1_dout[31:0] + 1;
							bram_1_din[35:32] <= 0;
							bram_1_din[67:36] <= stage_2_bram_1_dout[67:36] - 8 + stage_2_bram_1_dout[35:32];
							bram_1_din[99:68] <= stage_2_bram_1_dout[99:68];
							bram_1_we <= 1;

							if( stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b1;
								stage_3_pointer_buf_1_data[199:100] <= stage_2_queue_data;
								bram_1_din[199:100] <= stage_2_queue_data;
							end

						end		
					end
					else begin

							if( bram_2_valid[stage_2_k_to_request] || stage_2_queue_data_valid ) begin
								bram_2_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b1;
								bram_1_we <= 1'b1;
								// data_transfer <= 1;
							end
							else begin
								bram_1_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b0; // buf_1 is correct
							end

							if( bram_2_valid[stage_2_k_to_request] ) begin
								stage_3_pointer_buf_1_data <= stage_2_bram_1_dout >> 100;
								bram_1_din <= stage_2_bram_1_dout >> 100;
							end
							else if( stage_2_queue_data_valid ) begin
								stage_3_pointer_buf_1_data <= stage_2_queue_data;
								bram_1_din <= stage_2_queue_data;
							end
						
						stage_3_ptr_consume_enable <= 1;

						// for stage 3, to generate the info data
						stage_3_pointer_buf_valid <= 1;
						stage_3_pointer_buf[99:68] <= stage_2_bram_1_dout[99:68];
						stage_3_pointer_buf[67:36] <= 0;
						stage_3_pointer_buf[35:32] <= 0;
						stage_3_pointer_buf[31:0] <= 0;
					end

				end	// end of all cases, when bram1 is filled

				
			end
			// queue data, which can be directly processed
			else if( !bram_1_valid[stage_2_k_to_request] && k_to_request_valid[stage_2_k_to_request] && stage_2_queue_data_valid ) begin
					stage_3_debug_case <= 8;

					if( stage_2_queue_data[67:36] > 0  ) begin
	
						// put data for the request
						stage_3_pointer_buf_valid <= 1;
						stage_3_request_valid <= 1;
						stage_3_pointer_buf <= stage_2_queue_data;

						if( stage_2_queue_data[67:36] <= 8 - stage_2_queue_data[35:32] ) begin
							
								// bram_1_valid[stage_2_k_to_request] <= 1'b0; not required
								stage_3_pointer_buf_1_valid <= 1'b0; // buf_1 is correct
								stage_3_ptr_consume_enable <= 1;
						end
						else begin
							stage_3_pointer_buf_1_valid <= 1; // buf_1 is correct
							
							// update the buffer
							stage_3_pointer_buf_1_data[31:0] <= stage_2_queue_data[31:0] + 1'b1 ; // increase the counter to the next cache line
							stage_3_pointer_buf_1_data[35:32] <= 0; // delete the offset
							stage_3_pointer_buf_1_data[67:36] <= stage_2_queue_data[67:36] - 8 + stage_2_queue_data[35:32];
							stage_3_pointer_buf_1_data[99:68] <= stage_2_queue_data[99:68];

							// update the BRAM
							bram_1_din[31:0] <= stage_2_queue_data[31:0] + 1'b1;
							bram_1_din[35:32] <= 0;
							bram_1_din[67:36] <= stage_2_queue_data[67:36] - 8 + stage_2_queue_data[35:32];

							bram_1_din[99:68] <= stage_2_queue_data[99:68];

							bram_1_we <= 1;
							bram_1_valid[stage_2_k_to_request] <= 1;
						end		
					end
					else begin

								// bram_1_valid[stage_2_k_to_request] <= 1'b0;
								stage_3_pointer_buf_1_valid <= 1'b0; // buf_1 is correct
								stage_3_ptr_consume_enable <= 1;

								// for stage 3, to generate the info data
								stage_3_pointer_buf_valid <= 1;
								stage_3_pointer_buf[99:68] <= stage_2_queue_data[99:68];
								stage_3_pointer_buf[67:36] <= 0;
								stage_3_pointer_buf[35:32] <= 0;
								stage_3_pointer_buf[31:0] <= 0;
					end

			end
			else begin
				stage_3_debug_case <= 9;

				if( stage_2_queue_data_valid ) begin

					stage_3_pointer_buf_1_valid <= 1;
					if( stage_3_pointer_buf_1_valid && stage_3_pointer_buf_1_k == stage_2_k_to_request ) begin
						stage_3_pointer_buf_1_data <= stage_3_pointer_buf_1_data;
						bram_1_din <= stage_3_pointer_buf_1_data;
						stage_3_pointer_buf_2_valid <= 0;
					end
					else if( stage_3_pointer_buf_2_valid && stage_3_pointer_buf_2_k == stage_2_k_to_request ) begin
						stage_3_pointer_buf_1_data <= stage_3_pointer_buf_2_data;
						bram_1_din <= stage_3_pointer_buf_2_data;
						stage_3_pointer_buf_3_valid <= 0;
					end
					else if( stage_3_pointer_buf_3_valid && stage_3_pointer_buf_3_k == stage_2_k_to_request ) begin
						stage_3_pointer_buf_1_data <= stage_3_pointer_buf_3_data;
						bram_1_din <= stage_3_pointer_buf_3_data;
						stage_3_pointer_buf_4_valid <= 0;
					end
					else if( stage_3_pointer_buf_4_valid && stage_3_pointer_buf_4_k ==  stage_2_k_to_request ) begin
						stage_3_pointer_buf_1_data <= stage_3_pointer_buf_4_data;
						bram_1_din<= stage_3_pointer_buf_4_data;
					end
					else begin
						stage_3_pointer_buf_1_data <= stage_2_bram_1_dout;
						bram_1_din <= stage_2_bram_1_dout;
					end


					stage_3_debug_case <= 1;
					bram_1_we <= 1;

					stage_3_pointer_buf_1_valid <= 1;
					if( bram_1_valid[stage_2_k_to_request] == 0 ) begin
						bram_1_valid[stage_2_k_to_request] <= 1'b1;
						bram_1_din <= stage_2_queue_data;
						stage_3_pointer_buf_1_data <= stage_2_queue_data;
						
					end
					else begin
						bram_2_valid[stage_2_k_to_request] <= 1'b1;
						bram_1_din[199:100] <= stage_2_queue_data;
						stage_3_pointer_buf_1_data[199:100] <= stage_2_queue_data;
					end
				end

			end
		end
	end

end


//---------------------- Stage 4: ISSUE REQUEST ---------------------------//


reg	[31:0]	stage_4_num_removed_pointers;		// counts the number of pointers we have thrown away from the BRAM
reg	[31:0]	stage_4_debug_num_flush_pointers;

reg [31:0] stage_4_num_total_edges_requested;
reg [31:0] stage_4_num_edge_requests;			// count the number of requests we have issued, such we do not overload the system

reg [K_BITS-1:0] 		stage_4_edge_added_k;
reg [3:0]				stage_4_edge_added_num;
reg						stage_4_edge_added_enable;

reg				stage_4_debug_epoch_enable;
reg [31:0]		stage_4_debug_epoch;
reg [31:0]		stage_4_debug_u;
reg [15:0]		stage_4_debug_case;
reg				stage_4_debug_ptr_consume_enable;
reg [8:0]		stage_4_debug_ptr_consumed_k;


reg [15:0]		stage_4_debug_bram_1_valid;
reg [15:0]		stage_4_debug_bram_2_valid;
reg [15:0]		stage_4_debug_rd_k;

always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin

		// output
		queue_write_enable <= 0;
		queue_request_write_enable <= 0;
		queue_request_buf_in <= 0;
		edge_requester_done <= 0;

		// tracking data
		stage_4_num_total_edges_requested <= 0;
		stage_4_num_edge_requests <= 0;
		stage_4_debug_num_flush_pointers <= 0;
		stage_4_num_removed_pointers <= 0;

		// for the sorter
		stage_4_edge_added_enable<= 0;

		stage_4_debug_epoch_enable <= 0;
		stage_4_debug_u <= 0;
		stage_4_debug_case <= 0;
		stage_4_debug_rd_k<= 0;
	end
	else begin
		stage_4_debug_bram_1_valid <= bram_1_valid;
		stage_4_debug_bram_2_valid <= bram_2_valid;
		stage_4_debug_rd_k <= rd1_k;

		stage_4_debug_epoch_enable <= 0;
		stage_4_debug_u <= 0;
		queue_write_enable <= 0;
		queue_request_write_enable <= 0;
		stage_4_edge_added_enable <= 0;
		stage_4_edge_added_k <= stage_3_k_to_request;
		stage_4_debug_case <= stage_3_debug_case;
		stage_4_debug_ptr_consume_enable <= stage_3_ptr_consume_enable;
		stage_4_debug_ptr_consumed_k<= stage_3_ptr_consumed_k;


		if( stage_3_ptr_consume_enable ) begin
			stage_4_num_removed_pointers <= stage_4_num_removed_pointers + 1;
		end

		if( stage_4_num_removed_pointers == num_ptrs) begin		
			edge_requester_done <= 1;
		end
		
		if( stage_3_pointer_buf_valid && ~stage_3_request_valid ) begin
			stage_4_debug_num_flush_pointers <= stage_4_debug_num_flush_pointers +1;
		end

		if( stage_3_pointer_buf_valid ) begin

			stage_4_edge_added_enable <= 1;

			// update tracking variables
			if( stage_3_pointer_buf[67:36] == 0 ) begin
				 stage_4_edge_added_num <=  1;
			end
			else begin
				 stage_4_edge_added_num <= ( ( stage_3_pointer_buf[67:36] >= 8 - stage_3_pointer_buf[35:32]) ? 8 - stage_3_pointer_buf[35:32] : stage_3_pointer_buf[67:36]) ;
			end
			
			// write infos to the "request info" queue
			// pass (0) the epoch, (1) k to compute the u index, (2) the offset, and (3) the number of edges in this cl
			queue_request_write_enable <= 1;
			queue_request_buf_in[127:96] <= stage_3_pointer_buf[99:68];
			queue_request_buf_in[95:64] <= stage_3_k_to_request;
			queue_request_buf_in[35:32] <= stage_3_pointer_buf[35:32];
			queue_request_buf_in[31:0] <= (stage_3_pointer_buf[67:36] >= 8 - stage_3_pointer_buf[35:32]) ? 
											(8 - stage_3_pointer_buf[35:32] ) : stage_3_pointer_buf[67:36] ;


		end

		if( stage_3_request_valid ) begin
			// update tracking variables
			stage_4_num_total_edges_requested <= stage_4_num_total_edges_requested + 
																( ( stage_3_pointer_buf[67:36] >= 8 - stage_3_pointer_buf[35:32]) ? 8 - stage_3_pointer_buf[35:32] : stage_3_pointer_buf[67:36]) ;

			// write request to the request queue
			stage_4_num_edge_requests <= stage_4_num_edge_requests + 1;
			queue_write_enable <= 1;
			queue_buf_in <= { {edges_base_addr + stage_3_pointer_buf[31:0] }, 8'b0} + EDGE_REQUEST_TAG; // compute request address and put tag
			stage_4_debug_epoch <= stage_3_pointer_buf[99:68];
			stage_4_debug_u <= (stage_3_pointer_buf[99:68] << K_BITS) + stage_3_k_to_request + 1;

		end
	end
end


reg [31:0]	num_ptrs_added;
reg [31:0]	num_ptrs_removed;
always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
		num_ptrs_added <= 0;
		num_ptrs_removed <= 0;
	end
	else begin
		if( stage_3_ptr_consume_enable) begin
			num_ptrs_removed <= num_ptrs_removed + 1;
		end

		if( ptr_add_enable) begin
			num_ptrs_added <= num_ptrs_added  + 1;
		end
	end
end


//---------------------- DETERMINE THE QUEUE WE USE TO GENERATE THE NEXT REQUEST ---------------------------//


generate
genvar ii;
begin
	
	wire [K_BITS-1:0] connect_k 	[0:K_BITS-1][0:2**K_BITS-1];
	wire [SORTER_FIRST_QUEUE_LENGTH_BITS:0] connect_edge_length	[0:K_BITS-1][0:2**K_BITS-1];
	wire connect_ptr_length	[0:K_BITS-1][0:2**K_BITS-1];


	if( K_BITS == 3) begin
		
					length_sorter_start8 #(.K_BITS(K_BITS), .QUEUE_1_K(0), .QUEUE_2_K(1), .QUEUE_3_K(2), .QUEUE_4_K(3), 
													.QUEUE_5_K(4), .QUEUE_6_K(5), .QUEUE_7_K(6), .QUEUE_8_K(7), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS)  ) 
						length_sorter_start8 (
							.clk						(clk),
		   					.rst_n						(rst_n),
							.um_done					(um_done),
							.um_activated				(um_activated),
							.edge_requester_done		(edge_requester_done),

							.queue_1_edge_consumed_bit		(sorter_edge_consumed[0]),
							.queue_2_edge_consumed_bit		(sorter_edge_consumed[1]),
							.queue_3_edge_consumed_bit		(sorter_edge_consumed[2]),
							.queue_4_edge_consumed_bit		(sorter_edge_consumed[3]),
							.queue_5_edge_consumed_bit		(sorter_edge_consumed[4]),
							.queue_6_edge_consumed_bit		(sorter_edge_consumed[5]),
							.queue_7_edge_consumed_bit		(sorter_edge_consumed[6]),
							.queue_8_edge_consumed_bit		(sorter_edge_consumed[7]),

							.queue_edge_added_enable	( stage_4_edge_added_enable ),
							.queue_edge_added_num		( stage_4_edge_added_num ),
							.queue_edge_added_k			( stage_4_edge_added_k ),

							.ptr_add_enable				( ptr_add_enable ),
							.ptr_added_k				( ptr_added_k ),
			
							.ptr_consume_enable			( stage_3_ptr_consume_enable ),
							.ptr_consumed_k				( stage_3_ptr_consumed_k ),
					
							.out_k						(k_to_request),
							.out_edge_length			( ),
							.out_ptr_length				( ),

							.k_to_request_valid_1		(k_to_request_valid[0]),
							.k_to_request_valid_2		(k_to_request_valid[1]),
							.k_to_request_valid_3		(k_to_request_valid[2]),
							.k_to_request_valid_4		(k_to_request_valid[3]),
							.k_to_request_valid_5		(k_to_request_valid[4]),
							.k_to_request_valid_6		(k_to_request_valid[5]),
							.k_to_request_valid_7		(k_to_request_valid[6]),
							.k_to_request_valid_8		(k_to_request_valid[7])
							);

	end
	if( K_BITS == 4 ) begin

		for(ii = 0; ii < 2; ii++ ) begin : edge_requester_outer_for_8
					// special case: the first row
					length_sorter_start8 #(.K_BITS(K_BITS), .QUEUE_1_K(8*ii), .QUEUE_2_K(8*ii+1), .QUEUE_3_K(8*ii+2), .QUEUE_4_K(8*ii+3), 
													.QUEUE_5_K(8*ii+4), .QUEUE_6_K(8*ii+5), .QUEUE_7_K(8*ii+6), .QUEUE_8_K(8*ii+7), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS) ) 
						length_sorter_start8 (
							.clk						(clk),
		   					.rst_n						(rst_n),
							.um_done					(um_done),
							.um_activated				(um_activated),
							.edge_requester_done		(edge_requester_done),

							.queue_1_edge_consumed_bit		(sorter_edge_consumed[8*ii]),
							.queue_2_edge_consumed_bit		(sorter_edge_consumed[8*ii+1]),
							.queue_3_edge_consumed_bit		(sorter_edge_consumed[8*ii+2]),
							.queue_4_edge_consumed_bit		(sorter_edge_consumed[8*ii+3]),
							.queue_5_edge_consumed_bit		(sorter_edge_consumed[8*ii+4]),
							.queue_6_edge_consumed_bit		(sorter_edge_consumed[8*ii+5]),
							.queue_7_edge_consumed_bit		(sorter_edge_consumed[8*ii+6]),
							.queue_8_edge_consumed_bit		(sorter_edge_consumed[8*ii+7]),

							.queue_edge_added_enable	( stage_4_edge_added_enable ),
							.queue_edge_added_num		( stage_4_edge_added_num ),
							.queue_edge_added_k			( stage_4_edge_added_k ),

							.ptr_add_enable				( ptr_add_enable ),
							.ptr_added_k				( ptr_added_k ),
			
							.ptr_consume_enable			( stage_3_ptr_consume_enable ),
							.ptr_consumed_k				( stage_3_ptr_consumed_k ),
					
							.out_k						(connect_k[0][ii]),
							.out_edge_length			(connect_edge_length[0][ii]),
							.out_ptr_length				(connect_ptr_length[0][ii]),

							.k_to_request_valid_1		(k_to_request_valid[8*ii]),
							.k_to_request_valid_2		(k_to_request_valid[8*ii+1]),
							.k_to_request_valid_3		(k_to_request_valid[8*ii+2]),
							.k_to_request_valid_4		(k_to_request_valid[8*ii+3]),
							.k_to_request_valid_5		(k_to_request_valid[8*ii+4]),
							.k_to_request_valid_6		(k_to_request_valid[8*ii+5]),
							.k_to_request_valid_7		(k_to_request_valid[8*ii+6]),
							.k_to_request_valid_8		(k_to_request_valid[8*ii+7])
							);
		end
					// special case: the last row
					length_sorter_interm #(.K_BITS(K_BITS), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS)) length_sorter_last_row2(
						.clk						(clk),
	   					.rst_n						(rst_n),
						.um_done					(um_done),
						.um_activated				(um_activated),
						.edge_requester_done		(edge_requester_done),
					
						.queue_1_k					(connect_k[0][0]),
						.queue_1_edge_length		(connect_edge_length[0][0]),
						.queue_1_ptr_length			(connect_ptr_length[0][0] ),

						.queue_2_k					(connect_k[0][1]),
						.queue_2_edge_length		(connect_edge_length[0][1]),
						.queue_2_ptr_length			(connect_ptr_length[0][1]),
					
						.out_k						(k_to_request),
						.out_edge_length			(),
						.out_ptr_length				()
					);
	end
	if( K_BITS == 5 ) begin

		for(ii = 0; ii < 4; ii++ ) begin : edge_requester_outer_for_8
					// special case: the first row
					length_sorter_start8 #(.K_BITS(K_BITS), .QUEUE_1_K(8*ii), .QUEUE_2_K(8*ii+1), .QUEUE_3_K(8*ii+2), .QUEUE_4_K(8*ii+3), 
													.QUEUE_5_K(8*ii+4), .QUEUE_6_K(8*ii+5), .QUEUE_7_K(8*ii+6), .QUEUE_8_K(8*ii+7), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS) ) 
						length_sorter_start8 (
							.clk						(clk),
		   					.rst_n						(rst_n),
							.um_done					(um_done),
							.um_activated				(um_activated),
							.edge_requester_done		(edge_requester_done),

							.queue_1_edge_consumed_bit		(sorter_edge_consumed[8*ii]),
							.queue_2_edge_consumed_bit		(sorter_edge_consumed[8*ii+1]),
							.queue_3_edge_consumed_bit		(sorter_edge_consumed[8*ii+2]),
							.queue_4_edge_consumed_bit		(sorter_edge_consumed[8*ii+3]),
							.queue_5_edge_consumed_bit		(sorter_edge_consumed[8*ii+4]),
							.queue_6_edge_consumed_bit		(sorter_edge_consumed[8*ii+5]),
							.queue_7_edge_consumed_bit		(sorter_edge_consumed[8*ii+6]),
							.queue_8_edge_consumed_bit		(sorter_edge_consumed[8*ii+7]),

							.queue_edge_added_enable	( stage_4_edge_added_enable ),
							.queue_edge_added_num		( stage_4_edge_added_num ),
							.queue_edge_added_k			( stage_4_edge_added_k ),

							.ptr_add_enable				( ptr_add_enable ),
							.ptr_added_k				( ptr_added_k ),
			
							.ptr_consume_enable			( stage_3_ptr_consume_enable ),
							.ptr_consumed_k				( stage_3_ptr_consumed_k ),
					
							.out_k						(connect_k[0][ii]),
							.out_edge_length			(connect_edge_length[0][ii]),
							.out_ptr_length				(connect_ptr_length[0][ii]),

							.k_to_request_valid_1		(k_to_request_valid[8*ii]),
							.k_to_request_valid_2		(k_to_request_valid[8*ii+1]),
							.k_to_request_valid_3		(k_to_request_valid[8*ii+2]),
							.k_to_request_valid_4		(k_to_request_valid[8*ii+3]),
							.k_to_request_valid_5		(k_to_request_valid[8*ii+4]),
							.k_to_request_valid_6		(k_to_request_valid[8*ii+5]),
							.k_to_request_valid_7		(k_to_request_valid[8*ii+6]),
							.k_to_request_valid_8		(k_to_request_valid[8*ii+7])
							);
		end

					length_sorter_interm4 #(.K_BITS(K_BITS), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS)) length_sorter_last_row4(
						.clk						(clk),
	   					.rst_n						(rst_n),
						.um_done					(um_done),
						.um_activated				(um_activated),
						.edge_requester_done		(edge_requester_done),
					
						.queue_1_k					(connect_k[0][0]),
						.queue_1_edge_length		(connect_edge_length[0][0]),
						.queue_1_ptr_length			(connect_ptr_length[0][0] ),

						.queue_2_k					(connect_k[0][1]),
						.queue_2_edge_length		(connect_edge_length[0][1]),
						.queue_2_ptr_length			(connect_ptr_length[0][1]),

						.queue_3_k					(connect_k[0][2]),
						.queue_3_edge_length		(connect_edge_length[0][2]),
						.queue_3_ptr_length			(connect_ptr_length[0][2] ),

						.queue_4_k					(connect_k[0][3]),
						.queue_4_edge_length		(connect_edge_length[0][3]),
						.queue_4_ptr_length			(connect_ptr_length[0][3]),

					
						.out_k						(k_to_request),
						.out_edge_length			(),
						.out_ptr_length				()
					);
	end
	if( K_BITS == 6 ) begin

		for(ii = 0; ii < 8; ii++ ) begin : edge_requester_outer_for_8
					// special case: the first row
					length_sorter_start8 #(.K_BITS(K_BITS), .QUEUE_1_K(8*ii), .QUEUE_2_K(8*ii+1), .QUEUE_3_K(8*ii+2), .QUEUE_4_K(8*ii+3), 
													.QUEUE_5_K(8*ii+4), .QUEUE_6_K(8*ii+5), .QUEUE_7_K(8*ii+6), .QUEUE_8_K(8*ii+7), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS) ) 
						length_sorter_start8 (
							.clk						(clk),
		   					.rst_n						(rst_n),
							.um_done					(um_done),
							.um_activated				(um_activated),
							.edge_requester_done		(edge_requester_done),

							.queue_1_edge_consumed_bit		(sorter_edge_consumed[8*ii]),
							.queue_2_edge_consumed_bit		(sorter_edge_consumed[8*ii+1]),
							.queue_3_edge_consumed_bit		(sorter_edge_consumed[8*ii+2]),
							.queue_4_edge_consumed_bit		(sorter_edge_consumed[8*ii+3]),
							.queue_5_edge_consumed_bit		(sorter_edge_consumed[8*ii+4]),
							.queue_6_edge_consumed_bit		(sorter_edge_consumed[8*ii+5]),
							.queue_7_edge_consumed_bit		(sorter_edge_consumed[8*ii+6]),
							.queue_8_edge_consumed_bit		(sorter_edge_consumed[8*ii+7]),

							.queue_edge_added_enable	( stage_4_edge_added_enable ),
							.queue_edge_added_num		( stage_4_edge_added_num ),
							.queue_edge_added_k			( stage_4_edge_added_k ),

							.ptr_add_enable				( ptr_add_enable ),
							.ptr_added_k				( ptr_added_k ),
			
							.ptr_consume_enable			( stage_3_ptr_consume_enable ),
							.ptr_consumed_k				( stage_3_ptr_consumed_k ),
					
							.out_k						(connect_k[0][ii]),
							.out_edge_length			(connect_edge_length[0][ii]),
							.out_ptr_length				(connect_ptr_length[0][ii]),

							.k_to_request_valid_1		(k_to_request_valid[8*ii]),
							.k_to_request_valid_2		(k_to_request_valid[8*ii+1]),
							.k_to_request_valid_3		(k_to_request_valid[8*ii+2]),
							.k_to_request_valid_4		(k_to_request_valid[8*ii+3]),
							.k_to_request_valid_5		(k_to_request_valid[8*ii+4]),
							.k_to_request_valid_6		(k_to_request_valid[8*ii+5]),
							.k_to_request_valid_7		(k_to_request_valid[8*ii+6]),
							.k_to_request_valid_8		(k_to_request_valid[8*ii+7])
							);
		end

					// special case: the last row
					length_sorter_interm8 #(.K_BITS(K_BITS), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS)) length_sorter_last_row8(
						.clk						(clk),
	   					.rst_n						(rst_n),
						.um_done					(um_done),
						.um_activated				(um_activated),
						.edge_requester_done		(edge_requester_done),
					
						.queue_1_k					(connect_k[0][0]),
						.queue_1_edge_length		(connect_edge_length[0][0]),
						.queue_1_ptr_length			(connect_ptr_length[0][0] ),

						.queue_2_k					(connect_k[0][1]),
						.queue_2_edge_length		(connect_edge_length[0][1]),
						.queue_2_ptr_length			(connect_ptr_length[0][1]),

						.queue_3_k					(connect_k[0][2]),
						.queue_3_edge_length		(connect_edge_length[0][2]),
						.queue_3_ptr_length			(connect_ptr_length[0][2]),

						.queue_4_k					(connect_k[0][3]),
						.queue_4_edge_length		(connect_edge_length[0][3]),
						.queue_4_ptr_length			(connect_ptr_length[0][3]),

						.queue_5_k					(connect_k[0][4]),
						.queue_5_edge_length		(connect_edge_length[0][4]),
						.queue_5_ptr_length			(connect_ptr_length[0][4] ),

						.queue_6_k					(connect_k[0][5]),
						.queue_6_edge_length		(connect_edge_length[0][5]),
						.queue_6_ptr_length			(connect_ptr_length[0][5]),

						.queue_7_k					(connect_k[0][6]),
						.queue_7_edge_length		(connect_edge_length[0][6]),
						.queue_7_ptr_length			(connect_ptr_length[ 0][6]),

						.queue_8_k					(connect_k[0][7]),
						.queue_8_edge_length		(connect_edge_length[0][7]),
						.queue_8_ptr_length			(connect_ptr_length[0][7]),
					
						.out_k						(k_to_request),
						.out_edge_length			(),
						.out_ptr_length				()
					);
	end
	if( K_BITS == 7 ) begin

		for(ii = 0; ii < 16; ii++ ) begin : edge_requester_outer_for_8
					// special case: the first row
					length_sorter_start8 #(.K_BITS(K_BITS), .QUEUE_1_K(8*ii), .QUEUE_2_K(8*ii+1), .QUEUE_3_K(8*ii+2), .QUEUE_4_K(8*ii+3), 
													.QUEUE_5_K(8*ii+4), .QUEUE_6_K(8*ii+5), .QUEUE_7_K(8*ii+6), .QUEUE_8_K(8*ii+7), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS) ) 
						length_sorter_start8 (
							.clk						(clk),
		   					.rst_n						(rst_n),
							.um_done					(um_done),
							.um_activated				(um_activated),
							.edge_requester_done		(edge_requester_done),

							.queue_1_edge_consumed_bit		(sorter_edge_consumed[8*ii]),
							.queue_2_edge_consumed_bit		(sorter_edge_consumed[8*ii+1]),
							.queue_3_edge_consumed_bit		(sorter_edge_consumed[8*ii+2]),
							.queue_4_edge_consumed_bit		(sorter_edge_consumed[8*ii+3]),
							.queue_5_edge_consumed_bit		(sorter_edge_consumed[8*ii+4]),
							.queue_6_edge_consumed_bit		(sorter_edge_consumed[8*ii+5]),
							.queue_7_edge_consumed_bit		(sorter_edge_consumed[8*ii+6]),
							.queue_8_edge_consumed_bit		(sorter_edge_consumed[8*ii+7]),

							.queue_edge_added_enable	( stage_4_edge_added_enable ),
							.queue_edge_added_num		( stage_4_edge_added_num ),
							.queue_edge_added_k			( stage_4_edge_added_k ),

							.ptr_add_enable				( ptr_add_enable ),
							.ptr_added_k				( ptr_added_k ),
			
							.ptr_consume_enable			( stage_3_ptr_consume_enable ),
							.ptr_consumed_k				( stage_3_ptr_consumed_k ),
					
							.out_k						(connect_k[0][ii]),
							.out_edge_length			(connect_edge_length[0][ii]),
							.out_ptr_length				(connect_ptr_length[0][ii]),

							.k_to_request_valid_1		(k_to_request_valid[8*ii]),
							.k_to_request_valid_2		(k_to_request_valid[8*ii+1]),
							.k_to_request_valid_3		(k_to_request_valid[8*ii+2]),
							.k_to_request_valid_4		(k_to_request_valid[8*ii+3]),
							.k_to_request_valid_5		(k_to_request_valid[8*ii+4]),
							.k_to_request_valid_6		(k_to_request_valid[8*ii+5]),
							.k_to_request_valid_7		(k_to_request_valid[8*ii+6]),
							.k_to_request_valid_8		(k_to_request_valid[8*ii+7])
							);
		end

		for(ii = 0; ii < 2; ii++ ) begin : edge_requester_inner_for_8
			length_sorter_interm8 #(.K_BITS(K_BITS), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS)) length_sorter_interm8(
						.clk						(clk),
	   					.rst_n						(rst_n),
						.um_done					(um_done),
						.um_activated				(um_activated),
						.edge_requester_done		(edge_requester_done),
					
						.queue_1_k					(connect_k[0][8*ii]),
						.queue_1_edge_length		(connect_edge_length[0][8*ii]),
						.queue_1_ptr_length			(connect_ptr_length[0][8*ii] ),

						.queue_2_k					(connect_k[0][8*ii+1]),
						.queue_2_edge_length		(connect_edge_length[0][8*ii+1]),
						.queue_2_ptr_length			(connect_ptr_length[0][8*ii+1]),

						.queue_3_k					(connect_k[0][8*ii+2]),
						.queue_3_edge_length		(connect_edge_length[0][8*ii+2]),
						.queue_3_ptr_length			(connect_ptr_length[0][8*ii+2]),

						.queue_4_k					(connect_k[0][8*ii+3]),
						.queue_4_edge_length		(connect_edge_length[0][8*ii+3]),
						.queue_4_ptr_length			(connect_ptr_length[0][8*ii+3]),

						.queue_5_k					(connect_k[0][8*ii+4]),
						.queue_5_edge_length		(connect_edge_length[0][8*ii+4]),
						.queue_5_ptr_length			(connect_ptr_length[0][8*ii+4] ),

						.queue_6_k					(connect_k[0][8*ii+5]),
						.queue_6_edge_length		(connect_edge_length[0][8*ii+5]),
						.queue_6_ptr_length			(connect_ptr_length[0][8*ii+5]),

						.queue_7_k					(connect_k[0][8*ii+6]),
						.queue_7_edge_length		(connect_edge_length[0][8*ii+6]),
						.queue_7_ptr_length			(connect_ptr_length[0][8*ii+6]),

						.queue_8_k					(connect_k[0][8*ii+7]),
						.queue_8_edge_length		(connect_edge_length[0][8*ii+7]),
						.queue_8_ptr_length			(connect_ptr_length[0][8*ii+7]),
					
						.out_k						(connect_k[1][ii]),
						.out_edge_length			(connect_edge_length[1][ii]),
						.out_ptr_length				(connect_ptr_length[1][ii])
					);
		end

					// special case: the last row
					length_sorter_interm #(.K_BITS(K_BITS), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS)) length_sorter_last_row2(
						.clk						(clk),
	   					.rst_n						(rst_n),
						.um_done					(um_done),
						.um_activated				(um_activated),
						.edge_requester_done		(edge_requester_done),
					
						.queue_1_k					(connect_k[1][0]),
						.queue_1_edge_length		(connect_edge_length[1][0]),
						.queue_1_ptr_length			(connect_ptr_length[1][0] ),

						.queue_2_k					(connect_k[1][1]),
						.queue_2_edge_length		(connect_edge_length[1][1]),
						.queue_2_ptr_length			(connect_ptr_length[1][1]),
					
						.out_k						(k_to_request),
						.out_edge_length			(),
						.out_ptr_length				()
					);

	end
	if( K_BITS == 8) begin
			for(ii = 0; ii < 32; ii++ ) begin : edge_requester_outer_for_8
					// special case: the first row
					length_sorter_start8 #(.K_BITS(K_BITS), .QUEUE_1_K(8*ii), .QUEUE_2_K(8*ii+1), .QUEUE_3_K(8*ii+2), .QUEUE_4_K(8*ii+3), 
													.QUEUE_5_K(8*ii+4), .QUEUE_6_K(8*ii+5), .QUEUE_7_K(8*ii+6), .QUEUE_8_K(8*ii+7), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS) ) 
						length_sorter_start8 (
							.clk						(clk),
		   					.rst_n						(rst_n),
							.um_done					(um_done),
							.um_activated				(um_activated),
							.edge_requester_done		(edge_requester_done),

							.queue_1_edge_consumed_bit		(sorter_edge_consumed[8*ii]),
							.queue_2_edge_consumed_bit		(sorter_edge_consumed[8*ii+1]),
							.queue_3_edge_consumed_bit		(sorter_edge_consumed[8*ii+2]),
							.queue_4_edge_consumed_bit		(sorter_edge_consumed[8*ii+3]),
							.queue_5_edge_consumed_bit		(sorter_edge_consumed[8*ii+4]),
							.queue_6_edge_consumed_bit		(sorter_edge_consumed[8*ii+5]),
							.queue_7_edge_consumed_bit		(sorter_edge_consumed[8*ii+6]),
							.queue_8_edge_consumed_bit		(sorter_edge_consumed[8*ii+7]),

							.queue_edge_added_enable	( stage_4_edge_added_enable ),
							.queue_edge_added_num		( stage_4_edge_added_num ),
							.queue_edge_added_k			( stage_4_edge_added_k ),

							.ptr_add_enable				( ptr_add_enable ),
							.ptr_added_k				( ptr_added_k ),
			
							.ptr_consume_enable			( stage_3_ptr_consume_enable ),
							.ptr_consumed_k				( stage_3_ptr_consumed_k ),
					
							.out_k						(connect_k[0][ii]),
							.out_edge_length			(connect_edge_length[0][ii]),
							.out_ptr_length				(connect_ptr_length[0][ii]),

							.k_to_request_valid_1		(k_to_request_valid[8*ii]),
							.k_to_request_valid_2		(k_to_request_valid[8*ii+1]),
							.k_to_request_valid_3		(k_to_request_valid[8*ii+2]),
							.k_to_request_valid_4		(k_to_request_valid[8*ii+3]),
							.k_to_request_valid_5		(k_to_request_valid[8*ii+4]),
							.k_to_request_valid_6		(k_to_request_valid[8*ii+5]),
							.k_to_request_valid_7		(k_to_request_valid[8*ii+6]),
							.k_to_request_valid_8		(k_to_request_valid[8*ii+7])
							);
		end

		for(ii = 0; ii < 4; ii++ ) begin : edge_requester_inner_for_8
			length_sorter_interm8 #(.K_BITS(K_BITS), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS)) length_sorter_interm8(
						.clk						(clk),
	   					.rst_n						(rst_n),
						.um_done					(um_done),
						.um_activated				(um_activated),
						.edge_requester_done		(edge_requester_done),
					
						.queue_1_k					(connect_k[0][8*ii]),
						.queue_1_edge_length		(connect_edge_length[0][8*ii]),
						.queue_1_ptr_length			(connect_ptr_length[0][8*ii] ),

						.queue_2_k					(connect_k[0][8*ii+1]),
						.queue_2_edge_length		(connect_edge_length[0][8*ii+1]),
						.queue_2_ptr_length			(connect_ptr_length[0][8*ii+1]),

						.queue_3_k					(connect_k[0][8*ii+2]),
						.queue_3_edge_length		(connect_edge_length[0][8*ii+2]),
						.queue_3_ptr_length			(connect_ptr_length[0][8*ii+2]),

						.queue_4_k					(connect_k[0][8*ii+3]),
						.queue_4_edge_length		(connect_edge_length[0][8*ii+3]),
						.queue_4_ptr_length			(connect_ptr_length[0][8*ii+3]),

						.queue_5_k					(connect_k[0][8*ii+4]),
						.queue_5_edge_length		(connect_edge_length[0][8*ii+4]),
						.queue_5_ptr_length			(connect_ptr_length[0][8*ii+4] ),

						.queue_6_k					(connect_k[0][8*ii+5]),
						.queue_6_edge_length		(connect_edge_length[0][8*ii+5]),
						.queue_6_ptr_length			(connect_ptr_length[0][8*ii+5]),

						.queue_7_k					(connect_k[0][8*ii+6]),
						.queue_7_edge_length		(connect_edge_length[0][8*ii+6]),
						.queue_7_ptr_length			(connect_ptr_length[0][8*ii+6]),

						.queue_8_k					(connect_k[0][8*ii+7]),
						.queue_8_edge_length		(connect_edge_length[0][8*ii+7]),
						.queue_8_ptr_length			(connect_ptr_length[0][8*ii+7]),
					
						.out_k						(connect_k[1][ii]),
						.out_edge_length			(connect_edge_length[1][ii]),
						.out_ptr_length				(connect_ptr_length[1][ii])
					);
		end


			length_sorter_interm4 #(.K_BITS(K_BITS), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS)) length_sorter_last_row4(
						.clk						(clk),
	   					.rst_n						(rst_n),
						.um_done					(um_done),
						.um_activated				(um_activated),
						.edge_requester_done		(edge_requester_done),
					
						.queue_1_k					(connect_k[1][0]),
						.queue_1_edge_length		(connect_edge_length[1][0]),
						.queue_1_ptr_length			(connect_ptr_length[1][0] ),

						.queue_2_k					(connect_k[1][1]),
						.queue_2_edge_length		(connect_edge_length[1][1]),
						.queue_2_ptr_length			(connect_ptr_length[1][1]),

						.queue_3_k					(connect_k[1][2]),
						.queue_3_edge_length		(connect_edge_length[1][2]),
						.queue_3_ptr_length			(connect_ptr_length[1][2] ),

						.queue_4_k					(connect_k[1][3]),
						.queue_4_edge_length		(connect_edge_length[1][3]),
						.queue_4_ptr_length			(connect_ptr_length[1][3]),

						.out_k						(k_to_request),
						.out_edge_length			(),
						.out_ptr_length				()
					);
		
	end
	if( K_BITS == 9 ) begin
		for(ii = 0; ii < 64; ii++ ) begin : edge_requester_outer_for_8
					// special case: the first row
					length_sorter_start8 #(.K_BITS(K_BITS), .QUEUE_1_K(8*ii), .QUEUE_2_K(8*ii+1), .QUEUE_3_K(8*ii+2), .QUEUE_4_K(8*ii+3), 
													.QUEUE_5_K(8*ii+4), .QUEUE_6_K(8*ii+5), .QUEUE_7_K(8*ii+6), .QUEUE_8_K(8*ii+7), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS) ) 
						length_sorter_start8 (
							.clk						(clk),
		   					.rst_n						(rst_n),
							.um_done					(um_done),
							.um_activated				(um_activated),
							.edge_requester_done		(edge_requester_done),

							.queue_1_edge_consumed_bit		(sorter_edge_consumed[8*ii]),
							.queue_2_edge_consumed_bit		(sorter_edge_consumed[8*ii+1]),
							.queue_3_edge_consumed_bit		(sorter_edge_consumed[8*ii+2]),
							.queue_4_edge_consumed_bit		(sorter_edge_consumed[8*ii+3]),
							.queue_5_edge_consumed_bit		(sorter_edge_consumed[8*ii+4]),
							.queue_6_edge_consumed_bit		(sorter_edge_consumed[8*ii+5]),
							.queue_7_edge_consumed_bit		(sorter_edge_consumed[8*ii+6]),
							.queue_8_edge_consumed_bit		(sorter_edge_consumed[8*ii+7]),

							.queue_edge_added_enable	( stage_4_edge_added_enable ),
							.queue_edge_added_num		( stage_4_edge_added_num ),
							.queue_edge_added_k			( stage_4_edge_added_k ),

							.ptr_add_enable				( ptr_add_enable ),
							.ptr_added_k				( ptr_added_k ),
			
							.ptr_consume_enable			( stage_3_ptr_consume_enable ),
							.ptr_consumed_k				( stage_3_ptr_consumed_k ),
					
							.out_k						(connect_k[0][ii]),
							.out_edge_length			(connect_edge_length[0][ii]),
							.out_ptr_length				(connect_ptr_length[0][ii]),

							.k_to_request_valid_1		(k_to_request_valid[8*ii]),
							.k_to_request_valid_2		(k_to_request_valid[8*ii+1]),
							.k_to_request_valid_3		(k_to_request_valid[8*ii+2]),
							.k_to_request_valid_4		(k_to_request_valid[8*ii+3]),
							.k_to_request_valid_5		(k_to_request_valid[8*ii+4]),
							.k_to_request_valid_6		(k_to_request_valid[8*ii+5]),
							.k_to_request_valid_7		(k_to_request_valid[8*ii+6]),
							.k_to_request_valid_8		(k_to_request_valid[8*ii+7])
							);
		end


		for(ii = 0; ii < 8; ii++ ) begin : edge_requester_inner_for_8
			length_sorter_interm8 #(.K_BITS(K_BITS), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS)) length_sorter_interm8(
						.clk						(clk),
	   					.rst_n						(rst_n),
						.um_done					(um_done),
						.um_activated				(um_activated),
						.edge_requester_done		(edge_requester_done),
					
						.queue_1_k					(connect_k[0][8*ii]),
						.queue_1_edge_length		(connect_edge_length[0][8*ii]),
						.queue_1_ptr_length			(connect_ptr_length[0][8*ii] ),

						.queue_2_k					(connect_k[0][8*ii+1]),
						.queue_2_edge_length		(connect_edge_length[0][8*ii+1]),
						.queue_2_ptr_length			(connect_ptr_length[0][8*ii+1]),

						.queue_3_k					(connect_k[0][8*ii+2]),
						.queue_3_edge_length		(connect_edge_length[0][8*ii+2]),
						.queue_3_ptr_length			(connect_ptr_length[0][8*ii+2]),

						.queue_4_k					(connect_k[0][8*ii+3]),
						.queue_4_edge_length		(connect_edge_length[0][8*ii+3]),
						.queue_4_ptr_length			(connect_ptr_length[0][8*ii+3]),

						.queue_5_k					(connect_k[0][8*ii+4]),
						.queue_5_edge_length		(connect_edge_length[0][8*ii+4]),
						.queue_5_ptr_length			(connect_ptr_length[0][8*ii+4] ),

						.queue_6_k					(connect_k[0][8*ii+5]),
						.queue_6_edge_length		(connect_edge_length[0][8*ii+5]),
						.queue_6_ptr_length			(connect_ptr_length[0][8*ii+5]),

						.queue_7_k					(connect_k[0][8*ii+6]),
						.queue_7_edge_length		(connect_edge_length[0][8*ii+6]),
						.queue_7_ptr_length			(connect_ptr_length[0][8*ii+6]),

						.queue_8_k					(connect_k[0][8*ii+7]),
						.queue_8_edge_length		(connect_edge_length[0][8*ii+7]),
						.queue_8_ptr_length			(connect_ptr_length[0][8*ii+7]),
					
						.out_k						(connect_k[1][ii]),
						.out_edge_length			(connect_edge_length[1][ii]),
						.out_ptr_length				(connect_ptr_length[1][ii])
					);
		end

					// special case: the last row
					length_sorter_interm8 #(.K_BITS(K_BITS), .SORTER_QUEUE_LENGTH_BITS( SORTER_FIRST_QUEUE_LENGTH_BITS ), 
													.PTR_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS)) length_sorter_last_row8(
						.clk						(clk),
	   					.rst_n						(rst_n),
						.um_done					(um_done),
						.um_activated				(um_activated),
						.edge_requester_done		(edge_requester_done),
					
						.queue_1_k					(connect_k[1][0]),
						.queue_1_edge_length		(connect_edge_length[1][0]),
						.queue_1_ptr_length			(connect_ptr_length[1][0] ),

						.queue_2_k					(connect_k[1][1]),
						.queue_2_edge_length		(connect_edge_length[1][1]),
						.queue_2_ptr_length			(connect_ptr_length[1][1]),

						.queue_3_k					(connect_k[1][2]),
						.queue_3_edge_length		(connect_edge_length[1][2]),
						.queue_3_ptr_length			(connect_ptr_length[1][2]),

						.queue_4_k					(connect_k[1][3]),
						.queue_4_edge_length		(connect_edge_length[1][3]),
						.queue_4_ptr_length			(connect_ptr_length[1][3]),

						.queue_5_k					(connect_k[1][4]),
						.queue_5_edge_length		(connect_edge_length[1][4]),
						.queue_5_ptr_length			(connect_ptr_length[1][4] ),

						.queue_6_k					(connect_k[1][5]),
						.queue_6_edge_length		(connect_edge_length[1][5]),
						.queue_6_ptr_length			(connect_ptr_length[1][5]),

						.queue_7_k					(connect_k[1][6]),
						.queue_7_edge_length		(connect_edge_length[1][6]),
						.queue_7_ptr_length			(connect_ptr_length[ 1][6]),

						.queue_8_k					(connect_k[1][7]),
						.queue_8_edge_length		(connect_edge_length[1][7]),
						.queue_8_ptr_length			(connect_ptr_length[1][7]),
					
						.out_k						(k_to_request),
						.out_edge_length			(),
						.out_ptr_length				()
					);
	end

end
endgenerate


// count the clock cycles, for which we could have issued some request
reg [31:0]	debug_writeable_clk;

always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin	
		debug_writeable_clk <= 0;
	end
	else begin

		// we can issue a request, if there is a valid queue, and there is a valid queue
		if( ~queue_buf_filled_next && ~queue_request_buf_filled_next && k_to_request_valid ) begin
			debug_writeable_clk <= debug_writeable_clk + 1;
		end

	end
end


always @(posedge clk) begin
	debug_edge_requester <= {32'b0, 32'b0,
								32'b0, 32'b0,
						16'b0, stage_3_debug_case, stage_2_valid_debug, stage_1_valid_debug,
						queue1_write, queue1_read,
						stage_4_debug_rd_k, stage_4_debug_case, stage_4_debug_bram_2_valid, stage_4_debug_bram_1_valid,
						stage_4_num_total_edges_requested, stage_4_num_edge_requests,
						stage_4_num_removed_pointers, stage_1_k_to_request_32b,
						stage_1_k_to_request_valid_32b, stage_4_debug_num_flush_pointers};
end


endmodule




`default_nettype wire






