
`default_nettype none

/**
	The Edge Writer gets a stream of edges that must be written back to main-memory, since they belong to the matching.
	This module takes them, puts them into a cache-line, and writes them to the Write Interface, if the cache-line is filled.
*/

module edge_writer  #(parameter OVERFLOW_PADDING = 6, parameter EDGE_OUT_TAG = 14, parameter BRAM_QUEUE_LENGTH_BITS= 6, parameter L_BITS = 3)  (
	//---------------------- GENERAL VARIABLES---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input wire 										um_done,						// true, if we are done processing (global)
	input wire 										um_activated,					// true, if the afu is activated
	input wire[57:0]								out_edges_base_addr,
	input wire[31:0]								out_edges_offset_addr,

	output reg										edge_writer_done,
	input wire										edge_processor_done,

	//---------------------- EDGES FROM EDGE PROCESSOR ---------------------------//
	input wire 										edge_queue_write_enable,
	input wire [128-1:0]							edge_queue_buf_in,
	output reg 										edge_queue_buf_filled_next,

	//---------------------- CL TO WRITE INTERFACE ---------------------------//
	output reg 										write_queue_write_enable,
	output reg [569:0]								write_queue_buf_in,
	input wire										write_queue_buf_filled_next,
	output reg[511:0]								debug_edge_writer

);

// put the stuff into a temp queue, then into the write interface queue -> use a shift register to handle the "cache-lining"


reg									edge_m_axis_ready;
wire								edge_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	edge_data_count;
wire [127:0]						edge_m_axis_data;

reg			edge_s_axis_write_enable;
reg	[127:0]	edge_s_axis_data;

reg			edge_buffer_1_valid;
reg [127:0]	edge_buffer_1;
reg			edge_buffer_2_valid;
reg [127:0]	edge_buffer_2;


nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(128), .PROG_FULL_THRESH(0 ) ) edge_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(edge_s_axis_write_enable),
    .s_axis_ready(),
    .s_axis_data(edge_s_axis_data),
    .s_axis_last(1'b0),
    .m_axis_valid(edge_m_axis_valid),
    .m_axis_ready(edge_m_axis_ready),
    .m_axis_data(edge_m_axis_data),
    .m_axis_last(),
    .data_count(edge_data_count)
);

reg [31:0] edge_ntotal_read;
reg [31:0] edge_ntotal_write;


always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin

		edge_ntotal_write <= 0;
		edge_queue_buf_filled_next <= 0;
		edge_s_axis_write_enable <= 0;
	end
	else begin
		edge_s_axis_write_enable<= 0;
		if( edge_queue_write_enable ) begin
			edge_s_axis_data <= { edge_queue_buf_in[31:0] - ( 32'd512 - 2**L_BITS)+1, edge_queue_buf_in[63:32], edge_queue_buf_in[95:64], edge_queue_buf_in[127:96]};
			edge_ntotal_write <= edge_ntotal_write +1;
			edge_s_axis_write_enable<= 1;
		end
		edge_queue_buf_filled_next <=  ( edge_data_count + OVERFLOW_PADDING) >= (2**BRAM_QUEUE_LENGTH_BITS  );
	end

end




reg bram_cl_1_re;
reg bram_cl_1_we;
reg [L_BITS-1:0] 		bram_cl_1_raddr;
reg [L_BITS-1:0] 		bram_cl_1_waddr;
reg [127:0]				bram_cl_1_din;
wire [127:0]			bram_cl_1_dout;
reg [2**L_BITS-1:0]		bram_cl_1_valid;

spl_sdp_mem_t #( .DATA_WIDTH(128), .ADDR_WIDTH(L_BITS) ) bram_cl_1 (
    .clk(clk),
    .we(bram_cl_1_we),
	.re(bram_cl_1_re),
    .raddr(bram_cl_1_raddr),
    .waddr(bram_cl_1_waddr),
    .din(bram_cl_1_din),
    .dout(bram_cl_1_dout)
);


reg bram_cl_2_re;
reg bram_cl_2_we;
reg [L_BITS-1:0] 		bram_cl_2_raddr;
reg [L_BITS-1:0] 		bram_cl_2_waddr;
reg [127:0]				bram_cl_2_din;
wire [127:0]			bram_cl_2_dout;
reg [2**L_BITS-1:0]		bram_cl_2_valid;

spl_sdp_mem_t #( .DATA_WIDTH(128), .ADDR_WIDTH(L_BITS) ) bram_cl_2 (
    .clk(clk),
    .we(bram_cl_2_we),
	.re(bram_cl_2_re),
    .raddr(bram_cl_2_raddr),
    .waddr(bram_cl_2_waddr),
    .din(bram_cl_2_din),
    .dout(bram_cl_2_dout)
);

reg bram_cl_3_re;
reg bram_cl_3_we;
reg [L_BITS-1:0] 		bram_cl_3_raddr;
reg [L_BITS-1:0] 		bram_cl_3_waddr;
reg [127:0]				bram_cl_3_din;
wire [127:0]			bram_cl_3_dout;
reg [2**L_BITS-1:0]		bram_cl_3_valid;

spl_sdp_mem_t #( .DATA_WIDTH(128), .ADDR_WIDTH(L_BITS) ) bram_cl_3 (
    .clk(clk),
    .we(bram_cl_3_we),
	.re(bram_cl_3_re),
    .raddr(bram_cl_3_raddr),
    .waddr(bram_cl_3_waddr),
    .din(bram_cl_3_din),
    .dout(bram_cl_3_dout)
);

/*
reg bram_cl_4_re;
reg bram_cl_4_we;
reg [L_BITS-1:0] 		bram_cl_4_raddr;
reg [L_BITS-1:0] 		bram_cl_4_waddr;
reg [127:0]				bram_cl_4_din;
wire [127:0]			bram_cl_4_dout;
reg [2**L_BITS-1:0]		bram_cl_4_valid;

spl_sdp_mem_t #( .DATA_WIDTH(128), .ADDR_WIDTH(L_BITS) ) bram_cl_4 (
    .clk(clk),
    .we(bram_cl_4_we),
	.re(bram_cl_4_re),
    .raddr(bram_cl_4_raddr),
    .waddr(bram_cl_4_waddr),
    .din(bram_cl_4_din),
    .dout(bram_cl_4_dout)
);

*/

reg bram_address_we;
reg bram_address_re;
reg [L_BITS-1:0] 		bram_address_raddr;
reg [L_BITS-1:0] 		bram_address_waddr;
reg [31:0]				bram_address_din;
wire [31:0]				bram_address_dout;

spl_sdp_mem_t #( .DATA_WIDTH(32), .ADDR_WIDTH(L_BITS) ) bram_address (
    .clk(clk),
    .we(bram_address_we),
	.re(bram_address_re),
    .raddr(bram_address_raddr),
    .waddr(bram_address_waddr),
    .din(bram_address_din),
    .dout(bram_address_dout)
);



reg								stage_1_last_edge_valid;
reg	[L_BITS-1:0]				stage_1_last_edge_cnt;
reg [31:0]						stage_1_last_edge_timeout;

reg [31:0]						stage_1_processed_edges;

reg [127:0]						stage_1_edge;
reg	[L_BITS-1:0]				stage_1_level;
reg 							stage_1_valid;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		// local variables
		stage_1_processed_edges <= 0;

		// handling last for flush
		stage_1_last_edge_valid <= 0;
		stage_1_last_edge_timeout <= 0;
		stage_1_last_edge_cnt <= 0;

		// queue vars
		// drain queues
		if( (um_done || ~um_activated) && rst_n ) begin
			edge_m_axis_ready <= 1;
		end
		else begin
			edge_m_axis_ready <= 0;
		end
		edge_buffer_1_valid <= 0;
		edge_buffer_2_valid <= 0;

		// to stage 2
		stage_1_valid <= 0;

		bram_cl_1_raddr <= 0;
		bram_cl_2_raddr <= 0;
		bram_cl_3_raddr <= 0;
		// bram_cl_4_raddr <= 0;

		bram_cl_1_re <= 0;
		bram_cl_2_re <= 0;
		bram_cl_3_re <= 0;
		// bram_cl_4_re <= 0;

		bram_address_re <= 0;
	end
	else begin
		stage_1_valid <= 0;

		bram_cl_1_re <= 0;
		bram_cl_2_re <= 0;
		bram_cl_3_re <= 0;
		// bram_cl_4_re <= 0;
		bram_address_re <= 0;

		if( ~edge_writer_done ) begin

			if( edge_processor_done && edge_ntotal_read == edge_ntotal_write  && edge_data_count == 0 && !edge_buffer_1_valid && !edge_buffer_2_valid  ) begin
				
				if( stage_1_last_edge_timeout < 10) begin
					stage_1_last_edge_timeout <= stage_1_last_edge_timeout +1;
				end
				else begin
					stage_1_last_edge_valid <= 1;
					stage_1_last_edge_cnt <= stage_1_last_edge_cnt  +1;

					stage_1_level <= stage_1_last_edge_cnt;
					stage_1_valid <= 1;
					bram_cl_1_raddr <= stage_1_last_edge_cnt;
					bram_cl_2_raddr <= stage_1_last_edge_cnt;
					bram_cl_3_raddr <= stage_1_last_edge_cnt ;
					// bram_cl_4_raddr <= stage_1_last_edge_cnt;
					bram_address_raddr <= stage_1_last_edge_cnt;
					
					bram_cl_1_re <= 1;
					bram_cl_2_re <= 1;
					bram_cl_3_re <= 1;
					// bram_cl_4_re <= 1;
					bram_address_re <= 1;
				end

			end
			// then, pass to stage 2
			else if( ~write_queue_buf_filled_next && edge_buffer_1_valid ) begin

						stage_1_edge <= edge_buffer_1;
						stage_1_level <= edge_buffer_1[127:96]-1;
						stage_1_valid <= 1;

						bram_cl_1_raddr <= edge_buffer_1[127:96]-1;
						bram_cl_2_raddr <= edge_buffer_1[127:96]-1;
						bram_cl_3_raddr <= edge_buffer_1[127:96]-1;
						// bram_cl_4_raddr <= edge_buffer_1[31:0];
						bram_address_raddr <= edge_buffer_1[127:96]-1;
					
						bram_cl_1_re <= 1;
						bram_cl_2_re <= 1;
						bram_cl_3_re <= 1;
						// bram_cl_4_re <= 1;
						bram_address_re <= 1;

						// handle new data and consumption of data
						if( edge_buffer_2_valid ) begin
							edge_buffer_1 <= edge_buffer_2;
							if (edge_m_axis_valid && edge_m_axis_ready ) begin
								edge_buffer_2 <= edge_m_axis_data;
								edge_m_axis_ready<= 0;
							end
							else begin
								edge_buffer_2_valid <= 0;
								edge_m_axis_ready<= 1;
							end
						end
						else begin
							if (edge_m_axis_valid && edge_m_axis_ready ) begin
								edge_buffer_1 <= edge_m_axis_data;
							end
							else begin
								edge_buffer_1_valid <= 0;
							end
						end

			end
			else begin

						// handle new data
						if( edge_m_axis_valid && edge_m_axis_ready ) begin
								if( ~edge_buffer_1_valid ) begin
									edge_buffer_1 <= edge_m_axis_data;
									edge_buffer_1_valid <= 1;
								end
								else if( ~edge_buffer_2_valid ) begin
									edge_buffer_2 <= edge_m_axis_data;
									edge_buffer_2_valid <= 1;
									edge_m_axis_ready<= 0;
								end
						end
						else if( ~edge_buffer_1_valid || ~edge_buffer_2_valid ) begin
							edge_m_axis_ready<= 1;
						end
			end
		end
	end
end

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		edge_ntotal_read <= 0;
	end
	else begin
		if (edge_m_axis_valid && edge_m_axis_ready ) begin
			edge_ntotal_read <= edge_ntotal_read   + 1;
		end
	end
end

reg				stage_2_valid;
reg				stage_2_last_edge_valid;
reg [127:0]		stage_2_edge;
reg	[L_BITS-1:0]		stage_2_level;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		stage_2_valid <= 0;
		stage_2_last_edge_valid <= 0;
	end
	else begin
		stage_2_valid <= 0;
		stage_2_last_edge_valid <= stage_1_last_edge_valid;
		stage_2_level <= stage_1_level;
		stage_2_valid <= stage_1_valid;
		stage_2_edge <= stage_1_edge;
	end
end



reg [127:0]		stage_3_edge_1_1;
reg [L_BITS-1:0]		stage_3_edge_1_1_level;
reg				stage_3_edge_1_1_valid;
reg [127:0]		stage_3_edge_1_2;
reg [L_BITS-1:0]		stage_3_edge_1_2_level;
reg				stage_3_edge_1_2_valid;

reg [127:0]		stage_3_edge_2_1;
reg [L_BITS-1:0]		stage_3_edge_2_1_level;
reg				stage_3_edge_2_1_valid;
reg [127:0]		stage_3_edge_2_2;
reg [L_BITS-1:0]		stage_3_edge_2_2_level;
reg				stage_3_edge_2_2_valid;

reg [127:0]		stage_3_edge_3_1;
reg [L_BITS-1:0]		stage_3_edge_3_1_level;
reg				stage_3_edge_3_1_valid;
reg [127:0]		stage_3_edge_3_2;
reg [L_BITS-1:0]		stage_3_edge_3_2_level;
reg				stage_3_edge_3_2_valid;

reg [31:0]				stage_3_address_1;
reg [L_BITS-1:0]		stage_3_address_1_level;
reg						stage_3_address_1_valid;
reg [31:0]				stage_3_address_2;
reg [L_BITS-1:0]		stage_3_address_2_level;
reg						stage_3_address_2_valid;

reg [127:0]		stage_3_edge_out_1;
reg [127:0]		stage_3_edge_out_2;
reg [127:0]		stage_3_edge_out_3;
reg [127:0]		stage_3_edge_out_4;
reg [31:0]		stage_3_address_out;
reg [L_BITS-1:0] 	stage_3_level;
reg				stage_3_valid;
reg				stage_3_last_edge_valid;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		stage_3_valid <= 0;
		stage_3_last_edge_valid <= 0;

		stage_3_edge_1_1_level <= 0;
		stage_3_edge_1_2_valid <= 0;
		stage_3_edge_2_1_level <= 0;
		stage_3_edge_2_2_level <= 0;
		stage_3_edge_3_1_level <= 0;
		stage_3_edge_3_2_level <= 0;
		
		stage_3_address_1_valid <= 0;
		stage_3_address_2_valid <= 0;

		bram_cl_1_valid <= 0;
		bram_cl_2_valid <= 0;
		bram_cl_3_valid <= 0;
		
		bram_address_din <= 0;
		bram_cl_1_din <= 0;
		bram_cl_2_din <= 0;
		bram_cl_3_din <= 0;

		if( (um_done || ~um_activated) && rst_n ) begin
			// reset bram
			bram_address_we <= 1'b1;
			bram_cl_1_we <= 1;
			bram_cl_2_we <= 1;
			bram_cl_3_we <= 1;

			bram_address_waddr <= bram_address_waddr + 1;
			bram_cl_1_waddr <= bram_cl_1_waddr+ 1;
			bram_cl_2_waddr <= bram_cl_2_waddr +1;
			bram_cl_3_waddr <= bram_cl_3_waddr +1;
		end
		else begin
			// bram address value initialization
			bram_address_we <= 0;
			bram_cl_1_we <= 0;
			bram_cl_2_we <= 0;
			bram_cl_3_we <= 0;

			bram_address_waddr <= 0;
			bram_cl_1_waddr <= 0;
			bram_cl_2_waddr <= 0;
			bram_cl_3_waddr <= 0;
		end
		
	end
	else begin
		stage_3_valid <= 0;
		stage_3_last_edge_valid <= 0;

		bram_cl_1_we <= 0;
		bram_cl_2_we <= 0;
		bram_cl_3_we <= 0;
		bram_address_we <= 0;

		bram_cl_1_waddr <= stage_2_level;
		bram_cl_2_waddr <= stage_2_level;
		bram_cl_3_waddr <= stage_2_level;
		bram_address_waddr <= stage_2_level;

		stage_3_level <= stage_2_level;

		if( stage_2_valid ) begin

			if( stage_2_last_edge_valid ) begin
				stage_3_valid <= 1;
				stage_3_last_edge_valid <= 1;
				stage_3_edge_out_1 <= bram_cl_1_dout;
				stage_3_edge_out_2 <= bram_cl_2_dout;
				stage_3_edge_out_3 <= bram_cl_3_dout;
				stage_3_edge_out_4 <= 128'b0;
				stage_3_address_out <= bram_address_dout;
				// just write out what we get - a timeout in the stage 1 makes sure that we get valid data from bram
			end
			else begin
				// case 1: all three are occupied, and we add the last for write out
				if( bram_cl_1_valid[ stage_2_level ] && bram_cl_2_valid[ stage_2_level ] && bram_cl_3_valid[ stage_2_level ] ) begin

					stage_3_valid <= 1;
					bram_cl_1_valid[ stage_2_level ] <= 0;
					bram_cl_2_valid[ stage_2_level ] <= 0;
					bram_cl_3_valid[ stage_2_level ] <= 0;

					if( stage_3_edge_1_1_valid && stage_3_edge_1_1_level == stage_2_level ) begin // then replace the element in bram 3
						stage_3_edge_1_1_valid <= 0;
						stage_3_edge_out_1 <= stage_3_edge_1_1;
					end
					else if( stage_3_edge_1_2_valid && stage_3_edge_1_2_level == stage_2_level ) begin
						stage_3_edge_1_2_valid <= 0;
						stage_3_edge_out_1 <= stage_3_edge_1_2;
					end
					else begin
						stage_3_edge_out_1 <= bram_cl_1_dout;
					end


					if( stage_3_edge_2_1_valid && stage_3_edge_2_1_level == stage_2_level ) begin
						stage_3_edge_2_1_valid <= 0;
						stage_3_edge_out_2 <= stage_3_edge_2_1;
					end
					else if( stage_3_edge_2_2_valid && stage_3_edge_2_2_level == stage_2_level ) begin
						stage_3_edge_2_2_valid <= 0;
						stage_3_edge_out_2 <= stage_3_edge_2_2;
					end
					else begin
						stage_3_edge_out_2 <= bram_cl_2_dout;
					end

				
					if( stage_3_edge_3_1_valid && stage_3_edge_3_1_level == stage_2_level ) begin
						stage_3_edge_3_1_valid <= 0;
						stage_3_edge_out_3 <= stage_3_edge_3_1;
					end
					else if( stage_3_edge_3_2_valid && stage_3_edge_3_2_level == stage_2_level ) begin
						stage_3_edge_3_2_valid <= 0;
						stage_3_edge_out_3 <= stage_3_edge_3_2;
					end
					else begin
						stage_3_edge_out_3 <= bram_cl_3_dout;
					end

					stage_3_edge_out_4 <= stage_2_edge;


					if( stage_3_address_1_valid && stage_3_address_1_level == stage_2_level ) begin
						stage_3_address_out <= stage_3_address_1;
						bram_address_we <= 1;
						bram_address_din <= stage_3_address_1 + 1;
						stage_3_address_1 <= stage_3_address_1 + 1;
					end
					else if( stage_3_address_2_valid && stage_3_address_2_level == stage_2_level ) begin
						stage_3_address_out <= stage_3_address_2;
						bram_address_we <= 1;
						bram_address_din <= stage_3_address_2 + 1;
						stage_3_address_2 <= stage_3_address_2 + 1;
					end
					else begin
						stage_3_address_out <= bram_address_dout;
						stage_3_address_1_valid <= 1;
						stage_3_address_1_level <= stage_2_level;
						stage_3_address_1 <= bram_address_dout+ 1;

						stage_3_address_2 <= stage_3_address_1;
						stage_3_address_2_level <= stage_3_address_1_level;
						stage_3_address_2_valid <= stage_3_address_1_valid;
	
						bram_address_we <= 1;
						bram_address_din <= bram_address_dout+ 1; // increment the address offset by 1
					end
				end

				// case 2: not all are occupied, so we write to the according bram, and shift the values
				else begin
					if( ~bram_cl_1_valid[ stage_2_level ] ) begin
						stage_3_edge_1_1 <= stage_2_edge;
						stage_3_edge_1_1_level <= stage_2_level;
						stage_3_edge_1_1_valid <= 1;
						stage_3_edge_1_2 <= stage_3_edge_1_1;
						stage_3_edge_1_2_level <= stage_3_edge_1_1_level;
						stage_3_edge_1_2_valid <= stage_3_edge_1_1_valid;

						bram_cl_1_valid[ stage_2_level ] <= 1;
						bram_cl_1_we <= 1;
						bram_cl_1_din <= stage_2_edge;
					end
					else if(~bram_cl_2_valid[ stage_2_level ] ) begin
						stage_3_edge_2_1 <= stage_2_edge;
						stage_3_edge_2_1_level <= stage_2_level;
						stage_3_edge_2_1_valid <= 1;
						stage_3_edge_2_2 <= stage_3_edge_2_1;
						stage_3_edge_2_2_level <= stage_3_edge_2_1_level;
						stage_3_edge_2_2_valid <= stage_3_edge_2_1_valid;

						bram_cl_2_valid[ stage_2_level ] <= 1;
						bram_cl_2_we <= 1;	
						bram_cl_2_din <= stage_2_edge;
					end
					else begin
						stage_3_edge_3_1 <= stage_2_edge;
						stage_3_edge_3_1_level <= stage_2_level;
						stage_3_edge_3_1_valid <= 1;
						stage_3_edge_3_2 <= stage_3_edge_3_1;
						stage_3_edge_3_2_level <= stage_3_edge_3_1_level;
						stage_3_edge_3_2_valid <= stage_3_edge_3_1_valid;

						bram_cl_3_valid[ stage_2_level ] <= 1;
						bram_cl_3_we <= 1;
						bram_cl_3_din <= stage_2_edge;
					end

					stage_3_address_1 <= bram_address_dout;
					stage_3_address_1_level <= stage_2_level;
					stage_3_address_1_valid <= 1;

					stage_3_address_2 <= stage_3_address_1;
					stage_3_address_2_level <= stage_3_address_1_level;
					stage_3_address_2_valid <= stage_3_address_1_valid;
				end
			end

		end
	end
end


reg[31:0]	stage_3b_edges_address_offset;
multiplier #(.L_BITS(L_BITS)) multiplier (
	.clk(clk),
	.offset(out_edges_offset_addr),
	.level(stage_3_level),
	.result(stage_3b_edges_address_offset)
);


reg					stage_3b_valid;
reg [127:0]			stage_3b_edge_out_4;
reg [127:0]			stage_3b_edge_out_3;
reg [127:0]			stage_3b_edge_out_2;
reg [127:0]			stage_3b_edge_out_1;
reg [L_BITS-1:0]	stage_3b_level;
reg [57:0]			stage_3b_address_out; // expand the offset from 32 bit to 58 bit
reg					stage_3b_last_edge_valid;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		stage_3b_valid <= 0;
		stage_3b_last_edge_valid <= 0;
		stage_3b_address_out <= 0;
	end
	else begin
		stage_3b_valid<= stage_3_valid;
		stage_3b_edge_out_4 <= stage_3_edge_out_4;
		stage_3b_edge_out_3 <= stage_3_edge_out_3;
		stage_3b_edge_out_2 <= stage_3_edge_out_2;
		stage_3b_edge_out_1 <= stage_3_edge_out_1;
		stage_3b_level <= stage_3_level;
		stage_3b_address_out <= stage_3_address_out;
		stage_3b_last_edge_valid <= stage_3_last_edge_valid;
	end
end



reg [31:0]		n_finish_num_written_levels;

reg [57:0]		stage_4_debug_address;
reg 			stage_4_debug_we;
reg [31:0]		stage_4_debug_level;
reg [511:0]		stage_4_debug_payload;
reg [31:0]		stage_4_num_writes;


always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		write_queue_write_enable <= 0;
		edge_writer_done <= 0;
		n_finish_num_written_levels<= 0;

		stage_4_debug_we <= 0;
		stage_4_debug_address <= 0;
		stage_4_debug_level <= 0;
		stage_4_num_writes <= 0;
	end
	else begin
		stage_4_debug_we <= 0; 
		write_queue_write_enable <= 0;
		if( stage_3b_valid && ~edge_writer_done) begin
			write_queue_buf_in <= { stage_3b_edge_out_4, stage_3b_edge_out_3, stage_3b_edge_out_2, stage_3b_edge_out_1, ( out_edges_base_addr + stage_3b_edges_address_offset + stage_3b_address_out)};
			write_queue_write_enable <= 1;
			stage_4_num_writes <= stage_4_num_writes + 1;

			if( stage_3b_last_edge_valid) begin
				n_finish_num_written_levels <= n_finish_num_written_levels  + 1;
				if( n_finish_num_written_levels +1 == 2** L_BITS) begin
					edge_writer_done <= 1;
				end
			end

			stage_4_debug_address <= ( out_edges_base_addr + stage_3b_edges_address_offset + stage_3b_address_out);
			stage_4_debug_we <= 1;
			stage_4_debug_level <= stage_3b_level;
			stage_4_debug_payload <= { stage_3b_edge_out_1, stage_3b_edge_out_2, stage_3b_edge_out_3, stage_3b_edge_out_4};
		end
	end
end



always @(posedge clk)
begin
	debug_edge_writer <= { 416'b0, edge_ntotal_read, edge_ntotal_write, 32'b0};
end





endmodule





`default_nettype wire





