`default_nettype none



module mcm_bit_receiver #(parameter BRAM_QUEUE_LENGTH_BITS  = 7, parameter MCM_BITS_REQUEST_TAG = 4 ) (
	//---------------------- GENERAL VARIABLES---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input 	wire 									um_done,						// true, if we are done processing (global)
	input 	wire 									um_activated,					// true, if the afu is activated

	output reg 										mcm_bit_receiver_done,					// true, if we are done processing data
	input wire [31:0]								mcm_bit_requester_num_requests_issued,
	input wire										mcm_bit_requester_done,
	input wire										mcm_bit_requester_epoch_done,
	output reg										epoch_done,

	//---------------------- GENERAL READ INPUT VARIABLES---------------------------//
	input   wire [7:0]                         	    um_rx_rd_tag,	// response to the read request: the tag
    input   wire [511:0]                       	    um_rx_data,		// the requested data
    input   wire                               	    um_rx_rd_valid,	// that tag and data are valid

	//---------------------- VARIABLES TO INDICATE THE CURRENT INPUT SPACE WE HAVE LEFT ---------------------------//
	output reg [31:0]								nelems_space, 						// the number of space left in our input queue
	output reg [31:0]								num_resp_received,

	//---------------------- FOR THE MCM OUTPUT QUEUE ---------------------------//
	output reg 										out_queue_write_enable,
	output reg [512 -1:0]							out_queue_buf_in,
	input wire										out_queue_filled_next,
	output reg[511:0]								debug_mcm_bit_receiver
);



reg									mcm_m_axis_ready;
wire								mcm_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	mcm_data_count;
wire [511:0]						mcm_m_axis_data;

reg			mcm_s_axis_write_enable;
reg	[511:0]	mcm_s_axis_data;

reg			mcm_buffer_1_valid;
reg [511:0]	mcm_buffer_1;
reg			mcm_buffer_2_valid;
reg [511:0]	mcm_buffer_2;


nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(512), .PROG_FULL_THRESH(0 ) ) mcm_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(mcm_s_axis_write_enable),
    .s_axis_ready(),
    .s_axis_data(mcm_s_axis_data),
    .s_axis_last(1'b0),
    .m_axis_valid(mcm_m_axis_valid),
    .m_axis_ready(mcm_m_axis_ready),
    .m_axis_data(mcm_m_axis_data),
    .m_axis_last(),
    .data_count(mcm_data_count)
);

reg [31:0] ntotal_read;
reg [31:0] ntotal_write;
reg debug_overflow_flag;


// ----- put the incoming mcm-bits (cache lines) in a BRAM queue ----- //
always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin

        // reset of output registers
        nelems_space <= 0;
		num_resp_received <= 0;
		ntotal_write <= 0;
		debug_overflow_flag<= 0;

		mcm_s_axis_write_enable <= 0;
    end
    else begin
		nelems_space <= (2**BRAM_QUEUE_LENGTH_BITS) - ( ntotal_write - ntotal_read);
		mcm_s_axis_write_enable <= 0;
		if( ~mcm_bit_receiver_done) begin
			if( um_rx_rd_valid & ( um_rx_rd_tag == MCM_BITS_REQUEST_TAG) ) begin
				mcm_s_axis_write_enable <= 1;
				mcm_s_axis_data <= um_rx_data;

				ntotal_write <= ntotal_write + 1;
				num_resp_received <= num_resp_received + 1;
			end
		end

		debug_overflow_flag <= ( ntotal_write - ntotal_read) + 2 > 2**BRAM_QUEUE_LENGTH_BITS;
	end
end

always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
		mcm_bit_receiver_done <= 0;
		out_queue_write_enable <= 0;
		out_queue_buf_in <= 0;
        
		ntotal_read <= 0;
		epoch_done <= 0;
		mcm_buffer_1_valid <= 0;
		mcm_buffer_2_valid <= 0;

		// drain queues
		if( (um_done || ~um_activated) && rst_n ) begin
			mcm_m_axis_ready <= 1;
		end
		else begin
			mcm_m_axis_ready <= 0;
		end
		
    end
	else begin
		out_queue_write_enable <= 0;
		epoch_done <= 0;
		
		if( ~mcm_bit_receiver_done ) begin
			if( ntotal_read == mcm_bit_requester_num_requests_issued ) begin
				if( mcm_bit_requester_epoch_done ) begin
					epoch_done <= 1;
				end
				if( mcm_bit_requester_done ) begin
					mcm_bit_receiver_done <= 1;
				end

				// handle queue data: reads
				if( mcm_m_axis_ready && mcm_m_axis_valid ) begin
					if( ~mcm_buffer_1_valid ) begin
						mcm_buffer_1 <= mcm_m_axis_data;
						mcm_buffer_1_valid <= 1;
					end
					else if( ~mcm_buffer_1_valid ) begin
						mcm_buffer_2 <= mcm_m_axis_data;
						mcm_buffer_2_valid <= 1;
						mcm_m_axis_ready <= 0;
					end
				end
				else if( ~mcm_buffer_1_valid ) begin
					mcm_m_axis_ready <= 1;
				end

			end
			else if( mcm_buffer_1_valid && ~out_queue_filled_next ) begin

				out_queue_write_enable <= 1;
				out_queue_buf_in <= mcm_buffer_1;
				ntotal_read <= ntotal_read +1;

				// handle queue data
				if( mcm_buffer_2_valid ) begin
					mcm_buffer_1 <= mcm_buffer_2;
					if( mcm_m_axis_ready && mcm_m_axis_valid ) begin
						mcm_buffer_2 <= mcm_m_axis_data;
						mcm_m_axis_ready <= 0;
					end
					else begin
						mcm_m_axis_ready <= 1;
						mcm_buffer_2_valid <= 0;
					end
				end
				else begin
					mcm_m_axis_ready <= 1;
					if( mcm_m_axis_ready && mcm_m_axis_valid ) begin
						mcm_buffer_1 <= mcm_m_axis_data;
					end
					else begin
						mcm_buffer_1_valid <= 0;
					end
				end


			end
			else begin
				
				// handle queue data
				if( mcm_m_axis_ready && mcm_m_axis_valid ) begin
					if( ~mcm_buffer_1_valid ) begin
						mcm_buffer_1 <= mcm_m_axis_data;
						mcm_buffer_1_valid <= 1;
					end
					else if( ~mcm_buffer_1_valid ) begin
						mcm_buffer_2 <= mcm_m_axis_data;
						mcm_buffer_2_valid <= 1;
						mcm_m_axis_ready <= 0;
					end
				end
				else if( ~mcm_buffer_1_valid ) begin
					mcm_m_axis_ready <= 1;
				end


			end
		end
	end
end


/*
// high, if we do not output data, but we could
reg				do_not_work_valid;
reg [31:0]		do_not_work_cycle_counter;
// high, if we hava data, but cannot output data
reg				can_not_output_valid;
reg [31:0]		can_not_output_cycle_counter;
// high, if we do not have data, but could output
reg				can_not_input_valid;
reg [31:0]		can_not_input_cycle_counter;


always @(posedge clk) begin
	if (~rst_n | um_done | ~um_activated) begin
		do_not_work_valid <= 0;
		do_not_work_cycle_counter <= 0;
		can_not_output_valid <= 0;
		can_not_output_cycle_counter <= 0;
		can_not_input_valid <= 0;
		can_not_input_cycle_counter <= 0;
	end
	else begin
		do_not_work_valid <= 0;
		can_not_output_valid <= 0;
		can_not_input_valid <= 0;

		if( ~out_queue_filled_next && mcm_data_count > 0 && !( out_queue_write_enable ) ) begin
			do_not_work_valid<= 1;
			do_not_work_cycle_counter <= do_not_work_cycle_counter + 1;
		end

		if( mcm_data_count > 0  && ( out_queue_filled_next ) ) begin
			can_not_output_valid <= 1;
			can_not_output_cycle_counter <= can_not_output_cycle_counter  + 1;
		end

		if(( mcm_data_count == 0 ) && ~out_queue_filled_next) begin
			can_not_input_valid <= 1;
			can_not_input_cycle_counter <= can_not_input_cycle_counter + 1;
		end
	end
end
*/


always @(posedge clk) begin
	debug_mcm_bit_receiver <= { 416'b0, ntotal_read, ntotal_write, num_resp_received};
end



endmodule









`default_nettype wire









