
`default_nettype none



/**

    The MCM bit requester issues read requests for the matching bits, so we get them from RAM
    Currently, for every edge, we issue a new request (expect, if the "v" value is shared). In this case, we send a special offset (all 1), to indicate that the previous value can be used.

    In a further refinement, we can track if the address used is actually the same, and we only have to send another offset.
    In this case, we need a bit to determine that the old matching bit must be kept, but maybe another offset must be used.

    Note that it is not a solution to write the offset into the tag, since we must not issue a request for every edge, and we can share data already received.
*/

module mcm_bit_requester #(parameter BRAM_QUEUE_LENGTH_BITS  = 6, parameter MCM_BITS_REQUEST_TAG = 3, parameter BRAM_OVERFLOW_PADDING= 6, parameter L_BITS = 5) (
	//---------------------- GENERAL VARIABLES (STATE) ---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input wire 										um_done,						// true, if we are done processing (global)
	input wire 										um_activated,					// true, if the afu is activated
	output reg										mcm_bit_requester_done,						// true, if the pointer requester is done
	output reg[31:0]								mcm_bit_requester_num_requests_issued,

	input wire[31:0]								global_epoch,
	output reg										epoch_done, // if the epoch he has is done or not

	//---------------------- ADDRESS VARIABLES---------------------------//
	input wire [57:0]								read_mcm_base_address,			// the base read address
	input wire [31:0]								num_edges,						// number of edges

	//---------------------- QUEUE VARIABLES FOR ISSUING THE REQUEST - TO REQUEST INTERFACE ---------------------------//
	output reg									   	req_queue_write_enable,				// wheter we want to write to the queue or not
	output reg [65:0] 							   	req_queue_buf_in,					// the queue input buffer (65:8 is the request address, 7:0 the tag we assign to the request)
	input wire 									   	req_queue_buf_filled_next,			// 1, if the queue is filled soon

	//---------------------- QUEUE VARIABLES FOR INPUT - DATA FROM SORTER---------------------------//
	input wire										sorter_queue_write_enable,
	input wire[127:0]								sorter_queue_buf_in,				// per entry we fetch: (epoch, u, v, w)
	output reg 									   	sorter_queue_buf_filled_next,
	
	//---------------------- QUEUE VARIABLES FOR OUTPUT (edge ready queue) - TO EDGE PROCESSOR ---------------------------//
	output reg										ready_queue_write_enable,
	output reg[127:0]								ready_queue_buf_in,
	input wire										ready_queue_buf_filled_next,
    
    //---------------------- QUEUE VARIABLES FOR OFFSET - GOING TO EDGE PROCESSOR ---------------------------//
    output reg                                      offset_queue_write_enable,
    output reg[(9 - L_BITS)+1:0]				    offset_queue_buf_in,
	input wire										offset_queue_buf_filled_next,

	output reg[511:0]								debug_mcm_bit_requester
);




reg									mcm_m_axis_ready;
wire								mcm_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	mcm_data_count;
wire [127:0]						mcm_m_axis_data;

reg  [127:0]						mcm_data_buffer_1;
reg									mcm_data_buffer_1_valid;
reg  [127:0]						mcm_data_buffer_2;
reg									mcm_data_buffer_2_valid;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(128), .PROG_FULL_THRESH(0 ) ) edge_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(sorter_queue_write_enable), // 
    .s_axis_ready(),
    .s_axis_data(sorter_queue_buf_in),
    .s_axis_last(1'b0),
    .m_axis_valid(mcm_m_axis_valid),
    .m_axis_ready(mcm_m_axis_ready),
    .m_axis_data(mcm_m_axis_data),
    .m_axis_last(),
    .data_count(mcm_data_count)
);

reg [31:0] ntotal_read;
reg [31:0] ntotal_write;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		sorter_queue_buf_filled_next <= 0;
		ntotal_write <= 0;
	end
	else begin
		if( sorter_queue_write_enable ) begin
			ntotal_write <=  ntotal_write + 1;
		end

		if( ntotal_write - ntotal_read + BRAM_OVERFLOW_PADDING >= 2**BRAM_QUEUE_LENGTH_BITS ) begin
			sorter_queue_buf_filled_next <= 1;
		end
		else begin
			sorter_queue_buf_filled_next <= 0;
		end
	end
end

reg [31:0] last_v_address;  // the last "v" of (u, v), which we have requested - to track if it is required to issue requests for this "v"
reg			last_v_address_valid;
reg [31:0] last_epoch;
reg [31:0] num_processed_edges; // count the number of edges we have requested

reg			debug_error_flag;
reg [5:0]	debug_case;


always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
        // local variables
        last_v_address <= 0;
		last_v_address_valid <= 0;
		mcm_bit_requester_done <= 0;
		num_processed_edges <= 0;

        // outgoing variables
		ready_queue_write_enable <= 0;
		ready_queue_buf_in <= 0;

		offset_queue_write_enable <= 0;
    	offset_queue_buf_in <= 0;

		req_queue_write_enable <= 0;
		req_queue_buf_in <= 0;

		ntotal_read <= 0;

		mcm_bit_requester_num_requests_issued <= 0;


		epoch_done <= 0;
		last_epoch <= 0;

		debug_error_flag <= 0;
		debug_case <= 0;

		mcm_data_buffer_1_valid <= 0;
		mcm_data_buffer_2_valid <= 0;

		if( (um_done || ~um_activated) && rst_n ) begin
			mcm_m_axis_ready <= 1;
		end
		else begin
			mcm_m_axis_ready <= 0;
		end
	end
	else begin
            offset_queue_write_enable <= 0;
            ready_queue_write_enable <= 0;
            req_queue_write_enable <= 0;

			epoch_done <= 0;

			debug_case <= 0;
			

		    if( ~mcm_bit_requester_done ) begin
				
		        if( num_edges == num_processed_edges) begin
					debug_case <= 1;
		            mcm_bit_requester_done <= 1;
		        end
				else if( ~mcm_data_buffer_1_valid ) begin
					debug_case <= 2;
					mcm_m_axis_ready <= 1;

					if( mcm_m_axis_ready && mcm_m_axis_valid ) begin
						mcm_data_buffer_1_valid<= 1;
						mcm_data_buffer_1 <= mcm_m_axis_data;
					end
				end
				else if( mcm_data_buffer_1[127:96] > global_epoch ) begin
					debug_case <= 2;
					epoch_done <= 1;

					// HANDLE ALREADY REQUESTED DATA
					if( mcm_m_axis_ready && mcm_m_axis_valid ) begin
						if( mcm_data_buffer_2_valid ) begin
							debug_error_flag <= 1;
						end
						mcm_data_buffer_2_valid<= 1;
						mcm_data_buffer_2 <= mcm_m_axis_data;
						mcm_m_axis_ready <= 0;
					end
				end
		        else if( mcm_data_buffer_1_valid && ( ~offset_queue_buf_filled_next) && ( ~ready_queue_buf_filled_next) ) begin // no check for "req_queue_buf_filled_next", since we might omit the request
					debug_case <= 3;
					// check, if we can a) drop the data element, since we have issued the request for this "v" already -> can improve by checking, if it is the same address as we have alredy issued for!, but then we need more information: the offset and a "use same data as before" bit!
					if( last_v_address_valid && last_v_address ==  ( ( (mcm_data_buffer_1[63:32] -1) << L_BITS ) >> 9 ) && last_epoch == mcm_data_buffer_1[127:96] ) begin
						debug_case <= 4;

						// do nothing - drop this element
						ntotal_read <= ntotal_read +1;
						num_processed_edges <= num_processed_edges +1;

						// write the edge to the pending queue
						ready_queue_write_enable <= 1;
						ready_queue_buf_in <= mcm_data_buffer_1;

						// write to the offset queue
		                offset_queue_write_enable <= 1;
						offset_queue_buf_in[(9 - L_BITS):0] <= ( ( ( mcm_data_buffer_1[63:32] -1) << L_BITS )  -  ( ( ( ( mcm_data_buffer_1[63:32] -1) << L_BITS ) >> 9 ) << 9 ) ) >> L_BITS;
		                offset_queue_buf_in[(9 - L_BITS)+1] <= 1'b1;

						// check for new data in the queue
						if( mcm_data_buffer_2_valid ) begin
							mcm_data_buffer_1 <= mcm_data_buffer_2;
							if( mcm_m_axis_ready && mcm_m_axis_valid ) begin
								mcm_data_buffer_2_valid<= 1;
								mcm_data_buffer_2 <= mcm_m_axis_data;
								mcm_m_axis_ready <= 0;
							end
							else begin
								mcm_data_buffer_2_valid <= 0;
								mcm_m_axis_ready <= 1;
							end
						end
						else begin
							mcm_m_axis_ready <= 1;
							if( mcm_m_axis_ready && mcm_m_axis_valid ) begin
								mcm_data_buffer_1_valid<= 1;
								mcm_data_buffer_1 <= mcm_m_axis_data;
							end
							else begin
								mcm_data_buffer_1_valid<= 0;
							end
						end

					end
					else begin
						debug_case <= 5;
						// issue a new request - if we can
						if( ~req_queue_buf_filled_next) begin
							debug_case <= 6;

							mcm_bit_requester_num_requests_issued <= mcm_bit_requester_num_requests_issued  + 1;
							ntotal_read <= ntotal_read +1;
							num_processed_edges <= num_processed_edges +1;
	
							// issue request
							req_queue_write_enable <= 1;
							req_queue_buf_in <= { read_mcm_base_address +  ( ( (mcm_data_buffer_1[63:32] -1) << L_BITS ) >> 9 ), 8'b0 } +  MCM_BITS_REQUEST_TAG ;
							
							// write edge to the pending queue and update the last seen v
							ready_queue_write_enable <= 1;
							ready_queue_buf_in <= mcm_data_buffer_1;
							last_v_address <= ( ( (mcm_data_buffer_1[63:32] -1) << L_BITS ) >> 9 ); // mcm_data_buffer_1[63:32];
							last_v_address_valid <= 1;
					
		                    // write the offset to the queue, such the mcm-bit receiver can use the value
		                    offset_queue_write_enable <= 1;
							// bit position of the mcm-bit in the stream MINUS the bit position of the cache-line (I think that the position of the cache-line can be rewritten by cutting off some least significant bits!)
		                    offset_queue_buf_in[(9 - L_BITS):0] <= ( ( ( mcm_data_buffer_1[63:32] -1) << L_BITS )  -  ( ( ( ( mcm_data_buffer_1[63:32] -1) << L_BITS ) >> 9 ) << 9 ) ) >> L_BITS;
							offset_queue_buf_in[(9 - L_BITS)+1] <= 1'b0;
						
							last_epoch <= mcm_data_buffer_1[127:96];


							// check for new data in the queue
							if( mcm_data_buffer_2_valid ) begin
								mcm_data_buffer_1 <= mcm_data_buffer_2;
								if( mcm_m_axis_ready && mcm_m_axis_valid ) begin
									mcm_data_buffer_2_valid<= 1;
									mcm_data_buffer_2 <= mcm_m_axis_data;
									mcm_m_axis_ready <= 0;
								end
								else begin
									mcm_data_buffer_2_valid <= 0;
									mcm_m_axis_ready <= 1;
								end
							end
							else begin
								mcm_m_axis_ready <= 1;
								if( mcm_m_axis_ready && mcm_m_axis_valid ) begin
									mcm_data_buffer_1_valid<= 1;
									mcm_data_buffer_1 <= mcm_m_axis_data;
								end
								else begin
									mcm_data_buffer_1_valid<= 0;
								end
							end

						end
						else begin
							debug_case <= 7;
							// handle data assignement


							if( mcm_m_axis_ready && mcm_m_axis_valid ) begin
								if( ~mcm_data_buffer_1_valid ) begin
									mcm_data_buffer_1_valid<= 1;
									mcm_data_buffer_1 <= mcm_m_axis_data;
								end
								else if( ~mcm_data_buffer_2_valid ) begin
									mcm_data_buffer_2_valid<= 1;
									mcm_data_buffer_2 <= mcm_m_axis_data;
									mcm_m_axis_ready <= 0;
								end
								mcm_data_buffer_2_valid<= 1;
								mcm_data_buffer_2 <= mcm_m_axis_data;
							end
							else if( ~mcm_data_buffer_1_valid ) begin
								mcm_m_axis_ready <= 1;
							end


						end
					end // end of else for reuse of matching bits or not
				end
				else begin // else for the different cases
							debug_case <= 8;

							if( mcm_m_axis_ready && mcm_m_axis_valid ) begin
								if( ~mcm_data_buffer_1_valid ) begin
									mcm_data_buffer_1_valid<= 1;
									mcm_data_buffer_1 <= mcm_m_axis_data;
								end
								else if( ~mcm_data_buffer_2_valid ) begin
									mcm_data_buffer_2_valid<= 1;
									mcm_data_buffer_2 <= mcm_m_axis_data;
									mcm_m_axis_ready <= 0;
								end
								mcm_data_buffer_2_valid<= 1;
								mcm_data_buffer_2 <= mcm_m_axis_data;
							end
							else if( ~mcm_data_buffer_1_valid ) begin
								mcm_m_axis_ready <= 1;
							end

				end

		    end // end of ~mcm_bit_requester_done

	end

end


/*

// high, if we do not output data, but we could
reg				do_not_work_valid;
reg [31:0]		do_not_work_cycle_counter;
// high, if we hava data, but cannot output data
reg				can_not_output_valid;
reg [31:0]		can_not_output_cycle_counter;
// high, if we do not have data, but could output
reg				can_not_input_valid;
reg [31:0]		can_not_input_cycle_counter;


always @(posedge clk) begin
	if (~rst_n | um_done | ~um_activated) begin
		do_not_work_valid <= 0;
		do_not_work_cycle_counter <= 0;
		can_not_output_valid <= 0;
		can_not_output_cycle_counter <= 0;
		can_not_input_valid <= 0;
		can_not_input_cycle_counter <= 0;
	end
	else begin
		do_not_work_valid <= 0;
		can_not_output_valid <= 0;
		can_not_input_valid <= 0;

		if( ~epoch_done && ~req_queue_buf_filled_next && ~ready_queue_buf_filled_next && ~offset_queue_buf_filled_next && mcm_data_count > 0 && !( ready_queue_write_enable || offset_queue_write_enable) ) begin
			do_not_work_valid<= 1;
			do_not_work_cycle_counter <= do_not_work_cycle_counter + 1;
		end

		if( mcm_data_count > 0  && ( req_queue_buf_filled_next || ready_queue_buf_filled_next || offset_queue_buf_filled_next || epoch_done ) ) begin
			can_not_output_valid <= 1;
			can_not_output_cycle_counter <= can_not_output_cycle_counter  + 1;
		end

		if(( mcm_data_count == 0 ) && ~req_queue_buf_filled_next && ~ready_queue_buf_filled_next && ~offset_queue_buf_filled_next) begin
			can_not_input_valid <= 1;
			can_not_input_cycle_counter <= can_not_input_cycle_counter + 1;
		end
	end
end
*/





always @(posedge clk) begin
	debug_mcm_bit_requester <= {384'b0, 
								32'b0, ntotal_read, 
								ntotal_write, num_processed_edges };
end

endmodule






`default_nettype wire










