


/**
	The MCM BIT WRITER gets a stream of [ vertex-v id, mcm-cache-line]. According to this, he collects the cache-line, and issues write requests to the write interface.
	REFINEMENT: COMPARE THE THING WE WRITE WITH THE INPUT - MAYBE WE CAN OMIT SOME WRITES, SINCE THE DATA DID NOT CHANGE!
*/


module mcm_bit_writer  #(parameter OVERFLOW_PADDING = 6, parameter MCM_OUT_TAG = 14, parameter BRAM_QUEUE_LENGTH_BITS= 6, parameter L_BITS = 4)  (
	//---------------------- GENERAL VARIABLES---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input wire 										um_done,						// true, if we are done processing (global)
	input wire 										um_activated,					// true, if the afu is activated
	input wire [57:0]								out_mcm_bits_base_addr,

	output reg										mcm_writer_done,
	output reg										mcm_writer_epoch_done,
	output reg [31:0] 								mcm_bit_writer_num_writes,

	input wire										edge_processor_done,
	input wire										edge_processor_epoch_done,

	//---------------------- MCMs FROM EDGE PROCESSOR ---------------------------//
	input wire 										mcm_queue_write_enable,
	input wire [32+512-1:0]							mcm_queue_buf_in,
	output reg 										mcm_queue_buf_filled_next,

	//---------------------- CL TO WRITE INTERFACE ---------------------------//
	output reg 										write_queue_write_enable,
	output reg [569:0]							write_queue_buf_in,
	input wire										write_queue_buf_filled_next,

	output reg[511:0]								debug_mcm_bit_writer
	
);


reg									mcm_m_axis_ready;
wire								mcm_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	mcm_data_count;
wire [512+32-1:0]					mcm_m_axis_data;

reg			mcm_s_axis_write_enable;
reg	[512+32-1:0]	mcm_s_axis_data;

reg			mcm_buffer_1_valid;
reg [512+32-1:0]	mcm_buffer_1;
reg			mcm_buffer_2_valid;
reg [512+32-1:0]	mcm_buffer_2;


nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(512+32), .PROG_FULL_THRESH(0 ) ) mcm_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(mcm_s_axis_write_enable),
    .s_axis_ready(),
    .s_axis_data(mcm_s_axis_data),
    .s_axis_last(1'b0),
    .m_axis_valid(mcm_m_axis_valid),
    .m_axis_ready(mcm_m_axis_ready),
    .m_axis_data(mcm_m_axis_data),
    .m_axis_last(),
    .data_count(mcm_data_count)
);


reg [31:0] mcm_ntotal_read;
reg [31:0] mcm_ntotal_write;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin

		mcm_ntotal_write <= 0;
		mcm_queue_buf_filled_next <= 0;
		mcm_s_axis_write_enable<= 0;

	end
	else begin
		mcm_s_axis_write_enable <= 0;
		if( mcm_queue_write_enable ) begin
			mcm_s_axis_write_enable<= 1;
			mcm_s_axis_data <= mcm_queue_buf_in;
			mcm_ntotal_write <= mcm_ntotal_write +1;
		end
		// needs more padding, since the edge-processor checks for the queue size only in the first stage!
		mcm_queue_buf_filled_next <=  (mcm_data_count + OVERFLOW_PADDING + 15) >= (2**BRAM_QUEUE_LENGTH_BITS  );
	end

end

reg [31:0] 		address_offset;
reg		   		mcm_cl_valid;
reg [512-1:0] 	mcm_cl;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		address_offset <= 0;
		write_queue_write_enable <= 0;
		mcm_cl<= 0;
		mcm_cl_valid <= 0;

		// outgoing variables
		mcm_writer_epoch_done <= 0;
		mcm_writer_done<= 0;

		mcm_ntotal_read <= 0;

		mcm_bit_writer_num_writes <= 0;

		mcm_buffer_1_valid<= 0;
		mcm_buffer_2_valid <= 0;

		// drain queues
		if( (um_done || ~um_activated) && rst_n ) begin
			mcm_m_axis_ready <= 1;
		end
		else begin
			mcm_m_axis_ready <= 0;
		end
		
	end
	else begin

		write_queue_write_enable <= 0;
		mcm_writer_epoch_done <= 0;

		if( ~mcm_writer_done ) begin
		
			
			if( ( edge_processor_epoch_done || edge_processor_done ) && mcm_ntotal_write  == mcm_ntotal_read ) begin
				if( mcm_cl_valid == 0 && edge_processor_epoch_done ) begin
					mcm_writer_epoch_done <= 1;
				end
				if( mcm_cl_valid == 0 && edge_processor_done ) begin
					mcm_writer_done <= 1;
				end

				if( mcm_cl_valid ) begin // if we have a valid v, we need to write it back
					if( ~write_queue_buf_filled_next ) begin

						write_queue_buf_in <= { mcm_cl, out_mcm_bits_base_addr +  address_offset };
						write_queue_write_enable <= 1;
						mcm_bit_writer_num_writes <= mcm_bit_writer_num_writes + 1;

						mcm_cl_valid <= 0; // set to 0 to prevent a further write
						if( edge_processor_epoch_done ) begin
							mcm_writer_epoch_done <= 1;
						end
						if( edge_processor_done ) begin
							mcm_writer_done <= 1;
						end
					end
				end

				// no new data arrives from the edge processor, and we have read all elements from the queue: so no read/write-handling is required

			end
			// if there is data in the queue to consume (and we can issue requests), then do so
			else if(  mcm_buffer_1_valid && ~write_queue_buf_filled_next ) begin
			
				mcm_ntotal_read<= mcm_ntotal_read + 1;

				// if it is the same cache-line, then we can merge with the old one
				if( mcm_cl_valid && (( (mcm_buffer_1[31:0] -1 ) << L_BITS)  >> 9 ) == address_offset ) begin
					mcm_cl <= mcm_cl | mcm_buffer_1[512+32-1:32];
				end
				else begin
					// write back old cl, if it is a valid "v"
					if( mcm_cl_valid ) begin
						write_queue_buf_in <= { mcm_cl, out_mcm_bits_base_addr + address_offset};
						write_queue_write_enable <= 1;
						mcm_bit_writer_num_writes <= mcm_bit_writer_num_writes+1;
					end

					// set new cache line
					mcm_cl <= mcm_buffer_1[512+32-1:32];
					address_offset <=  (( (mcm_buffer_1[31:0] -1 ) << L_BITS)  >> 9 );
					mcm_cl_valid <= 1;
				end

				// consume from the queue, shift elements
				if( mcm_buffer_2_valid ) begin
					mcm_buffer_1 <= mcm_buffer_2;
					if( mcm_m_axis_ready &&  mcm_m_axis_valid ) begin
						mcm_buffer_2 <= mcm_m_axis_data;
						mcm_m_axis_ready <= 0;
					end
					else begin
						mcm_buffer_2_valid <= 0;
						mcm_m_axis_ready <= 1;
					end
				end
				else begin
					mcm_m_axis_ready <= 1;
					if( mcm_m_axis_ready &&  mcm_m_axis_valid ) begin
						mcm_buffer_1 <= mcm_m_axis_data;
					end
					else begin
						mcm_buffer_1_valid <= 0;
					end
				end


			end
			else begin
				// in this case: read from the queue
				if( mcm_m_axis_ready &&  mcm_m_axis_valid ) begin
					if( ~mcm_buffer_1_valid ) begin
						mcm_buffer_1 <= mcm_m_axis_data;
						mcm_buffer_1_valid <= 1;
					end
					else if( ~mcm_buffer_2_valid ) begin
						mcm_buffer_2 <= mcm_m_axis_data;
						mcm_buffer_2_valid <= 1;
						mcm_m_axis_ready <= 0;
					end
				end
				else if ( ~mcm_buffer_1_valid || ~mcm_buffer_2_valid ) begin
					mcm_m_axis_ready <= 1;
				end

			end


		end
	end
end


always @(posedge clk) begin
	debug_mcm_bit_writer <= {416'b0, mcm_ntotal_read, mcm_ntotal_write, mcm_bit_writer_num_writes };
end


endmodule




