`default_nettype none


`define D_POINTER_REQUEST_TAG 1
`define D_EDGE_REQUEST_TAG 2
`define D_MCM_BITS_REQUEST_TAG 3
`define D_BRAM_MCM_BITS_REQUEST_TAG 4  

`define D_EDGE_WRITE_TAG 10
`define D_MCM_WRITE_TAG 14
`define D_NUM_WRITES_IN_FLIGHT 256

`define D_BRAM_OVERFLOW_PADDING 8

`define POINTER_RECEIVER_QUEUE_LENGTH_BITS 7  	// 512*128 = 65536bits	(4 BRAMS)
`define REQUEST_INTERFACE_QUEUE_LENGTH_BITS 9	// 66*512 = 33792bits	(2 BRAMS)
`define EDGE_REQUESTER_QUEUE_LENGTH_BITS 7		// 100*128 = 12800bits (*4, since there are 4 queues)	(4 BRAMS)
`define EDGE_RECEIVER_QUEUE_LENGTH_BITS 9		// 512*512 + 512*128 = 262144 + 65536 = 327680bits (17 BRAMS)
`define SORTER_FIRST_QUEUE_LENGTH_BITS 9		// 128*512 = 32768bits (4 BRAMS)
`define SORTER_INNER_QUEUE_LENGTH_BITS 7		// 128*128 = 16384 bits (1 BRAM)
`define MCM_BIT_REQUESTER_QUEUE_LENGTH_BITS 12	// 131072 bits (7 BRAMS)
`define MCM_BIT_RECEIVER_QUEUE_LENGTH_BITS 9	// 512*512 = 262144bits (14 BRAMS)

`include "l_bits.vh"
`include "k_bits.vh"

module mwm_crouch_csr_ram #(parameter NUM_USER_STATE_COUNTERS = 10, parameter BRAM_QUEUE_LENGTH_BITS  = 8, parameter BRAM_QUEUE_LENGTH_BITS_LARGE = 12, parameter EPSILON = 0.03 )(
    input   wire                                   clk,				// cloc
    input   wire                                   rst_n,			// Stays low as long as AFU is idle
    //-------------------------------------------------//
    input   wire                                   start_um,		// active hight 8 cycles to trigger AFU
    input   wire [511:0]                           um_params,		// params (config data structure)
    output  wire                                   um_done,			// indicate that we are done

	// output  wire [NUM_USER_STATE_COUNTERS*32-1:0]  um_state_counters,
    // output  wire                                   um_state_counters_valid,

    // TX RD
    output  wire  [57:0]                            um_tx_rd_addr,	// read req. address (accesses 512bits)
    output  wire  [7:0]                             um_tx_rd_tag,	// read req. tag
    output  wire                                    um_tx_rd_valid,	// indicated that tag and addr. are valid
    input   wire                                    um_tx_rd_ready,	// framework tells that we can issue request
    // TX WR
    output  wire  [57:0]                            um_tx_wr_addr,	// address to write result
    output  wire  [7:0]                             um_tx_wr_tag,	// tag to write the result
    output  wire                                    um_tx_wr_valid,	// tell that tag and address are valid
    output  wire  [511:0]                           um_tx_data,		// the data to write
    input   wire                                    um_tx_wr_ready,	// that the framework is read to get data
    // RX RD
    input   wire [7:0]                             um_rx_rd_tag,	// response to the read request: the tag
    input   wire [511:0]                           um_rx_data,		// the requested data
    input   wire                                   um_rx_rd_valid,	// that tag and data are valid
    output  wire                                   um_rx_rd_ready,	// that we are ready for new data
    // RX WR 
    input   wire                                   um_rx_wr_valid,	// that the tag below, that we have once used in the write, is valid
    input   wire [7:0]                             um_rx_wr_tag		// the tag to confirm the write we issued
);


//---------------------- DEBUG  ---------------------------//
wire[511:0]								debug_bram_mcm_bit_receiver;
wire[511:0]								debug_bram_mcm_bit_requester;
wire[511:0]								debug_edge_processor;
wire[511:0]								debug_edge_receiver;
wire[511:0]								debug_edge_requester;
wire[511:0]								debug_edge_writer;
wire[511:0]								debug_mcm_bit_ack_receiver;
wire[511:0]								debug_mcm_bit_receiver;
wire[511:0]								debug_mcm_bit_requester;
wire[511:0]								debug_mcm_bit_writer;
reg[511:0]								debug_mwm_crouch_csr_ram;
wire[511:0]								debug_pointer_receiver;
wire[511:0]								debug_pointer_requester;
wire[511:0]								debug_request_interface;
wire[511:0]								debug_sorter;
wire[511:0]								debug_state_controller;
wire[511:0]								debug_write_interface;

wire 									debug_cl_queue_write_enable;
wire[569:0]							    debug_cl_queue_buf_in;
wire									debug_cl_queue_buf_filled_next;


//---------------------- HANDLE START AND END STATE  ---------------------------//
reg                         um_activated;
reg  [511:0]                um_params_reg;
reg[31:0]					num_ptrs;			// number of pointers (the number of vertices)
reg[31:0]					num_epochs;
reg[31:0]					edge_offset;
reg[57:0]					ptr_base_addr;		// the base address for the pointers
reg[31:0]					num_edges;			// the number of edges
reg[57:0]					edges_base_addr;	// the base address for the edges
reg[57:0]					out_edges_base_addr;		// the base address to write out the edges
reg[57:0]					out_mcm_bits_base_addr;		// the base address to write the matching bits
reg[57:0]					out_edge_debug_base_addr;
reg[57:0]					debug_base_addr;


//---------------------- EPOCH STATE  ---------------------------//
wire[31:0]					global_epoch;
wire 						global_epoch_change;

always @(posedge clk) begin
    if (~rst_n) begin
        // reset
        um_activated <= 1'b0;
    end
    else if(um_done) begin
        um_activated <= 1'b0;
    end
    else if (start_um ) begin
        um_activated <= 1'b1;
    end
end

always @(posedge clk) begin

	if( ~rst_n) begin
		um_params_reg <= 512'b0;

		ptr_base_addr <= 0;	
		num_ptrs <= 0;
		
		edges_base_addr <= 0;
		num_edges <= 0;

		out_edges_base_addr <= 0;
		out_mcm_bits_base_addr <= 0;

		num_epochs <= 0;
		
		out_edge_debug_base_addr <= 0;

		debug_base_addr <= 0;
		edge_offset <= 0;
	end
	else begin
		if (start_um) begin
		    um_params_reg <= um_params;
			debug_mwm_crouch_csr_ram = um_params;

			ptr_base_addr <= um_params[63:6];
			edges_base_addr <= um_params[127:70];
			out_edges_base_addr <= um_params[191:134];
			out_mcm_bits_base_addr <= um_params[255:198];
			out_edge_debug_base_addr <= um_params[319:262];
			debug_base_addr <= um_params[383:326];
			
			num_ptrs <= um_params[415:384];
			num_edges <= um_params[447:416];
			edge_offset <= um_params[480:448];
			
			num_epochs <= um_params[415:384] >> `K_BITS;
		end
	end
end


//---------------------- POINTER REQUESTER  ---------------------------//
wire pointer_requester_queue_write_enable;
wire [65:0] pointer_requester_queue_buf_in;
wire pointer_requester_queue_buf_filled_next;
wire prq_done;

pointer_requester #( .POINTER_REQUEST_TAG( `D_POINTER_REQUEST_TAG )  ) pointer_requester (
	.clk						(clk),
	.rst_n						(rst_n),
	.um_done					(um_done),
	.prq_done					(prq_done),
	.um_activated				(um_activated),
	.read_address				(ptr_base_addr),
	.num_ptrs					(num_ptrs),
	.queue_req_write_enable			(pointer_requester_queue_write_enable),
	.queue_req_buf_in				(pointer_requester_queue_buf_in),
	.queue_req_buf_filled_next		(pointer_requester_queue_buf_filled_next),

	.debug_pointer_requester		(debug_pointer_requester)
);


//---------------------- POINTER RECEIVER  ---------------------------//

// ----- K many queues for the pointers - GOING TO EDGE REQUESTER ----- //



wire pointer_receiver_queue_write_enable;
wire [127:0] pointer_receiver_queue_buf_in;
wire pointer_receiver_queue_buf_filled_next;


wire[31:0] pointer_receiver_space_left;		// the amount of space left in the queue of the pointer receiver
wire[31:0] pointer_receiver_num_resp_received;		// the amount of responses the pointer receiver has gotten

wire ptr_rec_done;

pointer_receiver #( .BRAM_QUEUE_LENGTH_BITS( `POINTER_RECEIVER_QUEUE_LENGTH_BITS ), .POINTER_REQUEST_TAG ( `D_POINTER_REQUEST_TAG ), .K_BITS(`K_BITS) ) pointer_receiver(
	.clk				(clk),
	.rst_n				(rst_n),
	.um_done			(um_done),
    .um_activated		(um_activated),

	.ptr_rec_done		(ptr_rec_done),
	.num_ptrs			(num_ptrs),
	
	.um_rx_rd_tag		(um_rx_rd_tag),
	.um_rx_data			(um_rx_data),
	.um_rx_rd_valid		(um_rx_rd_valid),

	.queue_write_enable		(pointer_receiver_queue_write_enable),
	.queue_buf_in			(pointer_receiver_queue_buf_in),
    .queue_filled_next   	(pointer_receiver_queue_buf_filled_next),
    
	.num_resp_received      (pointer_receiver_num_resp_received),    
    .space_left             (pointer_receiver_space_left),
	
	.debug_pointer_receiver		(debug_pointer_receiver)
);



//---------------------- EDGE REQUESTER  ---------------------------//

// ----- The queue, to which we put requests, that are then consumed by the  request interface ----- //
wire edge_requester_queue_write_enable;
wire [65:0] edge_requester_queue_buf_in;
wire edge_requester_queue_buf_filled_next;

// ----- The queue, to which we put request information - GOING TO REQUEST INTERFACE ----- //
wire edge_requester_info_queue_write_enable;
wire [127:0] edge_requester_info_queue_buf_in;
wire edge_requester_info_queue_buf_filled_next;

wire sorter_edge_consumed[0:2**`K_BITS-1];

// ----- K many queues for the edges ----- //

wire edge_receiver_queue_write_enable;
wire [95:0] edge_receiver_queue_buf_in;
wire [`K_BITS-1:0] edge_receiver_queue_buf_k_to_write;

wire edge_receiver_queue_buf_filled_next[0:2**`K_BITS-1];

wire edge_requester_done;

edge_requester #( .K_BITS( `K_BITS ), .EDGE_REQUEST_TAG(`D_EDGE_REQUEST_TAG), .QUEUE_THRESHOLD(`D_BRAM_OVERFLOW_PADDING), .BRAM_QUEUE_LENGTH_BITS(`EDGE_REQUESTER_QUEUE_LENGTH_BITS), 
					.SORTER_FIRST_QUEUE_LENGTH_BITS(`SORTER_FIRST_QUEUE_LENGTH_BITS)  )  edge_requester (
	.clk							(clk),
	.rst_n							(rst_n),
    .um_done						(um_done),
	.um_activated					(um_activated),

	.edges_base_addr				(edges_base_addr),
	.num_ptrs						(num_ptrs),
	.edge_requester_done			(edge_requester_done),

	.queue_write_enable				(edge_requester_queue_write_enable),
	.queue_buf_in					(edge_requester_queue_buf_in),
	.queue_buf_filled_next			(edge_requester_queue_buf_filled_next),

	.queue_ptr_write_enable			(pointer_receiver_queue_write_enable),
	.queue_ptr_buf_in				(pointer_receiver_queue_buf_in),
	.queue_ptr_buf_filled_next		(pointer_receiver_queue_buf_filled_next),
	
	.queue_request_write_enable		(edge_requester_info_queue_write_enable),	
	.queue_request_buf_in			(edge_requester_info_queue_buf_in),
	.queue_request_buf_filled_next	(edge_requester_info_queue_buf_filled_next),	

	.debug_edge_requester				(debug_edge_requester),

	.sorter_edge_consumed			(sorter_edge_consumed)
);


//---------------------- EDGE RECEIVER  ---------------------------//


wire edge_receiver_done;
wire[31:0] edge_receiver_num_resp_received;
wire [31:0] edge_receiver_space_left;

edge_receiver #(.BRAM_QUEUE_LENGTH_BITS(`EDGE_RECEIVER_QUEUE_LENGTH_BITS), .K_BITS(`K_BITS), .EDGE_REQUEST_TAG( `D_EDGE_REQUEST_TAG) ) edge_receiver (
	.clk							(clk),
    .rst_n							(rst_n),
	.um_done						(um_done),
	.um_activated					(um_activated),
	.edge_receiver_done				(edge_receiver_done),

	.num_edges						(num_edges),
	.edge_requester_done			(edge_requester_done),

	.um_rx_rd_tag					(um_rx_rd_tag),
	.um_rx_data						(um_rx_data),
	.um_rx_rd_valid					(um_rx_rd_valid),
	
	.queue_write_enable				(edge_receiver_queue_write_enable),
	.queue_buf_in					(edge_receiver_queue_buf_in),
	.queue_write_enable_k			(edge_receiver_queue_buf_k_to_write),
	
	.space_left							(edge_receiver_space_left),
	.edge_receiver_num_resp_received	(edge_receiver_num_resp_received),

	.queue_request_write_enable		(edge_requester_info_queue_write_enable),
	.queue_request_buf_in			(edge_requester_info_queue_buf_in),	
	.queue_request_filled_next		(edge_requester_info_queue_buf_filled_next),

	.debug_edge_receiver			(debug_edge_receiver)
);


//---------------------- SORTER  ---------------------------//
wire sorter_queue_write_enable;
wire [127:0] sorter_queue_buf_in;
wire sorter_queue_buf_filled_next;

wire sorter_done;

wire								out_edge_debug_write_enable;
wire [569:0]						out_edge_debug_buf_in;
wire								out_edge_debug_buf_filled_next;

sorter #( .K_BIT_2 (`K_BITS), .K_BITS(`K_BITS), .SORTER_FIRST_QUEUE_LENGTH_BITS(`SORTER_FIRST_QUEUE_LENGTH_BITS), .BRAM_OVERFLOW_PADDING(`D_BRAM_OVERFLOW_PADDING) ) sorter (
	.clk							(clk),
    .rst_n							(rst_n),
	.um_done						(um_done),
	.um_activated					(um_activated),
	.sorter_done					(sorter_done),

	.in_queue_write_enable			(edge_receiver_queue_write_enable),
	.in_queue_buf_in				(edge_receiver_queue_buf_in),
	.in_queue_write_enable_k		(edge_receiver_queue_buf_k_to_write),

	.edge_receiver_done				(edge_receiver_done),

	.out_queue_write_enable			(sorter_queue_write_enable),
	.out_queue_buf_in				(sorter_queue_buf_in),
	.out_queue_buf_filled_next		(sorter_queue_buf_filled_next),
	
	.out_edge_debug_buf_in			(out_edge_debug_buf_in),
	.out_edge_debug_write_enable	(out_edge_debug_write_enable),
	.out_edge_debug_buf_filled_next	(out_edge_debug_buf_filled_next),
	.edge_debug_address				(out_edge_debug_base_addr),
		
	.sorter_edge_consumed			(sorter_edge_consumed),

	.debug_sorter					(debug_sorter)
	
);


//---------------------- MCM BIT REQUESTER ---------------------------//

// ----- The queue to which the mcm-bit requester places the requests ----- //
wire mcm_bit_requester_queue_write_enable;
wire [65:0] mcm_bit_requester_queue_buf_in;
wire mcm_bit_requester_queue_buf_filled_next;


// ----- The queue to which the mcm-bit requester places the informations about the requests made (offset) ----- //
wire mcm_bit_requester_info_queue_write_enable;
wire [(9 - `L_BITS)+1:0] mcm_bit_requester_info_queue_buf_in;
wire mcm_bit_requester_info_queue_buf_filled_next;


wire mcm_bit_requester_ready_queue_write_enable;
wire [127:0] mcm_bit_requester_ready_queue_buf_in;
wire mcm_bit_requester_ready_queue_buf_filled_next;

wire[31:0] mcm_bit_requester_num_requests_issued;

wire mcm_bit_requester_epoch_done;

wire mcm_bit_ack_receiver_epoch_done;

wire mcm_bit_requester_done;

mcm_bit_requester #(.BRAM_QUEUE_LENGTH_BITS(`MCM_BIT_REQUESTER_QUEUE_LENGTH_BITS), .MCM_BITS_REQUEST_TAG(`D_MCM_BITS_REQUEST_TAG), .BRAM_OVERFLOW_PADDING(`D_BRAM_OVERFLOW_PADDING),  .L_BITS(`L_BITS) ) mcm_bit_requester (

	.clk									(clk),
    .rst_n									(rst_n),
	.um_done								(um_done),
	.um_activated							(um_activated),
	.mcm_bit_requester_done					(mcm_bit_requester_done),
	.mcm_bit_requester_num_requests_issued	(mcm_bit_requester_num_requests_issued),

	.global_epoch							(global_epoch),
	.epoch_done								(mcm_bit_requester_epoch_done),

	.read_mcm_base_address				(out_mcm_bits_base_addr),
	.num_edges							(num_edges),

	.req_queue_write_enable				(mcm_bit_requester_queue_write_enable),
	.req_queue_buf_in					(mcm_bit_requester_queue_buf_in),
	.req_queue_buf_filled_next			(mcm_bit_requester_queue_buf_filled_next),

	.sorter_queue_write_enable			(sorter_queue_write_enable),
	.sorter_queue_buf_in				(sorter_queue_buf_in),
	.sorter_queue_buf_filled_next		(sorter_queue_buf_filled_next),

	.ready_queue_write_enable			(mcm_bit_requester_ready_queue_write_enable),
	.ready_queue_buf_in					(mcm_bit_requester_ready_queue_buf_in),
	.ready_queue_buf_filled_next		(mcm_bit_requester_ready_queue_buf_filled_next),
	
	.offset_queue_write_enable			(mcm_bit_requester_info_queue_write_enable),
	.offset_queue_buf_in				(mcm_bit_requester_info_queue_buf_in),
	.offset_queue_buf_filled_next		(mcm_bit_requester_info_queue_buf_filled_next),

	.debug_mcm_bit_requester			(debug_mcm_bit_requester)
);

wire [31:0]		mcm_bit_receiver_space_left;
wire [31:0]		mcm_bit_receiver_num_resp_received;



//---------------------- MCM BIT RECEIVER ---------------------------//
wire mcm_bit_receiver_bits_queue_write_enable;
wire [512-1:0] mcm_bit_receiver_bits_queue_buf_in;
wire mcm_bit_receiver_bits_queue_buf_filled_next;

wire mcm_bit_receiver_done;
wire mcm_bit_receiver_epoch_done;

mcm_bit_receiver #(.BRAM_QUEUE_LENGTH_BITS(`MCM_BIT_RECEIVER_QUEUE_LENGTH_BITS), .MCM_BITS_REQUEST_TAG(`D_MCM_BITS_REQUEST_TAG) ) mcm_bit_receiver  (
	.clk							(clk),
    .rst_n							(rst_n),
	.um_done						(um_done),
	.um_activated					(um_activated),

	.mcm_bit_receiver_done						(mcm_bit_receiver_done),
	.mcm_bit_requester_num_requests_issued		(mcm_bit_requester_num_requests_issued),
	.mcm_bit_requester_done						(mcm_bit_requester_done),
	.mcm_bit_requester_epoch_done				(mcm_bit_requester_epoch_done),
	.epoch_done									(mcm_bit_receiver_epoch_done),

	.um_rx_rd_tag					(um_rx_rd_tag),
    .um_rx_data						(um_rx_data),
    .um_rx_rd_valid					(um_rx_rd_valid),

	.nelems_space					(mcm_bit_receiver_space_left), 						// the number of space left in our input queue
	.num_resp_received				(mcm_bit_receiver_num_resp_received),

	.out_queue_write_enable			(mcm_bit_receiver_bits_queue_write_enable),
	.out_queue_buf_in				(mcm_bit_receiver_bits_queue_buf_in),
	.out_queue_filled_next			(mcm_bit_receiver_bits_queue_buf_filled_next),
	.debug_mcm_bit_receiver			(debug_mcm_bit_receiver)
);

//---------------------- EDGE PROCESSOR ---------------------------//
wire 			edge_processor_done;
wire 			edge_processor_epoch_done;

wire 					queue_mcmbram_write_enable;
wire[2**`L_BITS -1:0]	queue_mcmbram_buf_in;
wire					queue_mcmbram_buf_filled_next;

wire 				edge_writer_edge_queue_write_enable;
wire [128-1:0]		edge_writer_edge_queue_buf_in;
wire				edge_writer_edge_queue_buf_filled_next;

wire				mcm_bit_writer_mcm_queue_write_enable;
wire [32+512-1:0]	mcm_bit_writer_mcm_queue_buf_in;
wire				mcm_bit_writer_mcm_buf_filled_next;

wire bram_mcm_bit_receiver_epoch_done;

edge_processor #(.BRAM_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS), .BRAM_QUEUE_LENGTH_BITS_LARGE(BRAM_QUEUE_LENGTH_BITS_LARGE), .K_BITS(`K_BITS), .L_BITS(`L_BITS), .OVERFLOW_PADDING(`D_BRAM_OVERFLOW_PADDING), .EPSILON(EPSILON) ) edge_processor (
	
	.clk								(clk),
    .rst_n								(rst_n),
	.um_done							(um_done),
	.um_activated						(um_activated),
	.edge_processor_done				(edge_processor_done),
	.num_edges							(num_edges),

	.mcm_bit_receiver_epoch_done		(mcm_bit_receiver_epoch_done),
	.bram_mcm_bit_receiver_epoch_done	(bram_mcm_bit_receiver_epoch_done),
	
	.global_epoch						(global_epoch),
	.global_epoch_change				(global_epoch_change),
	.epoch_done							(edge_processor_epoch_done),

	.queue_mcmbram_write_enable			(queue_mcmbram_write_enable),
	.queue_mcmbram_buf_in				(queue_mcmbram_buf_in),	
	.queue_mcmbram_buf_filled_next		(queue_mcmbram_buf_filled_next),

	.queue_offset_write_enable			(mcm_bit_requester_info_queue_write_enable),
	.queue_offset_buf_in				(mcm_bit_requester_info_queue_buf_in),
	.queue_offset_buf_filled_next		(mcm_bit_requester_info_queue_buf_filled_next),

    .queue_mcmin_write_enable			(mcm_bit_receiver_bits_queue_write_enable),					
	.queue_mcmin_buf_in					(mcm_bit_receiver_bits_queue_buf_in),		
	.queue_mcmin_buf_filled_next		(mcm_bit_receiver_bits_queue_buf_filled_next),

    .queue_edgein_write_enable			(mcm_bit_requester_ready_queue_write_enable),
	.queue_edgein_buf_in				(mcm_bit_requester_ready_queue_buf_in),
	.queue_edgein_buf_filled_next		(mcm_bit_requester_ready_queue_buf_filled_next),

	.queue_mcmout_write_enable			(mcm_bit_writer_mcm_queue_write_enable),
	.queue_mcmout_buf_in				(mcm_bit_writer_mcm_queue_buf_in),
	.queue_mcmout_buf_filled_next		(mcm_bit_writer_mcm_buf_filled_next),

	.queue_edgeout_write_enable				(edge_writer_edge_queue_write_enable),
	.queue_edgeout_buf_in					(edge_writer_edge_queue_buf_in),
	.queue_edgeout_buf_filled_next			(edge_writer_edge_queue_buf_filled_next),

	.debug_edge_processor					(debug_edge_processor)

);

//---------------------- BRAM MCM-BITS REQUESTER ---------------------------//

wire bram_mcm_bit_requester_done;
wire				bram_mcm_bit_requester_epoch_done;

wire									bram_mcm_bit_requester_queue_buf_filled_next;
wire [65:0]								bram_mcm_bit_requester_queue_buf_in;
wire 									bram_mcm_bit_requester_queue_write_enable;



bram_mcm_bit_requester #(.BRAM_MCM_BITS_REQUEST_TAG (`D_BRAM_MCM_BITS_REQUEST_TAG ), .L_BITS(`L_BITS), .K_BITS(`K_BITS) ) bram_mcm_bit_requester (
	.clk									(clk),
    .rst_n									(rst_n),
	.um_done								(um_done),
	.um_activated							(um_activated),
	
	.bram_mcm_bit_requester_done			(bram_mcm_bit_requester_done),
	.epoch_done								(bram_mcm_bit_requester_epoch_done),
	.num_epochs								(num_epochs),

    .global_epoch							(global_epoch),
	.global_epoch_change					(global_epoch_change),
	
	.mcm_bit_base_address					(out_mcm_bits_base_addr),
	
	.req_queue_write_enable					(bram_mcm_bit_requester_queue_write_enable),
	.req_queue_buf_in						(bram_mcm_bit_requester_queue_buf_in),
	.req_queue_buf_filled_next				(bram_mcm_bit_requester_queue_buf_filled_next),

	.debug_bram_mcm_bit_requester			(debug_bram_mcm_bit_requester)
	
);


//---------------------- BRAM MCM-BITS RECEIVER ---------------------------//
wire[31:0]								bram_mcm_bit_receiver_num_resp_received;
wire[31:0]								bram_mcm_bit_receiver_space_left;


wire bram_mcm_bit_receiver_done;

bram_mcm_bit_receiver #(.BRAM_MCM_BITS_REQUEST_TAG(`D_BRAM_MCM_BITS_REQUEST_TAG), .OVERFLOW_PADDING(`D_BRAM_OVERFLOW_PADDING),  .L_BITS(`L_BITS), .K_BITS(`K_BITS), .BRAM_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS)) bram_mcm_bit_receiver (
	
	.clk								(clk),
    .rst_n								(rst_n),
	.um_done							(um_done),
	.um_activated						(um_activated),

	.bram_mcm_bit_receiver_done			(bram_mcm_bit_receiver_done),
	.bram_mcm_bit_receiver_epoch_done	(bram_mcm_bit_receiver_epoch_done),
	.global_epoch_change				(global_epoch_change),
	.bram_mcm_bit_requester_done		(bram_mcm_bit_requester_done),

	.um_rx_rd_tag						(um_rx_rd_tag),
    .um_rx_data							(um_rx_data),
    .um_rx_rd_valid						(um_rx_rd_valid),

	.nelems_space						(bram_mcm_bit_receiver_space_left),
	.num_resp_received					(bram_mcm_bit_receiver_num_resp_received),

	.out_queue_write_enable				(queue_mcmbram_write_enable),
	.out_queue_buf_in					(queue_mcmbram_buf_in),
	.in_queue_buf_filled_next			(queue_mcmbram_buf_filled_next),

	.debug_bram_mcm_bit_receiver		(debug_bram_mcm_bit_receiver)
);



//---------------------- ACK RECEIVER ---------------------------//
wire mcm_bit_ack_receiver_done;

wire mcm_bit_writer_done;
wire mcm_bit_writer_epoch_done;

wire[31:0]			mcm_bit_writer_num_writes;

mcm_bit_ack_receiver #(. MCM_BIT_WRITE_TAG (`D_MCM_WRITE_TAG), .L_BITS(`L_BITS) ) mcm_bit_ack_receiver (
    .clk								(clk),
    .rst_n								(rst_n),
	.um_done							(um_done),
	.um_activated						(um_activated),

	.mcm_bit_ack_receiver_done			(mcm_bit_ack_receiver_done),
	.mcm_bit_ack_receiver_epoch_done	(mcm_bit_ack_receiver_epoch_done),

	.mcm_bit_writer_done				(mcm_bit_writer_done),
	.mcm_bit_writer_epoch_done			(mcm_bit_writer_epoch_done),
	.mcm_bit_writer_num_writes			(mcm_bit_writer_num_writes),

    .um_rx_wr_valid						(um_rx_wr_valid),
    .um_rx_wr_tag						(um_rx_wr_tag),
	.debug_mcm_bit_ack_receiver			(debug_mcm_bit_ack_receiver)
);



//---------------------- MCM BIT WRITER ---------------------------//


wire 				mcm_writer_out_queue_write_enable;
wire [569:0]  	mcm_writer_out_queue_buf_in;
wire 				mcm_writer_out_queue_buf_filled_next;


mcm_bit_writer  #( .OVERFLOW_PADDING(`D_BRAM_OVERFLOW_PADDING), .MCM_OUT_TAG(`D_MCM_WRITE_TAG), .BRAM_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS), .L_BITS(`L_BITS) ) mcm_bit_writer (
	.clk							(clk),
    .rst_n							(rst_n),
	.um_done						(um_done),
	.um_activated					(um_activated),
	.out_mcm_bits_base_addr			(out_mcm_bits_base_addr),

	.mcm_writer_done				(mcm_bit_writer_done),
	.mcm_writer_epoch_done			(mcm_bit_writer_epoch_done),
	.mcm_bit_writer_num_writes		(mcm_bit_writer_num_writes),
	.edge_processor_done			(edge_processor_done),
	.edge_processor_epoch_done		(edge_processor_epoch_done),

	.mcm_queue_write_enable			(mcm_bit_writer_mcm_queue_write_enable),
	.mcm_queue_buf_in				(mcm_bit_writer_mcm_queue_buf_in),
	.mcm_queue_buf_filled_next		(mcm_bit_writer_mcm_buf_filled_next),

	.write_queue_write_enable		(mcm_writer_out_queue_write_enable),
	.write_queue_buf_in				(mcm_writer_out_queue_buf_in),
	.write_queue_buf_filled_next	(mcm_writer_out_queue_buf_filled_next),

	.debug_mcm_bit_writer			(debug_mcm_bit_writer)
);



//---------------------- EDGE WRITER ---------------------------//
wire edge_writer_done;

wire 				edge_writer_out_queue_write_enable;
wire [569:0]  		edge_writer_out_queue_buf_in;
wire 				edge_writer_out_queue_buf_filled_next;

edge_writer  #(.OVERFLOW_PADDING (`D_BRAM_OVERFLOW_PADDING), .EDGE_OUT_TAG(`D_EDGE_WRITE_TAG), .BRAM_QUEUE_LENGTH_BITS ( BRAM_QUEUE_LENGTH_BITS ), .L_BITS(`L_BITS) ) edge_writer  (
	
	.clk					(clk),
    .rst_n					(rst_n),
	.um_done				(um_done),
	.um_activated			(um_activated),
	.out_edges_base_addr	(out_edges_base_addr),

	.edge_writer_done		(edge_writer_done),
	.edge_processor_done	(edge_processor_done),

	.edge_queue_write_enable	(edge_writer_edge_queue_write_enable),
	.edge_queue_buf_in			(edge_writer_edge_queue_buf_in),
	.edge_queue_buf_filled_next	(edge_writer_edge_queue_buf_filled_next),

	.write_queue_write_enable			(edge_writer_out_queue_write_enable),
	.write_queue_buf_in					(edge_writer_out_queue_buf_in),
	.write_queue_buf_filled_next		(edge_writer_out_queue_buf_filled_next),
	.debug_edge_writer					(debug_edge_writer),
	.out_edges_offset_addr				(edge_offset)
);

//---------------------- REQUEST INTERFACE  ---------------------------//

request_interface #(.L_BITS(`L_BITS), .BRAM_QUEUE_LENGTH_BITS(`REQUEST_INTERFACE_QUEUE_LENGTH_BITS) ) request_interface(
	
	.clk				(clk),
    .rst_n				(rst_n),
	.um_done			(um_done),
	.um_activated		(um_activated),	
    .um_tx_rd_addr		(um_tx_rd_addr),
    .um_tx_rd_tag 		(um_tx_rd_tag),
    .um_tx_rd_valid		(um_tx_rd_valid),
    .um_tx_rd_ready		(um_tx_rd_ready),
	.um_rx_rd_valid		(um_rx_rd_valid),

	.ptr_base_addr				(ptr_base_addr),
	.edges_base_addr			(edges_base_addr),
	.out_mcm_bits_base_addr		(out_mcm_bits_base_addr),	
	.num_ptrs					(num_ptrs),
	.num_edges					(num_edges),

	.pointer_requester_queue_buf_filled_next	(pointer_requester_queue_buf_filled_next),
	.pointer_requester_queue_buf_in				(pointer_requester_queue_buf_in),
	.pointer_requester_queue_write_enable		(pointer_requester_queue_write_enable),
	.pointer_receiver_num_resp_received			(pointer_receiver_num_resp_received),
	.pointer_receiver_space_left				(pointer_receiver_space_left),
	
	.edge_requester_queue_buf_filled_next		(edge_requester_queue_buf_filled_next),
	.edge_requester_queue_buf_in				(edge_requester_queue_buf_in),
	.edge_requester_queue_write_enable			(edge_requester_queue_write_enable),
	.edge_receiver_num_resp_received			(edge_receiver_num_resp_received),
	.edge_receiver_space_left					(edge_receiver_space_left),

	.mcm_bit_requester_queue_buf_filled_next	(mcm_bit_requester_queue_buf_filled_next),
	.mcm_bit_requester_queue_buf_in				(mcm_bit_requester_queue_buf_in),
	.mcm_bit_requester_queue_write_enable		(mcm_bit_requester_queue_write_enable),
	.mcm_bit_receiver_num_resp_received			(mcm_bit_receiver_num_resp_received),
	.mcm_bit_receiver_space_left				(mcm_bit_receiver_space_left),

	.bram_mcm_bit_requester_queue_buf_filled_next	(bram_mcm_bit_requester_queue_buf_filled_next),
	.bram_mcm_bit_requester_queue_buf_in			(bram_mcm_bit_requester_queue_buf_in),
	.bram_mcm_bit_requester_queue_write_enable		(bram_mcm_bit_requester_queue_write_enable),
	.bram_mcm_bit_receiver_num_resp_received		(bram_mcm_bit_receiver_num_resp_received),
	.bram_mcm_bit_receiver_space_left				(bram_mcm_bit_receiver_space_left),

	.debug_request_interface						(debug_request_interface)
);

//---------------------- WRITE INTERFACE  ---------------------------//

wire write_interface_done;


write_interface #(.OVERFLOW_PADDING (`D_BRAM_OVERFLOW_PADDING), .BRAM_QUEUE_LENGTH_BITS(BRAM_QUEUE_LENGTH_BITS), .L_BITS(`L_BITS), 
								.NUM_WRITES_IN_FLIGHT(`D_NUM_WRITES_IN_FLIGHT), .EDGE_WRITE_TAG(`D_EDGE_WRITE_TAG), .MCM_WRITE_TAG(`D_MCM_WRITE_TAG) ) write_interface (
	
	.clk								(clk),
    .rst_n								(rst_n),
	.um_done							(um_done),
	.um_activated						(um_activated),

	.ack_receiver_done					(mcm_bit_ack_receiver_done),
	.mcm_bit_writer_done				(mcm_bit_writer_done),
	.edge_writer_done					(edge_writer_done),
	.write_interface_done				(write_interface_done),

	.ptr_base_addr						(ptr_base_addr),
	.edges_base_addr					(edges_base_addr),
	.out_edges_base_addr				(out_edges_base_addr),
	.out_mcm_bits_base_addr				(out_mcm_bits_base_addr),
	.out_edge_debug_base_addr			(out_edge_debug_base_addr),
	.debug_base_addr					(debug_base_addr),
	.num_ptrs							(num_ptrs),
	.num_edges							(num_edges),

	.um_tx_wr_addr						(um_tx_wr_addr),
    .um_tx_wr_tag						(um_tx_wr_tag),
    .um_tx_wr_valid						(um_tx_wr_valid),
    .um_tx_data							(um_tx_data),
    .um_tx_wr_ready						(um_tx_wr_ready),

	.um_rx_wr_valid						(um_rx_wr_valid),
    .um_rx_wr_tag						(um_rx_wr_tag),

	.edge_queue_write_enable			(edge_writer_out_queue_write_enable),
	.edge_queue_buf_in					(edge_writer_out_queue_buf_in),
	.edge_queue_buf_filled_next			(edge_writer_out_queue_buf_filled_next),

	.mcm_queue_write_enable				(mcm_writer_out_queue_write_enable),
	.mcm_queue_buf_in					(mcm_writer_out_queue_buf_in),
	.mcm_queue_buf_filled_next			(mcm_writer_out_queue_buf_filled_next),

	.edge_debug_queue_write_enable		(out_edge_debug_write_enable),
	.edge_debug_queue_buf_in			(out_edge_debug_buf_in),
	.edge_debug_queue_buf_filled_next	(out_edge_debug_buf_filled_next),

	.debug_cl_queue_write_enable		(debug_cl_queue_write_enable),
	.debug_cl_queue_buf_in				(debug_cl_queue_buf_in),
	.debug_cl_queue_buf_filled_next		(debug_cl_queue_buf_filled_next),

	.debug_write_interface				(debug_write_interface)
);



state_controller state_controller (
	.clk							(clk),
    .rst_n							(rst_n),
	.um_activated					(um_activated),	
	.um_done						(um_done),


	.num_epochs						(num_epochs),
	.pointer_requester_done			(prq_done),
	.pointer_receiver_done			(ptr_rec_done),
	.edge_requester_done			(edge_requester_done),
	.edge_receiver_done				(edge_receiver_done),
	.sorter_done					(sorter_done),

	.mcm_bit_requester_done			(mcm_bit_requester_done),
	.mcm_bit_receiver_done			(mcm_bit_receiver_done),

	.bram_mcm_bit_requester_done	(bram_mcm_bit_requester_done),
	.bram_mcm_bit_receiver_done		(bram_mcm_bit_receiver_done),		
	.edge_processor_done			(edge_processor_done),

	.write_interface_done			(write_interface_done),

	.mcm_bit_ack_receiver_done		(mcm_bit_ack_receiver_done),
	.edge_writer_done				(edge_writer_done),

	.mcm_bit_requester_epoch_done	(mcm_bit_requester_epoch_done),
	.mcm_bit_receiver_epoch_done	(mcm_bit_receiver_epoch_done),
	.edge_processor_epoch_done		(edge_processor_epoch_done),
	.mcm_writer_epoch_done			(mcm_bit_writer_epoch_done),
	.bram_mcm_bit_requester_epoch_done	(bram_mcm_bit_requester_epoch_done),
	.bram_mcm_bit_receiver_epoch_done	(bram_mcm_bit_receiver_epoch_done),
	.mcm_bit_ack_receiver_epoch_done	(mcm_bit_ack_receiver_epoch_done),
	
	// for epoch handling
	.global_epoch						(global_epoch),
	.global_epoch_change				(global_epoch_change),

	.debug_state_controller				(debug_state_controller)
);


mwm_debug  #(.NUM_USER_STATE_COUNTERS(NUM_USER_STATE_COUNTERS) ) mwm_debug(
	
	.clk							(clk),
    .rst_n							(rst_n),
	.um_done						(um_done),
	.um_activated					(um_activated),

	.debug_base_addr				(debug_base_addr),

	.debug_bram_mcm_bit_receiver	(debug_bram_mcm_bit_receiver),	// 0
	.debug_bram_mcm_bit_requester	(debug_bram_mcm_bit_requester),	// 1
	.debug_edge_processor			(debug_edge_processor),			// 2
	.debug_edge_receiver			(debug_edge_receiver),			// 3
	.debug_edge_requester			(debug_edge_requester),			// 4
	.debug_edge_writer				(debug_edge_writer),				// 5
	.debug_mcm_bit_ack_receiver		(debug_mcm_bit_ack_receiver),		// 6
	.debug_mcm_bit_receiver			(debug_mcm_bit_receiver),			// 7
	.debug_mcm_bit_requester		(debug_mcm_bit_requester),		// 8
	.debug_mcm_bit_writer			(debug_mcm_bit_writer),			// 9
	.debug_mwm_crouch_csr_ram		(debug_mwm_crouch_csr_ram),		// 10
	.debug_pointer_receiver			(debug_pointer_receiver),			// 11
	.debug_pointer_requester		(debug_pointer_requester),		// 12
	.debug_request_interface		(debug_request_interface),		// 13
	.debug_sorter					(debug_sorter),					// 14
	.debug_state_controller			(debug_state_controller),			// 15
	.debug_write_interface			(debug_write_interface),			// 16

	.debug_cl_queue_write_enable	(debug_cl_queue_write_enable),
	.debug_cl_queue_buf_in			(debug_cl_queue_buf_in),
	.debug_cl_queue_buf_filled_next	(debug_cl_queue_buf_filled_next)
);



//---------------------- OTHER INTERFACE STUFF  ---------------------------//
assign um_rx_rd_ready = 1; // we can always consume incoming data


endmodule




`default_nettype wire


