

/**
	The "pointer receiver" gets the responses from the framework. In each cache line, we have 5 pointer entries (96 bits each).
	Therefore, we require 5 cycles to process an entry and must not care about the "2-cycle latency" of the bram queues.

	We put data only in the output queues, if there is enough space left in those queues.
	Since we request sequentiall the pointers, the edge-receiver assigns the received pointers in a round robin fashion.
	The edge requester processes the pointers in a "epoch-step-wise" fashion. Therefore, the queue's length differs by at most 1, and it is sufficient to check the length of only one queue.

	TODO: CHECK THAT!!!!
	-> PROBLEM IF THE EDGE-REQUESTER DOES SO: WE HAVE A FULL ROW, A 1 ROW, A 1 ROW -> WE MUST ALLOW THAT WE ARE IN DIFFERENT EPOCHS!!!! OTHERWISE, WE HAVE HOLES IN THE SORTER, WHICH ARE NOT FILLED!
	-> can problems occur, if we check and stall, if a queue overflows?

	NOTE: WHAT ABOUT 1 QUEUE, WITH 95*K WIDTH? -> maybe would not synthesize!

todo: recheck!!!!
*/

module pointer_receiver #(parameter BRAM_QUEUE_LENGTH_BITS  = 7, parameter K_BITS = 3, parameter POINTER_REQUEST_TAG = 1 ) (
	//---------------------- GENERAL VARIABLES---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input 	wire 									um_done,						// true, if we are done processing (global)
	input 	wire 									um_activated,					// true, if the afu is activated
	//---------------------- TERMINATION ---------------------------//
	output reg 										ptr_rec_done,					// true, if we are done here
	input [31:0]									num_ptrs,                    // the amount of cache-lines we must process

	//---------------------- GENERAL READ INPUT VARIABLES---------------------------//
	input   wire [7:0]                         	    um_rx_rd_tag,	// response to the read request: the tag
    input   wire [511:0]                       	    um_rx_data,		// the requested data
    input   wire                               	    um_rx_rd_valid,	// that tag and data are valid
	
	//---------------------- VARIABLES FOR THE K MANY POINTER QUEUES ---------------------------//
	output reg 						queue_write_enable,					// wheter we want to write to the queue or not
	output reg [127:0] 				queue_buf_in,						// row_pointer, cache-line offset, number of edges
	input wire 						queue_filled_next,	                // since the edge requester processes the pointers "epoch-step-wise", every queue differs at most 1 in the length compared to the other queues. So we only need the value of one queue to determine, if it is soon filled - BUT WE DO IT THAT WAY LATER ON!

	//---------------------- VARIABLES TO INDICATE THE CURRENT INPUT SPACE WE HAVE LEFT ---------------------------//
	output 	reg[31:0]								num_resp_received,			// the number of pointers (in cache-lines) received so far
	output reg [31:0]								space_left,                 // the space left in the input queue

	output reg[511:0]								debug_pointer_receiver
);


//---------------------- RECEIVE DATA AND PUT INTO QUEUE AND TRACK THE AVAILABLE INPUT SPACE  ---------------------------//
reg in_queue_m_axis_ready;
reg in_queue_m_axis_valid;
reg [511:0] in_queue_buf_in;
wire [511:0] in_queue_buf_out;
wire [BRAM_QUEUE_LENGTH_BITS-1:0] in_queue_nelems;
reg [31:0] in_queue_ntotal_read;
reg [31:0] in_queue_ntotal_write;
reg in_queue_write_enable;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(512), .PROG_FULL_THRESH(0 ) ) nlb_axififo	(
    .clk (clk),
    .rst(~rst_n),
    
    .s_axis_valid(in_queue_write_enable),
    .s_axis_ready(),
    .s_axis_data(in_queue_buf_in),
    .s_axis_last(1'b0),
    .m_axis_valid(in_queue_m_axis_valid),
    .m_axis_ready(in_queue_m_axis_ready),
    .m_axis_data(in_queue_buf_out),
    .m_axis_last(),
    .data_count(in_queue_nelems)
);


//---------------------- STORE DATA IN THE QUEUE  ---------------------------//
always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
        // reset of output registers
        space_left <= 0;
		num_resp_received <= 0;

		// reset of local variables
		in_queue_write_enable <= 0;
		in_queue_buf_in <= 0;
		in_queue_ntotal_write <= 0;
    end
    else begin
		space_left <= (2**BRAM_QUEUE_LENGTH_BITS-1) - in_queue_nelems - 10;

		if( ~ptr_rec_done) begin
			if( um_rx_rd_valid & ( um_rx_rd_tag == POINTER_REQUEST_TAG) ) begin
				in_queue_buf_in <= um_rx_data;
				in_queue_write_enable <= 1;
				in_queue_ntotal_write <= in_queue_ntotal_write  + 1;

				num_resp_received <= num_resp_received +1;
			end
			else begin
				in_queue_write_enable <= 0;
				in_queue_buf_in <= 0;
			end
		end
	end
end

//---------------------- Small wrapper to always have a new value in a register  ---------------------------//

reg wrapper_read;
reg wrapper_valid;
reg [511:0] wrapper_data;
always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
		in_queue_m_axis_ready <= 0;
		in_queue_ntotal_read<= 0;
		wrapper_valid <= 0;

		// drain the queue
		if( (um_done || ~um_activated) && rst_n ) begin
			in_queue_m_axis_ready <= 1;
		end
		else begin
			in_queue_m_axis_ready <= 0;
		end
	end
	else begin
		if( in_queue_m_axis_ready && in_queue_m_axis_valid ) begin
			in_queue_m_axis_ready <= 0;
			wrapper_data <= in_queue_buf_out;
			wrapper_valid<= 1;
			in_queue_ntotal_read <= in_queue_ntotal_read  + 1;
		end
		else if( ~wrapper_valid || wrapper_read ) begin
			in_queue_m_axis_ready <= 1;
			wrapper_valid<= 0;
		end
	end
end


//---------------------- DETERMINE WHERE THE RECEIVED DATA MUST GO TO AND UPDATE K  ---------------------------//

reg [K_BITS:0] k_cnt;

reg [511:0] input_buf;
reg input_buf_empty;

reg [4:0]  offset_cnt;	// a cl holds 5 pointers, so we need to count the offset

reg [31:0] 	ptr_processed_cnt;	// number of processed ptrs
reg [31:0]	debug_num_flush_pointers;
reg [31:0] 	debug_edges_count;

reg trigger;

always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
		// reset of output variables
		ptr_rec_done <= 0;
		queue_write_enable <= 0;
		queue_buf_in <= 0;

		// reset of local state-vars
		ptr_processed_cnt <= 0;
		debug_num_flush_pointers <= 0;
		debug_edges_count <= 0;
		k_cnt <= 0;
		input_buf_empty <= 1;
		input_buf <= 0;
		offset_cnt <= 0;

		// wrapper
		wrapper_read <= 0;
		trigger <= 0;
    end
    else begin

		if( in_queue_nelems > 0) begin
			trigger <= 1;
		end


		wrapper_read <= 0;
		queue_write_enable <= 0;
		
		if( ~ptr_rec_done && trigger == 1) begin		
			// if we have processed all cache lines, we are done. check that input_buf_empty holds! otherwise we have a bug	
			if( ptr_processed_cnt == num_ptrs ) begin
				ptr_rec_done <= 1;
			end
			// if the input buffer is empty, get the next data	
			else if( input_buf_empty ) begin
				if( wrapper_valid ) begin
					input_buf <= wrapper_data;
					input_buf_empty <= 0;
					wrapper_read <= 1;
				end
			end
			// if there is enough space in our output queues, and we have input
			else if( ~queue_filled_next ) begin

				if( input_buf[95:64] == 0 ) begin
					debug_num_flush_pointers <= debug_num_flush_pointers  + 1;
				end
				debug_edges_count<= debug_edges_count  + input_buf[95:64];

				queue_buf_in[95:0] <=  input_buf[95:0];
				queue_buf_in[127:96] <= ptr_processed_cnt >> K_BITS;
				queue_write_enable <= 1;

				ptr_processed_cnt <= ptr_processed_cnt+1;

				// grab new data, if we have read the last element
				if( offset_cnt +1 >= 4'd5) begin
					offset_cnt <= 0;

					if( wrapper_valid ) begin
						input_buf <= wrapper_data;
						input_buf_empty <= 0;
						wrapper_read <= 1;			
					end
					else begin
						input_buf <= 0;
						input_buf_empty <= 1;
					end
				end
				else begin
					// else, go to the next element via shifting
					offset_cnt <= offset_cnt + 1;
					input_buf <= (input_buf >> 96);
				end

				// update of k
				if( k_cnt +1 >= 2**K_BITS) begin
					k_cnt <= 0;
				end
				else begin
					k_cnt <= k_cnt+1;
				end

			end
		end
	end
end

// high, if we do not output data, but we could
reg				do_not_work_valid;
reg [31:0]		do_not_work_cycle_counter;
// high, if we hava data, but cannot output data
reg				can_not_output_valid;
reg [31:0]		can_not_output_cycle_counter;
// high, if we do not have data, but could output
reg				can_not_input_valid;
reg [31:0]		can_not_input_cycle_counter;

always @(posedge clk) begin
	if (~rst_n | um_done | ~um_activated) begin
		do_not_work_valid <= 0;
		do_not_work_cycle_counter <= 0;
		can_not_output_valid <= 0;
		can_not_output_cycle_counter <= 0;
		can_not_input_valid <= 0;
		can_not_input_cycle_counter <= 0;
	end
	else begin
		do_not_work_valid <= 0;
		can_not_output_valid <= 0;
		can_not_input_valid <= 0;

		if( ~queue_filled_next && in_queue_nelems > 0 && ~queue_write_enable) begin
			do_not_work_valid<= 1;
			do_not_work_cycle_counter <= do_not_work_cycle_counter + 1;
		end
		if( in_queue_nelems > 0 && queue_filled_next ) begin
			can_not_output_valid <= 1;
			can_not_output_cycle_counter <= can_not_output_cycle_counter + 1;
		end
		if( ~queue_filled_next && in_queue_nelems  == 0 ) begin
			can_not_input_valid <= 1;
			can_not_input_cycle_counter <= can_not_input_cycle_counter +1;
		end
	end
end


always @(posedge clk) begin
	debug_pointer_receiver <= {352'b0, debug_edges_count, in_queue_ntotal_read, in_queue_ntotal_write, ptr_processed_cnt, debug_num_flush_pointers};
end




endmodule






