
/**
	The "pointer requester" issues requests to the main memory to get the row-pointers.
	Since the pointers are stored consecutively in main memory, we can issue requests sequential.

	We can issue the requests to the request-queue as long as it is not filled. The request interface knows the real amount of data in flight and has to handle
	the fact that we want only data in flight that we can receive.
*/

module pointer_requester #(parameter POINTER_REQUEST_TAG = 1) (
	//---------------------- GENERAL VARIABLES---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input wire 										um_done,						// true, if we are done processing (global)
	input wire 										um_activated,					// true, if the afu is activated
	output reg										prq_done,						// true, if the pointer requester is done
	//---------------------- ADDRESS VARIABLES---------------------------//
	input wire [57:0]								read_address,					// the base read address
	input [31:0]									num_ptrs,						// number of pointers we can fetch
	//---------------------- QUEUE VARIABLES---------------------------//
	output reg									   	queue_req_write_enable,				// wheter we want to write to the queue or not
	output reg [65:0] 							   	queue_req_buf_in,					// the queue input buffer (65:8 is the request address, 7:0 the tag we assign to the request)
	input wire 									   	queue_req_buf_filled_next,			// true, if the queue is filled soon

	output reg[511:0]								debug_pointer_requester
);

reg [31:0] prq_num_req;

//---------------------- GENERATE READ REQUESTS ---------------------------//
always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
        // reset of output registers
        queue_req_write_enable <= 0;
		queue_req_buf_in <= 0;
		prq_done <= 0;

		// reset of local variables
		prq_num_req <= 0;
    end
    else begin
		
		queue_req_write_enable <= 0;
		queue_req_buf_in <= 0;
		
		// if we have not generated all requests...
		if (~prq_done) begin
			// if the request queue is not filled in the future...
		    if( ~queue_req_buf_filled_next ) begin
				
				prq_num_req <= prq_num_req + 1;
				queue_req_buf_in <= { {read_address + prq_num_req } , 8'b0} + POINTER_REQUEST_TAG; // address and 1 tag
				queue_req_write_enable <= 1;

				// if this was our last read, say that generating read requests is done
				if( prq_num_req*5 +5 >=  num_ptrs) begin 
					prq_done <= 1;
				end
			end
		end
	end
end

/*
// high, if we do not output data, but we could
reg				do_not_work_valid;
reg [31:0]		do_not_work_cycle_counter;
// high, if we hava data, but cannot output data
reg				can_not_output_valid;
reg [31:0]		can_not_output_cycle_counter;
// high, if we do not have data, but could output
reg				can_not_input_valid;
reg [31:0]		can_not_input_cycle_counter;


always @(posedge clk) begin
	if (~rst_n | um_done | ~um_activated) begin
		do_not_work_valid <= 0;
		do_not_work_cycle_counter <= 0;
		can_not_output_valid <= 0;
		can_not_output_cycle_counter <= 0;
		can_not_input_valid <= 0;
		can_not_input_cycle_counter <= 0;
	end
	else begin
		do_not_work_valid <= 0;
		can_not_output_valid <= 0;
		can_not_input_valid <= 0;

		if( ~queue_req_buf_filled_next && ~queue_req_write_enable) begin
			do_not_work_valid<= 1;
			do_not_work_cycle_counter <= do_not_work_cycle_counter + 1;
		end
	end
end
*/

always @(posedge clk) begin
	debug_pointer_requester <= {480'b0, prq_num_req};
end

endmodule









