
`default_nettype none

/**
	The sorter node start is attached to two BRAM queue's. He actively reads from a queue, and puts the data into its local queue. He then pushes data to the sorter node intermediate, if
	space is available.
*/


module sorter_node_start #(parameter BRAM_QUEUE_LENGTH_BITS = 7, parameter OVERFLOW_PADDING= 8, parameter K_BITS = 3 ) (
	//---------------------- GENERAL VARIABLES---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input 	wire 									um_done,						// true, if we are done processing (global)
	input 	wire 									um_activated,					// true, if the afu is activated
	input 	wire									sorter_done,
	input 	wire 									edge_receiver_done,
	output reg										local_sorter_done,

	//---------------------- VARIABLES FOR THE INPUT EDGE (what we need from the input queue) ---------------------------//
	input wire									in_queue_write_enable_0,
	input wire[95:0] 							in_queue_buf_in_0,	
	// output reg									in_queue_filled_next_0,
	output reg									sorter_edge_consumed_0,

	input wire									in_queue_write_enable_1,
	input wire[95:0] 							in_queue_buf_in_1,	
	// output reg									in_queue_filled_next_1,
	output reg									sorter_edge_consumed_1,

	//---------------------- VARIABLES FOR THE OUTPUT ---------------------------//
	output reg									out_queue_write_enable,
	output reg[95:0]							out_queue_in,
	input wire									out_queue_filled_next,

	output reg[15:0] 									debug_q1_length,
	output reg[15:0] 									debug_q2_length
);


`ifndef SORTER_EDGE_MIN_2

`define SORTER_EDGE_MIN_2(v1, v2) ( (v1[95:64+K_BITS] < v2[95:64+K_BITS] ) ? 0 :  ( ( v1[95:64+K_BITS] == v2[95:64+K_BITS] ) & ( v1[63:32] < v2[63:32] ) ) ? 0 : ( ( v1[95:64+K_BITS] == v2[95:64+K_BITS] ) & ( v1[63:32] == v2[63:32] ) & ( v1[95:64] < v2[95:64] )  ) ? 0 : 1 )

`endif


reg error_flag;

wire q1_m_axis_valid;
reg q1_m_axis_ready;
wire [95:0] q1_m_axis_data;
wire [BRAM_QUEUE_LENGTH_BITS-1:0] q1_data_count;
reg q1_data_buffer_1_valid;
reg q1_data_buffer_2_valid;
reg q1_data_buffer_3_valid;
reg [95:0] q1_data_buffer_1;
reg [95:0] q1_data_buffer_2;
reg [95:0] q1_data_buffer_3;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(96), .PROG_FULL_THRESH(0 ) ) q1_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(in_queue_write_enable_0),
    .s_axis_ready(),
    .s_axis_data(in_queue_buf_in_0),
    .s_axis_last(1'b0),
    .m_axis_valid(q1_m_axis_valid),
    .m_axis_ready(q1_m_axis_ready),
    .m_axis_data(q1_m_axis_data),
    .m_axis_last(),
    .data_count(q1_data_count)
);

wire q2_m_axis_valid;
reg q2_m_axis_ready;
wire [95:0] q2_m_axis_data;
wire [BRAM_QUEUE_LENGTH_BITS-1:0] q2_data_count;
reg q2_data_buffer_1_valid;
reg q2_data_buffer_2_valid;
reg q2_data_buffer_3_valid;

reg [95:0] q2_data_buffer_1;
reg [95:0] q2_data_buffer_2;
reg [95:0] q2_data_buffer_3;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(96), .PROG_FULL_THRESH(0 ) ) q2_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(in_queue_write_enable_1),
    .s_axis_ready(),
    .s_axis_data(in_queue_buf_in_1),
    .s_axis_last(1'b0),
    .m_axis_valid(q2_m_axis_valid),
    .m_axis_ready(q2_m_axis_ready),
    .m_axis_data(q2_m_axis_data),
    .m_axis_last(),
    .data_count(q2_data_count)
);


reg [31:0]	debug_num_edges;
reg [5:0]	debug_case;
always @(posedge clk) begin
    if (~rst_n | um_done | ~um_activated) begin
		out_queue_write_enable <= 0;
		out_queue_in <= 0;

        local_sorter_done <= 0;
		debug_num_edges <= 0;

		sorter_edge_consumed_0 <= 0;
		sorter_edge_consumed_1 <= 0;

		q1_data_buffer_1_valid <= 0;
		q1_data_buffer_2_valid <= 0;
		q1_data_buffer_3_valid <= 0;
		q2_data_buffer_1_valid <= 0;
		q2_data_buffer_2_valid <= 0;
		q2_data_buffer_3_valid <= 0;

		// drain queues
		if( (um_done || ~um_activated) && rst_n ) begin
			q1_m_axis_ready <= 1;
			q2_m_axis_ready <= 1;
		end
		else begin
			q1_m_axis_ready <= 0;
			q2_m_axis_ready <= 0;
		end

		debug_case <= 0;

		debug_q1_length <= 0;
		debug_q2_length <= 0;
		error_flag <= 0;
	end
	else begin
		out_queue_write_enable <= 0;
		sorter_edge_consumed_0 <= 0;
		sorter_edge_consumed_1 <= 0;

		debug_q1_length <= ( q1_data_count + q1_data_buffer_1_valid+ q1_data_buffer_2_valid + q1_data_buffer_3_valid)  ;
		debug_q2_length <= ( q2_data_count + q2_data_buffer_1_valid + q2_data_buffer_2_valid + q2_data_buffer_3_valid) ;

		if( ~sorter_done ) begin

            if( ( q1_data_count + q1_data_buffer_1_valid+ q1_data_buffer_2_valid + q1_data_buffer_3_valid) == 0 &&
				 ( q2_data_count + q2_data_buffer_1_valid + q2_data_buffer_2_valid + q2_data_buffer_3_valid) == 0 && ( edge_receiver_done )  ) begin
                local_sorter_done <= 1;
				debug_case<= 1;
            end
			// if a) we cannot write out or b) a not-done input-queue has 0 input, then we must wait till we get more data: handle the updates
			else if( out_queue_filled_next || 
					(!edge_receiver_done && ( q1_data_count + q1_data_buffer_1_valid+ q1_data_buffer_2_valid + q1_data_buffer_3_valid) == 0 ) ||
					 ( !edge_receiver_done && ( q2_data_count + q2_data_buffer_1_valid + q2_data_buffer_2_valid + q2_data_buffer_3_valid) == 0) )begin
				
				debug_case<= 2;

				if( q1_m_axis_valid && q1_m_axis_ready ) begin
					if( ~q1_data_buffer_1_valid ) begin
						q1_data_buffer_1_valid <= 1;
						q1_data_buffer_1 <= q1_m_axis_data;
					end
					else if( ~q1_data_buffer_2_valid ) begin
						q1_data_buffer_2_valid <= 1;
						q1_data_buffer_2 <= q1_m_axis_data;
					end
					else if( ~q1_data_buffer_3_valid ) begin
						q1_data_buffer_3_valid <= 1;
						q1_data_buffer_3 <= q1_m_axis_data;
						q1_m_axis_ready <= 0;
					end
				end

				if( q2_m_axis_valid && q2_m_axis_ready ) begin
					if( ~q2_data_buffer_1_valid ) begin
						q2_data_buffer_1_valid <= 1;
						q2_data_buffer_1 <= q2_m_axis_data;
					end
					else if( ~q2_data_buffer_2_valid ) begin
						q2_data_buffer_2_valid <= 1;
						q2_data_buffer_2 <= q2_m_axis_data;
					end
					else if( ~q2_data_buffer_3_valid ) begin
						q2_data_buffer_3_valid <= 1;
						q2_data_buffer_3 <= q2_m_axis_data;
						q2_m_axis_ready <= 0;
					end
				end
			end
			// we have data in both: decide what we push next
			else if( ( q1_data_count + q1_data_buffer_1_valid) > 0 && ( q2_data_count + q2_data_buffer_1_valid) > 0  ) begin

				debug_case <= 3;

				if( q1_data_buffer_1_valid && q2_data_buffer_1_valid ) begin
					if(  ( `SORTER_EDGE_MIN_2(q1_data_buffer_1, q2_data_buffer_1)  ) == 0   ) begin
						// output from q1
						out_queue_in <= q1_data_buffer_1;
						out_queue_write_enable <= 1;
						debug_num_edges <= debug_num_edges  + 1;
						sorter_edge_consumed_0<= 1;

						// handle update for q1
						if( q1_data_buffer_2_valid ) begin
							q1_data_buffer_1 <= q1_data_buffer_2;
							if( q1_data_buffer_3_valid ) begin
								q1_data_buffer_2 <= q1_data_buffer_3;
								if(  q1_m_axis_valid && q1_m_axis_ready ) begin
									q1_data_buffer_3 <= q1_m_axis_data;
									q1_m_axis_ready <= 0;
								end
								else begin
									q1_data_buffer_3_valid <= 0;
									q1_m_axis_ready <= 1;
								end
							end
							else begin
								q1_m_axis_ready <= 1;
								if(  q1_m_axis_valid && q1_m_axis_ready ) begin
									q1_data_buffer_2 <= q1_m_axis_data;
								end
								else begin
									q1_data_buffer_2_valid <= 0;
								end
							end
						end
						else begin
							q1_m_axis_ready <= 1;
							if(  q1_m_axis_valid && q1_m_axis_ready ) begin
								q1_data_buffer_1 <= q1_m_axis_data;
							end
							else begin
								q1_data_buffer_1_valid <= 0;
							end
						end


						// handle update for q2
						if( q2_m_axis_valid && q2_m_axis_ready ) begin
							if( ~q2_data_buffer_1_valid ) begin
								q2_data_buffer_1_valid <= 1;
								q2_data_buffer_1 <= q2_m_axis_data;
							end
							else if( ~q2_data_buffer_2_valid ) begin
								q2_data_buffer_2_valid <= 1;
								q2_data_buffer_2 <= q2_m_axis_data;
							end
							else if( ~q2_data_buffer_3_valid ) begin
								q2_data_buffer_3_valid <= 1;
								q2_data_buffer_3 <= q2_m_axis_data;
								q2_m_axis_ready <= 0;
							end
						end
						
						
					end
					else begin
							// output from q2
							out_queue_in <= q2_data_buffer_1;
							out_queue_write_enable <= 1;
							debug_num_edges <= debug_num_edges + 1;
							sorter_edge_consumed_1 <= 1;

							// handle update for q1
							if( q1_m_axis_valid && q1_m_axis_ready ) begin
								if( ~q1_data_buffer_1_valid ) begin
									q1_data_buffer_1_valid <= 1;
									q1_data_buffer_1 <= q1_m_axis_data;
								end
								else if( ~q1_data_buffer_2_valid ) begin
									q1_data_buffer_2_valid <= 1;
									q1_data_buffer_2 <= q1_m_axis_data;
								end
								else if( ~q1_data_buffer_3_valid ) begin
									q1_data_buffer_3_valid <= 1;
									q1_data_buffer_3 <= q1_m_axis_data;
									q1_m_axis_ready <= 0;
								end
							end


							// handle update for q2
							if( q2_data_buffer_2_valid ) begin
								q2_data_buffer_1 <= q2_data_buffer_2;
								if( q2_data_buffer_3_valid ) begin
									q2_data_buffer_2 <= q2_data_buffer_3;
									if(  q2_m_axis_valid && q2_m_axis_ready ) begin
										q2_data_buffer_3 <= q2_m_axis_data;
										q2_m_axis_ready <= 0;
									end
									else begin
										q2_data_buffer_3_valid <= 0;
										q2_m_axis_ready <= 1;
									end
								end
								else begin
									q2_m_axis_ready <= 1;
									if(  q2_m_axis_valid && q2_m_axis_ready ) begin
										q2_data_buffer_2 <= q2_m_axis_data;
									end
									else begin
										q2_data_buffer_2_valid <= 0;
									end
								end
							end
							else begin
								q2_m_axis_ready <= 1;
								if(  q2_m_axis_valid && q2_m_axis_ready ) begin
									q2_data_buffer_1 <= q2_m_axis_data;
								end
								else begin
									q2_data_buffer_1_valid <= 0;
								end
							end


					end

				end
				else begin
					// nothing is written out, since we have not data for both
					//q1_m_axis_ready <= 1;
					//q2_m_axis_ready <= 1;

					if( q1_m_axis_valid && q1_m_axis_ready ) begin
						if( ~q1_data_buffer_1_valid ) begin
							q1_data_buffer_1_valid <= 1;
							q1_data_buffer_1 <= q1_m_axis_data;
						end
						else if( ~q1_data_buffer_2_valid ) begin
							q1_data_buffer_2_valid <= 1;
							q1_data_buffer_2 <= q1_m_axis_data;
						end
						else if( ~q1_data_buffer_3_valid ) begin
							q1_data_buffer_3_valid <= 1;
							q1_data_buffer_3 <= q1_m_axis_data;
							q1_m_axis_ready <= 0;
						end
					end
					else if( ~q1_data_buffer_3_valid || ~q1_data_buffer_2_valid || ~q1_data_buffer_1_valid ) begin
						q1_m_axis_ready <= 1;
					end

					if( q2_m_axis_valid && q2_m_axis_ready ) begin
						if( ~q2_data_buffer_1_valid ) begin
							q2_data_buffer_1_valid <= 1;
							q2_data_buffer_1 <= q2_m_axis_data;
						end
						else if( ~q2_data_buffer_2_valid ) begin
							q2_data_buffer_2_valid <= 1;
							q2_data_buffer_2 <= q2_m_axis_data;
						end
						else if( ~q2_data_buffer_3_valid ) begin
							q2_data_buffer_3_valid <= 1;
							q2_data_buffer_3 <= q2_m_axis_data;
							q2_m_axis_ready <= 0;
						end
					end
					else if( ~q2_data_buffer_3_valid || ~q2_data_buffer_2_valid || ~q2_data_buffer_1_valid ) begin
						q2_m_axis_ready <= 1;
					end


				end
				
			end
			
			// if edge receiver is done, and the other queue has no data, we can simply take from the other one -> note the order of the cases!!
			else if( ( q1_data_count + q1_data_buffer_1_valid) > 0 ) begin // then, we can read from queue 1

						debug_case <= 4;

					if( q1_data_buffer_1_valid) begin

						// output from q1
						out_queue_in <= q1_data_buffer_1;
						out_queue_write_enable <= 1;
						debug_num_edges <= debug_num_edges  + 1;
						sorter_edge_consumed_0<= 1;

						// handle update for q1
						if( q1_data_buffer_2_valid ) begin
							q1_data_buffer_1 <= q1_data_buffer_2;
							if( q1_data_buffer_3_valid ) begin
								q1_data_buffer_2 <= q1_data_buffer_3;
								if(  q1_m_axis_valid && q1_m_axis_ready ) begin
									q1_data_buffer_3 <= q1_m_axis_data;
									q1_m_axis_ready <= 0;
								end
								else begin
									q1_data_buffer_3_valid <= 0;
									q1_m_axis_ready <= 1;
								end
							end
							else begin
								q1_m_axis_ready <= 1;
								if(  q1_m_axis_valid && q1_m_axis_ready ) begin
									q1_data_buffer_2 <= q1_m_axis_data;
								end
								else begin
									q1_data_buffer_2_valid <= 0;
								end
							end
						end
						else begin
							q1_m_axis_ready <= 1;
							if(  q1_m_axis_valid && q1_m_axis_ready ) begin
								q1_data_buffer_1 <= q1_m_axis_data;
							end
							else begin
								q1_data_buffer_1_valid <= 0;
							end
						end

					end
					else begin
						if( q1_m_axis_valid && q1_m_axis_ready ) begin
							if( ~q1_data_buffer_1_valid ) begin
								q1_data_buffer_1_valid <= 1;
								q1_data_buffer_1 <= q1_m_axis_data;
							end
							else if( ~q1_data_buffer_2_valid ) begin
								q1_data_buffer_2_valid <= 1;
								q1_data_buffer_2 <= q1_m_axis_data;
							end
							else if( ~q1_data_buffer_3_valid ) begin
								q1_data_buffer_3_valid <= 1;
								q1_data_buffer_3 <= q1_m_axis_data;
								q1_m_axis_ready <= 0;
							end
						end
						else if( ~q1_data_buffer_3_valid || ~q1_data_buffer_2_valid || ~q1_data_buffer_1_valid ) begin
							q1_m_axis_ready <= 1;
						end
					end
			

 
			end
			else if( ( q2_data_count + q2_data_buffer_1_valid) > 0 ) begin // then, we read from queue 2

					debug_case <= 5;
					if( q2_data_buffer_1_valid ) begin

							// output from q2
							out_queue_in <= q2_data_buffer_1;
							out_queue_write_enable <= 1;
							debug_num_edges <= debug_num_edges + 1;
							sorter_edge_consumed_1 <= 1;

							// handle update for q2
							if( q2_data_buffer_2_valid ) begin
								q2_data_buffer_1 <= q2_data_buffer_2;
								if( q2_data_buffer_3_valid ) begin
									q2_data_buffer_2 <= q2_data_buffer_3;
									if(  q2_m_axis_valid && q2_m_axis_ready ) begin
										q2_data_buffer_3 <= q2_m_axis_data;
										q2_m_axis_ready <= 0;
									end
									else begin
										q2_data_buffer_3_valid <= 0;
										q2_m_axis_ready <= 1;
									end
								end
								else begin
									q2_m_axis_ready <= 1;
									if(  q2_m_axis_valid && q2_m_axis_ready ) begin
										q2_data_buffer_2 <= q2_m_axis_data;
									end
									else begin
										q2_data_buffer_2_valid <= 0;
									end
								end
							end
							else begin
								q2_m_axis_ready <= 1;
								if(  q2_m_axis_valid && q2_m_axis_ready ) begin
									q2_data_buffer_1 <= q2_m_axis_data;
								end
								else begin
									q2_data_buffer_1_valid <= 0;
								end
							end
					end
					else begin

							if( q2_m_axis_valid && q2_m_axis_ready ) begin
								if( ~q2_data_buffer_1_valid ) begin
									q2_data_buffer_1_valid <= 1;
									q2_data_buffer_1 <= q2_m_axis_data;
								end
								else if( ~q2_data_buffer_2_valid ) begin
									q2_data_buffer_2_valid <= 1;
									q2_data_buffer_2 <= q2_m_axis_data;
								end
								else if( ~q2_data_buffer_3_valid ) begin
									q2_data_buffer_3_valid <= 1;
									q2_data_buffer_3 <= q2_m_axis_data;
									q2_m_axis_ready <= 0;
								end
							end
							else if( ~q2_data_buffer_3_valid || ~q2_data_buffer_2_valid || ~q2_data_buffer_1_valid ) begin
								q2_m_axis_ready <= 1;
							end

					end


			end
			else begin
				error_flag <= 1;
			end
			

		end
	end
end




endmodule




`default_nettype wire








