

module state_controller (
    input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input 	wire 									um_activated,	

	output  reg 									um_done,


	// for global end state
	input wire[31:0]								num_epochs,

	input wire										pointer_requester_done,
	input wire										pointer_receiver_done,


	input wire										edge_requester_done,
	input wire										edge_receiver_done,			
	input wire										sorter_done,

	input wire										mcm_bit_requester_done,
	input wire										mcm_bit_receiver_done,
	input wire										bram_mcm_bit_requester_done,
	input wire										bram_mcm_bit_receiver_done,
	input wire										edge_processor_done,
	input wire										write_interface_done,
	input wire										mcm_bit_ack_receiver_done,
	input wire										edge_writer_done,

	input wire										mcm_bit_requester_epoch_done,
	input wire										mcm_bit_receiver_epoch_done,
	input wire										edge_processor_epoch_done,
	input wire										mcm_writer_epoch_done,
	input wire										bram_mcm_bit_requester_epoch_done,
	input wire										bram_mcm_bit_receiver_epoch_done,
	input wire										mcm_bit_ack_receiver_epoch_done,

	// for epoch handling
	output reg[31:0]								global_epoch,
	output reg										global_epoch_change,


	output reg[511:0]								debug_state_controller
);



reg [6:0]	epoch_change_timeout;// since it might take some time till the others notify the change, we have a small timeout of 32 cycles, till we can change the epoch again
reg [63:0] 	clk_counter;

reg [31:0] 	done_timeout;
reg [31:0]	mcm_ack_cycle_count;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		global_epoch <= 0;
		global_epoch_change <= 0;
		um_done <= 0;
		epoch_change_timeout <= 0;
		clk_counter<= 0;

		done_timeout <= 0;
		mcm_ack_cycle_count <= 0;
	end
	else begin
		clk_counter <= clk_counter + 1;
		// um_done <= 0;

		// update the timeout
		if( epoch_change_timeout < 7'd4 ) begin
			epoch_change_timeout <= epoch_change_timeout + 1;
		end
		
		if( ( epoch_change_timeout >= 7'd4) && mcm_bit_requester_epoch_done && mcm_bit_receiver_epoch_done && 
				edge_processor_epoch_done &&
				bram_mcm_bit_requester_epoch_done && bram_mcm_bit_receiver_epoch_done &&
				mcm_bit_ack_receiver_epoch_done && mcm_writer_epoch_done) begin

				global_epoch_change <= 1;
				global_epoch <= global_epoch + 1;
				epoch_change_timeout <= 0;
		end
		else begin
			global_epoch_change <= 0;
		end

		// count the number of cycles we must wait for the ack
		if( mcm_bit_requester_epoch_done && mcm_bit_receiver_epoch_done && edge_processor_epoch_done && bram_mcm_bit_requester_epoch_done && bram_mcm_bit_receiver_epoch_done && mcm_writer_epoch_done) begin
			mcm_ack_cycle_count <= mcm_ack_cycle_count + 1; 
		end
		
		// we do not fully end the last epoch, we go directly to the done state - therefore the +1 at the global epoch
		if( // pointer_requester_done && pointer_receiver_done && // (num_epochs <= global_epoch +1) &&  // since we might not go through all epochs, we also might not fetch all pointers!
			// edge_requester_done && // similarly: the edge requester might want to push the "empty" pointers. But the edge receiver has already gotten all data - so we must not care about the edge requester!
			
			edge_receiver_done && sorter_done && mcm_bit_requester_done && mcm_bit_receiver_done &&
			// bram_mcm_bit_requester_done && bram_mcm_bit_receiver_done &&		// we must not wait till they have finished: they might try to load data for epochs, which do not need to be processed
			edge_processor_done && write_interface_done && mcm_bit_ack_receiver_done && edge_writer_done ) begin
				// done_timeout <= done_timeout + 1;
				um_done <= 1;
		end

		/*
		if( done_timeout > 0 ) begin
			um_done <= 1;
		end
		*/


	end
end


wire[15:0]	pointer_requester_done_16b;
wire[15:0]	pointer_receiver_done_16b;
wire[15:0]	edge_requester_done_16b;
wire[15:0]	edge_receiver_done_16b;
wire[15:0] 	sorter_done_16b;
wire[15:0]	mcm_bit_requester_done_16b;
wire[15:0]	mcm_bit_receiver_done_16b;
wire[15:0]  bram_mcm_bit_requester_done_16b;
wire[15:0] 	bram_mcm_bit_receiver_done_16b;
wire[15:0]	edge_processor_done_16b;
wire[15:0]	write_interface_done_16b;
wire[15:0]	mcm_bit_ack_receiver_done_16b;
wire[15:0]	edge_writer_done_16b;

wire[15:0]	mcm_bit_requester_epoch_done_16b;
wire[15:0]	mcm_bit_receiver_epoch_done_16b;
wire[15:0]	edge_processor_epoch_done_16b;
wire[15:0]	mcm_writer_epoch_done_16b;
wire[15:0]	bram_mcm_bit_requester_epoch_done_16b;
wire[15:0]	bram_mcm_bit_receiver_epoch_done_16b;
wire[15:0]  mcm_bit_ack_receiver_epoch_done_16b;

assign pointer_requester_done_16b = pointer_requester_done;
assign pointer_receiver_done_16b = pointer_receiver_done;
assign edge_requester_done_16b = 0; // edge_requester_done;
assign edge_receiver_done_16b = 0; // edge_receiver_done;
assign sorter_done_16b = 0; // sorter_done;
assign mcm_bit_requester_done_16b = 0; // mcm_bit_requester_done;
assign mcm_bit_receiver_done_16b = 0; // mcm_bit_receiver_done;
assign bram_mcm_bit_requester_done_16b = 0; //bram_mcm_bit_requester_done;
assign bram_mcm_bit_receiver_done_16b = 0;// bram_mcm_bit_receiver_done;
assign edge_processor_done_16b = 0; // edge_processor_done;
assign write_interface_done_16b = write_interface_done;
assign mcm_bit_ack_receiver_done_16b = 0; // mcm_bit_ack_receiver_done;
assign edge_writer_done_16b = 0; //edge_writer_done;

assign mcm_bit_requester_epoch_done_16b= 0; // mcm_bit_requester_epoch_done;
assign mcm_bit_receiver_epoch_done_16b = 0; // mcm_bit_receiver_epoch_done;
assign edge_processor_epoch_done_16b = 0; // edge_processor_epoch_done;
assign mcm_writer_epoch_done_16b = 0; // mcm_writer_epoch_done;
assign bram_mcm_bit_requester_epoch_done_16b = 0; // bram_mcm_bit_requester_epoch_done;
assign bram_mcm_bit_receiver_epoch_done_16b = 0; // bram_mcm_bit_receiver_epoch_done;
assign mcm_bit_ack_receiver_epoch_done_16b = 0; //mcm_bit_ack_receiver_epoch_done;

always @(posedge clk) begin
	debug_state_controller <= { mcm_ack_cycle_count, global_epoch,
								clk_counter,
								pointer_requester_done_16b, pointer_receiver_done_16b, edge_requester_done_16b, edge_receiver_done_16b,
								sorter_done_16b, 16'b0, mcm_bit_requester_done_16b, mcm_bit_receiver_done_16b, 
								bram_mcm_bit_requester_done_16b, bram_mcm_bit_receiver_done_16b, edge_processor_done_16b, write_interface_done_16b,
								mcm_bit_ack_receiver_done_16b, edge_writer_done_16b, 16'b0, 16'b0,
								
								mcm_bit_requester_epoch_done_16b, mcm_bit_receiver_epoch_done_16b, edge_processor_epoch_done_16b, mcm_writer_epoch_done_16b,
								bram_mcm_bit_requester_epoch_done_16b, bram_mcm_bit_receiver_epoch_done_16b, mcm_bit_ack_receiver_epoch_done_16b, 16'b0 };
end


/*

always @(posedge clk) begin
	debug_state_controller <= { 32'b0, 32'b0,
								clk_counter,
								16'b0, 16'b0, 16'b0, 16'b0,
								16'b0, 16'b0, 16'b0, 16'b0, 
								16'b0, 16'b0, 16'b0, 16'b0,
								16'b0, 16'b0, 16'b0, 16'b0,
								
								16'b0, 16'b0, 16'b0, 16'b0,
								16'b0, 16'b0, 16'b0, 16'b0 };
end
*/

endmodule




