
`default_nettype none


module write_interface #(parameter OVERFLOW_PADDING= 8, parameter BRAM_QUEUE_LENGTH_BITS = 5, parameter L_BITS = 3, parameter NUM_WRITES_IN_FLIGHT = 1024,
							parameter  EDGE_WRITE_TAG = 3, parameter MCM_WRITE_TAG = 4 ) (
	//---------------------- GENERAL VARIABLES---------------------------//
	input   wire                                   	clk,							// clock
    input   wire                                   	rst_n,							// Stays low as long as AFU is idle
	input wire 										um_done,						// true, if we are done processing (global)
	input wire 										um_activated,					// true, if the afu is activated

	//---------------------- VARIABLES FOR DEBUG PURPOSE ---------------------------//
	input wire[57:0]								ptr_base_addr,
	input wire[57:0]								edges_base_addr,
	input wire[57:0]								out_edges_base_addr,
	input wire[57:0]								out_mcm_bits_base_addr,
	input wire[57:0]								out_edge_debug_base_addr,
	input wire[57:0]								debug_base_addr,
	input wire[31:0]								num_ptrs,
	input wire[31:0]								num_edges,

	//---------------------- State variables ---------------------------//
	input wire										ack_receiver_done,
	input wire										mcm_bit_writer_done,
	input wire										edge_writer_done,
	output reg										write_interface_done,

	//---------------------- WRITE INTERFACE ---------------------------//
	output  reg  [57:0]                            um_tx_wr_addr,	// address to write result
    output  reg  [7:0]                             um_tx_wr_tag,	// tag to write the result
    output  reg                                    um_tx_wr_valid,	// tell that tag and address are valid
    output  reg  [511:0]                           um_tx_data,		// the data to write
    input   wire                                   um_tx_wr_ready,	// that the framework is read to get data

	input   wire                                   um_rx_wr_valid,
    input   wire [7:0]                             um_rx_wr_tag,
	
	//---------------------- QUEUE VARIABLES TO WRITE EDGES ---------------------------//
	input wire									   	edge_queue_write_enable,
	input wire [569:0] 								edge_queue_buf_in,					// the queue input buffer (65:8 is the request address, 7:0 the tag we assign to the request)
	output reg 									   	edge_queue_buf_filled_next,			// 1, if the queue is filled soon
	
	//---------------------- QUEUE VARIABLES TO WRITE MCM-BITS ---------------------------//
	input wire									   	mcm_queue_write_enable,
	input wire [569:0] 								mcm_queue_buf_in,
	output reg 									   	mcm_queue_buf_filled_next,

	//---------------------- QUEUE VARIABLES FOR EDGES FROM THE SORTER ---------------------------//
	input wire										edge_debug_queue_write_enable,
	input wire [569:0]								edge_debug_queue_buf_in,
	output reg										edge_debug_queue_buf_filled_next,

	//---------------------- QUEUE VARIABLES FOR DEBUG ---------------------------//
	input wire										debug_cl_queue_write_enable,
	input wire[569:0]								debug_cl_queue_buf_in,
	output reg 										debug_cl_queue_buf_filled_next,

	output reg[511:0]								debug_write_interface
);


reg [31:0] num_issued_writes;		// count the number of issued read requests
reg [31:0] num_received_acks;	// count the number of received read responses



reg									mcm_m_axis_ready;
wire								mcm_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	mcm_data_count;
wire [569:0]						mcm_m_axis_data;
reg  [569:0]						mcm_data_buffer;
reg									mcm_data_buffer_valid;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(570), .PROG_FULL_THRESH(0 ) ) mcm_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(mcm_queue_write_enable),
    .s_axis_ready(),
    .s_axis_data(mcm_queue_buf_in),
    .s_axis_last(1'b0),
    .m_axis_valid(mcm_m_axis_valid),
    .m_axis_ready(mcm_m_axis_ready),
    .m_axis_data(mcm_m_axis_data),
    .m_axis_last(),
    .data_count(mcm_data_count)
);




reg									edge_m_axis_ready;
wire								edge_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	edge_data_count;
wire [569:0]						edge_m_axis_data;
reg  [569:0]						edge_data_buffer;
reg									edge_data_buffer_valid;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(570), .PROG_FULL_THRESH(0 ) ) edge_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(edge_queue_write_enable),
    .s_axis_ready(),
    .s_axis_data(edge_queue_buf_in),
    .s_axis_last(1'b0),
    .m_axis_valid(edge_m_axis_valid),
    .m_axis_ready(edge_m_axis_ready),
    .m_axis_data(edge_m_axis_data),
    .m_axis_last(),
    .data_count(edge_data_count)
);


reg									debug_m_axis_ready;
wire								debug_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	debug_data_count;
wire [569:0]						debug_m_axis_data;
reg  [569:0]						debug_data_buffer;
reg									debug_data_buffer_valid;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(570), .PROG_FULL_THRESH(0 ) ) debug_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(1'b0), // debug_cl_queue_write_enable
    .s_axis_ready(),
    .s_axis_data(debug_cl_queue_buf_in),
    .s_axis_last(1'b0),
    .m_axis_valid(debug_m_axis_valid),
    .m_axis_ready(debug_m_axis_ready),
    .m_axis_data(debug_m_axis_data),
    .m_axis_last(),
    .data_count(debug_data_count)
);


reg									edge_debug_m_axis_ready;
wire								edge_debug_m_axis_valid;
wire [BRAM_QUEUE_LENGTH_BITS-1:0]	edge_debug_data_count;
wire [569:0]						edge_debug_m_axis_data;
reg  [569:0]						edge_debug_data_buffer;
reg									edge_debug_data_buffer_valid;

nlb_axififo #(.BUS_SIZE_ADDR(BRAM_QUEUE_LENGTH_BITS), .BUS_SIZE_DATA(570), .PROG_FULL_THRESH(0 ) ) edge_debug_axififo	(
    .clk (clk),
    .rst(~rst_n),
    .s_axis_valid(1'b0), // 1'b0        edge_debug_queue_write_enable        
    .s_axis_ready(),
    .s_axis_data(edge_debug_queue_buf_in),
    .s_axis_last(1'b0),
    .m_axis_valid(edge_debug_m_axis_valid),
    .m_axis_ready(edge_debug_m_axis_ready),
    .m_axis_data(edge_debug_m_axis_data),
    .m_axis_last(),
    .data_count(edge_debug_data_count)
);




//---------------------- Handle and track the filled state ---------------------------//
always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		debug_cl_queue_buf_filled_next <= 0;
		edge_debug_queue_buf_filled_next <= 0;
		edge_queue_buf_filled_next <= 0;
		mcm_queue_buf_filled_next<= 0;
	end
	else begin
		debug_cl_queue_buf_filled_next <= ( debug_data_count + OVERFLOW_PADDING ) > ( 2**BRAM_QUEUE_LENGTH_BITS );
		edge_debug_queue_buf_filled_next <= ( edge_debug_data_count + OVERFLOW_PADDING) > (2**BRAM_QUEUE_LENGTH_BITS);
		edge_queue_buf_filled_next <= ( edge_data_count + OVERFLOW_PADDING) > (2**BRAM_QUEUE_LENGTH_BITS);
		mcm_queue_buf_filled_next <= ( mcm_data_count + OVERFLOW_PADDING) > ( 2**BRAM_QUEUE_LENGTH_BITS );
	end
end

// vars to track the reads and writes on the queue
reg [31:0] edge_ntotal_read;
reg [31:0] edge_ntotal_write;

reg [31:0] mcm_ntotal_read;
reg [31:0] mcm_ntotal_write;

reg [31:0] edge_debug_ntotal_read;
reg [31:0] edge_debug_ntotal_write;

reg [31:0] debug_ntotal_read;
reg [31:0] debug_ntotal_write;


//---------------------- Handle and track writes and reads to the queue for debug ---------------------------//
always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		edge_ntotal_write <= 0;
		mcm_ntotal_write <= 0;
		edge_debug_ntotal_write <= 0;
		debug_ntotal_write <= 0;
	end
	else begin
		if( debug_cl_queue_write_enable ) begin
			debug_ntotal_write <= debug_ntotal_write + 1;
		end
		if( edge_debug_queue_write_enable ) begin
			edge_debug_ntotal_write <= edge_debug_ntotal_write  + 1;
		end

		if( edge_queue_write_enable ) begin
			edge_ntotal_write <= edge_ntotal_write  + 1;
		end

		if( mcm_queue_write_enable ) begin
			mcm_ntotal_write <= mcm_ntotal_write + 1;
		end
	end
end

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		mcm_ntotal_read <= 0;
		edge_ntotal_read<= 0;
		edge_debug_ntotal_read <= 0;
		debug_ntotal_read <= 0;
	end
	else begin
		if( debug_m_axis_valid && debug_m_axis_ready ) begin
			debug_ntotal_read <= debug_ntotal_read + 1;
		end
		if( edge_debug_m_axis_valid && edge_debug_m_axis_ready ) begin
			edge_debug_ntotal_read <= edge_debug_ntotal_read  + 1;
		end
		if(  mcm_m_axis_valid && mcm_m_axis_ready ) begin
			mcm_ntotal_read <= mcm_ntotal_read + 1;
		end
		if( edge_m_axis_valid && edge_m_axis_ready  ) begin
			edge_ntotal_read <= edge_ntotal_read + 1;
		end
	end
end


reg[5:0]		write_case;

//---------------------- Write to RAM ---------------------------//
always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		write_interface_done <= 0;
		um_tx_wr_valid <= 0;
		um_tx_wr_tag <= 0;
		um_tx_data<= 0;
		um_tx_wr_addr <= 0;

		edge_data_buffer_valid <= 0;
		debug_data_buffer_valid <= 0;
		edge_debug_data_buffer_valid <= 0;
		mcm_data_buffer_valid <= 0;

		if( (um_done || ~um_activated) && rst_n ) begin
			edge_m_axis_ready <= 1;
			debug_m_axis_ready <= 1;
			edge_debug_m_axis_ready<= 1;
			mcm_m_axis_ready <= 1;
		end
		else begin
			edge_m_axis_ready <= 0;
			debug_m_axis_ready <= 0;
			edge_debug_m_axis_ready<= 0;
			mcm_m_axis_ready <= 0;
		end
		
		
		write_case <= 0;
	end
	else begin
		um_tx_wr_valid <= 0;

		// check, if we can issue a write
		if( ack_receiver_done && mcm_bit_writer_done && edge_writer_done && mcm_ntotal_write == mcm_ntotal_read && edge_ntotal_write == edge_ntotal_read ) begin
			write_interface_done <= 1;
		end


		//---------------------- Debug Data ---------------------------//
		if( 
			(num_issued_writes <  (num_received_acks + NUM_WRITES_IN_FLIGHT)  && um_tx_wr_ready ) && 
			(  debug_data_buffer_valid || ( debug_m_axis_ready &&  debug_m_axis_valid) )
		 ) begin

			write_case <= 1;

			if( debug_data_buffer_valid ) begin
				um_tx_data <= debug_data_buffer[569:58];
				um_tx_wr_addr <= debug_data_buffer[57:0];
				um_tx_wr_tag <= 0;
				um_tx_wr_valid <= 1;
				
				// sub-case: if we have valid data on the wire, move to the buffer
				if( debug_m_axis_valid && debug_m_axis_ready) begin
					debug_data_buffer <= debug_m_axis_data;
					debug_data_buffer_valid <= 1;
					debug_m_axis_ready<= 0;
				end

				// sub-case: we do not have valid data on the wire, set buffer to invalid
				if( !(debug_m_axis_valid && debug_m_axis_ready) ) begin
					debug_data_buffer_valid <= 0;
					debug_m_axis_ready<= 1;
				end
			end
			else begin
				um_tx_data <= debug_m_axis_data[569:58];
				um_tx_wr_addr <= debug_m_axis_data[57:0];
				um_tx_wr_tag <= 0;
				um_tx_wr_valid <= 1;
				debug_data_buffer_valid <= 0;
				debug_m_axis_ready<= 1;
			end
		end
		else begin
				if( debug_m_axis_valid && debug_m_axis_ready) begin
					debug_data_buffer <= debug_m_axis_data;
					debug_data_buffer_valid <= 1;
					debug_m_axis_ready <= 0;	
				end
				else if( !debug_data_buffer_valid && !debug_m_axis_valid  ) begin
					debug_m_axis_ready <= 1;
				end
		end

		//---------------------- Debug Edges ---------------------------//
		if( !( (num_issued_writes <  (num_received_acks + NUM_WRITES_IN_FLIGHT)  && um_tx_wr_ready ) && 
			( debug_m_axis_valid || ( debug_m_axis_ready &&  debug_data_buffer_valid) )
		 	)
			&&
			( (num_issued_writes <  (num_received_acks + NUM_WRITES_IN_FLIGHT) && um_tx_wr_ready ) && 
				( edge_debug_data_buffer_valid || ( edge_debug_m_axis_valid && edge_debug_m_axis_ready  ) )
			)
			) begin

				write_case <= 2;

				if( edge_debug_data_buffer_valid ) begin
					um_tx_data <= edge_debug_data_buffer[569:58];
					um_tx_wr_addr <= edge_debug_data_buffer[57:0];
					um_tx_wr_tag <= 0;
					um_tx_wr_valid <= 1;

					// sub-case: if we have valid data on the wire, move to the buffer
					if( edge_debug_m_axis_valid && edge_debug_m_axis_ready) begin
						edge_debug_data_buffer <= edge_debug_m_axis_data;
						edge_debug_data_buffer_valid <= 1;
						edge_debug_m_axis_ready <= 0;
					end
					// sub-case: we do not have valid data on the wire, set buffer to invalid
					if( !(edge_debug_m_axis_valid && edge_debug_m_axis_ready) ) begin
						edge_debug_data_buffer_valid <= 0;
						edge_debug_m_axis_ready<= 1;
					end

				end
				else begin
					um_tx_data <= edge_debug_m_axis_data[569:58];
					um_tx_wr_addr <= edge_debug_m_axis_data[57:0];
					um_tx_wr_tag <= 0; // edge_debug_m_axis_data[7:0];
					um_tx_wr_valid <= 1;
					edge_debug_data_buffer_valid <= 0;
					edge_debug_m_axis_ready<= 1;
				end
		end
		else begin
				if( edge_debug_m_axis_valid && edge_debug_m_axis_ready ) begin
					edge_debug_data_buffer <= edge_debug_m_axis_data;
					edge_debug_data_buffer_valid <= 1;
					edge_debug_m_axis_ready <= 0;
				end
				else begin

				end
				if( !edge_debug_data_buffer_valid && !edge_debug_m_axis_valid  ) begin
					edge_debug_m_axis_ready <= 1;
				end
		end


		//---------------------- Writing Edges belonging to a matching ---------------------------//
		if( !( (num_issued_writes <  (num_received_acks + NUM_WRITES_IN_FLIGHT)  && um_tx_wr_ready ) && 
			( debug_m_axis_valid || ( debug_m_axis_ready &&  debug_data_buffer_valid) )
		 	)
			&&
			!( (num_issued_writes <  (num_received_acks + NUM_WRITES_IN_FLIGHT) && um_tx_wr_ready ) && 
				( edge_debug_data_buffer_valid || ( edge_debug_m_axis_valid && edge_debug_m_axis_ready  ) )
			)
			&&
			( (num_issued_writes <  (num_received_acks + NUM_WRITES_IN_FLIGHT) && um_tx_wr_ready ) && 
			  ( edge_data_buffer_valid || ( edge_m_axis_valid && edge_m_axis_ready  ) )
			)
			) begin

				write_case <= 3;

				if( edge_data_buffer_valid ) begin
					um_tx_data <= edge_data_buffer[569:58];
					um_tx_wr_addr <= edge_data_buffer[57:0];
					um_tx_wr_tag <= EDGE_WRITE_TAG;
					um_tx_wr_valid <= 1;

					// sub-case: if we have valid data on the wire, move to the buffer
					if( edge_m_axis_valid && edge_m_axis_ready) begin
						edge_data_buffer <= edge_m_axis_data;
						edge_data_buffer_valid <= 1;
						edge_m_axis_ready <= 0;
					end
					// sub-case: we do not have valid data on the wire, set buffer to invalid
					if( !(edge_m_axis_valid && edge_m_axis_ready) ) begin
						edge_data_buffer_valid <= 0;
						edge_m_axis_ready<= 1;
					end

				end
				else begin
					um_tx_data <= edge_m_axis_data[569:58];
					um_tx_wr_addr <= edge_m_axis_data[57:0];
					um_tx_wr_tag <= EDGE_WRITE_TAG; // edge_m_axis_data[7:0];
					um_tx_wr_valid <= 1;
					edge_data_buffer_valid <= 0;
					edge_m_axis_ready<= 1;
				end
		end
		else begin
				if( edge_m_axis_valid && edge_m_axis_ready ) begin
					edge_data_buffer <= edge_m_axis_data;
					edge_data_buffer_valid <= 1;
					edge_m_axis_ready <= 0;
				end
				else begin

				end
				if( !edge_data_buffer_valid && !edge_m_axis_valid  ) begin
					edge_m_axis_ready <= 1;
				end
		end



		//---------------------- Writing MCM BITS ---------------------------//
		if( !( (num_issued_writes <  (num_received_acks + NUM_WRITES_IN_FLIGHT)  && um_tx_wr_ready ) && 
			( debug_m_axis_valid || ( debug_m_axis_ready &&  debug_data_buffer_valid) )
		 	)
			&&
			!( (num_issued_writes <  (num_received_acks + NUM_WRITES_IN_FLIGHT) && um_tx_wr_ready ) && 
				( edge_debug_data_buffer_valid || ( edge_debug_m_axis_valid && edge_debug_m_axis_ready  ) )
			)
			&&
			!( (num_issued_writes <  (num_received_acks + NUM_WRITES_IN_FLIGHT) && um_tx_wr_ready ) && 
			  ( edge_data_buffer_valid || ( edge_m_axis_valid && edge_m_axis_ready  ) )
			)
			&&
			( (num_issued_writes <  (num_received_acks + NUM_WRITES_IN_FLIGHT) && um_tx_wr_ready ) && 
				( mcm_data_buffer_valid || ( mcm_m_axis_valid && mcm_m_axis_ready  ) )
			)
			) begin

				write_case <= 4;

				if( mcm_data_buffer_valid ) begin
					um_tx_data <= mcm_data_buffer[569:58];
					um_tx_wr_addr <= mcm_data_buffer[57:0];
					um_tx_wr_tag <= MCM_WRITE_TAG; // mcm_data_buffer[7:0];
					um_tx_wr_valid <= 1;

					// sub-case: if we have valid data on the wire, move to the buffer
					if( mcm_m_axis_valid && mcm_m_axis_ready) begin
						mcm_data_buffer <= mcm_m_axis_data;
						mcm_data_buffer_valid <= 1;
						mcm_m_axis_ready <= 0;
					end
					// sub-case: we do not have valid data on the wire, set buffer to invalid
					if( !(mcm_m_axis_valid && mcm_m_axis_ready) ) begin
						mcm_data_buffer_valid <= 0;
						mcm_m_axis_ready<= 1;
					end

				end
				else begin
					um_tx_data <= mcm_m_axis_data[569:58];
					um_tx_wr_addr <= mcm_m_axis_data[57:0];
					um_tx_wr_tag <= MCM_WRITE_TAG; // mcm_m_axis_data[7:0];
					um_tx_wr_valid <= 1;
					mcm_data_buffer_valid <= 0;
					mcm_m_axis_ready<= 1;
				end
		end
		else begin
				if( mcm_m_axis_valid && mcm_m_axis_ready ) begin
					mcm_data_buffer <= mcm_m_axis_data;
					mcm_data_buffer_valid <= 1;
					mcm_m_axis_ready <= 0;
				end
				else begin

				end
				if( !mcm_data_buffer_valid && !mcm_m_axis_valid  ) begin
					mcm_m_axis_ready <= 1;
				end
		end


	end
end

reg [31:0]		debug_address_error32;
reg	[31:0]		debug_case_error;

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		debug_address_error32 <= 0;
		debug_case_error <= 0;
	end
	else begin
		if( um_tx_wr_valid ) begin

			if( um_tx_wr_addr >= edges_base_addr && um_tx_wr_addr <= (edges_base_addr + (num_edges >> 3) ) ) begin
				// then ok
			end
			else if( um_tx_wr_addr >= out_edges_base_addr && um_tx_wr_addr <= (out_edges_base_addr +  ( ( (num_edges  >> 2) +1 ) << L_BITS +1 )      ) ) begin
				// then ok
			end
			else if( um_tx_wr_addr >= out_mcm_bits_base_addr && um_tx_wr_addr<= ( out_mcm_bits_base_addr + ( (num_ptrs << 9) >> L_BITS) + 1 ) ) begin
				// then ok
			end
			else if( um_tx_wr_addr >= out_edge_debug_base_addr && um_tx_wr_addr <= ( out_edge_debug_base_addr  + (num_edges >> 2) ) ) begin
				// then ok
			end
			else if( um_tx_wr_addr >= debug_base_addr &&  um_tx_wr_addr <= (debug_base_addr + 64 ) ) begin
				// then ok
			end
			else begin
				debug_address_error32 <= 1;
				debug_case_error <= write_case;
			end

		end
	end
end

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		num_issued_writes <= 0;
	end
	else begin
		if( um_tx_wr_valid ) begin
			num_issued_writes <= num_issued_writes +1;
		end
	end
end

always @(posedge clk)
begin
	if (~rst_n | um_done | ~um_activated) begin
		num_received_acks <= 0;
	end
	else begin
		if( um_rx_wr_valid ) begin
			num_received_acks <= num_received_acks + 1;
		end
	end
end


always @(posedge clk) begin
	debug_write_interface <= { 128'b0,
								num_received_acks, num_issued_writes,
								debug_case_error, debug_address_error32,
								debug_ntotal_read, debug_ntotal_write,
								edge_ntotal_read, edge_ntotal_write,
								mcm_ntotal_read, mcm_ntotal_write,
								edge_debug_ntotal_read,	edge_debug_ntotal_write };
end

endmodule




`default_nettype wire










