
#ifdef FPGA_FLAG


#include "mwm_crouch_fpga.hpp"

MWMCrouchFPGA::MWMCrouchFPGA(FPGA* fpga, int afu_id)
{
    this->myfpga = fpga;
    this->afu_id = afu_id;
}

MWMCrouchFPGA::~MWMCrouchFPGA()
{
    // empty on purpose
}

void* MWMCrouchFPGA::solve(void* pinput, Timing* timer, RunParams* rp)
{
    uint32_t setup_timer = timer->startTiming();
    WEdgeList* input = (WEdgeList*) pinput;

    uint64_t num_cache_lines;

    // 1. convert to a format that is suitable for the FPGA: in the "cache-line" format
    uint32_t* cache_line_format = Util::convert_to_cache_line_format(input, &num_cache_lines);

    const uint32_t num_in_edges = num_cache_lines *5;
    const uint32_t size_in_buffer_in_ints = num_cache_lines*512/32;
    const uint32_t size_in_buffer_in_bytes = size_in_buffer_in_ints*sizeof(uint32_t);

    const uint32_t num_out_edges =  num_in_edges;
    const uint32_t size_out_buffer_in_ints = num_out_edges*(1+1+1+1);
    const uint32_t size_out_buffer_in_bytes = size_out_buffer_in_ints*sizeof(uint32_t);

#ifndef PERFORMANCE
    std::cout << "afu_id="<<this->afu_id<< std::endl;
    std::cout << "num_cache_lines="<<num_cache_lines<< std::endl;
    std::cout << "num_in_edges="<<num_in_edges << std::endl;
    std::cout << "size_in_buffer_in_ints="<<size_in_buffer_in_ints<< std::endl;
    std::cout << "size_in_buffer_in_bytes="<<size_in_buffer_in_bytes<< std::endl;

    std::cout << "num_out_edges="<< num_out_edges<< std::endl;
    std::cout << "size_out_buffer_in_ints="<<size_out_buffer_in_ints<< std::endl;
    std::cout << "size_out_buffer_in_bytes="<<size_out_buffer_in_bytes<< std::endl;
#endif

    // allocate the buffer on the fpga
    uint32_t* data_in   = reinterpret_cast<uint32_t*>( myfpga->malloc(size_in_buffer_in_bytes ));
    uint32_t* data_out  = reinterpret_cast<uint32_t*>( myfpga->malloc(size_out_buffer_in_bytes + 8*512/8));

    for(uint32_t i = 0; i < size_in_buffer_in_ints; i++)
    {
        data_in[i] = cache_line_format[i];
        data_out[i] = 0;
    }

    free(cache_line_format);

    // 2. create and call the function
    MWM_AFU_CONFIG * afu_cfg = (struct MWM_AFU_CONFIG*)(myfpga->malloc( sizeof(MWM_AFU_CONFIG) ));

    afu_cfg->pSource = data_in;
    afu_cfg->pDest   = data_out;
    afu_cfg->num_cl  = num_cache_lines;

#ifndef PERFORMANCE
    printf("afu cfg struct: %p, data_in: %p, data_out: %p\n", afu_cfg, data_in, data_out); fflush(stdout);
    printf("Create Op\n"); fflush(stdout);
    auto start_time = std::chrono::high_resolution_clock::now();
#endif

    // destruction of record is done by Fthread
    auto mwmThreadRecOp =  new FthreadRec(myfpga, this->afu_id, reinterpret_cast<unsigned char*>(afu_cfg), sizeof(MWM_AFU_CONFIG) );
    timer->endTimingSetup(setup_timer);
    uint32_t processing_timer =  timer->startTiming();
    Fthread mwmThreadOp(mwmThreadRecOp);
    mwmThreadOp.join();
    timer->endTimingMainProcessing(processing_timer);
    timer->setULong("fpga_processing_cycles", mwmThreadOp.FRecord->get_status()->exec_cycles, false);
    timer->setULong("fpga_processing_reads", mwmThreadOp.FRecord->get_status()->reads, false);
    timer->setULong("fpga_processing_writes", mwmThreadOp.FRecord->get_status()->writes, false);

#ifndef PERFORMANCE
    //-----------------------------------------------------------------------------------------------------------------------------------//
    auto   end_time    = std::chrono::high_resolution_clock::now();
    double execTime    = (std::chrono::duration_cast<std::chrono::microseconds>(end_time - start_time).count())/1000.0;
    double opLocalTime = mwmThreadOp.timing();
    double throughput  = 1000.0*num_cache_lines/execTime;



    printf("Process done: exec time: %.10f, %.10f, %.10f\n", execTime, opLocalTime, throughput);  fflush(stdout);

    for(uint32_t i = 0; i < size_out_buffer_in_ints; i++)
    {
        printf("%d ", data_out[i]);
    }
    printf("\n");
#endif

    uint32_t merge_timing = timer->startTiming();
    WEdgeList* result = (WEdgeList*) malloc(sizeof(WEdgeList));
    result->wedge = new std::vector<WEdge>();
    auto bv = new BitVector(input->n + 10);

    // greedily create the matching, starting with the highest level

    for(int32_t l = rp->num_levels; l >= 0; l--)
    {
        uint32_t zero_count = 0;
        for(uint32_t i = 0; i < num_out_edges; i++)
        {
            uint32_t eu = data_out[i*4];
            uint32_t ev = data_out[i*4+1];
            uint32_t ew = data_out[i*4+2];
            uint32_t emax_pos = data_out[i*4+3];

            // early break
            if( eu == 0 && ev == 0)
            {
                zero_count = zero_count + 1;
            }

            if( zero_count > 8)
            {
                break;
            }

            if(emax_pos == l && eu != 0 && ev != 0 && bv->get(eu) == 0 && bv->get(ev) == 0 && eu <= input->n && ev <= input->n) // w.weight == l
            {
                WEdge we; we.u = eu; we.v = ev; we.weight = ew;
                bv->set_true(eu);
                bv->set_true(ev);
                result->wedge->push_back(we);
 #ifndef PERFORMANCE
                std::cout << "e.u="<<eu<<", e.v="<<ev<<", e.weight="<<ew<<", e.max_pos="<<emax_pos<<std::endl;
 #endif
            }
        }
    }
    timer->setULong("result_size", result->wedge->size(), false);
    timer->endTimingMergeLoop(merge_timing);
    uint32_t timer_teardown = timer->startTiming();

    result->m = result->wedge->size();
    result->n = input->n;

    delete(bv);

#ifndef PERFORMANCE
     mwmThreadOp.printStatusLine();
#endif

    myfpga->free(data_in);
    myfpga->free(data_out);

    timer->endTimingTeardown(timer_teardown);

    return result;
}




double MWMCrouchFPGA::get_guarantee(RunParams* rp)
{
    return 4.0f + rp->epsilon;
}

std::string MWMCrouchFPGA::get_solver_name()
{
    return "mwmw_crouch_fpga";
}

double MWMCrouchFPGA::get_min_epsilon()
{
    return 0.8f;
}

double MWMCrouchFPGA::get_min_delta()
{
    return 0.0f;
}


#endif
