
#ifdef FPGA_FLAG


#include "microbench.hpp"

MicroBench::MicroBench()
{
    this->myfpga = new FPGA();
    this->afu_id = MICROBENCH_OP;
}

MicroBench::~MicroBench()
{
    if(this->myfpga != nullptr)
    {
        delete(this->myfpga);
        this->myfpga = nullptr;
    }
}

void MicroBench::run(uint32_t num_cache_lines, uint32_t nitems_latency, uint32_t timeout, Latency *l )
{
    uint64_t size_of_buffer_in_bytes = (uint64_t)num_cache_lines * 512/(uint64_t)8 + 1024;
    uint64_t size_of_buffer_in_longs = size_of_buffer_in_bytes /8;

    std::cout << "afu_id="<<this->afu_id<< std::endl;
    std::cout << "num_cache_lines="<<num_cache_lines<< std::endl;
    std::cout << "size_of_buffer_in_bytes="<<size_of_buffer_in_bytes << std::endl;
    std::cout << "nitems_latency="<< nitems_latency << std::endl;
    std::cout << "timeout="<<timeout<< std::endl;

    // allocate the buffer on the fpga
    uint64_t* data_in   = reinterpret_cast<uint64_t*>( myfpga->malloc(size_of_buffer_in_bytes ));
    uint64_t* data_out  = reinterpret_cast<uint64_t*>( myfpga->malloc(size_of_buffer_in_bytes));

    for(uint32_t i = 0; i < size_of_buffer_in_longs; i++)
    {
        data_in[i] = rand();
        data_out[i] = 0;
    }


    // create and call the function
    MICROBENCH_AFU_CONFIG * afu_cfg = (struct MICROBENCH_AFU_CONFIG*)(myfpga->malloc( sizeof(MICROBENCH_AFU_CONFIG) ));

    afu_cfg->pSource = data_in;
    afu_cfg->pDest   = data_out;
    afu_cfg->num_cl  = num_cache_lines;
    afu_cfg->nitems_latency = nitems_latency;
    afu_cfg->timeout = timeout;

    printf("afu cfg struct: %p, data_in: %p, data_out: %p\n", afu_cfg, data_in, data_out); fflush(stdout);

    puts("Create Op"); fflush(stdout);
    auto start_time = std::chrono::high_resolution_clock::now();

    // destruction of record is done by Fthread
    auto mwmThreadRecOp =  new FthreadRec(myfpga, this->afu_id, reinterpret_cast<unsigned char*>(afu_cfg), sizeof(MICROBENCH_AFU_CONFIG) );
    Fthread mwmThreadOp(mwmThreadRecOp);

    int sleeps = 0;
    unsigned char state = 0;
    do
    {
        state = mwmThreadRecOp->get_status()->state;

        SleepNano( 500*1000*1000 ); // sleep for 0.5 sec
        sleeps++;
        if (sleeps >= 10*2) { // 600sec = 10 min
            puts("Killing AFU");
            myfpga->f_platform->srvHndle->m_pALIMMIOService->mmioWrite64(8 * (32 + 6), AAL::btUnsigned64bitInt(0));
            SleepMilli(1000);
            break;
        }
    }while(state != OPERATOR_DONE_STATE);

    mwmThreadOp.printStatusLine();


    //-----------------------------------------------------------------------------------------------------------------------------------//
    auto   end_time    = std::chrono::high_resolution_clock::now();
    double execTime    = (std::chrono::duration_cast<std::chrono::microseconds>(end_time - start_time).count())/1000.0;
    double opLocalTime = mwmThreadOp.timing();
    double throughput  = 1000.0*num_cache_lines/execTime;


    printf("Process done: exec time: %.10f, %.10f, %.10f\n", execTime, opLocalTime, throughput);  fflush(stdout);

    for(uint32_t i = 0; i < 16; i++)
    {
        std::cout << data_out[i] << ", ";
    }
    std::cout << std::endl;

    uint32_t tput_cycles_read = data_out[0];
    uint32_t tput_cl_read = data_out[1];
    uint32_t tput_cycles_write = data_out[2];
    uint32_t tput_cl_write = data_out[3];
    uint32_t tput_cycles_read_write = data_out[4];
    uint32_t tput_cl_read_write = data_out[5];
    l->tput_cycles_read.push_back(tput_cycles_read);
    l->tput_cl_read.push_back(tput_cl_read);
    l->tput_cycles_write.push_back(tput_cycles_write);
    l->tput_cl_write.push_back(tput_cl_write);
    l->tput_cycles_read_write.push_back(tput_cycles_read_write);
    l->tput_cl_read_write.push_back(tput_cl_read_write);

    uint32_t latency_cycles_read = data_out[8];
    uint32_t latency_cl_read = data_out[9];
    uint32_t latency_cycles_write = data_out[10];
    uint32_t latency_cl_write = data_out[11];
    uint32_t latency_cycles_read_write = data_out[12];
    uint32_t latency_cl_read_write = data_out[13];
    l->latency_cycles_read.push_back(latency_cycles_read);
    l->latency_cl_read.push_back(latency_cl_read);
    l->latency_cycles_write.push_back(latency_cycles_write);
    l->latency_cl_write.push_back(latency_cl_write);
    l->latency_cycles_read_write.push_back(latency_cycles_read_write);
    l->latency_cl_read_write.push_back(latency_cl_read_write);

    uint32_t tput_rand_cycles_read = data_out[16];
    uint32_t tput_rand_cl_read = data_out[17];
    uint32_t tput_rand_cycles_write = data_out[18];
    uint32_t tput_rand_cl_write = data_out[19];
    uint32_t tput_rand_cycles_read_write = data_out[20];
    uint32_t tput_rand_cl_read_write = data_out[21];
    l->tput_rand_cycles_read.push_back(tput_rand_cycles_read);
    l->tput_rand_cl_read.push_back(tput_rand_cl_read);
    l->tput_rand_cycles_write.push_back(tput_rand_cycles_write);
    l->tput_rand_cl_write.push_back(tput_rand_cl_write);
    l->tput_rand_cycles_read_write.push_back(tput_rand_cycles_read_write);
    l->tput_rand_cl_read_write.push_back(tput_rand_cl_read_write);

    uint32_t latency_rand_cycles_read = data_out[24];
    uint32_t latency_rand_cl_read = data_out[25];
    uint32_t latency_rand_cycles_write = data_out[26];
    uint32_t latency_rand_cl_write = data_out[27];
    uint32_t latency_rand_cycles_read_write = data_out[28];
    uint32_t latency_rand_cl_read_write = data_out[29];
    l->latency_rand_cycles_read.push_back(latency_rand_cycles_read);
    l->latency_rand_cl_read.push_back(latency_rand_cl_read);
    l->latency_rand_cycles_write.push_back(latency_rand_cycles_write);
    l->latency_rand_cl_write.push_back(latency_rand_cl_write);
    l->latency_rand_cycles_read_write.push_back(latency_rand_cycles_read_write);
    l->latency_rand_cl_read_write.push_back(latency_rand_cl_read_write);

    std::cout << "throughput sequential"<< std::endl;
    std::cout << "tput_cycles_read="<<tput_cycles_read <<", tput_cl_read="<<tput_cl_read << ", read bytes/cycle="<< ( ( (double)tput_cl_read*512/8)/(double)tput_cycles_read)<< std::endl;
    std::cout << "tput_cycles_write="<< tput_cycles_write <<", tput_cl_write="<< tput_cl_write <<", write bytes/cycle="<< ( ( (double) tput_cl_write *512/8 )/ (double) tput_cycles_write ) << std::endl;
    std::cout << "tput_cycles_read_write="<< tput_cycles_read_write << ", tput_cl_read_write="<< tput_cl_read_write <<", read/write bytes/cycle="<< ( ( (double) tput_cl_read_write * (double)512/8  ) / (double) tput_cycles_read_write) << std::endl;

    std::cout << "latency sequential"<< std::endl;
    std::cout << "latency_cycles_read="<< latency_cycles_read <<", latency_cl_read="<<latency_cl_read<<", avg. read latency (cycles)=" << (double) latency_cycles_read / (double) latency_cl_read<< std::endl;
    std::cout << "latency_cycles_write="<< latency_cycles_write<< ", latency_cl_write="<< latency_cl_write << ", avg. write latency (cycles)="<< (double) latency_cycles_write/ (double) latency_cl_write << std::endl;
    std::cout << "latency_cycles_read_write="<< latency_cycles_read_write <<", latency_cl_read_write="<< latency_cl_read_write<<", avg. read/write latency (cycles)="<< (double) latency_cycles_read_write / (double)latency_cl_read_write << std::endl;

    std::cout << "throughput random"<< std::endl;
    std::cout << "tput_rand_cycles_read="<<tput_rand_cycles_read <<", tput_rand_cl_read="<<tput_rand_cl_read << ", rand read bytes/cycle="<< (((double)tput_rand_cl_read*512/8)/(double)tput_rand_cycles_read)<< std::endl;
    std::cout << "tput_rand_cycles_write="<< tput_rand_cycles_write <<", tput_rand_cl_write="<< tput_rand_cl_write <<", rand write bytes/cycle="<< ( ( (double) tput_rand_cl_write *512/8 )/ (double) tput_rand_cycles_write ) << std::endl;
    std::cout << "tput_rand_cycles_read_write="<< tput_rand_cycles_read_write << ", tput_rand_cl_read_write="<< tput_rand_cl_read_write <<", rand read/write bytes/cycle="<< (( (double) tput_rand_cl_read_write * 512/8  ) / (double) tput_rand_cycles_read_write) << std::endl;

    std::cout << "latency random"<< std::endl;
    std::cout << "latency_rand_cycles_read="<< latency_rand_cycles_read <<", latency_rand_cl_read="<<latency_rand_cl_read<<", avg. read latency (cycles)=" << (double) latency_rand_cycles_read / (double) latency_rand_cl_read<< std::endl;
    std::cout << "latency_rand_cycles_write="<< latency_rand_cycles_write<< ", latency_rand_cl_write="<< latency_rand_cl_write << ", avg. write latency (cycles)="<< (double) latency_rand_cycles_write/ (double) latency_rand_cl_write << std::endl;
    std::cout << "latency_rand_cycles_read_write="<< latency_rand_cycles_read_write <<", latency_rand_cl_read_write="<< latency_rand_cl_read_write<<", avg. read/write latency (cycles)="<< (double) latency_rand_cycles_read_write / (double)latency_rand_cl_read_write << std::endl;


    mwmThreadOp.printStatusLine();

    myfpga->free(data_in);
    myfpga->free(data_out);

    return;
}

void MicroBench::compute_results(Latency *l)
{
    std::cout <<"-----OVERALL RESULTS-----"<< std::endl;

    // PERFORMANCE SEQUENTIAL
    {
        double tput_cycles_read_mean;
        double tput_cycles_read_std;
        compute_mean_std_cycles(l->tput_cycles_read, &tput_cycles_read_mean, &tput_cycles_read_std);
        double read_bytes_per_cycle_mean;
        double read_bytes_per_cycle_std;
        compute_mean_std_bytes_per_cycles(l->tput_cycles_read, l->tput_cl_read.at(0)*512/8, &read_bytes_per_cycle_mean, &read_bytes_per_cycle_std);

        std::cout << "throughput sequential read"<< std::endl;
        std::cout << "tput_cycles_read_mean="<<tput_cycles_read_mean<< " ("<< tput_cycles_read_std<< ")";
        std::cout <<", tput_cl_read="<< l->tput_cl_read.at(0);
        std::cout << ", read bytes/cycle="<< read_bytes_per_cycle_mean << "(" << read_bytes_per_cycle_std << ")" << std::endl;
    }
    {
        double tput_cycles_write_mean;
        double tput_cycles_write_std;
        compute_mean_std_cycles(l->tput_cycles_write, &tput_cycles_write_mean, &tput_cycles_write_std);
        double write_bytes_per_cycle_mean;
        double write_bytes_per_cycle_std;
        compute_mean_std_bytes_per_cycles(l->tput_cycles_write, l->tput_cl_write.at(0)*512/8, &write_bytes_per_cycle_mean, &write_bytes_per_cycle_std);
        std::cout << "throughput sequential write"<< std::endl;
        std::cout << "tput_cycles_write_mean="<< tput_cycles_write_mean << " ("<< tput_cycles_write_std<< ")";
        std::cout <<", tput_cl_write="<< l->tput_cl_write.at(0);
        std::cout <<", write bytes/cycle="<< write_bytes_per_cycle_mean << " ("<< write_bytes_per_cycle_std << ")"<< std::endl;
    }
    {
        double tput_cycles_read_write_mean;
        double tput_cycles_read_write_std;
        std::cout << "throughput sequential read/write"<< std::endl;
        compute_mean_std_cycles(l->tput_cycles_read_write, &tput_cycles_read_write_mean, &tput_cycles_read_write_std);
        double read_write_bytes_per_cycle_mean;
        double read_write_bytes_per_cycle_std;
        compute_mean_std_bytes_per_cycles(l->tput_cycles_read_write, l->tput_cl_read_write.at(0)*512/8, &read_write_bytes_per_cycle_mean, &read_write_bytes_per_cycle_std);
        std::cout << "tput_cycles_read_write="<< tput_cycles_read_write_mean << " ("<< tput_cycles_read_write_std << ")";
        std::cout << ", tput_cl_read_write="<< l->tput_cl_read_write.at(0);
        std::cout <<", read/write bytes/cycle="<< read_write_bytes_per_cycle_mean<< " ("<< read_write_bytes_per_cycle_std <<")"  << std::endl;
    }


    // LATENCY SEQUENTIAL
    {
        double latency_cycles_read_mean;
        double latency_cycles_read_std;
        std::cout << "latency sequential read"<< std::endl;
        compute_mean_std_latency(l->latency_cycles_read, l->latency_cl_read.at(0), &latency_cycles_read_mean, &latency_cycles_read_std);
        std::cout << "latency_cycles_read="<< latency_cycles_read_mean << " (" << latency_cycles_read_std << ")" << std::endl;
    }
    {
        double latency_cycles_write_mean;
        double latency_cycles_write_std;
        std::cout << "latency sequential write"<< std::endl;
        compute_mean_std_latency(l->latency_cycles_write, l->latency_cl_write.at(0), &latency_cycles_write_mean, &latency_cycles_write_std);
        std::cout << "latency_cycles_write="<< latency_cycles_write_mean << " (" << latency_cycles_write_std << ")" << std::endl;
    }
    {
        double latency_cycles_read_write_mean;
        double latency_cycles_read_write_std;
        std::cout << "latency sequential write"<< std::endl;
        compute_mean_std_latency(l->latency_cycles_read_write, l->latency_cl_read_write.at(0), &latency_cycles_read_write_mean, &latency_cycles_read_write_std);
        std::cout << "latency_cycles_read_write="<< latency_cycles_read_write_mean << " (" << latency_cycles_read_write_std << ")" << std::endl;
    }

    // PERFORMANCE RANDOM
    {
        double tput_cycles_read_mean;
        double tput_cycles_read_std;
        compute_mean_std_cycles(l->tput_rand_cycles_read, &tput_cycles_read_mean, &tput_cycles_read_std);
        double read_bytes_per_cycle_mean;
        double read_bytes_per_cycle_std;
        compute_mean_std_bytes_per_cycles(l->tput_rand_cycles_read, l->tput_rand_cl_read.at(0)*512/8, &read_bytes_per_cycle_mean, &read_bytes_per_cycle_std);

        std::cout << "throughput random read"<< std::endl;
        std::cout << "tput_rand_cycles_read_mean="<<tput_cycles_read_mean<< " ("<< tput_cycles_read_std<< ")";
        std::cout <<", tput_rand_cl_read="<< l->tput_rand_cl_read.at(0);
        std::cout << ", read bytes/cycle="<< read_bytes_per_cycle_mean << "(" << read_bytes_per_cycle_std << ")" << std::endl;
    }
    {
        double tput_cycles_write_mean;
        double tput_cycles_write_std;
        compute_mean_std_cycles(l->tput_rand_cycles_write, &tput_cycles_write_mean, &tput_cycles_write_std);
        double write_bytes_per_cycle_mean;
        double write_bytes_per_cycle_std;
        compute_mean_std_bytes_per_cycles(l->tput_rand_cycles_write, l->tput_rand_cl_write.at(0)*512/8, &write_bytes_per_cycle_mean, &write_bytes_per_cycle_std);
        std::cout << "throughput rand write"<< std::endl;
        std::cout << "tput_rand_cycles_write_mean="<< tput_cycles_write_mean << " ("<< tput_cycles_write_std<< ")";
        std::cout <<", tput_rand_cl_write="<< l->tput_rand_cl_write.at(0);
        std::cout <<", write bytes/cycle="<< write_bytes_per_cycle_mean << " ("<< write_bytes_per_cycle_std << ")"<< std::endl;
    }
    {
        double tput_cycles_read_write_mean;
        double tput_cycles_read_write_std;
        std::cout << "throughput rand read/write"<< std::endl;
        compute_mean_std_cycles(l->tput_rand_cycles_read_write, &tput_cycles_read_write_mean, &tput_cycles_read_write_std);
        double read_write_bytes_per_cycle_mean;
        double read_write_bytes_per_cycle_std;
        compute_mean_std_bytes_per_cycles(l->tput_rand_cycles_read_write, l->tput_rand_cl_read_write.at(0)*512/8, &read_write_bytes_per_cycle_mean, &read_write_bytes_per_cycle_std);
        std::cout << "tput_rand_cycles_read_write="<< tput_cycles_read_write_mean << " ("<< tput_cycles_read_write_std << ")";
        std::cout << ", tput_rand_cl_read_write="<< l->tput_rand_cl_read_write.at(0);
        std::cout <<", read/write bytes/cycle="<< read_write_bytes_per_cycle_mean<< " ("<< read_write_bytes_per_cycle_std <<")"  << std::endl;
    }

    // LATENCY RANDOM
    {
        double latency_cycles_read_mean;
        double latency_cycles_read_std;
        std::cout << "latency rand read"<< std::endl;
        compute_mean_std_latency(l->latency_rand_cycles_read, l->latency_rand_cl_read.at(0), &latency_cycles_read_mean, &latency_cycles_read_std);
        std::cout << "latency_rand_cycles_read="<< latency_cycles_read_mean << " (" << latency_cycles_read_std << ")" << std::endl;
    }
    {
        double latency_cycles_write_mean;
        double latency_cycles_write_std;
        std::cout << "latency rand write"<< std::endl;
        compute_mean_std_latency(l->latency_rand_cycles_write, l->latency_rand_cl_write.at(0), &latency_cycles_write_mean, &latency_cycles_write_std);
        std::cout << "latency_rand_cycles_write="<< latency_cycles_write_mean << " (" << latency_cycles_write_std << ")" << std::endl;
    }
    {
        double latency_cycles_read_write_mean;
        double latency_cycles_read_write_std;
        std::cout << "latency rand write"<< std::endl;
        compute_mean_std_latency(l->latency_rand_cycles_read_write, l->latency_rand_cl_read_write.at(0), &latency_cycles_read_write_mean, &latency_cycles_read_write_std);
        std::cout << "latency_rand_cycles_read_write="<< latency_cycles_read_write_mean << " (" << latency_cycles_read_write_std << ")" << std::endl;
    }
}

void MicroBench::compute_mean_std(std::vector<double> values, double* mean, double* std)
{
    double m = 0;
    for(int i = 0; i < values.size(); i++)
    {
        m = m + values[i];
    }

    m = m /  values.size();
    *mean = m;

    double d = 0;
    for(int i = 0; i < values.size(); i++)
    {
        d = ( values[i] -m )*(values[i] -m );
    }

    d = sqrt(d / (values.size() -1));
    *std = d;
}

void MicroBench::compute_mean_std_bytes_per_cycles(std::vector<uint32_t> values, uint32_t bytes, double* mean, double* std)
{
    std::vector<double> dvalues;
    for(int i = 0; i < values.size(); i++)
    {
        dvalues.push_back((double)bytes / values[i] );
    }
    double dmean;
    double dstd;
    compute_mean_std(dvalues, &dmean, &dstd);
    *mean = dmean;
    *std = dstd;
}

void MicroBench::compute_mean_std_latency(std::vector<uint32_t> values, uint32_t trials, double* mean, double* std)
{
    std::vector<double> dvalues;
    for(int i = 0; i < values.size(); i++)
    {
        dvalues.push_back( values[i] / trials );
    }
    double dmean;
    double dstd;
    compute_mean_std(dvalues, &dmean, &dstd);
    *mean = dmean;
    *std = dstd;
}


void MicroBench::compute_mean_std_cycles(std::vector<uint32_t> values, double* mean, double* std)
{
    std::vector<double> dvalues;
    for(int i = 0; i < values.size(); i++)
    {
        dvalues.push_back(values[i]);
    }
    double dmean;
    double dstd;
    compute_mean_std(dvalues, &dmean, &dstd);
    *mean = dmean;
    *std = dstd;
}


#endif
