

#include "mwmghaffariopenmp.hpp"



MWMGhaffariOpenMP::MWMGhaffariOpenMP()
{
    // empty on purpose
}

MWMGhaffariOpenMP::~MWMGhaffariOpenMP()
{
    // empty on purpose
}


void* MWMGhaffariOpenMP::solve(void* pinput, Timing* timer, RunParams* rp)
{

    WEdgeList* input = (WEdgeList*) pinput;
    double epsilon = rp->epsilon;
    int nthreads = rp->nthreads;

    // setup datastructure
    const int n_threads = nthreads;
    int n_regions = 2* n_threads;
    if(nthreads == 1)
    {
        n_regions = 1;
    }

    const double beta = ( 5* log( 1 / epsilon) ) / epsilon;
    const uint32_t n = input->n;
    const uint32_t m = input->m;

    // the stack, as list to allow easy arbitrary removal of items (edges)
    std::list<WEdge> s;
    std::unordered_map<WEdge, std::list<WEdge>::iterator, WEdgeHash > iterator_map; // hashmap to keep track of where edges are stored

    std::vector<std::queue<WEdge>> q(n+1); // per vertex a queue
    std::vector<uint32_t> phi(n+1, 0); // the phi variable, n+1 due to 1-based indexing of vertex id

    std::vector<std::pair<int, int>>* run_pairs = this->generate_run_pairs(n_threads);

    // for all edges
    for(uint32_t i = 0; i < m; )
    {
        // run n edges in parallel, determine the regions they belong to
        uint32_t slice = m;

        uint32_t remaining = m - i;
        if(slice > remaining)
        {
            slice = remaining;
        }

        // create for each thread its own datastructure to hold the parts
        std::vector<WEdge> buckets[n_threads][n_regions][n_regions];

#pragma omp parallel num_threads(n_threads) default(none) shared(buckets, input, i, slice, n_regions)
        {
            int tid = omp_get_thread_num();

#pragma omp for
            for(uint32_t k = i; k < i + slice; k++)
            {
                WEdge e = input->wedge->at(k);
                int region1 = (e.u % n_regions);
                int region2 = (e.v % n_regions);
                if(region2 < region1)
                {
                    int t = region1;
                    region1 = region2;
                    region2 = t;
                }
                buckets[tid][region1][region2].push_back(e);
            }
        }


        // run the regions in parallel, according to the run-pairs
        for(int k = 0; k < run_pairs->size(); k+= n_threads )
        {
#pragma omp parallel num_threads(n_threads) shared(phi, q, iterator_map, run_pairs, buckets)
            {
                int tid = omp_get_thread_num();
                std::pair<int, int> p = run_pairs->at(k + tid);

                for(int tiditer = 0; tiditer < n_threads; tiditer++)
                {
                    std::vector<WEdge> edges = buckets[tiditer][p.first][p.second];
                    for(int kk = 0; kk < edges.size(); kk++)
                    {
                        // get all information
                        const WEdge e = edges[kk];

                        const uint32_t u = e.u;
                        const uint32_t uu = e.v;
                        const uint32_t weight = e.weight;

                        // test if we can drop or not
                        if(  weight <= (1 + epsilon) * (  phi[u] + phi[uu]  )   )
                        {
                            continue;
                        }

                        // calculate remaining values and update phi
                        uint32_t wremaining = weight - ( phi[u] + phi[uu] );
                        phi[u] = phi[u] + wremaining;
                        phi[uu] = phi[uu] + wremaining;

                        // puhs on stack and store, where the edge is stored in the list
#pragma omp critical(dataupdate)
                        {
                            s.push_front(e);
                            iterator_map.emplace(e, s.begin());
                        }

                        uint32_t varr[2] = { uu, u};
                        for(uint32_t j = 0; j < 2; j++)
                        {
                            // push on the queue
                            uint32_t v = varr[j];
                            q[v].push(e);

                            if( q[v].size() >= beta)
                            {
                                // get element from the queue and remove
                                WEdge ee = q[v].front();
                                q[v].pop();

#pragma omp critical(dataupdate)
                                {
                                    // search in the stack (the list) and remove the edge there
                                    auto got = iterator_map.find(ee);
                                    if( got != iterator_map.end()) // should always be true!
                                    {
                                        std::list<WEdge>::iterator it = got->second;
                                        s.erase(it);
                                        iterator_map.erase(got);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        i = i + slice;
    }

    delete(run_pairs);


    // preamble with constructing the matching

    // iterate over the list and process the edges, such that we still have a matching
    std::vector<WEdge>* M = new std::vector<WEdge>();
    std::vector<bool> vertex_set(input->n+1, 0);
    uint32_t result_n = 0;
    while( !s.empty() )
    {
        WEdge e = s.front();
        s.pop_front();
        // add only to M, if we have a matching
        if( vertex_set[e.u] == 0 && vertex_set[e.v] == 0)
        {
            vertex_set[e.u] = 1;
            vertex_set[e.v] = 1;
            result_n = result_n+2;
            M->push_back(e);
        }
    }

    WEdgeList* result = (WEdgeList*) malloc(sizeof(WEdgeList));
    result->m = M->size();
    result->n = result_n;
    result->wedge = M;

    return result;
}

std::string MWMGhaffariOpenMP::get_solver_name()
{
    return "mwm_openmp_v1";
}



double MWMGhaffariOpenMP::get_guarantee(RunParams* rp)
{
    return (2+ rp->epsilon*2);
}


std::vector<std::pair<int, int>>* MWMGhaffariOpenMP::generate_run_pairs(int n_proc)
{

    if(n_proc == 1)
    {
        std::vector<std::pair<int, int>>* result = new std::vector<std::pair<int, int>>();
        std::pair<int, int> p;
        p.first = 0; p.second = 0;
        result->push_back(p);
        return result;
    }

    // ONLY ALLOW EVEN NUMBERS!!!
    if(n_proc %2 == 1)
    {
        return nullptr;
    }


    int n_teams = n_proc*2;
    int* team_list = (int*) malloc(sizeof(int)* n_teams);
    for(int i = 0; i < n_teams; i++)
    {
        team_list[i] = i;
    }


    std::vector<std::pair<int, int>>* result = new std::vector<std::pair<int, int>>();

    for(int rounds = 0; rounds < n_teams-1; rounds++)
    {
        // store the matches
        for(int i = 0; i < n_teams/2; i++)
        {
            std::pair<int, int> p;

            int min = std::min(team_list[i], team_list[n_teams-1-i]);
            int max = std::max(team_list[i], team_list[n_teams-1-i]);

            p.first = min;
            p.second = max;

            result->push_back(p);
        }

        // move the teams accordingly
        int dest_value_now = team_list[1];
        for(int i = 1; i < n_teams;i++)
        {
            int dest_index = (i + 1) % n_teams;
            if(dest_index == 0)
                dest_index = 1;
            int dest_value_next = team_list[dest_index]; // value we will delete
            team_list[dest_index] = dest_value_now;
            dest_value_now = dest_value_next;
        }
    }

    //store the pairs (i, i)
    for(int i = 0; i < n_teams; i++)
    {
        std::pair<int, int> p;
        p.first = i;
        p.second = i;
        result->push_back(p);
    }

    free(team_list);

    return result;
}


double MWMGhaffariOpenMP::get_min_epsilon()
{
    return 0.8f;
}

double MWMGhaffariOpenMP::get_min_delta()
{
    return 0.01f;
}
