
#ifndef __MWM_OPENMP_V1_HPP__
#define __MWM_OPENMP_V1_HPP__

#include <vector>
#include <utility>
#include <vector>
#include <cmath>
#include <ctgmath>
#include <algorithm>

#include <omp.h>

#include "mwmsolver.hpp"
#include "../solver/solver.hpp"
#include "../runconfig/commonconfig.hpp"


/**
 * @brief The MWMOpenMPV1 class
 *
 * Maximum weighted matching solver, using OpenMP.
 *
 * Idea is as follows: Assume, we use P threads. We split the graph into 2*P regions, where each region is determined by
 * vertex_id % 2*P.
 *
 * We take O(n) edges, and determine the "bucket", to which an edge belongs, by determining the region in which the two
 * vertices are. Since we have 2*P regions, we will always be able to find P-many pairs of buckets (region_a, reigon_b), such that
 * we can run P-many pairs in parallel. The edges which are in this bucket are processed in parallel. Global datastructure is protected
 * by openmp specific barriers.
 *
 * Note that the overhead induced by this complicated approach is higher than the actually achieved speedup. Therefore,
 * the sequential variant is still faster.
 *
 */
class MWMGhaffariOpenMP : public MWMSolver
{
public:
    MWMGhaffariOpenMP();
    ~MWMGhaffariOpenMP();

    void* solve(void* pinput, Timing* timer, RunParams* rp);
    std::string get_solver_name();
    double get_guarantee(RunParams*);

    double get_min_epsilon();

    double get_min_delta();

private:
    /**
     * @brief generate_run_pairs
     * Implements the Round-Robin Tournament algorithm (https://en.wikipedia.org/wiki/Round-robin_tournament),
     * to determine regions that we can run in parallel
     *
     * @param n_proc
     * @return
     */
    std::vector<std::pair<int, int>>* generate_run_pairs(int n_proc);

};

#endif
