
#include "mwmserial.hpp"

MWMGhaffariSerial::MWMGhaffariSerial()
{
    // empty on purpose
}


MWMGhaffariSerial::~MWMGhaffariSerial()
{
    // empty on purpose
}

std::string MWMGhaffariSerial::get_solver_name()
{
    return "mwm_ghaffari_serial";
}


double MWMGhaffariSerial::get_guarantee(RunParams* rp)
{
    return 2*(1+ 2* (rp->epsilon) );
}

void* MWMGhaffariSerial::solve(void* pinput, Timing* timer, RunParams* rp)
{
    double epsilon = rp->epsilon;
    WEdgeList* input = (WEdgeList*) pinput;

    uint32_t timing_setup_id = timer->startTiming();

    // the beta constant
    const double beta = ( 5* log( 1 / epsilon) ) / epsilon;
    const uint32_t n = input->n;
    const uint32_t m = input->m;

    // the stack, as list to allow easy arbitrary removal of items (edges)
    std::list<WEdge> s;
    std::unordered_map<WEdge, std::list<WEdge>::iterator, WEdgeHash > iterator_map; // hashmap to keep track of where edges are stored

    std::vector<std::queue<WEdge>> q(n+1); // per vertex a queue
    std::vector<int64_t> phi(n+1, 0); // the phi variable, n+1 due to 1-based indexing of vertex id

    timer->endTimingSetup(timing_setup_id);
    uint32_t timing_l1_id = timer->startTiming();

    // loop over the elements
    for(uint32_t i = 0; i < m; i++)
    {

        // get all information
        const WEdge e = (*input->wedge)[i];
        const uint32_t u = e.u;
        const uint32_t uu = e.v;
        const uint32_t weight = e.weight;

        // test if we can drop or not
        if(  weight <= (1 + epsilon) * (  phi[u] + phi[uu]  )   )
        {
            continue;
        }

        // calculate remaining values and update phi
        int64_t wremaining = weight - ( phi[u] + phi[uu] );
        phi[u] = phi[u] + wremaining;
        phi[uu] = phi[uu] + wremaining;

        // puhs on stack and store, where the edge is stored in the list
        s.push_front(e);
        iterator_map.emplace(e, s.begin());

        uint32_t varr[2] = { uu, u};
        for(uint32_t j = 0; j < 2; j++)
        {
            // push on the queue
            uint32_t v = varr[j];
            q[v].push(e);

            if( q[v].size() >= beta)
            {
                // get element from the queue and remove
                WEdge ee = q[v].front();
                q[v].pop();

                // search in the stack (the list) and remove the edge there
                auto got = iterator_map.find(ee);
                if( got != iterator_map.end()) // should always be true!
                {
                    auto it = got->second;
                    s.erase(it);
                    iterator_map.erase(got); // erase the element from the map
                }
            }
        }
    }


#ifdef RUN_STATISTICS
    timer->setUInt("stack_size", s.size());
    for(uint32_t i = 0; i < q.size(); i++)
    {
        timer->setUInt("queue_size_"+std::to_string(i), q[i].size());
    }
#endif

    timer->endTimingProcessingLoop(timing_l1_id);
    uint32_t timing_l2_id = timer->startTiming();

    // iterate over the list and process the edges, such that we still have a matching
    auto M = new std::vector<WEdge>();
    std::vector<bool> vertex_set(input->n+1, false);

    uint32_t result_n = 0;
    while( !s.empty() )
    {
        WEdge e = s.front();

        // add only to M, if we have a matching
        if( !vertex_set[e.u] && !vertex_set[e.v] )
        {
            vertex_set[e.u] = true;
            vertex_set[e.v] = true;
            result_n = result_n+2;

            M->push_back(e);
        }
        s.pop_front();
    }

    timer->endTimingMergeLoop(timing_l2_id);
    uint32_t timing_preamble_id = timer->startTiming();

    WEdgeList* result = (WEdgeList*) malloc(sizeof(WEdgeList));
    result->m = M->size();
    result->n = result_n;
    result->wedge = M;
    timer->endTimingTeardown(timing_preamble_id);

    return result;
}

double MWMGhaffariSerial::get_min_epsilon()
{
    return 0.8;
}

double MWMGhaffariSerial::get_min_delta()
{
    return 0.0;
}
