
#include "performanceconfig.hpp"

const std::string PerformanceConfig::FILE_EXTENSION = ".performance";

PerformanceConfig::PerformanceConfig(std::string file_path, uint32_t seed) : CommonConfig(file_path, seed)
{
    std::ifstream ifs(file_path);
    rapidjson::IStreamWrapper isw(ifs);
    rapidjson::Document d;
    d.ParseStream(isw);

    this->fpga_frequency = 200;
    this->delete_folder = true;



#ifdef HWAFU
    if(d.HasMember("output_folder_harp"))
    {
        rapidjson::Value& ofolder = d["output_folder_harp"];
        this->output_folder = ofolder.GetString();
    }
    else
    {
        if( d.HasMember("output_folder"))
        {
            rapidjson::Value& ofolder = d["output_folder"];
            this->output_folder = ofolder.GetString();
            std::cout << "use the output_folder to write out performance data!" << std::endl;
        }
        else
        {
            throw std::invalid_argument("Did not find member output_folder in file "+ file_path);
        }
    }

#else
    if(d.HasMember("output_folder"))
    {
        rapidjson::Value& ofolder = d["output_folder"];
        this->output_folder = ofolder.GetString();
    }
    else
    {
        throw std::invalid_argument("Did not find member output_folder in file "+ file_path);
    }
#endif

    if(d.HasMember("fpga_frequency"))
    {
        rapidjson::Value& ofolder = d["fpga_frequency"];
        this->fpga_frequency = ofolder.GetUint();
    }

    if(d.HasMember("delete_folder"))
    {
        rapidjson::Value& ofolder = d["delete_folder"];
        this->delete_folder = ofolder.GetBool();
    }
}


PerformanceConfig::~PerformanceConfig()
{
    // empty on purpose
}


void PerformanceConfig::execute()
{
    InputGenerator ig;
    Solver* first_solver = solvers.at(0);

    std::string output_folder = Timing::get_output_folder( this->output_folder, first_solver->get_problem_name() );

    // clean the folder, to which we write all output
    if(this->delete_folder)
    {
        if(Util::is_dir(output_folder.c_str()))
        {
            Util::cleanFolder(output_folder);
        }
    }

    // iterate over all different performance-configurations
    for(auto itp = this->run_params.begin(); itp != this->run_params.end(); ++itp)
    {
        RunParams rp = *itp;
        std::cout << "Config: rep="<< rp.rep << ", max_w="<< rp.max_w <<", rand_weights="<< rp.rand_weights
                  << ", epsilon="<< rp.epsilon << ", delta="<< rp.delta<< ", nthreads="<< rp.nthreads
                  <<", num_levels="<< rp.num_levels<< ", k="<< rp.k << std::endl;


        uint32_t n = 0;
        uint32_t m = 0;

        void* list = nullptr;
        if(first_solver->get_input_type() == Solver::INPUT_TYPE_DYNAMIC_UNWEIGHTED || first_solver->get_input_type() == Solver::INPUT_TYPE_UNWEIGHTED)
        {
            list = (void*)ig.uload(this->file_path);

            if(list == nullptr)
            {
                return;
            }

            Util::print_edge_list( (EdgeList*)list, 10);
            EdgeList* l = (EdgeList*)list;
            n = l->n;
            m = l->m;
        }
        else
        {
            uint32_t k = (*itp).k;
            list = ig.load(this->file_path, (*itp).rand_weights, (*itp).max_w);

            if(list == nullptr)
            {
                return;
            }

            {
                WEdgeList* l = (WEdgeList*)list;
                n = l->n;
                m = l->m;
            }

            if( k >= 1)
            {
                WEdgeList* ltemp = (WEdgeList*) list;
                std::cout << "start convert to k-merged"<< std::endl;
                WEdgeList* l = Util::convert_wedgelist_to_k_merged_wedgelist(ltemp, k);
                std::cout << "end convert to k-merged"<< std::endl;
                list = (void*)l;

                n = l->n;
                m = l->m;

                delete(ltemp->wedge);
                ltemp->wedge = nullptr;
                free(ltemp);
                ltemp = nullptr;
            }
        }

        if(list == nullptr)
        {
            std::cout << "Loading of file returned empty list: "<< this->file_path << std::endl;
        }

        // call all solvers on this performance-configuration
        for(auto it = this->solvers.begin(); it != this->solvers.end(); ++it)
        {
            Solver* s = (*it);

            double orig_epsilon = rp.epsilon;
            double orig_delta = rp.delta;
            rp.m = m;
            rp.n = n;

            std::cout << "--- Start performance: " << (*it)->get_solver_name() << ", nthreads=" << rp.nthreads <<", m="<< rp.m<< ", n="<< rp.n << std::endl;

            Timing t( &rp, this->file_path, (*it)->get_solver_name() );
            t.setUInt("seed", this->seed);
            t.setUInt("fpga_frequency", this->fpga_frequency);

            t.startAllRep();
            // loop over the number of repetitions
            for(int i = 0; i< rp.rep; i++)
            {
                t.startRep();
                void* result =  s->solve( list, &t, &rp);
                t.endRep();

                if(result != nullptr)
                {
                    s->delete_output(result);
                }
#ifdef FPGA_TIMEOUT
                std::cout << "sleep shortly"<< std::endl;
                // this are 200mio cycles @200MHz which we wait, to remove all the data in the BRAMs
                // note that we measure each repetition - therefore, waiting for a second does not
                // affect our measurements
                sleep(1);
#endif
            }
            t.endAllRep();

            t.writeFile(output_folder);

            rp.delta = orig_delta;
            rp.epsilon = orig_epsilon;
            std::cout << "--- End performance" << std::endl;
        }

        // sleep for 1ms, to get a new file-name for sure
        usleep(1000*1);

        first_solver->delete_input(list);
    }
}
