
#include "timing.hpp"


Timing::Timing(RunParams* rp, std::string file_path, std::string method_name)
{
    setUInt("num_n", rp->n );
    setUInt("num_m", rp->m);
    setUInt("nrep", rp->rep);
    setString("file_path", file_path);
    std::string file_name = Util::getFileName(file_path);
    setString("file_name", file_name);
    setString("method_name", method_name);
    setString("time", Util::currentDateTime());
    setDouble("delta", rp->delta);
    setDouble("epsilon", rp->epsilon);
    setInt("k", rp->k);
    setInt("max_w", rp->max_w);
    setInt("nthreads", rp->nthreads);
    setInt("num_levels", rp->num_levels);
    setInt("rand_weights", (int)rp->rand_weights);

    long int t = this->get_time_ms();
    output_file_name = method_name +"_"+file_name+"_"+Util::currentDateTime()+"_"+ std::to_string(t) +".txt";

    this->num_rep = 0;
}


Timing::~Timing()
{
    // empty on purpose
}

void Timing::setInt(std::string key, int32_t value)
{
    std::pair<std::string, std::string> p ( key, std::to_string(value) );
    this->map.push_back(p);
}

void Timing::setUInt(std::string key, uint32_t value)
{
    std::pair<std::string, std::string> p ( key, std::to_string(value) );
    this->map.push_back(p);
}

void Timing::setULong(std::string key, uint64_t value)
{
    std::pair<std::string, std::string> p(key, std::to_string(value));
    this->map.push_back(p);
}

void Timing::setULong(std::string key, uint64_t value, bool unique)
{
    std::string prefix = "";
    if(!unique)
    {
        prefix = "rep_"+std::to_string(this->num_rep)+"_";
    }
    std::pair<std::string, std::string> p(prefix +key, std::to_string(value));
    this->map.push_back(p);
}

void Timing::setDouble(std::string key, double value)
{
    std::pair<std::string, std::string> p ( key, std::to_string(value) );
    this->map.push_back(p);
}

void Timing::setString(std::string key, std::string value)
{
    std::pair<std::string, std::string> p ( key, value);
    this->map.push_back(p);
}


void Timing::startRep()
{
    this->timing_id_rep = this->startTiming();
}

void Timing::endRep()
{
    this->endTiming(this->timing_id_rep, "one_rep", false);
    this->num_rep++;
}


void Timing::startAllRep()
{
    this->timing_id_global = this->startTiming();
}

void Timing::endAllRep()
{
    this->endTiming(this->timing_id_global, "all_reps", true);
}

uint32_t Timing::startTiming()
{

    myInt64 cycles = start_tsc();
    double msecs = this->get_time();

    std::pair<double, myInt64> p(msecs, cycles);

    this->timings.push_back(p);
    return (this->timings.size() -1);
}

void Timing::endTiming(uint32_t identifier, std::string timing_name, bool unique)
{
    std::pair<double, myInt64> p = this->timings[identifier];

    myInt64 cycles = stop_tsc(p.second); // directly returns the amount of elapsed cycles
    double msecs = this->get_time() - p.first;

    std::string prefix = "";
    if( !unique )
    {
        prefix = "rep_"+std::to_string(this->num_rep)+"_";
    }

    this->setDouble(prefix+"msecs_"+timing_name, msecs);
    this->setULong(prefix+"cycles_"+timing_name, cycles);
}

void Timing::endTimingSetup(uint32_t identifier)
{
    this->endTiming(identifier, "setup", false);
}

void Timing::endTimingTeardown(uint32_t identifier)
{
    this->endTiming(identifier, "teardown", false);
}

void Timing::endTimingProcessingLoop(uint32_t identifier)
{
    this->endTiming(identifier, "processing_loop", false);
}

void Timing::endTimingMergeLoop(uint32_t identifier)
{
    this->endTiming(identifier, "merge_loop", false);
}

void Timing::endTimingMainProcessing(uint32_t identifier)
{
    this->endTiming(identifier, "main_processing", false);
}


std::string Timing::get_output_folder(std::string folder_path, std::string problem_name)
{
    std::string needle = "$problem$";
    if(folder_path.find(needle) != 0)
    {
        Util::replace( folder_path,needle, problem_name);
    }
    else
    {
        folder_path=folder_path+"/"+problem_name;
    }

    return folder_path;
}

void Timing::writeFile(std::string folder_path)
{
    if( !Util::createFolder(folder_path) )
    {
        std::cout << "could not create folder: "<< folder_path << std::endl;
        return;
    }


    std::ofstream myfile;
    std::string full_file_path;
    if(Util::has_suffix(folder_path, "/"))
    {
        full_file_path = folder_path + this->output_file_name;
    }
    else
    {
        full_file_path = folder_path+"/"+this->output_file_name;
    }

    myfile.open (full_file_path);

    if(myfile.bad() || myfile.fail())
    {
        std::cout << "Cannot open file for writing logs: "<< full_file_path<< std::endl;
        return;
    }

    std::cout << "Write into file: "<< full_file_path << std::endl;

    for(auto it = this->map.begin(); it != this->map.end(); it++)
    {
        auto p = *it;
        myfile<< p.first <<", "<< p.second<< "\r\n";
    }

    myfile.close();
}


// time in ms, but as double
double Timing::get_time()
{
    struct timespec tp;
    clock_gettime(CLOCK_MONOTONIC_RAW, &tp);
    double ret = (tp.tv_sec * 1e6 + tp.tv_nsec / 1e3) / 1e6; // time in sec
    return (ret*1000); // time in ms
}

// time in ms, but as integer!
long int Timing::get_time_ms()
{
    struct timespec tp;
    clock_gettime(CLOCK_MONOTONIC_RAW, &tp);
    long int ms = (tp.tv_sec * 1e3 + tp.tv_nsec / 1e6); // time in ms
    return ms;
}

uint32_t Timing::get_repetition()
{
    return this->num_rep;
}
